(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_achain.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2007-10-23 15:10:38 $
 * Version: $Revision: 1.25 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. AChain/def_achain.ml pour les infos sur les types achain et acell
 *          
 *
 *=========================================================================
 *)

open Global;;
open Def_achain;;
open Valeur;;
open Dispatch_sig;;

__register_version "$RCSfile: core_achain.ml,v $" "$Revision: 1.25 $" "$Date: 2007-10-23 15:10:38 $";;


(**************************************************************************
 ** Fonctions de manipulation des acells
 **************************************************************************)

(* --- fct internes : ne pas utiliser dehors ! ------------------------ *)
let internal_add_face f s = if not(List.mem f s.faces_acell) then s.faces_acell <- f::s.faces_acell
and internal_add_coface cf s = if not(List.mem cf s.cofaces_acell) then s.cofaces_acell <- cf::s.cofaces_acell
;;

(* --- Dictionnaire---------------------------------------------------- *)

let acell_dico = ((Hash.create 1223) : (int, acell) Hash.t)
;;

(* --- Accesseurs ----------------------------------------------------- *)

let get_dim_from_acell c = c.dim_acell ;;

let get_uid_from_acell c = c.uid_acell ;;

let get_modifiable_from_acell c = c.modifiable_acell ;;

let get_acell_from_uid uid = Hash.find acell_dico uid ;;

let set_modifiable_of_acell c = c.modifiable_acell <- false ;;

(* --- Cration ------------------------------------------------------- *)

(** generateur d'uid de acell unique. On commence  compter  1. 
 ** Si on retombe sur 0 c'est qu'on a trop de acells. 
 **)
let cpt_acell = ref (0)
and default_acell = { uid_acell = 0; dim_acell = -1; faces_acell = []; cofaces_acell = []; modifiable_acell = true }
;;

let new_uid() = 
  let uid = (incr cpt_acell; !cpt_acell)  in 
    if uid = 0 then raise (Erreur "Error: internal error: too much acells !!!") else uid
;;

(** creation d'une acell de dimension d  partir d'une liste de faces et d'une liste de co-faces.
 ** Il faut mettre  jour les faces et les cofaces ; ATTENTION : vrifier que le
 ** champ modifiable_acell est mis  true !!!
 **
 ** int -> acell list -> acell list -> acell
 **
 **)

let new_acell_force d fl cfl =
  let dim_face   = d-1
  and dim_coface = d+1 in
    begin
      (if (List.exists
	     (fun ac -> (ac.dim_acell <> dim_face))
	     fl)
       then
	 let err = "Error: acell of dimension "^(string_of_int d)
		   ^" cannot have this faces list (dimension mismatch).\n"
	 in
	   prerr_endline err;
	   prerr_newline();
	   raise (Erreur err);
       else if (List.exists
		  (fun ac -> (ac.dim_acell <> dim_coface))
		  cfl)
       then
	 let err = "Error: acell of dimension "^(string_of_int d)
		   ^" cannot have this cofaces list (dimension mismatch).\n"
         in
	   prerr_endline err;
	   prerr_newline();
	   raise (Erreur err)
      );
      let nuid = new_uid() in
      let nsimp = { uid_acell = nuid; dim_acell = d; faces_acell = fl; cofaces_acell = cfl; modifiable_acell = true } in
	begin
	  List.iter (internal_add_face nsimp) cfl;
	  List.iter (internal_add_coface nsimp) fl;
	  Hash.add acell_dico nsimp.uid_acell nsimp;
	  nsimp
	end
    end
;;

let new_acell d fl cfl =
  let dim_face   = d-1
  and dim_coface = d+1 in
    begin
      (if (List.exists
	     (fun ac ->
		   (ac.dim_acell <> dim_face)
		|| (not ac.modifiable_acell))
	     fl)
       then
	 let err = "Error: acell of dimension "^(string_of_int d)
		   ^" cannot have this faces list (dimension mismatch or forbidden topological modification).\n"
	 in
	   prerr_endline err;
	   prerr_newline();
	   raise (Erreur err);
       else if (List.exists
		  (fun ac ->
		        (ac.dim_acell = dim_coface)
		     || (not ac.modifiable_acell))
		  cfl)
       then
	 let err = "Error: acell of dimension "^(string_of_int d)
		   ^" cannot have this cofaces list (dimension mismatch or forbidden topological modification).\n"
         in
	   prerr_endline err;
	   prerr_newline();
	   raise (Erreur err)
      );
      let nuid = new_uid() in
      let nsimp = { uid_acell = nuid; dim_acell = d; faces_acell = fl; cofaces_acell = cfl; modifiable_acell = true } in
	begin
	  List.iter (internal_add_face nsimp) cfl;
	  List.iter (internal_add_coface nsimp) fl;
	  Hash.add acell_dico nsimp.uid_acell nsimp;
	  nsimp
	end
    end
;;

let new_vertex() = new_acell 0 [] []
and new_edge v1 v2 = new_acell 1 [v1; v2] []
and new_face e1 e2 e3 = new_acell 2 [e1; e2; e3] []
;;

(* --- Manipulations ------------------------------------------------- *)

(* Aussitot qu'une acell apparait dans une chaine, on doit la geler pour
 * interdire les changements de topologie...
 *)
let rec freeze_acell c =
  if (c.modifiable_acell) then (
    c.modifiable_acell <- false ;
    List.iter freeze_acell (c.faces_acell) ;
    List.iter freeze_acell (c.cofaces_acell))

(* --- Requetes ------------------------------------------------------ *)

let acelldim s = s.dim_acell ;;

let acelluid s = s.uid_acell ;;

let compare_acell c1 c2 = Pervasives.compare c1.uid_acell c2.uid_acell ;;

module SAcell = Set.Make(struct type t = acell let compare = compare_acell end)
;;

let sacellMap f ss = SAcell.fold (fun e suite -> SAcell.add (f e) suite) ss SAcell.empty
;;




(* Rsultat sous forme d'ensemble SAcell *)

let rec faces_sacell s   = List.fold_left (fun acc x -> (SAcell.add x acc)) SAcell.empty s.faces_acell

and cofaces_sacell s = List.fold_left (fun acc x -> (SAcell.add x acc)) SAcell.empty s.cofaces_acell

and icells_sacell s i =
  let d = s.dim_acell in
  let fonc,diff = if (i > d) then cofaces_sacell,(i-d) else faces_sacell,(d-i) in
  let rec icells_ acc n = match n with
    | 0 -> acc
    | n ->
	(
	  let acc' = SAcell.fold
		       (fun x a -> let fonc_x = fonc x in SAcell.union fonc_x a)
		       acc
		       SAcell.empty
	  in
	    icells_ acc' (n-1)
	)
  in
    icells_ (SAcell.add s SAcell.empty) diff

and ccells_sacell s i =
  let d  = s.dim_acell in
  let ic = icells_sacell s i in
    SAcell.diff
      (SAcell.fold (fun x a -> SAcell.union (icells_sacell x d) a) ic SAcell.empty)
      (SAcell.add s SAcell.empty)

and transitive_faces_sacell s = tfaces SAcell.empty s
and tfaces set s = List.fold_left (fun set x -> tfaces (SAcell.add x set) x) set s.faces_acell

and transitive_cofaces_sacell s = tcofaces SAcell.empty s
and tcofaces set s = List.fold_left (fun set x -> tcofaces (SAcell.add x set) x) set s.cofaces_acell

and common_faces_sacell s1 s2 = SAcell.inter (faces_sacell s1) (faces_sacell s2)

and common_cofaces_sacell s1 s2 = SAcell.inter (cofaces_sacell s1) (cofaces_sacell s2)

and common_icells_sacell s1 s2 i = SAcell.inter (icells_sacell s1 i) (icells_sacell s2 i)

and common_ccells_sacell s1 s2 i = SAcell.inter (ccells_sacell s1 i) (ccells_sacell s2 i)

and common_transitive_faces_sacell s1 s2 =
  let tf1 = transitive_faces_sacell s1
  and tf2 = transitive_faces_sacell s2
  in (SAcell.inter tf1 tf2)

and common_transitive_cofaces_sacell s1 s2 =
  let tf1 = transitive_cofaces_sacell s1
  and tf2 = transitive_cofaces_sacell s2
  in (SAcell.inter tf1 tf2)


(* Rsultat sous forme de liste *)

and faces s    = s.faces_acell

and cofaces s  = s.cofaces_acell

and icells s i = SAcell.elements(icells_sacell s i)

and ccells s i = SAcell.elements (ccells_sacell s i)

and transitive_faces s = SAcell.elements (transitive_faces_sacell s)

and transitive_cofaces s = SAcell.elements (transitive_cofaces_sacell s)

and common_faces s1 s2 = SAcell.elements(common_faces_sacell s1 s2)

and common_cofaces s1 s2 = SAcell.elements(common_cofaces_sacell s1 s2)

and common_icells s1 s2 i = SAcell.elements(common_icells_sacell s1 s2 i)

and common_ccells s1 s2 i = SAcell.elements(common_ccells_sacell s1 s2 i)

and common_faces_cofaces s1 s2 = List.filter (fun  x -> List.mem x s1.cofaces_acell) s1.faces_acell

and common_transitive_faces   s1 s2 = SAcell.elements (common_transitive_faces_sacell   s1 s2)

and common_transitive_cofaces s1 s2 = SAcell.elements (common_transitive_cofaces_sacell s1 s2)
;;





(**************************************************************************
 ** Fonctions de manipulation des chains
 **************************************************************************)

 (** valeur qu'on affecte par defaut aux simplexes qu'on cree 
  ** pour assurer la coherence de la structure de complexe.
  **)
let default_value = Valeur.mgsundef
;;

(* --- Manipulation de chaines ------------------------------------------------------- *)

let map fct ch =
  let ret = Hash.copy ch in
    begin
      Hash.iter (fun uid (sp, v) -> Hash.replace ret uid (sp, (fct sp v))) ret;
      ret
    end

let map_d dim fct ch =
  let ret = Hash.copy ch in
    begin
      Hash.iter (fun uid (sp, v) -> if (acelldim sp = dim) then Hash.replace ret uid (sp, (fct sp v))) ret;
      ret
    end
    
let iter fct ch =
  Hash.iter (fun uid (sp, v) -> fct sp v) ch
  
let iter_d vdefault dim fct ch =
  Hash.iter (fun uid (sp, v) -> if (acelldim sp = dim) then fct sp v else vdefault) ch
  
let fold fct ch zero =
  Hash.fold (fun uid (sp, v) suite -> fct sp v suite) ch zero

let fold_d dim fct ch zero = 
  Hash.fold (fun uid (sp, v) suite -> if (acelldim sp = dim) then fct sp v suite else suite) ch zero

let forall fct ch =
  Hash.forall (fun uid (sp, v) -> fct sp v) ch
  
let forall_d dim fct ch =
  Hash.forall (fun uid (sp, v) -> ((acelldim sp <> dim) || (fct sp v)) ) ch
  
let exists fct ch =
  Hash.exists (fun uid (sp, v) -> fct sp v ) ch

let exists_d dim fct ch =
  Hash.exists (fun uid (sp, v) -> ((acelldim sp = dim) && (fct sp v)) ) ch

let size ch =
  Hash.size ch

let size_d dim ch =
  let nb = ref 0 in
    Hash.iter (fun uid (sp, v) -> if (acelldim sp = dim) then nb := !nb + 1) ch ;
    !nb

let find ch sp =
  snd (Hash.find ch (acelluid sp))

let nth n ch =
  snd (Hash.nth n ch)

exception FoundAcell

let nth_d dim n ch =
  let value = ref None in
  try
    let pos = ref n in
      Hash.iter
        (fun uid (sp, v) ->
           if (acelldim sp = dim) then
             if (!pos = 0) then (
	       value := Some (sp, v) ;
               raise FoundAcell
             ) else
               pos := !pos - 1
        )
        ch ;
      raise Not_found
  with
    | FoundAcell -> !value

let remove ch sp = Hash.remove ch (acelluid sp)

let copy ch = Hash.copy ch

let hd ch = let (_, (sp, v)) = Hash.hd ch in (sp,v)

let last ch = let (_, (sp, v)) = Hash.last ch in (sp,v)

let tl ch = Hash.tl ch

let mem ch s = Hash.mem ch (acelluid s)

let compare valcmp ch1 ch2 = Hash.compare (fun (_,v1) (_,v2) -> valcmp v1 v2) ch1 ch2

let to_list ch = fold (fun sp v s -> (sp,v)::s) ch []

(** dimension **)
let get_dim_from_achain ch = Hash.fold (fun _ (acell, _) acc -> max acell.dim_acell acc) ch (-1) ;;

let get_acell_from_uid_in_achain ch uid = fst(Hash.find ch uid)

let get_value_from_uid_in_achain ch uid = snd(Hash.find ch uid)

let get_uid_from_val_in_achain ch v =
  Hash.fold (fun uid (_, vv) acc -> if v = vv then uid::acc else acc) ch []

let get_value_from_acell_in_achain ch sp = get_value_from_uid_in_achain ch (acelluid sp)

let get_acell_from_val_in_achain ch v =
  Hash.fold (fun uid (sp, vv) acc -> if v = vv then sp::acc else acc) ch []

let get_acells_of_dim_in_achain dim ch =
  let h = Hash.create 11
  in
    iter (fun s v -> if ((acelldim s) = dim) then Hash.add h (acelluid s) (s,v) ) ch;
    h

(** Renvoie un tableau dont l'lment i contient 
 ** la liste des acells de dimension i d'une chaine 
 **)
let skelette ch =
  let skel = Array.make (1+(get_dim_from_achain ch)) []
  in
    Hash.iter (fun _ (acell,v) -> skel.(acell.dim_acell) <- (acell,v)::skel.(acell.dim_acell)) ch;
    skel

let faces_achain ch sp =
  if (Hash.mem ch (acelluid sp)) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (faces sp)
      []
  else
    let err = "Error: Core_achain.faces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let cofaces_achain ch sp =
  if (Hash.mem ch (acelluid sp)) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (cofaces sp)
      []
  else
    let err = "Error: Core_achain.cofaces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let ccells_achain ch sp i =
  if (Hash.mem ch (acelluid sp)) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (ccells sp i)
      []
  else
    let err = "Error: Core_achain.ccells_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let icells_achain ch sp i =
  if (Hash.mem ch (acelluid sp)) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (icells sp i)
      []
  else
    let err = "Error: Core_achain.icells_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let transitive_faces_achain ch sp =
  if (Hash.mem ch (acelluid sp)) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (transitive_faces sp)
      []
  else
    let err = "Error: Core_achain.transitive_faces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let transitive_cofaces_achain ch sp =
  if (Hash.mem ch (acelluid sp)) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (transitive_cofaces sp)
      []
  else
    let err = "Error: Core_achain.transitive_cofaces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let common_faces_achain ch sp1 sp2 =
  if ( (Hash.mem ch (acelluid sp1)) && (Hash.mem ch (acelluid sp2)) ) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (common_faces sp1 sp2)
      []
  else
    let err = "Error: Core_achain.common_faces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let common_cofaces_achain ch sp1 sp2 =
  if ( (Hash.mem ch (acelluid sp1)) && (Hash.mem ch (acelluid sp2)) ) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (common_cofaces sp1 sp2)
      []
  else
    let err = "Error: Core_achain.common_cofaces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let common_transitive_faces_achain ch sp1 sp2 =
  if ( (Hash.mem ch (acelluid sp1)) && (Hash.mem ch (acelluid sp2)) ) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (common_transitive_faces sp1 sp2)
      []
  else
    let err = "Error: Core_achain.common_transitive_faces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)

let common_transitive_cofaces_achain ch sp1 sp2 =
  if ( (Hash.mem ch (acelluid sp1)) && (Hash.mem ch (acelluid sp2)) ) then
    List.fold_right
      (fun fp acc -> try let (c,v) = Hash.find ch (acelluid fp) in (c,v)::acc with Not_found -> acc)
      (common_transitive_cofaces sp1 sp2)
      []
  else
    let err = "Error: Core_achain.common_transitive_cofaces_achain: acell not in achain"
    in prerr_endline err; raise(Erreur err)




(* --- Construction de achain ------------------------------------------------------- *)

let new_achain () = Hash.create 11

let create_achain d = Hash.create d

(** --- une somme --------------- **)

(** quand on ajoute un acell a une achain, on suppose que les faces
 ** et co-faces des acells de la achains sont deja a jour. C'est le cas
 ** si on utilise new_acell pour creer le acell avant de l'ajouter.
 **
 ** Par contre, on ne suppose pas que les faces appartiennent dj  la achain. 
 ** Et on les rajoutes si ncessaire.
 ** Mais on ne rajoute pas les cofaces !!! 
 ** (c'est voulu, on inclus vers le bas les constituant, mais on ne
 **  fait pas apparaitre les structures dans lesquelles on apparait).
 **)

let add achain s v = Hash.add achain (acelluid s) (s,v)

let replace achain s v = Hash.replace achain (acelluid s) (s,v)

let rec add_acell achain s v =
  begin
    freeze_acell s;
    List.iter (fun f -> add_one_acell achain f default_value) s.faces_acell;
    add_force_one_acell achain s v;
    achain
  end

and add_one_acell achain s v =
  if not(Hash.mem achain s.uid_acell) then ignore(add_acell achain s v)

and add_force_one_acell achain s v =
  Hash.replace achain s.uid_acell (s,v)
;;


(** rajoute une acell ou bien combine la valeur si la acell est deja present **)
let update_achain ch combine v acell =
  try
    let (_, v0) = Hash.find ch acell.uid_acell
    in Hash.replace ch acell.uid_acell (acell, (combine v0 v))
  with Not_found -> ignore(add_acell ch acell v)
;;


(** Generation d'une filtration... **)

exception FiltrationFindInSK of acell * int ;;
exception FiltrationEnd ;;
let filtration ch fct =

  let sk = skelette ch in

  let cmp (c,v) (c',v') =
    let d = get_dim_from_acell c in
      fct d v v'
  in

  let tmp_ch = new_achain () in
  let sk = Array.map
	     (List.map (fun (c,v) -> add tmp_ch c (List.fold_left (fun i f ->
								     match find ch f with
								       | VAL_undef _ -> i
								       | _           -> i+1
								  ) 0 c.faces_acell,ref 0
						  ) ; c ))
	     (Array.map (fun ski -> List.sort cmp (List.filter (function (_, VAL_undef _) -> false | _ -> true) ski)) sk)
  in

  let l = (Array.length sk) - 1 in
  let ret = Array.make (size ch) default_acell in
  let current_ret_case = ref 0 in
    try
      while true do
	try
	  for i = l downto 0 do
	    if (sk.(i) <> []) then
	      let c = List.hd(sk.(i)) in
	      let (nb_faces, nb_traites) = find tmp_ch c in
		if (nb_faces = !nb_traites) then
		  raise (FiltrationFindInSK (c,i))
	  done ;
	  raise FiltrationEnd
	with
	  | FiltrationFindInSK (c,i) -> (
	      List.iter (fun cf -> let (nb,t) = find tmp_ch cf in incr t) c.cofaces_acell ;
	      sk.(i) <- List.tl(sk.(i)) ;
	      ret.(!current_ret_case) <- c ;
	      incr current_ret_case
	    )
      done ;
      let err = "Error: Core_achain: filtration: internal major failure"
      in prerr_endline err; raise(Erreur err)
    with
      | FiltrationEnd -> Array.sub ret 0 (!current_ret_case)
;;


(** Qmification d'une chaine et chainification d'une qmf **)

open Def_qmf ;;

let to_qmf (funct : 'a -> float * float * float ) (c : 'a achain) (((g,ha,d) as q):'a qmf) default_value = q
;;

(*
let from_qmf (((g,ha,d) as q):'a qmf) = new_achain ()
*)


let from_qmf (((g,ha,d) as q):'a qmf) =
  let cp = Hash.create 11 in
  let ch = new_achain () in
    for i = 0 to d do
      Core_qmf.iter_d
	i
	( fun nc v -> 
	    let ac = new_acell_force
	               i 
		       (List.map (fun c -> try Hash.find cp (Core_qmf.normal_form c) with Not_found -> raise (Erreur "gloups ")) (Core_qmf.faces nc))
		       []
	    in
	      Hash.add cp nc ac ;
	      ignore(add_acell ch ac v)
	)
	q
    done ;
    ch
;;

open Def_del ;;

let from_delaunay (((telements,_) as d):'a tdel) =
  let cp = Hash.create 11 in
  let ch = new_achain () in
    Array.iteri
      (fun i ( (value,_,voisins) as elt) ->
	 let cell =
	   try
	     let c = Hash.find cp i in
	       ignore(add_acell ch c value) ;
	       c
	   with Not_found ->
	     let ac = new_acell_force 0 [] [] in
	       Hash.add cp i ac ;
	       ignore(add_acell ch ac value) ;
	       ac
	 in
	   List.iter
	     (fun p ->
		try
		  let c_vois = Hash.find cp p in
		  let arc = new_acell_force 1 [cell;c_vois] [] in
		    ignore(add_acell ch arc (VAL_undef "Achainify from delaunay dummy value"))
		with Not_found -> ()		    
	     )
	     voisins
      )
      telements
    ;
    ch
;;

let from_seq to_mgs_val s =
  let vertices = Hash.create 11 in
  let ch = new_achain () in
    Array.iteri
      (fun i v ->
	 let cell =
	   try
	     let c = Hash.find vertices i in
	       ignore(add_acell ch c v) ;
	       c
	   with Not_found ->
	     let ac = new_acell_force 0 [] [] in
	       Hash.add vertices i ac ;
	       ignore(add_acell ch ac v) ;
	       ac
	 in
	   (try
	      let c_vois = Hash.find vertices (i-1) in
	      let arc = new_acell_force 1 [cell;c_vois] [] in
		ignore(add_acell ch arc (
			 let ev = Core_rec.empty in
			 let ev = Core_rec.add (VAL_symbol "`vi") (VAL_acell c_vois) ev in
			 let ev = Core_rec.add (VAL_symbol "`vj") (VAL_acell cell) ev in
			   to_mgs_val ev
		       )
		      )
	    with Not_found -> ()) ;
	   (try
	      let c_vois = Hash.find vertices (i+1) in
	      let arc = new_acell_force 1 [cell;c_vois] [] in
		ignore(add_acell ch arc (
			 let ev = Core_rec.empty in
			 let ev = Core_rec.add (VAL_symbol "`vi") (VAL_acell cell) ev in
			 let ev = Core_rec.add (VAL_symbol "`vj") (VAL_acell c_vois) ev in
			   to_mgs_val ev
		       ))
	    with Not_found -> ())
      )
      s
    ;
    ch

;;
(*
let to_qmf (funct : 'a -> float * float * float ) (c : 'a achain) (((g,ha,d) as q):'a qmf) default_value =
  let values = Hash.create 11 in
  let cpy = Hash.create 11 in

  let rec make_darts = function
    | []         -> []
    | vol::suite ->
	let face = faces vol in
	let fdarts = make_darts face in
	let vfdarts = if (fdarts = []) then [[vol]] else List.map (fun x -> x@[vol]) fdarts in
(*	  Hash.replace values vol (try find c vol with Not_found -> default_value) ;*)
	  vfdarts@(make_darts(suite))
  in

  let rec cmp_dart d1 d2 = cmp_dart_ (fst d1) (fst d2) 0
  and cmp_dart_ d1 d2 i = match (d1,d2) with
    | [], [] -> -1
    | c1::dd1, c2::dd2 ->
	if (c1 <> c2) then
	  if (dd1 = dd2) then
	    i
	  else
	    -1
	else
	  cmp_dart_ dd1 dd2 (i+1)
    | _ -> 
	let err = "Error: Core_achain.to_qmf: cmp_dart: length"
	in prerr_endline err; raise(Erreur err)
  in

  let dim_c = get_dim_from_achain c in
    if (dim_c > 3) then
      let err = "Error: Core_achain.to_qmf: dimension greater than 3"
      in prerr_endline err; raise(Erreur err)
    else (
      let greater_objects = fold_d dim_c (fun ac v acc -> ac::acc) c [] in
      let darts = Array.of_list( make_darts greater_objects) in
      let idarts = Array.mapi (fun i dart -> dart,[|i;i;i;i|]) darts in
      let l = (Array.length idarts) - 1 in
	for i = 0 to l do
	  let ((dart1,alpha1) as d1) = idarts.(i) in
	    for j = (i+1) to l do
	      let ((dart2,alpha2) as d2) = idarts.(j) in
		match (cmp_dart d1 d2) with
		  | -1 -> ()
		  | n -> 
		      if (n>dim_c) then
			let err = "Error: Core_achain.to_qmf: dimension error"
			in prerr_endline err; raise(Erreur err)
		      else if (alpha1.(n) = i) && (alpha2.(n) = j) then (
			alpha1.(n) <- j ;
			alpha2.(n) <- i ;
		      ) else
			let err = "Error: Core_achain.to_qmf: not a quasimanifold"
			in prerr_endline err; raise(Erreur err)
	    done 
	done ;
	(**
	  Printf.printf "dimension: %i\n" dim_c ; flush stdout ;
	  Printf.printf "nb objects: %i\n" (List.length greater_objects) ; flush stdout ;
	  Printf.printf "nb darts: %i\ndarts:\n" (Array.length darts) ; flush stdout ;
	  Array.iteri (fun i (dart,alpha) ->
	  Printf.printf "\t[%2i](" i;
	  List.iter (fun c -> Printf.printf "%2i, " (get_uid_from_acell c)) dart ;
	  Printf.printf ") " ;
	  Array.iteri (fun j aj ->
	  Printf.printf "[%2i]:(%2i) " j aj;
	  ) alpha ;
	  Printf.printf "\n"; flush stdout ;
	  ) idarts ;
	**)
	let idarts = Array.mapi
		       (fun i (dart,alpha) ->
			  let (x,y,z) = funct (try find c (List.hd dart) with Not_found -> default_value) in
			  let dd = Core_qmf.newDart g x y z in
			    List.iter (fun c -> Hash.replace cpy c dd) dart;
			    (dart,alpha,dd)
		       )
		       idarts
	in
	  for i = 0 to l do
	    let ((n1,alpha1,dart1) as d1) = idarts.(i) in
	      for j = 0 to ((Array.length alpha1) - 1) do
		if (i < alpha1.(j)) then
		  let ((n2,alpha2,dart2) as d2) = idarts.(alpha1.(j)) in
		    Core_qmf.sew g dart1 dart2 j
	      done
	  done ;


	  (** Plongement et cration de la QMF... **)
	  let q = Core_qmf.extends_to_dim q dim_c in
	    Hash.iter
	      (fun ac d ->
		 let n = get_dim_from_acell ac in
		 let v = (try find c ac with Not_found -> default_value) in
		   Core_qmf.add q (Core_qmf.normal_form (g,d,n)) v
	      )
	      cpy ;
	    q
    )


;;

*)




(** --- un produit --------------- **)

let rec cartesian_prod combine ch1 ch2 =
  let sz = (Hash.size ch1) * (Hash.size ch2)
  in (* on cree un dictionnaire ou il y a une entree par simplex du produit *)
  let nch = Hash.create sz in
    begin
      Hash.iter (prd_complex_simplex nch combine ch2) ch1;
      update_faces nch;
      Hash.fold
        (fun (splx1, splx2) (splx, v) tab ->
	   Hash.replace tab (get_uid_from_acell splx) (splx, v); tab
	)
        nch
	(Hash.create sz)
    end

and prd_complex_simplex nch combine ch2 uid1 (splx1, v1) =
  Hash.iter (prd_simplex_simplex nch combine splx1 v1) ch2

and prd_simplex_simplex nch combine splx1 v1 uid2 (splx2, v2) =
  let nsimp = new_acell ((get_dim_from_acell splx1) + (get_dim_from_acell splx2)) [] []
  in Hash.replace nch (splx1, splx2) (nsimp, combine v1 v2)

and update_faces nch = Hash.iter (update_f nch) nch

and update_f nch (splx1, splx2) (splx, v) =
  List.iter (update_f2 nch splx splx2) (faces splx1);
  List.iter (update_f3 nch splx splx1) (faces splx2)

and update_f2 nch splx splx2 splx1f =
  try
    let (f,_) = Hash.find nch (splx1f, splx2)
    in splx.faces_acell <- f::splx.faces_acell;
       f.cofaces_acell <- splx::f.cofaces_acell
  with Not_found -> failwith "Error in chain cartesian-product\n"

and update_f3 nch splx splx1 splx2f =
  try
    let (f,_) = Hash.find nch (splx1, splx2f)
    in splx.faces_acell <- f::splx.faces_acell;
       f.cofaces_acell <- splx::f.cofaces_acell
  with Not_found -> failwith "Error in chain cartesian-product\n"
;;
(***
(** --- normalisation ------------- **)

(** On reconstruit un complexe en identifiant les simplexes qui ont les m%G�%@mes faces. 
 ** On propage par dimension croissante, afin d'eviter qu'une
 ** identification oblige a revenir sur les simplexes d%G�%@j%G�%@ calcul%G�%@s.
 ** Ca ressemble a du "zip" ou de la soudure.
 **)

let rebuild_cofaces ch =
  Hash.iter (fun _ (splx, _) -> splx.cofaces <- []) ch;
  Hash.iter (fun _ (splx, _) -> List.map (fun s -> s.cofaces <- splx::s.cofaces) splx.faces) ch;
  ch

and replace_face nsplx splx =
  List.iter
    (fun cosplx -> cosplx.faces <- List.map (fun s -> if s = splx then nsplx else s) cosplx.faces)
    splx.cofaces
;;

let normalise ch combine =
  let dim_ch = dimension ch in
  let skel = skelette ch
  and dico_subst = ((Hash.create (Hash.size ch)) : (int, int) Hash.t)
  and dico_faces = ((Hash.create (Hash.size ch)) : (int list, simplex) Hash.t)
  and nch  = new_chain() in
  let alias uid = try Hash.find dico_subst uid with Not_found -> uid in
  let normalized_faces splx =
    without_double(List.sort Pervasives.compare (List.map (fun s -> alias s.uid_splx) splx.faces))
  in
    begin
      (** on enregistre les points **)
      List.iter (fun (splx, v) -> add_force_one_simplex nch splx v) skel.(0);      for i = 1 to dim_ch do
        List.iter
          (fun (splx, v) ->
             let nfaces = normalized_faces splx in
               try
                 let norm_simp = Hash.find dico_faces nfaces
                 in
                   (* OK norm_simp me remplace: on met a jour les faces des cofaces *)
                   replace_face norm_simp splx;
                   (* update de la valeur du simplexe *)
                   update_chain nch combine v norm_simp;
                   (* finalement on se met dans le dico des substitutions *)
                   Hash.replace dico_subst splx.uid_splx norm_simp.uid_splx
               with Not_found ->
                 (Hash.replace dico_faces nfaces splx;
                  Hash.replace dico_subst splx.uid_splx splx.uid_splx;
                  update_chain nch combine v splx))
          skel.(i)
      done;
      (* reste a mettre %G�%@ jour les cofaces *)
      rebuild_cofaces nch
    end
;;

(** --- un quotient --------------- **)

let rec quotient ch combine subst =
  let nch = new_chain()
  and skel =  skelette ch
  and dico = ((Hash.create (Hash.size ch)) : (int, simplex) Hash.t) in
    begin
      for i = 0 to (dimension ch) do
        List.iter (fun x -> ignore(quotient_simplex ch combine subst nch dico x)) skel.(i)
      done;
      rebuild_cofaces nch
    end

and quotient_simplex ch combine subst nch dico (splx, v) =
  (** printf "-- je quotiente le simplexe: %d\n" splx.uid_splx; **)
  (* si le simplex a deja ete traites, i.e. il est present dans dico, on passe *)
  try Hash.find dico splx.uid_splx
  with Not_found ->
    begin
      (** printf "  -- simplexe %d pas encore traite\n" splx.uid_splx; **)
      (* on regarde si le splx doit etre identifie a un autre simplex *)
      if List.mem_assoc splx.uid_splx subst then
        begin (* oui, le simplex est quotiente *)
          (** printf "  -- simplexe %d doit etre quotiente\n" splx.uid_splx; **)
          let uid = List.assoc splx.uid_splx subst in
            if Hash.mem dico uid then
              let nsplx = Hash.find dico uid in  (* le simplex quotiente existe deja *)
                begin
                  (** printf "  -- simplexe %d doit etre quotiente avec %d qui est deja traite en %d\n" splx.uid_splx uid nsplx.uid_splx; **)
                  (* reste a mettre a jour la valeur dans la nouvelle chaine *)
                  update_chain nch combine v nsplx;
                  (* et %G�%@ introduire le splx dans le dico *)
                  Hash.add dico splx.uid_splx nsplx;
                  nsplx
                end
            else (* le simplex quotiente n'a pas encore ete traite *)
              begin
                (** printf "  -- simplexe %d doit etre quotiente avec %d qui est pas encore traite\n" splx.uid_splx uid; **)
                (* on le cree *)
                quotient_simplex ch combine subst nch dico (Hash.find ch uid);
                (* et maintenant on peut traiter le splx courant *)
                let nsplx = Hash.find dico uid in
                  begin
                    (* reste a mettre a jour la valeur dans la nouvelle chaine *)
                    update_chain nch combine v nsplx;
                    (* et %G�%@ introduire le splx dans le dico *)
                    Hash.add dico splx.uid_splx nsplx;
                    nsplx
                  end
              end
        end
      else (* le splx ne doit pas %G�%@tre identifi%G�%@ %G�%@ un autre : on le clone *)
        begin
          (** printf "  -- simplexe %d va etre clone\n" splx.uid_splx; **)
          let nsplx = clone_simplex ch combine subst nch dico splx
          in (add_force_one_simplex nch nsplx v;
              nsplx)
        end
    end

and clone_simplex ch combine subst nch dico splx =
  let ns = new_splx  
             splx.dim_splx
             (List.map 
                (fun s -> 
                   (** printf "    -- clonage simplexe %d traitement face %d\n" splx.uid_splx s.uid_splx; **)
                   quotient_simplex ch combine subst nch dico (Hash.find ch s.uid_splx))
                splx.faces)
             []
  in (Hash.add dico splx.uid_splx ns;
      ns)

;;
***)
