(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: kernel_collection.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-21 20:51:16 $
 * Version: $Revision: 1.14 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Type;;
open Dispatch_sig;;
open Environnement;;

open Dispatch;;

__register_version "$RCSfile: kernel_collection.ml,v $" "$Revision: 1.14 $" "$Date: 2006-04-21 20:51:16 $";;


print_string "(* loading kernel functions on COLLECTION into user-space ... ";
flush Pervasives.stdout;;


(* === CHAPITRES de la doc =============================================== *) 
(*!*)

(*** 
  @module intro 

  <b>A very short introduction to MGS.</b>

  For a more comprehensive documentation, a list of examples, and an
  exposition of the MGS motivations and background, please visit the
  URL [http://mgs.lami.univ-evry.fr]. This chapter is only a survival
  guide and does not cover the specificities of the MGS
  language. Refer to the chapter [Collections] and [Functional] for a
  presentation of the core concepts of MGS.

  MGS is a untyped functionnal language that enables the case-based
  definition of functions on <i>collections</i>. These functions are
  called <i>transformation</i>. The language includes also imperatives
  features like globale variables.

  A collection is an aggregate of values structured by topological
  relationships. See the chapter [Collections] in this documentation
  for more details about topological collections.  See the chapter
  [Functional] for more details about the notion of transformations.

  <b>The top-level <i>read-eval-print</i> loop:</b>
  The MGS interpreter reads its input, parses the stream of characters
  to recognize an expression, evaluates the expression and prints the
  result. To force the end of an expression, one may use the
  terminator [;;] (the double semi-column). In this manual pages, we
  often omit the terminator and the examples of expression must be
  completed by the [;;]. In addition, when using the interpreter on a
  standard unix terminal, the input are buffered (by unix) and a final
  carriage-return must be entered to complete the input.

  The top-level accepts expression and also <i>commands</i>. Commands
  are order directed to the interpreter and are not part of the MGS
  programming language. Commands always begin with a [!]. Type the
  command
  {[!help;;]}
  at the top level to have a list of the available commands.

  <b>Identifier:</b> 
  identifiers come in <i>simple</i> or <i>quoted</i> form. A quoted
  from is a string of characters that begins with a quote [']. Quoted
  identifiers are predefined variables. The value of a quoted
  identifier cannot be changed. A simple identifier is a string of
  characters that does not begin with a quote. Simple identifier can
  be used by the programmer as the name of a global variable, as the
  name of a function parameter or as the name of a variable introduced
  by a let. 

  The quote of a quoted name can often be omitted. Indeed, when a
  simple identifier is encountered, the evaluation process looks first
  if a value is attached to this identifier. If not, the evaluation
  process looks if the corresponding quoted name is defined. So
  expression [sin(0.77)] is interpreted as ['sin(0.77)] until the
  variable [sin] is defined. When this variable is defined, it is
  still possible to access the sin function using the (full) quoted
  name ['sin].

  <b>Global and let variables and their imperative assignation:</b>
  Some identifiers refer to an imperative variable. Such variables
  are global to the program or visible only in a scope introduced by a
  let construct. 

  The [:=] operator is used to change the value of a global variable:
  {[a := 3]}
  links the variable identified by [a] to the value [3] until the next
  affectation. The reference to a global variable that has not been 
  previously affected returns a special value called [<undef>]. The
  evaluation of the expression:
  {[ an_identifier_not_previously_used;; ]}
  returns [<undef: uninitialized var: an_identifier_not_previously_used >]. 
  In contrary to the usual handling of let variables in
  functional languages, the variables introduced by a let can be
  assigned in MGS. For example:
  {[let x = 1 in x:=x+2]}
  is a valid expression that returns [3]. The same let can be used to
  introduce several local variables:
  {[let x = 1 and y = 2 in x+y]}
  The scoping rules are the standard ones: the variable [x] introduced
  by the let is not usable in the expression defining the value of [y]
  and both [x] and [y] are unknown outside the [in] clause.


  <b>Functions:</b> 
  Function are first-class value that can be given as argument to
  another function or returned as the value of a function application.
  Functions are defined in a syntax similar to the
  usual lambda-calculus. For example, [\x.x] defines the identity
  function. The application of a function to its argument needs
  parenthesis:
  {[(\x.x)(3)]}
  denotes the application of the identity to the argument 3. 


  <i>All functions in MGS are curryfied</i>. However, there are some
  shortcuts to make the notation more lighter: [\x.\y.x(y)]
  can be simplified in
  {[\x,y.x(y)]}
  In the same spirit, the application of a function [f] to two (or
  more) arguments can be written
  {[f(1,2, ...)]}
  instead of the more heavier [f(2)(3)...]. Beware that the
  <i>comma</i> is also the name of an infix operator used to build
  sequence. For example [1, 2, 3] is an expression that builds a
  sequence of 3 elements. In function application, the comma is
  interpreted in favor of the application to multiple arguments. To
  force the application of a function to a sequence build with the
  comma operator, parenthesis can be used:
  {[f(1, 2, 3)]}
  denotes the application of function [f] to three arguments while:
  {[g((1, 2, 3))]}
  denotes the application of the function [g] to one argument which is
  an expression that build a sequence, and 
  {[h((1, 2, 3), 4)]}
  denotes the application of the function [h] to two arguments (the
  first is a list and the second is the integer [4]).

  To give a name to a function, we can use the <i>let</i> construct that
  binds locally an identifier to a value, or the <i>assignment</i>
  that changes the value linked to a global imperative variable. For
  example
  {[let id = \x.x in (id(id))(3) ]}
  returns [3] because [id(id)] is equivalent in the scope of the let,
  to the expression [(\x.x)(\x.x)] which evaluates to the identity
  function. Outside the [in] clause, the identifier [id] is unknown.

  The [:=] operator can used to give a global name to a function:
  {[id := \x.x]}
  links the identity function to the identifier [id] until the next
  assignment. There is a short-cut used to define a function and
  simultaneously to link it to a global identifier:
  {[fun id(x) = x]}
  is equivalent to
  {[id := \x.x]}
  The shortcut for the handling of multiple arguments is also
  provided:
  {[fun f(x, y, z) = x + y*z]}
  A function defined with the [fun] construct is called a <i>named
  function</i> by opposition with anonymous function defined using the
  lambda-notation. There is a special feature linked with the use of
  [fun] to define a function: such function can have optional arguments.
  
  Predefined functions, i.e. functions that are "magically" defined
  when one enters the MGS top-level, have all a quoted name that
  cannot be redefined. For example, [fun 'f(x) = x], ['f := ...] or
  [let 'f = ...] are incorrect expressions and are rejected (at the
  parsing phase).

  Some predefined functions have a special infix syntax, like the
  operator [+]: the syntax
  {[1+2]}
  is an alias for the expression
  {['add(1, 2)]}

  There is another way to define functions:
  <i>transformation</i>. Visit the [Functional] chapter for a
  description of transformations.

  <b>Recursion:</b>
  The name of a named function can be used in the definition of this
  function. This allows the direct expression of the recursion. For
  example:
  {[fun fact(x) = if x < 1 then 1 else x*fact(x-1) fi]}
  defines the standard factorial function.
  

  <b>Variable capture and imperative feature:</b>
  The partial application of the anonymous function [\x,y.x+y] to the
  argument [1] returns a function that can be used to compute the
  successor
  {[let succ = (\x,y.x+y)(1) in succ(2)]}
  returns [3]. It is customary to say that the argument [x] that refers
  to the value [1] is captured in the function [\y.x+y] and the result
  is a <i>cloture</i>, i.e. a function with an environment that binds some
  variables.

  In contrary to the usual handling of let variable in
  functional languages, in MGS the variable introduced by a let can be
  assigned. In conjonction with the capture of variables, this
  features enables intersting behaviors. For example:
  {[f := let x = 1 in \z.x:=x+z]}
  is a valid expression. The global variable [f] is linked to a
  cloture that embeeds the let variable [x]. Each time the
  function [f] is called, the store associated to [x] is updated. So
  [f(10)] returns the value [11] and a consecutive call [f(10)] returns
  [21].


  <b>Optional arguments of a named function:</b>
  not yet described

  <b>Iteration of the function application:</b>
  not yet described

  <b>Tracing a function:</b> the call to a named function [f] can be
  traced using the command [!trace f] or the expression [trace("f")]
  (this function takes a string as argument. The command [!untrace f] or
  the function [untrace("f")] can be used to cancel the trace of
  [f]. See the function the definition of the function [trace] in the
  [System] chapter. Only named function can be traced, however this
  includes predefined functions.

  <b>Switch statement:</b> the keyword [switch] can be used to avoid a
  lot of nested [if ... then ... else ... fi] constructions. The use
  of [switch] looks very similar as its equivalent construction in C:

  {[switch exp]}

  {[  case scl_1 : exp_1]}

  {[  ...]}

  {[  case scl_n : exp_n]}

  {[  default    : exp_d]}

  {[endswitch]}

  compares the evaluation of the expression [exp] with each scalar
  [scli] in the order given the user, and returns the evaluation of
  the associated expression [exp_i]. If no [scl_i] matches with [exp]
  the [default] expression [exp_d] is evaluated and returned. The
  [default] case is optional ; the returned default value is

  [<undef: macro: make_switch: switch default case>]

  The expression [exp] is evaluated only one time, at the beginning
  of the evaluation of the [switch] construction. In fact, a [switch]
  is a syntactic sugar for:

  {[let new_id = exp in]}

  {[  if      (new_id == scl_1) then exp_1]}

  {[  else if ...]}

  {[  else if (new_id == scl_n) then exp_n]}

  {[  else exp_d fi ... fi fi]}

  <b>Exception handling:</b> the structures [raise ...] and [try ...
  with ...] allows the use of exceptions in MGS. The syntax looks
  like the Ocaml one.

  To raise an exception, the keywords [raise] has to be used as
  following:

  [raise `symbol]

  The exception is identified by an MGS symbol (see the chapter
  [Scalar] for details), here [`symbol]. A sequence of arguments can
  be associated with these exceptions:

  [raise (`symbol arg_1, arg_2, ...)]

  where parentheses can be added around the arguments.

  These exceptions can be caught using a [try ... with ...] structure :

  {[try exp]}

  {[with]}

  {[  | `symbol1 (arg1_1, ...) -> exp_1]}

  {[  | `symbol2 (arg2_1, ...) -> exp_2]}

  {[  ...]}

  where the first pipe is optional. If the expression [exp] raises
  an exception, the exception value is
  matched against the list [`symbol1, `symbol2, ...]. If the raised
  exception corresponds to [`symbol_i], [exp_i] is evaluated where the
  free variables [arg_1, ...] are replaced by the arguments of the
  exception. Note the number of arguments has to matched for an
  exception to be caught. For example :

  {[try]}

  {[  raise (`S 1,2,3)]}

  {[with]}

  {[  | `S (x,y) -> x*y]}

  {[  | `S (x,y,z) -> x+y+z ;;]}

  returns [6].

  <b>The rest of the online-documentation:</b>
  the on-line documentation is divided in several chapters:
  {ul
  {- [Intro] is this short introduction to MGS.}
  {- [System] describes the system-related functions: input/output,
  process spawning, etc.}
  {- [Collections] introduces the notions of collections in MGS. In
  this chapter, only the overall structure of the collections types is
  described, as well as generic functions that can handle any
  collection types. Each specific type of collection are described in
  a dedicated chapter.}
  {- [Scalar] covers the scalar types and the related
  functions. Scalar type corresponds to indecomposable values. Example
  of scalar type are integer, boolean or function.}
  {- [Math] lists the arithmetic operators and the mathematical
  constants and functions}
  {- [Functional] introduces some functional constants and the
  specification of transformations, which are a special kind of
  functions that act on collections.}
  {- [Record] describes the record collection type.}
  {- [Monoidal] describes the set, multiset and sequence collection
  types. }
  {- [GBF] describes collections defined by a uniform
  neighborhood. Arrays are a special kind of GBF. }
  {- [Graph] describes the graph collection type.}
  {- [Delaunay] describes the delaunay collection type. }
  {- [Chain] describes the (forthcoming) abstract chain complex collection type. }
  }

*)

(*!*)

addenv_impure "'raise" (fun_2 __raiseFct) ;;
addenv_impure "'try" (fun_2 __tryFct) ;;

(*!*)


(* =============================================================== *)






(* === COLLECTION =================================================== *)



(*** @module collections     Generic functions acting on collections.

  A collection is an aggregate of values with a {i neighborhood
  relationship} between its elements. This determine a {b topology}
  and induces a notion of position, connected elements,
  path, boundaries, etc. This topology is then used in the
  definition of {i transformations}.
  
  Each collection type [t] as a companion type [p] used to denote the
  positions of the topology of [t]. A collection can be thought as an
  association of values to the positions of a topology. The expression
  [take(c, p)] can be used to get the value associated to a position [p]
  in the collection [c]. Positions can be dependent to a given
  collection (i.e. to a given value) or they can be dependent only to
  the collection type. See the function {i leibniz} and {i newton}.
  The following types are used to represent the positions of a
  collection:
  {ul
  {- the positions of a [record] are [string]. These strings are the
  name of the fields of the record.}
  {- the positions of a [monoidal] collection are integers ([seq]
  sequence, [set] set and [bag] multiset are monoidal collections).}
  {- the positions of a [GBF] are [posgbf]. These values represent the
  elements of the group underlying the GBF.}
  {- the position of a [grf] (a graph) are integers that label
  arbitrarily the vertices of the graph.}
  {- the positions of a [del] (Delaunay graph) are integers  that label
  arbitrarily the vertices of the graph.}
  }
  Given a collection [c], the function ['pospredicate_from_coll] returns a 
  predicate that checks if its argument has the type of a position for the
  collection [c]. The function ['predicate_from_coll] returns a predicate 
  that checks if its argument has the same type as [c]. 


  Any collection type [t] can be {i subtyped} into a collection type
  [t'] of exactly the same topology. The corresponding declaration
  begins with the keyword [collection]. This implies that the
  predicate ['collection] is denoted only by its quoted name
  ['collection].

  The declaration
  {[collection NAME = t;;]}
  where [t] is the name of a collection type, introduces a new collection
  type of name [NAME] and with the same topology as [t]. Sometimes
  [NAME] is refered as a [t] with a {i different color}. As for any
  other type, [NAME] is also a predicate to check if a value belongs
  to this type. A value of type [NAME] is also a value of type [t]
  (the subtyping relationship between [t] and [NAME]); however, a
  value of type [t] is not a value of type [NAME]. In other words
  [NAME(v)] implies [t(v)] but not the reverse.

  If [t] is a collection type, then [():t] or [t:()] denotes the empty 
  collection of type [t] and ['empty_from_coll(c)] returns an empty collection 
  with the same type as the collection [c].


  @voir 'leibniz  'leibniz
  @voir 'newton   'newton
  @voir 'monoidal 'monoidal
  @voir 'record   'record
  @voir 'set      'set
  @voir 'bag      'bag
  @voir 'seq      'seq
  @voir 'gbf      'gbf
  @voir 'graph    'graph
  @voir 'delaunay 'delaunay
  @voir 'chain    'chain
  @voir 'qmf      'qmf
  
*)



(** 
  @module collections
  @name 'leibniz
  @type any_bool
  Leibniz type and predicate.
  [leibniz] is the name of an {i abstract type}: it does not exist
  a value that have precisely this type. [leibniz] rather
  represents the kind of the collection types that admits free
  transformations. Like any other type, [leibniz] is also a predicate
  that can be used to check if a value has a leibnizian type.

  The collection types are divided into [leibnizian] and [newtonian]
  kinds. These names are after the philosophical positions advocated
  by Newton and Leibniz on the nature of [space].

  In a leibnizian space, the space is formed only by the presence of
  the objects in it: it is the positional relationships between
  material objects that are present together. In this view, the
  concept of the places or of the position of an empty space is
  inconceivable. 

  In a newtonian space, the space preexists to the objects and play the
  role of a scene where the objects can be placed. Though as a box,
  it is always possible to speak and designate the positions of an
  empty space. 

  Examples of leibnizian collection types are: [monoidal] collections
  (set, bag and seq) and [delaunay]. Leibnizian collections are always
  [total] data-structure, that is, each element in the collection has
  a well-defined value. Examples of newtonian types include [record]
  and [gbf]. Newtonian collection can be [partial] data-structure
  (e.g. arrays with hole). 

  Practically this two opposed views of the concept of space raise two
  different kind of spaces that exhibits different behaviors:
  {ul
  {- {i [<undef>] values in the collection}:
  A leibnizian collection cannot have [<undef>] elements. For
  instance, the expression {[1, <undef>, 2, <undef>, 3]} build the
  sequence [1, 2, 3]. In other word, leibnizian collections are total
  data-structure.

  On the other hand, a newtonian collection can be partial. For
  example, an infinite GBF of a type like [G = <a, b>] holds an
  infinite number of undefined values.

  The characterization {i leibniz = total} and {i newton = partial} is
  a rough approximation : it is possible to build in special
  circumstance a leibnizian collection including [<undef>]
  elements. Refer to the MGS tutorial for a full development.
  }
  {- {i Positions}: Each collection type [t] as a companion type [p]
  used to denote the positions of the topology of [t]. For example,
  a value of type [p] is used as an argument of the [take] function. 
  For a leibnizian type, a position has a value relevant for a given
  collection and cannot be used accross collections. For newtonian
  type, the positions preexist to any collection of this type and can
  be used across collections. 

  For example, integers are used to access the elements in a set (a
  leibnizian collection). However, accessing the 100th element of a
  set of 3 elements has no meaning. Even if there is enough elements,
  a position is related to a unique collection. For example, let [s1]
  and [s2] two sets that are equal: [s1 == s2], but built
  differently (they are the result of different computations). Then,
  we can have [take(s1, 1) <> take(s2, 1)] even if there is more than
  one element in these equal sets.
 
  In the other hand, let [G] be a GBF defined by [G = < a, b >] (GBF
  are newtonian collection). The positions of the GBF [G] are a new
  scalar type and [p == 2*|a> + 3*|b>] is a position related to the
  collection type [G]. Any collection of type [G] can be queryed for the
  element at position [p]. Eventually this value is [<undef>]. But for
  any GBF [g1] and [g2] such that [g1 == g2], we have [take(g1, p) ==
  take(g2, p)] for any position [p].
  }
  {- {i Deletion and erasure rules}: A rule of form
  {[... => <undef>]}
  in a transformation is interpreted as:
  {ul 
  {- a [deletion] rule if it is applied to a leibnizian collection:
  the elements matched by the left hand side of the rule are deleted
  in the result.
  }
  {- an [erasure] rule if it is applied to a newtonian collection:
  the values associated to the positions matched by the left hand side
  of the rule are replaced by [<undef>].
  }
  }
  For example, the rule [3 => <undef>] applied to the sequence [1, 2,
  3, 4, 3, 5] returns the shorter sequence [1, 2, 4, 5]. Applied to a
  GBF, this rule replaces each value [3] by [<undef>].
  }
  }
  
  @voir 'newton     'newton
  @voir 'collection 'collection
*)
(** 
  @module collections
  @name 'newton
  @type any_bool
  Newton type and predicate.
  [newton] is the name of an {i abstract type}: it does not exist
  a value that have precisely this type. [newton] rather
  represents the kind of the collection types that represent a
  space independent of the values it holds. Examples of newtonian types
  include [record] and [gbf]. Newtonian collection can be [partial]
  data-structure (e.g. arrays with hole). See [leibniz] for a more
  detailed explanation.

  @voir 'leibniz    'leibniz
  @voir 'collection 'collection
*)


(** 
  @module collections
  @name 'collection
  @type any_bool
  Collection type and predicate.
  [collection] is the name of an {i abstract type}: it does not exist
  a value that have precisely this type. [collection] rather
  represents the type of any collection type. Like any other type,
  [collection] is also a predicate. This predicate can be used to
  distinguish between atomic (or scalar) values (like integers, floats,
  booleans, strings...) and collection values (like records,
  sequences, sets, gbfs, graphs, etc.)


  @voir 'leibniz  'leibniz
  @voir 'newton   'newton
  @voir 'monoidal 'monoidal
  @voir 'record   'record
  @voir 'set      'set
  @voir 'bag      'bag
  @voir 'seq      'seq
  @voir 'gbf      'gbf
  @voir 'graph    'graph
  @voir 'delaunay 'delaunay
  @voir 'chain    'chain
  @voir 'qmf      'qmf
*)


(** On n'ajoute pas ce predicat  la main : 
 ** il sera ajoute automatiquement via la table des descripteurs de types 
 **)
Hash.iter (fun name desc -> addenv_pure name desc.predicate) type_tbl;;





(** 
  @module collections
  @name 'collection_name
  @type collection_string
  Return the type of its argument as a string. The argument must be
  a collection.
*)
(addenv_impure "'collection_name" (fun_1 collection_name));;

(** 
  @module collections
  @name 'pospredicate_from_coll
  @type collection_funct
  Return the type predicate of the positions of a collection.
  The argument is a collection and this function return the
  predicate that answer true if its argument represent a position in
  the collection type. For example, for ['seq] the predicate returned
  is ['int].
*)
(addenv_impure "'pospredicate_from_coll" (fun_1 pospredicate_from_coll));;

(** 
  @module collections
  @name 'empty_from_coll
  @type collection_collection
  Return an empty collection of the same type has its argument. The argument must be
  a collection.
*)
(addenv_impure "'empty_from_coll"  (fun_1 empty_from_coll));;

(** 
  @module collections
  @name 'predicate_from_coll
  @type collection_funct
  Return the type predicate of the type of its argument. The argument must be
  a collection.
*)
(addenv_impure "'predicate_from_coll"  (fun_1 predicate_from_coll));;

(**
  @module collections
  @name 'set_color
  @type collection_collection_collection
  Return its first argument with its color changed for the second one or raise
  an error if it is not possible.
*)

addenv_impure "'set_color" (fun_2 __setColor);;

(** 
  @module collections 
  @name 'size
  @type any_int
    Returns the number of elements in a collection (i.e. a positive or
    null integer). If the argument is not a collection, then it is a
    scalar value and the integer returned is strictly negative with
    the following convention:
    {ul
  	{- [-1] for an [undef]  }
  	{- [-2] for an [int]    }
  	{- [-3] for a [bool]    }
  	{- [-4] for a [float]   }
  	{- [-5] for a [string]  }
  	{- [-6] for a [funct]   }
  	{- [-7] for a [gbfpos]  }
  	{- [-8] for a [splx]    }
  	{- [-9] for a [gmap]    }
  	{- [-10] for an [ncell] }
  	{- [-11] for an [bint]  }
    }
*)
addenv_pure "'size" (fun_1 __size);;


(** 
  @type collection_bool
  Returns [true] if the argument is an empty collection and [false]
  if the argument is a non-empty collection. In other words,
  [empty(c) == (size(c) > 0)] if [c] is a collection but it is sometimes 
  more efficient to call [empty] rather than [size].
  If the argument is not a collection, then an [undef] value 
  is returned.
*)
addenv_pure "'empty" (fun_1 __empty);;


(** 
  @module collections
  @name 'extension
  @type string_seq_collection

  Build a collection of a specified type from the listing of its
  elements and positions.

  
  ['extension("type", (P1,V1)::(P2,V2):: ... ::(Pn,Vn)::():seq)]
  build a collection of type [type] with element [Vj] at position
  [Pj].
  The first argument is a string that represents the name of the type
  of the collection to be build. The second argument is a sequence of
  couples (a couple is just a sequence of two elements). Such a
  couple [(p,v)] specifies that the value [v] is at position [p] in
  the collection to be build.

  This function has a more convenient infix notation :
  {[ type:{| P1 = V1, ..., Pn = Vn |}]}
  or
  {[ {| P1 = V1, ..., Pn = Vn |}:type ]}
  is equivalent to
  ['extension("type", (P1,V1)::(P2,V2):: ... ::(Pn,Vn)::():seq)]

*)
addenv_impure "'extension" (fun_2 __extension);;




(* --- Acces ------------------------------------------- *)

(**
  @module collections
  @name 'hd
  @type collection_any

  Pick one element in a collection.
  The result is  [<undef>] if the collection is empty.
  For a sequence [l], the expression [hd(l)] returns the head of 
  the sequence. For a monoidal collection [c], it is ensured that [c
  == hd(c),tl(c)]. For non-monoidal collection types, an element is 
  returned without assuming any special property except a relationship
  with the [tl] function: [hd(c)] is element element removed
  from [c] in [tl(c)]. Note that it cannot be assumed that 
  the element returned is randomly chosen in the collection.

  @voir 'tl tl
  @voir 'take take
*)
addenv_impure "'hd"   (fun_1 __hd);;

(**  
  @module collections
  @name 'tl
  @type collection_collection

  Return the collection minus one element.
  The result is  [<undef>] if the collection is empty.
  For a sequence [l], the expression [tl(l)] returns the tail of 
  the sequence. For a monoidal collection [c], it is ensured that [c
  == hd(c),tl(c)]. For non-monoidal collection types, an element is 
  returned without assuming any special property except a relationship
  with the [hd] function: [hd(c)] is the element removed from [c] in 
  [tl(c)]. Note that it cannot be assumed that the supressed element is
  randomly chosen in the collection.

  @voir 'tl tl
  @voir 'take take
*)
addenv_impure "'tl"   (fun_1 __tl);;

(**
  @module collections
  @name 'last
  @type collection_any

  Return the last element in a sequence. For other collection, an
  element is returned.

  @voir 'hd hd
*)
addenv_impure "'last" (fun_1 __last);;


(**
  @module collections
  @name 'take
  @type seq_collection_any
  @type collection_int_any
  @type collection_any_any

  [take(c, p)] returns the element at position [p] in [c].
  See the entry [collections] for the representation of a position 
  for a given collection type. If there is no element at
  the given position, the value [<undef>] is returned.

  In any case, [p] can be an integer. In this case, 
  {[take(c, n) == hd(tl\['iter=n\](c))]}
  For sequences, a position is an integer but there is no ambiguity 
  since the previous relation holds.

  If [p] is a collection, then it is assumed that each element in this collection
  is an integer. Then [take(s, p)] returns a collection of the same kind as [p]
  where an element [n] has been replaced by [take(s, n)]. Note that a position is a scalar
  value, and not a collection; so there is no ambiguity with the previous meaning.

  [e.(p)] is an infix syntax for [take(e, p)]

  @example
  [(4, 5, 6, 7).(1, 2, 3, set:())] returns [(5, 6, 7):'set]

  @voir 'tl tl
  @voir 'hd hd
*)
addenv_impure "'take" (fun_2 __take);;



(**
  @module collections
  @name 'sort
  @type funct_collection_seq

  Sort a collection in increasing order according to a comparison
  function.  The comparison function must return 0 if it arguments
  compare as equal, a positive integer if the first is greater,
  and a negative integer if the first is smaller.

  @example
  {['sort((\x,y.if x> y then -1 else if x == y then 0 else 1 fi fi), (1, 2, 3, 4))]}
  returns
  {[(4, 3, 2, 1):'seq]}
*)
addenv_impure "'sort" (fun_2 __sort);;



(* --- Ordre superieur ------------------------------------------- *)

(** 
  @module collections
  @name 'map
  @type funct_collection_collection
  Map a function over the elements of a collection.

  [map] is a generic operator that acts over any collection type. 
  The results is a collection of the same type has the argument. 

  @example
  {[map((\x.x+1), (0, 1, 2))]}
  returns the sequence [1, 2, 3].

  @voir 'iter iter
  @voir 'fold fold
  @voir 'map_indexed map_indexed
*)
addenv_impure "'map"          (fun_2 __map);;


(** 
  @module collections
  @name 'fold
  @type funct_any_collection
  Fold a function over the elements of a collection.

  [fold] is a generic operator that acts over any collection type. 
  If [c1, c2, ..., cn] are the elements of a collection [c], then [fold(f,
  zero, c)] computes [f(c1, f(c2, ... f(cn, zero)))]. If the
  collection has no element, then [zero] is returned. The computation order of 
  the elements in the fold is not specified and may vary even within the
  same collection type and size. The accumulating function [f] takes 2
  arguments: the first is the current element and the second is the
  current result.

  [fold] also accept an integer instead of a collection:
  [fold(f, zero, n)] computes [f(n, f(n-1, ... f(1, zero)))].

  @example
  {[fold((\x,y.1+y), 0, c)]} 
  returns the number of elements in the collection [c].
  {[fold((\x,y.x+y), 0, 133)]} 
  computes the sum of the integers up to [133].


  @voir 'iter iter
  @voir 'map map
  @voir 'fold_indexed fold_indexed
*)
addenv_impure "'fold"         (fun_3 __fold);;


(**
  @module collections
  @name 'iter
  @type funct_collection_undef
  Iter a function over the elements of a collection.

  [iter] is a generic operator that acts over any collection type. 
  The results is undef. It is used for the side-effect of the function
  applied to the collection's elements. 

  @example
  {[a :=0; iter((\x.a:=a+x), (1, 2, 3))]}
  returns the value [<undef>] but the value of the imperative variable
  [a] is now [6].

  @voir 'map map
  @voir 'fold fold
  @voir 'iter_indexed iter_indexed
*)
addenv_impure "'iter"         (fun_2 __iter);;


(**
  @module collections
  @name 'iter_indexed
  @type funct_collection_undef
  Iter a function over the elements of a collection with their position.

  [iter_indexed(f, c)] is an expression that iterates the function
  [f] over the elements of the collection [c]. The function [f] takes
  2 arguments: the first one is a position [p] and the second is the value
  at position [p] in the collection [c]. 

  [iter_indexed] is a generic operator that acts over any collection type. 
  The results is undef. It is used for the side-effect of the function
  applied to the collection's elements. For the representation of a
  position of a collection, see the function {i collection} and also
  the function {i leibniz} and {i newton}. 
  

  @voir 'iter iter
  @voir 'map_indexed map_indexed
  @voir 'fold_indexed fold_indexed
*)
addenv_impure "'iter_indexed" (fun_2 __iterindexed);;




(**
  @module collections
  @name 'map_indexed
  @type funct_collection_collection
  Map a function over the elements of a collection with their position.

  [map_indexed(f, c)] is an expression that maps the function
  [f] over the elements of the collection [c]. The function [f] takes
  2 arguments: the first one is a position [p] and the second is the value
  at position [p] in the collection [c]. 

  [map_indexed] is a generic operator that acts over any collection type. 
  The results is the collection of the results returned by the
  application of [f]. For the representation of a
  position of a collection, see the function {i collection} and also
  the function {i leibniz} and {i newton}. 
  
  @example
  [map_indexed((\pos, value.pos), (4, 5, 6, 7))] returns the sequence
  of the positions of the elements of the list [4, 5, 6, 7]. Thus, the
  returned value is the sequence [0, 1, 2, 3].

  @voir 'iter_indexed iter_indexed
  @voir 'fold_indexed fold_indexed
  @voir 'map map
*)
addenv_impure "'map_indexed"  (fun_2 __mapindexed);;



(**
  @module collections
  @name 'fold_indexed
  @type funct_any_collection_any
  Fold a function over the elements of a collection with their position.

  [fold_indexed(f, zero, coll)] is an expression that fold the function
  [f] over the elements of the collection [coll]. The function [f] takes
  3 arguments: the first one is a position [p], the second is the value
  at position [p] in the collection [coll] and the third is the
  accumulator representing the result of the previous applications of
  [f].  The argument [zero] is the initial value of the
  accumulator. It is also the value returned if the collection [coll]
  is empty.

  [fold_indexed] is a generic operator that acts over any collection type. 
  For the representation of a position of a collection, see the
  function {i collection} and also the function {i leibniz} and {i newton}.
  
  @voir 'fold fold
  @voir 'iter_indexed iter_indexed
  @voir 'map_indexed map_indexed
*)
addenv_impure "'fold_indexed" (fun_3 __foldindexed);;



(**
  @module collections
  @name 'forall
  @type funct_collection_bool

  Universal predicate.
  ['forall(p, c)] checks if all elements of the collection [c]
  satisfy the predicate [p]. That is, it returns
  [p(c1) && p(c2) && ... && p(cn)] where [c]{_ n} are the elements of [c].
  If the collection is empty, [true] is returned.
*)
addenv_impure "'forall" (fun_2 __forall);;

(**
  @module collections
  @name 'exists
  @type funct_collection_bool

  Existential predicate. 
  ['exists(p, c)] checks if at least one element of
  the collection satisfies the predicate [p]. That is, it returns
  [p(c1) || p(c2) || ... || p(cn)] where [c]{_ n} are the elements of [c].
  If the collection is empty, [false] is returned.
*)
addenv_impure "'exists" (fun_2 __exists);;



(**
  @module collections
  @name 'count
  @type any_collection_int

  Counts the number of occurences of an element in a collection.

  @example 
  ['count(1, (2, 3, 1, 1, 5, 7, 1, 8))] returns [3].
*)
addenv_impure "'count"   (fun_2 __count);;



(**
  @module collections
  @name 'member
  @type any_collection_bool

  Membership predicate.

  ['member(e, c)] returns [true] if the element [e] appears in the
  collection [c] and [false] elsewhere.
*)
addenv_impure "'member"  (fun_2 __member);;



(**
  @module collections
  @name 'shape
  @type collection_funct

  Rebuild a collection from a sequence.

  
  ['shape(c)] (where [c] is a collection) returns a function [f] that
  expect just one argument. This function can be used to reconstruct a
  collection with the same {i shape} as the collection [c] but
  with other elements. The argument of the function [f] is a sequence
  of elements that replaces the sequence of elements of [c]. It is
  ensured that 
  {['shape(c)(sequify(c)) == c]}

  @example 
  First we build a GBF:
  {[gbf GG = <X, Y>;;]}
  {[array := iota(4, seq:()) following |X>+|Y>;;]}
  This GBF is an array made of a diagonal. The printing of this array
  gives:
  {[   0  ---  ---  ---]}
  {[ ---    1  ---  ---]}  
  {[ ---  ---    2  ---]}
  {[ ---  ---  ---    3]}
  The elements in this array are enumerated in some order by the
  [sequify] function: [sequify(array)] returns
  {[1, 2, 3, 0]}
  Now, we can use the [shape] primitive to replace all the elements by
  some others. The following expression permutes the elements using
  the [sort] function:
  {[array2 := 'shape(array)(sort(compare, sequify(array)))]}
  The printing of [array2] gives:
  {[  3  ---  ---  ---]}
  {[---    0  ---  ---]}
  {[---  ---    1  ---]}
  {[---  ---  ---    2]}
  and the sequification of [array2] corresponds to the sequence
  {[0, 1, 2, 3]}
*)
addenv_impure "'shape"   (fun_1 __shape);;





(**
  @module collections
  @name 'neighbors
  @type collection_any_seq

  The neighbors of an element in a collection.

  ['neighbors(c, p)] returns, in a sequence, the neighbors of the
  element at position [p] in the collection [c].
  Cf. the entry ['collection] for the representation of the positions
  associated to a collection type.

  @voir 'collection 'collection
  @voir 'leibniz    'leibniz
  @voir 'monoidal   'monoidal
*)
addenv_impure "'neighbors"    (fun_2 __neighbors);;

(**
  @module collections
  @name 'neighborspos
  @type collection_any_seq

  The neighbors position of an element in a collection.

  ['neighborspos(c, p)] returns, in a sequence, the neighbors position of the
  element at position [p] in the collection [c].
  Cf. the entry ['collection] for the representation of the positions
  associated to a collection type.

  @voir 'collection 'collection
  @voir 'leibniz    'leibniz
  @voir 'monoidal   'monoidal
*)


(* FIXME addenv_impure "'neighborspos" (fun_2 __neighborspos);; *)



(**
  @module collections
  @name 'neighbor
  @type chain_splx_int
  @type chain_splx_bool
  @type gbf_posgbf_any
  @type gbf_seq_seq

  Specific neighbors in a GBF or a chain.

  {i If} the first argument is a GBF, then ['neighbor(c, p, d)] returns
  the value of the neighbor following direction [d] of the element at
  position [p] in the GBF [c]. Both [p] and [d] are [posgbf].

  ['neighbor(c, p, l)] where [l] is a sequence of directions, returns
  the neighbors following the directions in the list [l] of the
  element at position [p] in the GBF [c].

  {i If} the first argument is a chain,  then ['neighbor(c, s, b)]
  returns the value of the simplices that are the faces (if the
  boolean [b] is false) or the cofaces (if [b] is true) of
  [s] in the chain [c]. Similarly,  ['neighbor(c, s, n)] where [n] is
  an integer, computes  ['neighbor(c, s, (n > 0))].
  

  @voir 'GBF chapter GBF
  @voir 'Chain chapter Chain
*)
(* FIXME addenv_impure "'neighbor"     (fun_3 __neighbor);; *)




(* =============================================================== *)




(* === MONOIDAL =================================================== *)

(*** @module monoidal  The {i monoidal} collection: set, bag and sequence. 
*)


(**   
  @module monoidal
  @name 'iota
  @type int_monoidal_monoidal
  
  Enumerate the [n] first integers.
  Given an integer [n] and a monoidal collection [m], the expression
  [iota(n+1, m)] returns the value [0, 1, 2, ..., n, m].

  @example [iota(3, (2, 3, 4))] returns the sequence 
  [0, 1, 2, 2, 3, 4] and [iota(3, (2, 3, 4, set:()))] returns the set 
  [0, 1, 2, 3, 4, set:()].
*)
addenv_impure "'iota"         (fun_2 __iota);;




(**
  @module monoidal
  @name 'filter
  @type funct_monoidal_monoidal

  ['filter(p, c)] returns all the elements of the monoidal collection [c]
  that satisfy the predicate [p].  If [c] is a sequence, the order of the elements
  in the input sequence is preserved. 
*)
addenv_impure "'filter" (fun_2 __filter);;



(**
  @module monoidal
  @name 'flatten
  @type monoidal_monoidal

  [flatten(col)] flattens a sequence of sequences, or a set of sets, or a bag of bags. 
  The flattening occurs only at the first level. The flattening of the [n] first nested levels
  of a monoid can be achieved by iteration: [flatten['iter=n](col)]
  By definition, [flatten(x) == x] if [x] is not a monoidal 
  collection.

  @example
  {[flatten(1::(2::22::222::(5, 55, 555, 5555)::():seq)::3::4::():seq)]}
  {[== (1, 2, 22, 222, (5, 55, 555, 5555):'seq, 3, 4):'seq]}
  and
  {[flatten['iter=2](1::(2::22::222::(5, 55, 555, 5555)::():seq)::3::4::():seq)]}
  {[(== 1, 2, 22, 222, 5, 55, 555, 5555, 3, 4):'seq]}
*)
addenv_impure "'flatten" (fun_1 __flatten);;




(**
  @module monoidal
  @name 'partition
  @type funct_monoidal_monoidal
  @type int_monoidal_monoidal

  [partition(n, s)], with [n] an integer, partitions a monoid [s] (i.e. a seq, a set or a bag) 
  into nonoverlapping submonoids of length [n] (except for one). 

  [partition(fct, s)], with [fct] a predicate, partitions a monoid [s] (i.e. a seq, a set or a bag) 
  into two nonoverlapping submonoids [s1] and [s2], where
  [s1] is the monoid of all the elements of [s] that satisfy the
  predicate [fct], and [s2] is the set of all the elements of
  [s] that do not satisfy [fct].

  @example
  {[partition(3, (1, 2, 3, 4, 5, 6, 7)) == ((1, 2, 3):'seq, (4, 5, 6):'seq, (7):'seq):'seq]}
  {[partition((\x.x<3), (1, 2, 3, 4, 5, 6, 7, ():set)) == (((1, 2):'set, (3, 4, 5, 6, 7):'set):'set]}

  @voir 'filter 'filter
  @voir 'partition_eq 'partition_eq
*)
addenv_impure "'partition" (fun_2 __partition);;




(**
  @module monoidal
  @name 'partition_eq
  @type funct_monoidal_monoidal

  [partition_eq(f, s)] partitions a sequence [s] into subsequences of elements of 
  the same equivalence class. Two elements [x] and [y] are of the same equivalence 
  class if [f(x, y)] is true.

  @example
  {[partition_eq((\x,y.true), (1, 2, 3, 4, 5)) == (5, 4, 3, 2, 1):'seq):'seq]}
  {[partition_eq((\x,y.false), (1, 2, 3, 4, 5, bag:())) == ((1):'bag, (2):'bag, (3):'bag, (4):'bag, (5):'bag):'bag]}
  {[partition_eq((\x,y.x%2==y%2), (1, 2, 3, 4, 5, set:())) == (((2, 4):'set, (1, 3, 5):'set):'set]}
*)
addenv_impure "'partition_eq" (fun_2 __partitioneq);;








(* --- Monoidal ---------------------------------------------------- *)


(**
  @module monoidal
  @name 'cons
  @type any_set_set
  @type any_bag_bag
  @type any_seq_seq

  Add an element to a monoidal collection.
*)
addenv_impure "'cons"   (fun_2 __cons);;

(**
  @module monoidal
  @name '::
  @type any_set_set
  @type any_bag_bag
  @type any_seq_seq

  Add an element to a monoidal collection.
*)
addenv_impure "'::"     (fun_2 __cons);;


(**
  @module monoidal
  @name 'append
  @type set_set_set
  @type bag_bag_bag
  @type seq_seq_seq

  Concatenates two monoidal collections.
*)
addenv_impure "'append" (fun_2 __append);;


(**
  @module monoidal
  @name 'join
  @type any_any_seq
  @type set_set_set
  @type bag_bag_bag
  @type seq_seq_seq
  @type any_set_set
  @type any_bag_bag
  @type any_seq_seq

  Join two values. If the argument are of the same monoidal collection
  types, the the result has the same collection type and corresponds
  to an [append] operation. 

  If one of the argument is a monoidal collection type [t] and the
  other argument is not a monoidal collection, the result is the same
  as a [cons] operation and has the type [t].

  For the other cases, the result is the sequence of the two arguments.
*)
addenv_impure "'join"   (fun_2 __join);;



(**
  @module monoidal
  @name 'inter
  @type set_set_set
  @type bag_bag_bag

  Set or bag intersection.
*)
addenv_impure "'inter"  (fun_2 __inter);;

(**
  @module monoidal
  @name 'diff
  @type set_set_set
  @type bag_bag_bag

  Set or bag difference. 
  This is not the symetric difference: [diff(A, B)] contains the
  elements of [A] that does not appear in [B]. For bag difference, the
  multiplicities in [B] are not taken into account.

  @example 
  [diff((1, 2, 3, set:()), (2, 3, 4, 5, set:()))] returns [(1):set]
  and 
  [diff((1, 1, 2, 2, 3, bag:()), (2, 3, 4, 5, bag:()))] returns 
  [(1, 1):bag].
*)
addenv_impure "'diff"   (fun_2 __diff);;

(**
  @module monoidal
  @name 'subset
  @type set_set_bool

  [subset(A, B)] returns true if [A] is included in [B]
*)
addenv_impure "'subset" (fun_2 __subset);;





(*!*)

(* --- fin de l'installation des fonctions de base --- *)

print_string "done *)";
print_newline();;

(* ======================================================= *)
