(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: printer.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-09-23 15:02:22 $
 * Version: $Revision: 1.5 $
 * -------------------------------------------------------------------
 *
 * Les fonctions d'impressions.
 *
 * Les fonctions d'impression disponibles:
 *
 *    prbxxx b v : print dans le buffer b de v: c'est la fonction de base a ecrire
 *    prfxxx f v : print dans le fichier f (f est un out-channel)
 *    prxxx  v   : print dans la sortie standard
 *    prsxxx v   : print dans une string qui est retournee
 *
 * xxx = val   : pour les mgs_value
 *       value : pour les mgs_value  (les 2 formes sont permises)
 *       term  : pour les termes
 *       code  : pour les byte_codes
 *
 * Si xxx = val ou value, 
 * on peut rajouter "unformatted_" ou juste "u" devant, ce qui  donne une impression non-formatte 
 * des valeurs. 
 * Exemple: uprsval ou bien unformatted_prsval ou bien uprsvalue ou bien unformatted_prsvalue
 * (ces 4 formes dsignent la mme fonction).
 *
 * On peut aussi rajouter un 2 derrire pour obtenir la fonction qui fait la mme 
 * chose mais qui retourne unit: prfval2.
 * Attention, il n'y a pas de prsval2 ou de prsvalue2, car pour cette fonction, il n'y a pas
 * d'effet de bord: une string est retourne.
 *
 * Pour rendre disponible les fonction prcdentes, il faut implmenter par type Xxx:
 *
 *  -  _uprbval_'outBuffer_xxx : impression non-formatte de xxx
 *
 *  -  _pfbval_'outBufferF_xxx : impression formatte de xxx
 *
 *=========================================================================
 *)

open Buffer;;
open Printf;;
open Format;;

open Global;;
open Valeur;;

open Dispatch_sig;;
open Environnement;;

__register_version "$RCSfile: printer.ml,v $" "$Revision: 1.5 $" "$Date: 2004-09-23 15:02:22 $";;


let valundef = Valeur.mgsundef
;;



(*** =======================================================
 *** === Impressions non formatts =========================
 *** ======================================================= 
 ***)



(** uprval b x : est une fonction qui imprime une valeur mgs x dans son buffer b.
 ** Cette fonction est la fonction principale d'affichage. L'impression dans une 
 ** string ou bien sur la sortie standard ou bien dans un fichier derivent de 
 ** cette fonction.
 ** Cette fonction est dispatche et elle est  implmenter dans 
 ** les fichiers Xxx/printer_xxx.ml respectifs. 
 **)
let uprbval = Dispatch_sig.uprbval
;;


(** le buffer extensible d'impression **)
let out_buffer = Buffer.create 2048
;;

let uprbval2 b x = ignore(uprbval b x)  (* version qui retourne unit *)

(** uprvalue: affichage sur la sortie standard.
 ** On cree un buffer, on imprime dedans puis on sort le buffer sur la sortie standard.
 **)

and uprvalue x = 
  clear out_buffer;
  ignore(uprbval out_buffer x); 
  output_buffer stdout out_buffer; 
  x

(** uprfvalue out x: sortie de x sur le fichier out **)
and uprfvalue out x = 
  clear out_buffer;
  ignore(uprbval out_buffer x); 
  output_buffer out out_buffer; 
  x

(** uprsvalue x : retourne une string qui represente la valeur x **)
and uprsvalue x =
  clear out_buffer;
  ignore(uprbval out_buffer x); 
  VAL_string(contents out_buffer)
;;

(** idem que uprfvalue, mais renvoie unit plutot que l'argument **)
let uprfvalue2 out x = ignore(uprfvalue out x)
and uprvalue2  x     = ignore(uprvalue x)
;;

(** mise a jour d'un wrapper forward de valeur.ml 
 ** Cette fct forward sert dans les traces. Je suppose que les impressions 
 ** non-formattes sont plus robustes aussi c'est celle la qu'on choisit.
 **)
forward_prsvalue := (function x -> to_string(uprsvalue x))
;;


(** alias **)

let uprbvalue  = uprbval
and uprbvalue2 = uprbval2
and unformatted_prbvalue  = uprbval
and unformatted_prbvalue2 = uprbval2
and unformatted_prfvalue  = uprfvalue
and unformatted_prfvalue2 = uprfvalue2
and unformatted_prsvalue  = uprsvalue
;;

let unformatted_prbval  = uprbval
and unformatted_prbval2 = uprbval2
and unformatted_prfval  = uprfvalue
and unformatted_prfval2 = uprfvalue2
and unformatted_prsval  = uprsvalue
and uprfval = uprfvalue
and uprfval2 = uprfvalue2
and uprsval = uprsvalue
;;



(*** =======================================================
 *** === Impressions FORMATTS =============================
 *** ======================================================= 
 ***)




(** --- Affichage generique dans un buffer ---- *)

(** _pfbval_'outBufferF_XXX (b, x) prend un buffer format [b] et une valeur [x] de type XXX. 
 ** Elle crit de faon formate [x] dans le buffer format [b], puis renvoie valundef
 **)

let pfbval = Dispatch_sig.pfbval  (** Renomer en pfbval, mais c'est en attendant de faire un gendispatch **)
;;


(** pfbvalue b x : 
 ** Cette fonction prend un buffer "normal" b et une valeur mgs. 
 ** Elle crit de faon formate x dans le buffer b. Puis, elle renvoie valundef
 **)
let pfbvalue = pfbval

(** pfbvalue2 b x : idem pfbvalue mais renvoie unit.
 **)
and pfbvalue2 b x = ignore(pfbval b x)
;;



(** pfvalue: affichage sur la sortie standard.
 ** On cree un buffer, on imprime dedans puis on sort le buffer sur la sortie standard.
 **)

(** le buffer extensible de formattage **)
let out_bufferF = Buffer.create 2048
;;
let bufferF = Format.formatter_of_buffer(out_bufferF)
;;

let pfvalue x =
  pp_print_flush bufferF ();
  clear out_bufferF;
  ignore(pfbvalue bufferF x);
  Format.pp_print_flush bufferF ();
  output_buffer stdout out_bufferF;
  x

(** prfvalue out x: sortie de x sur le fichier out **)
and pffvalue out x =
  pp_print_flush bufferF ();
  clear out_bufferF;
  ignore(pfbvalue bufferF x);
  Format.pp_print_flush bufferF ();
  output_buffer out out_bufferF;
  x

(** prsvalue x : retourne une string qui represente la valeur x **)
and pfsvalue x = 
  pp_print_flush bufferF ();
  clear out_bufferF;
  ignore(pfbvalue bufferF x);
  Format.pp_print_flush bufferF ();
  VAL_string(contents out_bufferF)
;;

(** idem que prfvalue, mais renvoie unit plutot que l'argument **)
let pffvalue2 out x = ignore(pffvalue out x)
and pfvalue2  x     = ignore(pfvalue x)
;;




(** alias **)

let prbvalue   = pfbvalue
and prfvalue   = pffvalue
and prvalue    = pfvalue 
and prsvalue   = pfsvalue 
and prbvalue2  = pfbvalue2
and prfvalue2  = pffvalue2
and prvalue2   = pfvalue2
;;

let prbval   = pfbvalue
and prfval   = pffvalue
and prval    = pfvalue 
and prsval   = pfsvalue 
and prbval2  = pfbvalue2
and prfval2  = pffvalue2
and prval2   = pfvalue2
;;



(** === Extraction de l'argument de VAL_funct avec affichage de l'erreur ========================== **)

let prsvalue2 _ a = to_string(prsvalue a)
;;

let erreur_pf msg x = 
  let errmsg = 	Printf.sprintf "Error: %s: expect a function, got\n\t%a\n" msg prsvalue2 x in
    begin
      if false == !Global.silent_error then 
	begin
	  Pervasives.flush_all();
	  Printf.fprintf stderr "%s" errmsg;
	end
    end;
    if !Global.abort_on_error then raise (Erreur errmsg)
    else (pf1 !current_error_handler) (VAL_string errmsg)
;;

let pf message  = function
  | VAL_funct f -> f  (** print_string("#trace : "^message^"\n"); f **)
  | x -> 
      begin
	match erreur_pf message x with 
	    (* s'il y a un retour, c'est que les erreurs sont grs
	     * par le handler d'erreur et il faut vrifiez que
	     * celui-ci retourne bien une fonction, sinon on ne peut
	     * pas continuer le traitement. Dans ce dernier cas, on
	     * aborte l'valuation. 
	     *) 
	  | VAL_funct f -> f
	  | _ -> raise (Erreur (message^"\n(irrecoverable error: the error handler does not return a function)"))
      end
;;

(* Redefinition de pf1 (le pf1 initial est dans valeur.ml)
 * afin d'avoir un meilleur affichage d'erreur.
 *)
let pf1 = pf "apply"
;;




(** Pour:
 *** les terme: voir Vm/printer_term.ml
 *** le byte-code: idem
 *** l'affichage de l'environnement: idem
 ***)





(*********************************************************************************************)
(*********************************************************************************************)
(*********************************************************************************************)
(*********************************************************************************************)

(** ---------------------------------------------------------------- **)
(** --- Implementation de print_collection formatt ---------------- **)



(**  pobval1_'outBufferF_int_string_string_string_funct_typemgs (b,t,pre,inter,post,f,x) : 
 ** Ecrit de faon formate l'lment de type mgs [x] dans le buffer format [b] 
 ** avec pour prfixe [pre], pour infixe [inter], pour suffixe [post], 
 ** et avec une indentation de longueur t. 
 ** De plus, elle applique la fonction f sur chaque lment de [x].
 ** Puis, elle renvoie le type valundef
 *)

(** pobval b pre inter post f x : 
 ** Cette fonction prend un buffer "normal" b, une structure mgs x, un prfixe pre, 
 ** un infixe inter, un suffixe post, une fonction f. Elle crit de faon formate x 
 ** dans le buffer b en appliquant la fonction f  chaque lment de x.
 ** Pour cela, elle transforme le prefixe, l'infixe, le suffixe en supprimant 
 ** les guillemets des VAL_string. Puis, elle calcule la taille du prfixe pour 
 ** l'indentation. Ensuite, elle appelle la fonction pobval1.
 ** Enfin, elle renvoie valundef
 **)
let rec pobval b  pre inter post f x= 
  let spre   = VAL_string(string_of_STRING(match (pfsvalue pre) with VAL_string(s)->s|_->"")) 
  and spost  = VAL_string(string_of_STRING(match (pfsvalue post) with VAL_string(s)->s|_->"")) 
  and sinter = VAL_string(string_of_STRING(match (pfsvalue inter) with VAL_string(s)->s|_->"")) 
  in
  let tpre = String.length (match spre with VAL_string(s)->s|_->"") in
    
  let bf = Format.formatter_of_buffer b in 
  let a= pobval1 bf (VAL_int tpre) spre sinter spost f x
  in Format.pp_print_newline bf ();a


and pobval2 b  pre inter post f x= 
  let spre   = VAL_string(string_of_STRING(match (pfsvalue pre) with VAL_string(s)->s|_->"")) 
  and spost  = VAL_string(string_of_STRING(match (pfsvalue post) with VAL_string(s)->s|_->"")) 
  and sinter = VAL_string(string_of_STRING(match (pfsvalue inter) with VAL_string(s)->s|_->"")) 
  in
    let tpre = String.length (match spre with VAL_string(s)->s|_->"") in
      pobval1 b  (VAL_int tpre) spre sinter spost f x; ()

and pobvalue b  pre inter post f x= pobval b  pre inter post f x
and pobvalue2 b  pre inter post f x= pobval2 b  pre inter post f x 



(** povalue: affichage sur la sortie standard.
 ** On cree un buffer, on imprime dedans puis on sort le buffer sur la sortie standard.
 **)
and povalue pre inter post f x =
  clear out_bufferF;
  pobval out_bufferF pre inter post f x;
  output_buffer stdout out_bufferF;
  x

(** prfvalue out x: sortie de x sur le fichier out **)
and pofvalue out pre inter post f x =
  clear out_bufferF;
  pobval out_bufferF pre inter post f x;
  output_buffer out out_bufferF;
  x

(** prsvalue x : retourne une string qui represente la valeur x **)
and posvalue pre inter post f x=
  clear out_bufferF;
  pobval out_bufferF pre inter post f x;
  VAL_string(contents out_bufferF)

    
;;

(** idem que po(f)value, mais renvoie unit plutot que l'argument **)
let pofvalue2 out pre inter post f x= (pofvalue out pre inter post f x; ())
and povalue2  pre inter post  f x= (povalue pre inter post f x; ())
;;


(** --------------------------------------------------------------------- **)
(** --- Implementation de uprint_collection non-formatt ---------------- **)



(**  upobval1_'outBufferF_int_string_string_string_funct_typemgs (b,t,pre,inter,post,f,x) : 
 ** Ecrit de faon non-formate l'lment de type mgs [x] dans le buffer [b] 
 ** avec pour prfixe [pre], pour infixe [inter], pour suffixe [post].
 ** De plus, elle applique la fonction f sur chaque lment de [x].
 ** Puis, elle renvoie le type valundef
 *)

(** upobval b pre inter post f x : 
 ** Cette fonction prend un buffer b, une structure mgs x, un prfixe pre, 
 ** un infixe inter, un suffixe post, une fonction f. Elle crit x 
 ** dans le buffer b en appliquant la fonction f  chaque lment de x.
 ** Pour cela, elle transforme le prefixe, l'infixe, le suffixe en supprimant 
 ** les guillemets des VAL_string. Ensuite, elle appelle la fonction pobval1.
 ** Enfin, elle renvoie valundef
 **)
let rec upobval b  pre inter post f x= 
  let spre   = VAL_string(string_of_STRING(match (uprsvalue pre) with VAL_string(s)->s|_->"")) 
  and spost  = VAL_string(string_of_STRING(match (uprsvalue post) with VAL_string(s)->s|_->"")) 
  and sinter = VAL_string(string_of_STRING(match (uprsvalue inter) with VAL_string(s)->s|_->"")) 
  in let a= upobval1 b spre sinter spost f x
  in Printf.bprintf b "\n";a


and upobval2 b  pre inter post f x= 
  let spre   = VAL_string(string_of_STRING(match (uprsvalue pre) with VAL_string(s)->s|_->"")) 
  and spost  = VAL_string(string_of_STRING(match (uprsvalue post) with VAL_string(s)->s|_->"")) 
  and sinter = VAL_string(string_of_STRING(match (uprsvalue inter) with VAL_string(s)->s|_->""))
  in upobval1 b  spre sinter spost f x; ()

and upobvalue b  pre inter post f x= upobval b  pre inter post f x
and upobvalue2 b  pre inter post f x= upobval2 b  pre inter post f x 



(** povalue: affichage sur la sortie standard.
 ** On cree un buffer, on imprime dedans puis on sort le buffer sur la sortie standard.
 **)
and upovalue pre inter post f x =
  clear out_buffer;
  upobval out_buffer pre inter post f x;
  output_buffer stdout out_buffer;
  x

(** prfvalue out x: sortie de x sur le fichier out **)
and upofvalue out pre inter post f x =
  clear out_buffer;
  upobval out_buffer pre inter post f x;
  output_buffer out out_buffer;
  x

(** prsvalue x : retourne une string qui represente la valeur x **)
and uposvalue pre inter post f x=
  clear out_buffer;
  upobval out_buffer pre inter post f x;
  VAL_string(contents out_buffer)

    
;;

(** idem que po(f)value, mais renvoie unit plutot que l'argument **)
let upofvalue2 out pre inter post f x= (upofvalue out pre inter post f x; ())
and upovalue2  pre inter post  f x= (upovalue pre inter post f x; ())
;;
