(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: type.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-21 20:51:16 $
 * Version: $Revision: 1.9 $
 * -------------------------------------------------------------------
 *
 * Les types collections MGS sont decrit par des instances du type CAML
 * mgs_type_def  qui est specifie dans valeur.ml.
 *
 * Ces instances sont relies par les champs derived_from et father_of.
 *
 * Ces instances, qui servent de descripteurs de types sont creees par la 
 * fonction create_type qui prend un mgs_type_def et qui met a jour les 
 * instances existantes.
 *
 *
 *
 * Addition des predicats de types dans la table des fonctions:
 * ca doit se faire plus tard: voir environnement.ml et kernel.ml 
 *
 * Le sommet de la hierarchie des types est preinstalle :
 *
 *
 *   'collection
 *   |
 *   +------ 'leibniz
 *   |       |
 *   |       +------ 'set
 *   |       +------ 'seq
 *   |       +------ 'bag
 *   |       +------ 'delaunay
 *   |       
 *   +------ 'newton
 *           |
 *           +------ 'record
 *           +------ 'gbf
 *           +------ 'graph
 *
 * Voir les fichier Xxx.type_xxx.ml
 * 
 *=========================================================================
 *)



open Global;;
open Valeur;;
open Environnement;;


__register_version "$RCSfile: type.ml,v $" "$Revision: 1.9 $" "$Date: 2006-04-21 20:51:16 $";;


(* --- Table des types et comparaison  ---------------------------------------- *)


(** La definition de :
 **
 **     mgs_type_def           (* le type caml d'un descripteur de type-mgs d'une collection *)
 **     type_tbl               (* le dico des type *)
 **     find_type_from_name    (* le descripteur d etype-mgs a partir du nom *)
 **
 ** sont dans valeur.ml 
 **)


let type_tbl = Valeur.type_tbl
and find_type_from_name = Valeur.find_type_from_name 
;;




module St = Set.Make(struct type t = mgs_type_def let compare=compare_type end);;
let uniquify l =  St.elements (List.fold_left (fun s x -> St.add x s) St.empty l)
;;

module Ss = Set.Make(struct type t = string let compare = Pervasives.compare end);;
let check_no_double l =
  (List.length (Ss.elements (List.fold_left (fun s x -> Ss.add x s) Ss.empty l))) = (List.length l)
;;

let type_uid = ref 0
;;
let get_new_uid() = let uid = !type_uid in (incr type_uid; uid)
;;

let empty_from_type_name id =
  try
    let desc = Hash.find type_tbl id
    in if desc.concrete then desc.empty
       else raise Not_found
  with Not_found ->
    begin
      try
	let desc = Hash.find type_tbl ("'"^id)
	in if desc.concrete then desc.empty
	   else raise Not_found
      with Not_found ->
	let err = "Error: identifier '"^id^"' is not a concrete type name";
	in prerr_endline err;
	   raise (Erreur err)
    end

(* --- Affichage ------------------------------------------------------- *)

let rec listing out = 
  let lst = List.sort compare_type (Hash.fold  (fun k v l -> v::l) type_tbl [])
  in List.iter (fun desc -> prftype out desc) lst

and prftype out t = 
  Printf.fprintf out "\t%s type  %s \tderives from %a\n"
    (if t.concrete then "concrete" else "abstract")
    (pad_name 8 t.name)
    prltypename t.derive_from

and prltypename out = function 
  | [] -> print_string "  nothing"
  | l -> List.iter (function t -> Printf.fprintf out "%s " t.name) l

and prtype x = prftype stdout x

and number_defined () = Hash.size type_tbl
;;

(* --- Creation d'un type  a la main ---------------------------------------- *)


let rec create_type r =
  (* on s'ajoute dans la table *)
  Hash.add type_tbl r.name r;
  (* on s'ajoute chez tous les fils recursivement *)
  add_sons r (get_transitive_sons r.father_of);
  (* on s'ajoute chez tous les peres recursivement *)
  add_fathers r (get_transitive_fathers r.derive_from);
  (* on trie la liste de derivees des sous-type vers la racine 'collection 
   * en n'oubliant pas de s'ajouter dans la liste des derivees !
   *)
  r.derive_from <- List.sort compare_type (r::r.derive_from);
  r

and get_transitive_sons l = uniquify(List.fold_left (fun ll t -> ll@t.father_of) [] l)

and get_transitive_fathers l = uniquify(List.fold_left (fun ll t -> ll@t.derive_from) [] l)

and add_sons r sl = List.iter (add_son1 r) sl

and add_fathers r sl = List.iter (add_father1 r) sl

and add_son1 r s =
  (if not(List.mem s r.father_of) then r.father_of <- s::r.father_of);
  (if not(List.mem r s.derive_from) then s.derive_from <- r::s.derive_from)
  
and add_father1 r f =
  (if not(List.mem f r.derive_from) then r.derive_from <- f::r.derive_from);
  (if not(List.mem r f.father_of) then f.father_of <- r::f.father_of)
;;


(* --- Plus grand ancetre commun --------------------------------------- *)

let rec common_type r r' = 
  if ((Pervasives.compare r r') = 0) then [r] else commont [] (r.derive_from, r'.derive_from)

and commont t = function
  | (a::l, b::l') -> if ((Pervasives.compare a b) = 0) then commont [a] (l, l') else t
  | _ -> t
;;

let commontype r r' = match common_type r r' with
  | [r] -> r
  | _ -> 
      let err = "Error: "^r.name^" and "^r'.name^" does not have a commonn ancestor !!!"
      in prerr_endline err; prtype r; prtype r'; raise (Erreur err)
;;

(* --- Changement de couleur ------------------------------------------------------------- *)

let _setColor_any_any (x,y) =
  if (is_collection x) & (is_collection y)
  then
    let tx = collection_type x
    and ty = collection_type y
    in
    let ct = commontype tx ty in
      if ct.concrete
      then
	change_type ty x
      else 
	let err = "Type: set_color: incompatible type"
	in prerr_endline err; raise (Erreur err)
  else
    let err = "Type: set_color: waiting for two collections"
    in prerr_endline err; raise (Erreur err)
;;

(* --- Relation de parente --------------------------------------------------------------- *)

(** r sous-type de r', r derive de r', r specialisation de r', r est un r' colore **)
let is_subtype r r' = List.mem r r'.derive_from
;;


(* --- Inspection des types --------------------------------------------------------------- *)



let collection_name v = VAL_string((collection_type v).name)
and pospredicate_from_coll v = (collection_type v).pospredicate
and empty_from_coll v = (collection_type v).empty
and predicate_from_coll v = (collection_type v).predicate
;;




(* --- Creation du sommet de la hierarchie ---------------------------------------------- *)

let wrap_pred f = VAL_funct(function x -> VAL_bool(f x))
;;

let pospredicate_dummy = 
  VAL_funct(function _ -> raise (Erreur "Internal error: pospred not defined in type descriptor"))

let predicate_dummy = 
  VAL_funct(function _ -> raise (Erreur "Internal error: pred not defined in type descriptor"))

and _intPredicate_int _ = mgstrue
and _intPredicate_any _ = mgsfalse

and _floatPredicate_float _ = mgstrue
and _floatPredicate_any _   = mgsfalse

and _boolPredicate_bool _ = mgstrue
and _boolPredicate_any _  = mgsfalse

and _stringPredicate_string _ = mgstrue
and _stringPredicate_any _    = mgsfalse

and _functPredicate_funct _ = mgstrue
and _functPredicate_any _   = mgsfalse

and _undefPredicate_undef _ = mgstrue
and _undefPredicate_any _   = mgsfalse

and _posgbfPredicate_posgbf _ = mgstrue
and _posgbfPredicate_any _    = mgsfalse

and _acellPredicate_acell _ = mgstrue
and _acellPredicate_any _  = mgsfalse

and _gmapPredicate_gmap _ = mgstrue
and _gmapPredicate_any _  = mgsfalse

and _ncellPredicate_ncell _ = mgstrue
and _ncellPredicate_any _   = mgsfalse

and _symbolPredicate_symbol _ = mgstrue
and _symbolPredicate_any _    = mgsfalse
;;


let xxxify_dummy tname =
   VAL_funct(function _ -> raise_error ("Error: function "^tname^"ify not defined"))

and make_dummy _ = raise (Erreur "Internal error: make not defined in type descriptor")

and pospredicate_int    = (fun_1 Dispatch_sig.intPredicate)
and pospredicate_string = (fun_1 Dispatch_sig.stringPredicate)
and pospredicate_posgbf = (fun_1 Dispatch_sig.posgbfPredicate)
and pospredicate_acell   = (fun_1 Dispatch_sig.acellPredicate)
and pospredicate_ncell  = (fun_1 Dispatch_sig.ncellPredicate)
;;





(** Construction gnrique d'une collection  partir d'une numration (extension) **)

let rec _extension_string_seq (s, (a, _)) =
  try
    let tdesc = (try find_type_from_name s with Not_found -> find_type_from_name ("'"^s)) in
      check_vplist s (to_funct tdesc.pospredicate) a;
      tdesc.make a
  with Not_found -> 
    let err = "Error: 'extension: unknown collection type '"^s^"'"
    in prerr_endline err; raise (Erreur err)


(* Predicat auxilliaires utilises pour vrifier les arguments *)

and check_vplist name pred a = 
  if not(array_forall (check_couple pred) a) then
    let err = "Error: extension: the (value, position) sequence is incorrect for building a "^name
    in prerr_endline err; raise (Erreur err)

and check_couple pred = function
  | VAL_seq(a, _) -> (2 = Array.length a) & to_bool(Dispatch_sig.val2bool(pred a.(0)))
  | _ -> false
;;



(** Cration d'une fonction xxxify par dfaut. 
 ** Utilise pour les types couleurs (gnr automatiquement),
 ** ainsi que les types abstrait.
 **)

let rec avant_dernier =  function
  | [] 
  | [_]
  -> raise (Erreur "Internal error: 'avant_dernier' in make_xxxfy")
  | [f; _] -> f
  | _::l -> avant_dernier l
  ;;

let make_xxxify t =
  let father = avant_dernier t.derive_from
  and name = t.name 
  and nameify = t.name^"ify"
  in
    begin
      (if  Hash.mem var_tbl nameify
       then  prerr_endline ("Warning: type related function   "^nameify
			    ^"  hiddes a function with same name"));
      (if t.concrete then
	 t.xxxify <- VAL_funct(fun x -> change_type t ((pf1 father.xxxify) x))
       else
	 t.xxxify <- VAL_funct(fun x -> raise_error ("Error: "
						     ^nameify
						     ^"cannot translate a "
						     ^(constructor_string x)
						     ^" into the abstract type "
						     ^name) ));
      addfct (nameify) t.xxxify
    end
;;






(** --- Cration de la racine des types collections : 'collection ------------------------- **)

let any_desc = create_type
            { name = "'any";
	      concrete = false;
	      leibniz = false;
	      predicate = VAL_funct(function _ -> VAL_bool(true));
	      pospredicate = pospredicate_dummy;
	      empty = VAL_undef "dummy for empty field in type descriptor 'any";
	      make = make_dummy;
	      xxxify = xxxify_dummy "'any";
	      derive_from = [];
	      father_of = [];
	      uid = get_new_uid();
	    }
;;

let coll_desc = create_type 
	     { name = "'collection";
	       concrete = false;
	       leibniz = false;
	       predicate = wrap_pred(is_collection);
	       pospredicate = pospredicate_dummy;
	       empty = VAL_undef "dummy for empty field in type descriptor 'collection";
	       make = make_dummy;
	       xxxify = xxxify_dummy "'collection";
	       derive_from = [any_desc];
	       father_of = [];
	       uid = get_new_uid();
	     }
;;





(** ce predicat permet de savoir si t est dans la liste des ancetres de x, x inclus *)
let coll_pred t x = 
  is_collection(x) && (let desc = collection_type(x) 
		       in ((Pervasives.compare desc t) = 0) or (List.mem t desc.derive_from))
;;

let coll_type_pred desc = ((Pervasives.compare desc coll_desc) = 0) or (List.mem coll_desc desc.derive_from)

let is_leibniz x = is_collection(x) & (collection_type x).leibniz
and is_newton x  = is_collection(x) & not((collection_type x).leibniz)
;;


(* --- Types abstraits : leibnitz, newton, monoidal -------------------------------- *)


let leibniz_desc = create_type
		 { name = "'leibniz";
		   concrete = false;
		   leibniz = true;
		   predicate = wrap_pred(is_leibniz);
		   empty = VAL_undef "dummy for empty field in type descriptor 'leibniz";
		   pospredicate = pospredicate_dummy;
		   make = make_dummy;
		   xxxify = xxxify_dummy "'leibniz";
		   derive_from = [coll_desc];
		   father_of = [];
		   uid = get_new_uid();
		 }

and newton_desc = create_type 
	        { name = "'newton";
		  concrete = false;
		  leibniz = false;
 		  predicate = wrap_pred(is_newton);
		  empty = VAL_undef "dummy for empty field in type descriptor 'newton";
		  pospredicate = pospredicate_dummy;
		  make = make_dummy;
		  xxxify = xxxify_dummy "'newton";
		  derive_from = [coll_desc];
		  father_of = [];
		  uid = get_new_uid();
	        }
;;

let monoidal_desc = create_type 
		  { name = "'monoidal";
		    concrete = false;
	 	    leibniz = true;
		    predicate = wrap_pred(is_leibniz);
		    empty = VAL_undef "dummy for empty field in type descriptor 'monoidal";
		    pospredicate = pospredicate_dummy;
		    make = make_dummy;
		    xxxify = xxxify_dummy "'monoidal";
		    derive_from = [leibniz_desc];
		    father_of = [];
		    uid = get_new_uid();
		  }
;;

make_xxxify leibniz_desc;;
make_xxxify newton_desc;;

monoidal_desc.predicate <- wrap_pred(coll_pred monoidal_desc);;

(*!*)
(** 
  @module collections
  @name 'monoidal
  @type any_bool
  Monoidal type and predicate. 
  [monoidal] is the name of an {i abstract type}: it does not exist
  a value that have precisely this type. [monoidal] rather represents
  the kind of the collection types that represent a space whose
  topology is based on a monoid: [set], [bag] (multiset) and [seq]
  (sequence or list). These three collection types are leibnizian.
  Like any other type, [monoidal] is also a predicate that can be used
  to check if the type of a value is a subtype of 
  [monoidal]. 

  Some functions are applicable to all monoidal types: [cons], [join],
  [append], etc. Cf. the section Monoidal.

  @voir Monoidal the chapter monoidal in the documentation.

  @voir 'set 'set
  @voir 'seq 'seq
  @voir 'bag 'bag
  @voir 'leibniz 'leibniz
*)
(*!*)






(* === Les couleurs ====================================== *)


let rec create_new_color nom fathername =
  try
    let t = Hash.find type_tbl nom
    in let err = "Error: type  "^nom^"  already defined."
    in prerr_endline err; prftype stderr t; raise (Erreur err)
  with Not_found ->
    try
      let t = Hash.find type_tbl fathername 
      in create_new_color_aux nom t
    with Not_found ->
      try
	let t = Hash.find type_tbl ("'"^fathername)
	in create_new_color_aux nom t
      with Not_found ->
	let err = "Error: type  "^fathername^"  is not a valid type to derive "^nom
	in prerr_endline err; raise (Erreur err)


and create_new_color_aux nom t =
  (if not(t.concrete) 
   then let err = "Error: cannot derive type "^nom^" from a non concrete type "^t.name
        in prerr_endline err; prftype stderr t; raise (Erreur err));
  (if not(coll_type_pred t) 
   then let err = "Error: cannot derive type "^nom^" from a non collection type "^t.name
        in prerr_endline err; prftype stderr t; raise (Erreur err));
  (if Hash.mem var_tbl nom
   then prerr_endline ("Warning: type predicate "^nom^" hiddes a function with same name"));
  let nt = create_type { name = nom;
		 	 concrete = true;
			 leibniz = t.leibniz;
			 predicate = t.predicate;
			 empty =  VAL_undef("transient dummy in color type creation "^nom);
			 pospredicate = t.pospredicate;
			 make = t.make;
			 xxxify = xxxify_dummy nom;
			 derive_from = [t];
			 father_of = [];
			 uid = get_new_uid();
		       }
  in
    begin
      nt.empty <- change_type nt t.empty;
      nt.predicate <- wrap_pred(coll_pred nt);
      nt.make <- extension_color t nt;
      ignore(make_xxxify nt);
      ignore(addfct nom nt.predicate);
      nt
    end

and extension_color t nt a = change_type nt (t.make a)
;;



let _raiseFct_symbol_any (s,v) =
  (raise (MGSExn (s, v)) ; VAL_undef ("Dummy return from 'raise")) ;;

let _tryFct_funct_funct (body,catched_rules) =
  try
    body (VAL_undef "")
  with MGSExn (s,v) ->
    let (i,a) = match v with
	VAL_seq (l,_) -> VAL_int(Array.length l), l
      | x -> VAL_int(1), [|x|]
    in
    let rule_found = to_funct (to_funct (catched_rules (VAL_symbol s)) i) v in
      Array.fold_left (fun acc vi -> (to_funct acc) vi) rule_found a
;;

(** --- Cration de la racine des types adt : 'adt ------------------------- *)
(*
let create_new_adt nom lcons =
  try
    let t = Hash.find type_tbl nom
    in let err = "Error: type "^nom^" already defined."
    in prerr_endline err; prftype stderr t; raise (Erreur err)
  with Not_found ->
    (if Hash.mem var_tbl nom
     then prerr_endline ("Warning: type predicate "^nom^" hiddes a function with same name"));
    let nt = create_type { name = nom;
		 	   concrete = false;
			   leibniz = false;
			   predicate = adt_desc.predicate;
			   pospredicate = adt_desc.pospredicate;
			   empty = VAL_undef("transient dummy in type creation "^nom);
			   make = adt_desc.make;
			   xxxify = xxxify_dummy nom;
			   derive_from = [adt_desc];
			   father_of = [];
			   uid = get_new_uid();
			 }
    in
    let lcons = List.map
		  (function
		       ADT_unknown s -> (if (s = nom)
					 then (ADT_color nt)
					 else (raise (Erreur "\"type ... and ... ;;\" not yet implemented")))
		     | x -> x
		  )
		  lcons
    in
    let f = fun v -> VAL_bool(
      List.exists 
        (function ADT_color c   -> (match c.predicate with VAL_funct f -> to_bool (f v) | _ -> raise (Erreur "Oups"))
	   |      ADT_scalar s  -> (Dispatch_sig.equal_mgs_value s v)
	   |      ADT_unknown _ -> raise (Erreur "Internal error here !!")
	)
	lcons
    ) in
      nt.predicate <- VAL_funct(f) ;
      ignore(addfct nom nt.predicate);
      nt

*)



let rec check_unknown_names_lcstrt lcstrt names = List.for_all (check_unknown_names_cstrt_elt names) lcstrt

and check_unknown_names_cstrt_elt names cstrt = match cstrt with
  | CSTRT_unknown (s, Some tree, _) -> (List.mem s names) && (check_unknown_names_cstrt_tree names tree)
  | CSTRT_unknown (s, _, _) -> (List.mem s names)
  | CSTRT_scalar _ -> true
  | CSTRT_var _ -> true
  | CSTRT_type (_,None,_) -> true
  | CSTRT_type (_,Some tree,_) -> check_unknown_names_cstrt_tree names tree

and check_unknown_names_cstrt_tree names tree = match tree with
  | TCSTRT_elt e -> check_unknown_names_cstrt_elt names e
  | TCSTRT_not a -> check_unknown_names_cstrt_tree names a
  | TCSTRT_and (a,b) -> (check_unknown_names_cstrt_tree names a) && (check_unknown_names_cstrt_tree names b)
  | TCSTRT_or  (a,b) -> (check_unknown_names_cstrt_tree names a) && (check_unknown_names_cstrt_tree names b)
;;

let create_dummy_constraint nom =
  let nt = create_type { name = nom;
		 	 concrete = false;
			 leibniz = false;
			 predicate = any_desc.predicate;
			 pospredicate = any_desc.pospredicate;
			 empty = VAL_undef("transient dummy in type creation "^nom);
			 make = any_desc.make;
			 xxxify = xxxify_dummy nom;
			 derive_from = [any_desc];
			 father_of = [];
			 uid = get_new_uid();
		       }
  in
    nt
;;

let rec replace_unknown_type lcstrt t = List.map (replace_unknown_type_cstrt_elt t) lcstrt

and replace_unknown_type_cstrt_elt t cstrt = match cstrt with
  | CSTRT_unknown (s, Some tree, b) -> CSTRT_type (t, Some (replace_unknown_type_cstrt_tree t tree), b)
  | CSTRT_unknown (s, None, b) -> CSTRT_type (t, None, b)
  | CSTRT_type (tt,Some tree,ii) -> CSTRT_type (tt, Some (replace_unknown_type_cstrt_tree t tree), ii)
  | x -> x

and replace_unknown_type_cstrt_tree t tree = match tree with
  | TCSTRT_elt e -> TCSTRT_elt (replace_unknown_type_cstrt_elt t e)
  | TCSTRT_not a -> TCSTRT_not (replace_unknown_type_cstrt_tree t a)
  | TCSTRT_and (a,b) -> TCSTRT_and (replace_unknown_type_cstrt_tree t a, replace_unknown_type_cstrt_tree t b)
  | TCSTRT_or  (a,b) -> TCSTRT_or  (replace_unknown_type_cstrt_tree t a, replace_unknown_type_cstrt_tree t b)
;;


let rec replaceDico_unknown_type dico lcstrt = List.map (replaceDico_unknown_type_cstrt_elt dico) lcstrt

and replaceDico_unknown_type_cstrt_elt dico cstrt = match cstrt with
  | CSTRT_unknown (s, Some tree, b) -> CSTRT_type (List.assoc s dico, Some (replaceDico_unknown_type_cstrt_tree dico tree), b)
  | CSTRT_unknown (s, None, b) -> CSTRT_type (List.assoc s dico, None, b)
  | CSTRT_type (tt,Some tree,ii) -> CSTRT_type (tt, Some (replaceDico_unknown_type_cstrt_tree dico tree), ii)
  | x -> x

and replaceDico_unknown_type_cstrt_tree dico tree = match tree with
  | TCSTRT_elt e -> TCSTRT_elt (replaceDico_unknown_type_cstrt_elt dico e)
  | TCSTRT_not a -> TCSTRT_not (replaceDico_unknown_type_cstrt_tree dico a)
  | TCSTRT_and (a,b) -> TCSTRT_and (replaceDico_unknown_type_cstrt_tree dico a, replaceDico_unknown_type_cstrt_tree dico b)
  | TCSTRT_or  (a,b) -> TCSTRT_or  (replaceDico_unknown_type_cstrt_tree dico a, replaceDico_unknown_type_cstrt_tree dico b)
;;

let rec build_constraint_pred lcstrt = VAL_funct(function v -> VAL_bool(List.exists (build_constraint_pred_elt v) lcstrt))

and build_constraint_pred_elt v cstrt =
  match cstrt with
    | CSTRT_unknown (s, _, _)  -> raise (Erreur ("Type: build_constraint: variable "^s^" undefined."))
    | CSTRT_scalar s   -> Dispatch_sig.equal_mgs_value s v
    | CSTRT_var (nf,f) -> to_bool((to_funct(f))(v))
    | CSTRT_type (t,None,None)       -> to_bool((to_funct(t.predicate))(v))
    | CSTRT_type (t,None, Some i)    -> to_bool((to_funct(t.predicate))(v)) & ( (to_int (Dispatch_sig.size(v))) = i)
    | CSTRT_type (t,Some tree, None) -> (
	to_bool((to_funct(t.predicate))(v)) &
	to_bool(Dispatch_sig.fold
		  (VAL_funct (fun e ->
				VAL_funct ( fun acc ->
					      VAL_bool( to_bool(acc) & (build_constraint_pred_tree e tree) )
					  )
			     )
		  )
		  (VAL_bool(true))
		  v)
      )
    | CSTRT_type (t,Some tree, Some i) -> (
	to_bool((to_funct(t.predicate))(v)) &
	( (to_int (Dispatch_sig.size(v))) = i) &
	to_bool(Dispatch_sig.fold
		  (VAL_funct (fun e ->
				VAL_funct ( fun acc ->
					      VAL_bool( to_bool(acc) & (build_constraint_pred_tree e tree) )
					  )
			     )
		  )
		  (VAL_bool(true))
		  v)
      )

and build_constraint_pred_tree v tree =
  match tree with
    | TCSTRT_elt e -> (build_constraint_pred_elt v e)
    | TCSTRT_not a -> (not (build_constraint_pred_tree v a))
    | TCSTRT_and (a,b) -> ((build_constraint_pred_tree v a) & (build_constraint_pred_tree v b))
    | TCSTRT_or  (a,b) -> ((build_constraint_pred_tree v a) or (build_constraint_pred_tree v b))

;;


let update_new_cstrt_node dico n id cstrt =
  let f = build_constraint_pred (replaceDico_unknown_type dico cstrt) in
    n.predicate <- f ;
    ignore(addfct id n.predicate);
    n
;;
