(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: valeur.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2006-04-23 05:22:18 $
 * Version: $Revision: 1.15 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *
 *=========================================================================
 *)

open Global;;
open Big_int;;

__register_version "$RCSfile: valeur.ml,v $" "$Revision: 1.15 $" "$Date: 2006-04-23 05:22:18 $";;


(** Les types des structures de donnes qui implmentent une collection xxx.
 ** - Chaque rpertoire Xxx fournit l'implmentation de la collection xxx. 
 ** - Le type ML correspond est txxx defini dans le fichier Xxx/def_xxx.ml 
 ** - Les fonctions sur ce type sont dans XXx/core_xxx.ml. 
 ** - La structure decrivant le type mgs xxx est dans Xxx/type_xxx.ml
 **)

open Def_rec;;    (* implantation des records                           *)
open Def_seq;;    (* implantation des sequences                         *)
open Def_ring;;   (* implantation des rings                             *)
open Def_set;;    (* implantation des ensembles                         *)
open Def_bag;;    (* implantation des multi-ensembles (bag)             *)
open Def_gab;;    (* implantation des multi-ensembles gnraliss (gab) *)
open Def_gbf;;    (* implantation des GBF                               *)
open Def_graph;;  (* implantation des graphes gnriques                *)
open Def_del;;    (* implantation des graphes de Delaunay               *)
open Def_prox;;   (* implantation des Graphes proximaux                 *)
open Def_achain;; (* implantation des chanes abstraites                *)
open Def_qmf;;    (* implantation des QMF                               *)



(** la description du type MGS des collections 
 ** concrete = il y a des valeurs qui ont ce type comme type le plus prcis.
 ** L'oppos c'est "abstrait". "Collection" qui est le type qui regroupe toutes les
 ** valeurs qui reprsentent des collections, est un type abstrait. Comme le type "Monoidal" 
 ** qui regroupe les ensembles, les bags et les squences.
 ** Tous les champs qui suivent le champ "concrete", n'ont de sens que si le type est concret.
 **)

type mgs_type_def = 
    {
      uid : int;                                   (* un entier unique croissant dans la hierarchie *)
      name : string;                               (* nom du type *)
      mutable predicate : mgs_value;               (* predicat qui teste si une valeur est de ce type *)
      mutable derive_from : mgs_type_def list;     (* de qui je suis sous-type, transitivement *)
      mutable father_of : mgs_type_def list;       (* qui sont mes sous-types (transitivement) *)

      concrete : bool;                             (* correspond a un type abstrait ou pas *)
      leibniz : bool;                              (* collection leibnizienne  *)
      mutable empty : mgs_value;                   (* la collection vide de ce type *)
      mutable make : mgs_value array -> mgs_value; (* creation  partir d'une enumration *)
      mutable pospredicate : mgs_value;            (* teste si une valeur est une position de ce type *)
      mutable xxxify : mgs_value;                  (* fct qui injecte une collection en ce type *)
    }

(* ****************************************** *)
(* Pour les constructions recursives de types *)
(* ****************************************** *)

and recTypeDef =
  | RecTypeDef_coll  of string * string
  | RecTypeDef_gbf   of string * (string list * gbf_expr list)
  | RecTypeDef_qmf   of string * gmap
  | RecTypeDef_rec   of string * (record_type_specification list) * (string list)
  | RecTypeDef_del   of string * int * bool * mgs_value
  | RecTypeDef_prox  of string * mgs_value
  | RecTypeDef_cstrt of string * (constraint_elt list)

(* ************************************* *)
(* Pour les types de donnes algbriques *)
(* ************************************* *)

and constraint_elt =
  | CSTRT_scalar  of mgs_value                                              (* predicat : la valeur doit tre gale               *)
  | CSTRT_var     of string * mgs_value                                     (* predicat : application de la fonction              *)
  | CSTRT_type    of mgs_type_def * (constraint_tree option) * (int option) (* predicat : application de .predicat                *)
  | CSTRT_unknown of string * (constraint_tree option) * (int option)       (* partie gauche en cours de dfinition ou une erreur *)

and constraint_tree =
  | TCSTRT_elt of constraint_elt
  | TCSTRT_not of constraint_tree
  | TCSTRT_and of constraint_tree * constraint_tree
  | TCSTRT_or  of constraint_tree * constraint_tree


(* ************************************* *)
(* Pour les records                      *)
(* ************************************* *)

and field_type = 
  | Present                      (* le champ doit tre prsent                    *)
  | Absent                       (* le champ doit tre absent                     *)
  | WithType  of mgs_value       (* prsent avec un type vrifiant le prdicat    *)
  | WithCstrt of constraint_tree (* prsent avec une contrainte de type  valuer *)
  | WithValue of mgs_value       (* prsent avec cette valeur                     *)
  | WithoutValue of mgs_value    (* prsent mais pas avec cette valeur            *)

and field_specification = string * field_type

(* une liste de: pattern_var_name, (field-name, field_type) *)
and fields_pattern_specification = (string option * field_specification) list 

and record_type_specification = 
  | Fields of field_specification list
  | RecordType of string


(* **************************** *)
(* Pour les symbols             *)
(* **************************** *)

and symbol = string

(* **************************** *)
(* Pour TOUS les graphes        *)
(* **************************** *)

and 'a graphe_element = 'a array

(* ****************************** *)
(* Definition des type collection *)
(* ****************************** *)

and 'a colors         = 'a * mgs_type_def

and rec_collection    = (mgs_value, mgs_value) trec
and seq_collection    = mgs_value tseq
and ring_collection   = mgs_value tring
and set_collection    = mgs_value tset
and bag_collection    = mgs_value tbag
and gab_collection    = mgs_value tgab
and gbf_collection    = mgs_value tgbf
and grf_collection    = mgs_value tgraf
and del_collection    = mgs_value tdel 
and prox_collection   = mgs_value tprox
and achain_collection = mgs_value achain
and qmf_collection    = mgs_value qmf


and mgs_value = 
  | VAL_undef  of string (* la string sert de message d'erreur si necessaire *)
  | VAL_int    of int
  | VAL_bint   of big_int
  | VAL_bool   of bool
  | VAL_float  of float
  | VAL_string of string
  | VAL_symbol of symbol
  | VAL_posgbf of posgbf
  | VAL_acell  of acell
  | VAL_gmap   of gmap
  | VAL_ncell  of ncell
  | VAL_funct  of (mgs_value -> mgs_value)

  | VAL_rec    of   rec_collection    colors
  | VAL_seq    of   seq_collection    colors
  | VAL_ring   of   ring_collection   colors
  | VAL_set    of   set_collection    colors
  | VAL_bag    of   bag_collection    colors
  | VAL_gab    of   gab_collection    colors
  | VAL_gbf    of   gbf_collection    colors
  | VAL_grf    of   grf_collection    colors
  | VAL_del    of   del_collection    colors
  | VAL_prox   of   prox_collection   colors
  | VAL_achain of   achain_collection colors
  | VAL_qmf    of   qmf_collection    colors

and direction_specification = AllDirections | SomeDirections of mgs_value array
;;


(** Une exception pour implementer les exceptions dans les transformations MGS **)
exception BreakTransformation of mgs_value;;

(** Une exception pour implementer les exceptions dans MGS tout court **)
exception MGSExn of string*mgs_value ;;


(** le dico des types-MGS correspondant  des collections **)

let type_tbl = ((Hash.create 37) : (string, mgs_type_def) Hash.t)
;;
let find_type_from_name s = Hash.find type_tbl s
;;


(** Traitement de l'egalite et de la comparaison **)

let mgsfalse = VAL_bool false
and mgstrue  = VAL_bool true
and mgszero  = VAL_int(0)
and mgsun    = VAL_int(1)
and mgsundef = VAL_undef("")
and current_error_handler = ref (VAL_undef "pre-init current_error_handler")
;;


let pequal x y    = try VAL_bool(0 = Pervasives.compare x y) 
                    with Invalid_argument s -> VAL_undef("funct comparaison '"^s^"'");;
let pnotequal x y = try VAL_bool(0 <> Pervasives.compare x y)
                    with Invalid_argument s -> VAL_undef("funct comparaison '"^s^"'");;

(*
 * compareMgsVal     : retourne un VAL_int
 * compare_mgs_value : retourne  un entier
 * equalMgsVal       : retourne un VAL_bool
 * equal_mgs_value   : retourne un boolen
 * notequalMgsVal    : retourne un VAL_bool
 * notequal_mgs_value: retourne un boolen
 *
 *    Toutes ces fcts sauf la premire sont dfinies  partir de compareMgsVal, qui est une 
 *    fonction dispatche, dans dispatch_sig o leur dfinition est insres par un 
 *    post-code dfini dans gendispatch.post_sig
 *)
let _compareMgsVal_undef_undef (x, y) = mgszero;;
let _compareMgsVal_funct_funct (x, y) = mgsun;;
let _compareMgsVal_any_any (x, y) = VAL_int(Pervasives.compare x y);; (** par defaut c'est la comparaison systeme **)

(** les autres cas de compareMgsVal sont dans les fichiers
 ** correspondant  l'implmentation des collections. La fonction
 ** d'galit est definie  partir de la fonction de comparaison dans le
 ** post-sig de dispatch_sig.ml.
 **)




(** Comparaison des types mgs
 **)
let compare_type r r' = Pervasives.compare r.uid  r'.uid
;;
let compare_type2 r r' = 0 - (compare_type r r')
;;


let constructor_string = function
  | VAL_undef  _ -> "undef"
  | VAL_bint   _ -> "bint"
  | VAL_int    _ -> "int"
  | VAL_bint   _ -> "bint"
  | VAL_bool   _ -> "bool"
  | VAL_float  _ -> "float"
  | VAL_string _ -> "string"
  | VAL_symbol _ -> "symbol"
  | VAL_posgbf _ -> "posgbf"
  | VAL_acell  _ -> "acell"
  | VAL_gmap   _ -> "gmap"
  | VAL_ncell  _ -> "ncell"
  | VAL_funct  _ -> "funct"

  | VAL_rec _    -> "rec"
  | VAL_seq _    -> "seq"
  | VAL_ring _   -> "ring"
  | VAL_set _    -> "set" 
  | VAL_bag _    -> "bag" 
  | VAL_gab _    -> "gab" 
  | VAL_gbf _    -> "gbf" 
  | VAL_grf _    -> "grf" 
  | VAL_del _    -> "del" 
  | VAL_prox _   -> "prox" 
  | VAL_achain _ -> "achain" 
  | VAL_qmf _    -> "qmf" 

(**
  | _ -> "<unknown constructor: please update 'constructor_string' in valeur.ml>"
**)
;;



(* --- Predicat et observateur sur les collections ------------------------ *)

let is_collection = function
  | VAL_undef  _
  | VAL_int    _
  | VAL_bint   _
  | VAL_bool   _
  | VAL_float  _
  | VAL_string _
  | VAL_symbol _
  | VAL_posgbf _
  | VAL_acell  _
  | VAL_gmap   _
  | VAL_ncell  _      
  | VAL_funct  _ -> false
  | _            -> true
;;

let true_coll = function
  | VAL_rec _
  | VAL_gab _ -> false
  | x -> is_collection x
;;
let true_collection = true_coll
;;

let collection_type = function
  | VAL_rec  (_, t)    -> t
  | VAL_seq  (_, t)    -> t
  | VAL_ring (_, t)    -> t
  | VAL_set  (_, t)    -> t
  | VAL_bag  (_, t)    -> t
  | VAL_gab  (_, t)    -> t
  | VAL_gbf  (_, t)    -> t
  | VAL_grf  (_, t)    -> t
  | VAL_del  (_, t)    -> t
  | VAL_prox (_, t)    -> t
  | VAL_achain (_, t)  -> t
  | VAL_qmf  (_, t)    -> t
  | v                  -> 
      let mesg = Printf.sprintf "Error: internal: collection_type: value of type %s is not a collection" (constructor_string v)
      in prerr_endline mesg; raise (Erreur mesg)

and change_type t = function
  | VAL_rec  (v, _)   -> VAL_rec   (v, t)
  | VAL_seq  (v, _)   -> VAL_seq   (v, t)
  | VAL_ring (v, _)   -> VAL_ring  (v, t)
  | VAL_set  (v, _)   -> VAL_set   (v, t)
  | VAL_bag  (v, _)   -> VAL_bag   (v, t)
  | VAL_gab  (v, _)   -> VAL_gab   (v, t)
  | VAL_gbf  (v, _)   -> VAL_gbf   (v, t)
  | VAL_grf  (v, _)   -> VAL_grf   (v, t)
  | VAL_del  (v, _)   -> VAL_del   (v, t)
  | VAL_prox (v, _)   -> VAL_prox  (v, t)
  | VAL_achain(v, _)  -> VAL_achain(v, t)
  | VAL_qmf(v, _)     -> VAL_qmf   (v, t)
  | x -> 
      let err = "Error: collection "^t.name^": cannot be subtyped (change_type)\n"
      in prerr_endline err; raise (Erreur err)

and is_same_coll_kind x y = match x, y with
  | (VAL_rec  _),  (VAL_rec  _)
  | (VAL_seq  _),  (VAL_seq  _) 
  | (VAL_ring _),  (VAL_ring _) 
  | (VAL_set  _),  (VAL_set  _) 
  | (VAL_bag  _),  (VAL_bag  _) 
  | (VAL_gab  _),  (VAL_gab  _) 
  | (VAL_gbf  _),  (VAL_gbf  _) 
  | (VAL_grf  _),  (VAL_grf  _) 
  | (VAL_del  _),  (VAL_del  _) 
  | (VAL_prox  _), (VAL_prox  _) 
  | (VAL_achain _), (VAL_achain _) 
  | (VAL_qmf _),   (VAL_qmf _)   -> true
  | _, _                         -> false

let is_same_kind x y = match x, y with
  | (VAL_undef  _), (VAL_undef  _)
  | (VAL_int    _), (VAL_int    _)
  | (VAL_bint   _), (VAL_bint   _)
  | (VAL_bool   _), (VAL_bool   _)
  | (VAL_float  _), (VAL_float  _)
  | (VAL_string _), (VAL_string _)
  | (VAL_symbol _), (VAL_symbol _)
  | (VAL_posgbf _), (VAL_posgbf _)
  | (VAL_acell  _), (VAL_acell  _)
  | (VAL_gmap   _), (VAL_gmap   _)
  | (VAL_ncell  _), (VAL_ncell  _)
  | (VAL_funct  _), (VAL_funct  _)     -> true
  | _                                  -> is_same_coll_kind x y
;;


(** mise a jour dans printer.ml avec prsvalue **)
let (forward_prsvalue : (mgs_value -> string) ref) = ref (function _ -> "dummy_forward_prsvalue")
;;

let error_bad_type expected x =
  let err = "Error: expect a "^expected^" and got :\n\t"^(!forward_prsvalue x)^"\n"
  in prerr_string err; raise (Erreur err)
;;



let is_int    = function VAL_int(_)    -> true | _ -> false
and is_bint   = function VAL_bint(_)   -> true | _ -> false
and is_string = function VAL_string(_) -> true | _ -> false
and is_symbol = function VAL_symbol(_) -> true | _ -> false
and is_undef  = function VAL_undef(_)  -> true | _ -> false
and is_bool   = function VAL_bool(_)   -> true | _ -> false
and is_float  = function VAL_float(_)  -> true | _ -> false
and is_posgbf = function VAL_posgbf(_) -> true | _ -> false
and is_acell  = function VAL_acell(_)  -> true | _ -> false
and is_gmap   = function VAL_gmap(_)   -> true | _ -> false
and is_ncell  = function VAL_ncell(_)  -> true | _ -> false
and is_funct  = function VAL_funct(_)  -> true | _ -> false

and is_rec    = function VAL_rec(_)    -> true | _ -> false
and is_seq    = function VAL_seq(_)    -> true | _ -> false
and is_ring   = function VAL_ring(_)   -> true | _ -> false
and is_set    = function VAL_set(_)    -> true | _ -> false
and is_bag    = function VAL_bag(_)    -> true | _ -> false
and is_gab    = function VAL_gab(_)    -> true | _ -> false
and is_gbf    = function VAL_gbf(_)    -> true | _ -> false
and is_grf    = function VAL_grf(_)    -> true | _ -> false
and is_del    = function VAL_del(_)    -> true | _ -> false
and is_prox   = function VAL_prox(_)   -> true | _ -> false
and is_achain = function VAL_achain(_) -> true | _ -> false
and is_qmf    = function VAL_qmf(_)    -> true | _ -> false
;;


let to_int    = function VAL_int(x)    -> x | m -> error_bad_type "int" m
and to_bint   = function VAL_bint(x)   -> x | m -> error_bad_type "bint" m
and to_string = function VAL_string(x) -> x | m -> error_bad_type "string" m
and to_symbol = function VAL_symbol(x) -> x | m -> error_bad_type "symbol" m
and to_undef  = function VAL_undef(x)  -> x | m -> error_bad_type "undef" m
and to_bool   = function VAL_bool(x)   -> x | m -> error_bad_type "bool" m
and to_float  = function VAL_float(x)  -> x | m -> error_bad_type "float" m
and to_posgbf = function VAL_posgbf(x) -> x | m -> error_bad_type "posgbf" m
and to_acell  = function VAL_acell(x)  -> x | m -> error_bad_type "acell" m
and to_gmap   = function VAL_gmap(x)   -> x | m -> error_bad_type "gmap" m
and to_ncell  = function VAL_ncell(x)  -> x | m -> error_bad_type "ncell" m
and to_funct  = function VAL_funct(x)  -> x | m -> error_bad_type "funct" m

and to_rec    = function VAL_rec(x)    -> fst x | m -> error_bad_type "rec"  m
and to_seq    = function VAL_seq(x)    -> fst x | m -> error_bad_type "seq"  m
and to_ring   = function VAL_ring(x)   -> fst x | m -> error_bad_type "ring" m
and to_set    = function VAL_set(x)    -> fst x | m -> error_bad_type "set" m
and to_bag    = function VAL_bag(x)    -> fst x | m -> error_bad_type "bag" m
and to_gab    = function VAL_gab(x)    -> fst x | m -> error_bad_type "gab" m
and to_gbf    = function VAL_gbf(x)    -> fst x | m -> error_bad_type "gbf" m
and to_grf    = function VAL_grf(x)    -> fst x | m -> error_bad_type "grf" m
and to_del    = function VAL_del(x)    -> fst x | m -> error_bad_type "del" m
and to_prox   = function VAL_prox(x)   -> fst x | m -> error_bad_type "prox" m
and to_achain = function VAL_achain(x) -> fst x | m -> error_bad_type "achain" m
and to_qmf    = function VAL_qmf(x)    -> fst x | m -> error_bad_type "qmf" m
;;


(** cette fonction est redefini dans printer.ml avec un meilleur affichage d'erreur **)
let pf1 = to_funct;;

let app_2 f x y = (match (f x) with
		     | VAL_funct g -> (g y)
		     | _           -> raise (Erreur "function of two arguments is required"))
;;		    
let app_3 f x y z = (match app_2 f x y with
		       | VAL_funct g -> g z
		       | _           -> raise (Erreur "function of three arguments is required"))
;;



(* --- Quelques fcts de services pour wrapper des fct caml en fct MGS ------------------------------- *)

let fun_0 arg = VAL_funct (function _ -> arg())
and fun_1 arg = VAL_funct arg
and fun_2 arg = VAL_funct (fun x -> VAL_funct (fun y -> arg x y))
and fun_3 arg = VAL_funct (fun x -> VAL_funct (fun y -> VAL_funct (fun z -> arg x y z)))
and fun_4 arg = VAL_funct (fun w -> VAL_funct (fun x -> VAL_funct (fun y -> VAL_funct (fun z -> arg w x y z))))
and fun_5 arg = VAL_funct (fun a -> VAL_funct (fun w -> VAL_funct (fun x -> VAL_funct (fun y -> VAL_funct (fun z -> arg a w x y z)))))
and fun_6 arg = VAL_funct (fun a -> VAL_funct (fun b -> VAL_funct (fun w -> VAL_funct (fun x -> VAL_funct (fun y -> VAL_funct (fun z -> arg a b w x y z))))))
;;



(* --- Traitement des erreurs ----------------------------------- *)

let current_error_handler = 
  (* un handler frustre, car on a pas encore les affichages. 
   * Cf. default_error_handler dans environement.ml pour le handler d'erreur par dfaut
   *)
  ref (VAL_funct (function _ -> Printf.fprintf stderr "pre-init current_error_handler";  exit(1)))
;;


let raise_error msg =
  begin
    if false == !Global.silent_error then 
      begin
	Pervasives.flush_all();
	Printf.fprintf stderr "Error: %s\n" msg;
      end
  end;
  if !Global.abort_on_error then raise (Erreur msg)
  else (pf1 !current_error_handler) (VAL_string msg)
;;



let abort_evaluation = function 
  | VAL_string s -> raise (Erreur s)
  | _ -> raise (Erreur "User aborted evaluation")
;;


let top_level = ref (None : (mgs_value -> mgs_value) option)
;;

let set_top_level f = 
  let old = !top_level in
    begin
      match f with
        | VAL_funct f -> top_level := Some f;
	| x           -> 
	    top_level :=  None;
	    Printf.eprintf  
	      "\nError: parameter top_level expects a function\nThe top_level is now set to the default top_level\n"
    end;
    match old with
      | Some ff -> VAL_funct ff
      | None -> VAL_undef "default top level"
;;

(* ======================================================= *)
