(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_gab.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-24 05:23:50 $
 * Version: $Revision: 1.3 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *     La representation d'un multi-ensemble (ou gab)
 *     reprsentation adapte  partir des arbres binaires balances (cf. set)
 *     pour tenir compte de l'occurence de chauqe lments.
 *
 *=========================================================================
 *)


open Def_gab;;
open Valeur;;
open Dispatch_sig;;


Global.__register_version "$RCSfile: core_gab.ml,v $" "$Revision: 1.3 $" "$Date: 2006-04-24 05:23:50 $";;



let compare_elt = compare_mgs_value
;;

let empty = Def_gab.Empty_gab
;;


(* Gabs are represented by balanced binary trees (the heights of the
   children differ by at most 2 *)

let height_g = function
    Empty_gab -> 0
  | Node_gab(_, _, _, _, h) -> h

(* Creates a new node with left son l, value x and right son r.
   l and r must be balanced and | height l - height r | <= 2.
   Inline expansion of height for better speed. *)

let create_g l x n r =                                           (* 
								    au lieu d'appeler height il reecrit le code *)
  let hl = match l with Empty_gab -> 0 | Node_gab(_,_,_,_,h) -> h in
  let hr = match r with Empty_gab -> 0 | Node_gab(_,_,_,_,h) -> h in
    Node_gab(l, x, n, r, (if hl >= hr then hl + 1 else hr + 1))

(* Same as create_g, but performs one step of rebalancing if necessary.
   Assumes l and r balanced.
   Inline expansion of create for better speed in the most frequent case
   where no rebalancing is required. *)

let bal_g l x n r =
  let hl = match l with Empty_gab -> 0 | Node_gab(_,_,_,_,h) -> h in
  let hr = match r with Empty_gab -> 0 | Node_gab(_,_,_,_,h) -> h in
    if hl > hr + 2 then begin
      match l with
          Empty_gab -> invalid_arg "Set.bal"
        | Node_gab(ll, lv,ln, lr, _) ->
	    if height_g ll >= height_g lr then
              create_g ll lv ln (create_g lr x n r)
            else begin
              match lr with
                  Empty_gab -> invalid_arg "Set.bal"
		| Node_gab(lrl, lrv,lrn, lrr, _)->
		    create_g (create_g ll lv ln lrl) lrv lrn (create_g lrr x n r)
            end
    end else if hr > hl + 2 then begin
      match r with
          Empty_gab -> invalid_arg "Set.bal"
        | Node_gab(rl, rv,rn , rr, _) ->
	    if height_g rr >= height_g rl then
              create_g (create_g l x n rl) rv rn rr
            else begin
              match rl with
                  Empty_gab -> invalid_arg "Set.bal"
		| Node_gab(rll, rlv,rln, rlr, _) ->
		    create_g (create_g l x n rll) rlv rln (create_g rlr rv rn rr)
            end
    end else
      Node_gab(l, x,n, r, (if hl >= hr then hl + 1 else hr + 1))

(* Same as bal_g, but repeat rebalancing until the final result
   is balanced. *)

let rec join_g l x n r =
  match bal_g l x n r with
      Empty_gab -> invalid_arg "Set.join"
    | Node_gab(l', x',n', r', _) as t' ->
	let d = height_g l' - height_g r' in
          if d < -2 or d > 2 then join_g l' x' n' r' else t'

(* Merge two trees l and r into one.
   All elements of l must precede the elements of r.
   Assumes | height_g l - height_g r | <= 2. *)

let rec merge_g t1 t2 =
  match (t1, t2) with
      (Empty_gab, t) -> t
    | (t, Empty_gab) -> t
    | (Node_gab(l1, v1,n1, r1, h1), Node_gab(l2, v2,n2, r2, h2)) ->
	bal_g l1 v1 n1 (bal_g (merge_g r1 l2) v2 n2 r2)

(* Same as merge_g, but does not assume anything about l and r. *)

let rec concat_g t1 t2 =
  match (t1, t2) with
      (Empty_gab, t) -> t
    | (t, Empty_gab) -> t
    | (Node_gab(l1, v1,n1, r1, h1), Node_gab(l2, v2,n2, r2, h2)) ->
	join_g l1 v1 n1(join_g (concat_g r1 l2) v2 n2 r2)

(* Splitting attention peut-etre probleme *)

let rec split_g x = function
    Empty_gab ->
      (Empty_gab, None,0, Empty_gab)
  | Node_gab(l, v,n, r, _) ->
      let c = compare_elt x v in
        if c = 0 then (l, Some v,n, r)
        else if c < 0 then
          let (ll, vl,nl, rl) = split_g x l in (ll, vl,nl, join_g rl v 
						  n r)
        else
          let (lr, vr,nr, rr) = split_g x r in (join_g l v n lr, 
						vr,nr, rr)

(* Implementation of the gab operations *)

let empty_g = Empty_gab
;;

let is_empty_g = function Empty_gab -> true | _ -> false
;;

let rec mem_g x = function
    Empty_gab -> false
  | Node_gab(l, v,n, r, _) ->
      let c = compare_elt x v in
        c = 0 || mem_g x (if c < 0 then l else r)
;;

let rec num_g x = function
    Empty_gab -> 0
  | Node_gab(l, v,n, r, _) ->
      let c = compare_elt x v in
        if  c = 0 then n else num_g x (if c < 0 then l else r)
;;

let rec add_g x = function
    Empty_gab -> Node_gab(Empty_gab, x,1, Empty_gab, 1)
  | Node_gab(l, v, n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then (if n + 1 = 0 then merge_g l r else Node_gab(l, v, n+1,
							      r, h) ) else
          if c < 0 then bal_g (add_g x l) v n r else bal_g l v n (add_g x r)
;;

let rec addn_g x m= function
    Empty_gab -> Node_gab(Empty_gab, x, m, Empty_gab, 1)
  | Node_gab(l, v,n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then (if (n + m = 0) then merge_g l r else Node_gab(l, v, n+m,
								     r, h) ) else
          if c < 0 then bal_g (addn_g x m l) v n r else bal_g l v n 
	    (addn_g x m r)
;;

let rec remove_g x = function
    Empty_gab -> Node_gab(Empty_gab, x, -1, Empty_gab, 1)
  | Node_gab(l, v, n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then (if (n - 1 = 0) then merge_g l r else Node_gab(l, v,n-1, 
							     r, h) ) else
          if c < 0 then bal_g (remove_g x l) v n r else bal_g l v n 
	    (remove_g x r)
;;

let rec removen_g x m = function
    Empty_gab -> Node_gab(Empty_gab, x, -m, Empty_gab, 1)
  | Node_gab(l, v, n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then (if (n - m = 0) then merge_g l r else Node_gab(l, v,n-m, 
							     r, h) ) else
          if c < 0 then bal_g (removen_g x m l) v n r else bal_g l v n 
	    (removen_g x m r)
;;


let singleton_g x = Node_gab(Empty_gab, x,1, Empty_gab, 1)
;;

let rec union_g s1 s2 =
  match (s1, s2) with
      (Empty_gab, t2) -> t2
    | (t1, Empty_gab) -> t1
    | (Node_gab(l1, v1,n1, r1, h1), Node_gab(l2, v2,n2, r2, h2)) ->
	if h1 >= h2 then
          if h2 = 1 then addn_g v2 n2 s1 else begin
            let (l2, _, nb, r2) = split_g v1 s2 in
	      if (n1+nb = 0) then
		merge_g (union_g l1 l2) (union_g r1 r2)
	      else
		join_g (union_g l1 l2) v1 (n1+nb) (union_g r1 r2)
          end
        else
          if h1 = 1 then addn_g v1 n1 s2 else begin
            let (l1, _, nb, r1) = split_g v2 s1 in
	      if (n2+nb = 0) then
		merge_g (union_g l1 l2) (union_g r1 r2)
	      else
		join_g (union_g l1 l2) v2 (n2+nb) (union_g r1 r2)
          end
;;

let rec inter_g s1 s2 =
  match (s1, s2) with
      (Empty_gab, t2) -> Empty_gab
    | (t1, Empty_gab) -> Empty_gab
    | (Node_gab(l1, v1, n1, r1, _), t2) ->
	match split_g v1 t2 with
            (l2, None, _, r2) ->
	      concat_g (inter_g l1 l2) (inter_g r1 r2)
          | (l2, Some _, n2, r2) when (n1 * n2 < 0) ->
	      concat_g (inter_g l1 l2) (inter_g r1 r2)
          | (l2, Some _, n2, r2) when (n1 * n2 > 0) ->
	      join_g (inter_g l1 l2) v1 (if n1 < 0 then max n1 n2 else min n1 n2) (inter_g r1 r2)
;;

let rec diff_g s1 s2 =
  match (s1, s2) with
      (Empty_gab, t2) -> Empty_gab
    | (t1, Empty_gab) -> t1
    | (Node_gab(l1, v1, n1, r1, _), t2) ->
	match split_g v1 t2 with
            (l2, None, _, r2) ->
	      join_g (diff_g l1 l2) v1 n1 (diff_g r1 r2)
          | (l2, Some _, n2, r2) ->
	      if (n1 - n2 = 0) then
		concat_g (diff_g l1 l2) (diff_g r1 r2)
	      else
		join_g (diff_g l1 l2) v1 (n1-n2) (diff_g r1 r2)
;;

let rec mult_g_aux p = function
    Empty_gab -> Empty_gab
  | Node_gab(l, v, n, r, h) -> Node_gab(mult_g_aux p l, v, p*n, mult_g_aux p r, h)

and mult_g p g = if (p = 0) then Empty_gab else mult_g_aux p g ;;

let rec map_occ_g f = function
    Empty_gab -> Empty_gab
  | Node_gab(l, v, n, r, h) ->
      let n' = f n in
	if n' = 0 then
	  concat_g (map_occ_g f l) (map_occ_g f r)
	else
	  join_g (map_occ_g f l) v n' (map_occ_g f r)
;;

let rec compare_g s1 s2 =  compare_aux_g [s1] [s2]
			   
and compare_aux_g l1 l2 =
  match (l1, l2) with
      ([], []) -> 0
    | ([], _)  -> -1
    | (_, []) -> 1
    | (Empty_gab :: t1, Empty_gab :: t2) -> compare_aux_g t1 t2
    | (Node_gab(Empty_gab, v1,n1, r1, _) :: t1, Node_gab(Empty_gab, v2,n2, r2, _) :: t2) ->
	if n1 = n2 (** On fait le test numrique avant de tester les elts, c'est plus rapide **)
	then let c = compare_elt v1 v2 in if c <> 0 then c else compare_aux_g (r1::t1) (r2::t2)
	else n1-n2 
    | (Node_gab(l1, v1,n1, r1, _) :: t1, t2) ->
	compare_aux_g (l1 :: Node_gab(Empty_gab, v1,n1, r1, 0) :: t1) t2
    | (t1, Node_gab(l2, v2,n2, r2, _) :: t2) ->
	compare_aux_g t1 (l2 :: Node_gab(Empty_gab, v2,n2, r2, 0) :: t2)

let equal_g s1 s2 = (0 = compare_g s1 s2)

(*
let rec subset_g s1 s2 =
  match (s1, s2) with
      Empty_gab, _ ->
	true
    | _, Empty_gab ->
	false
    | Node_gab (l1, v1,n1, r1, _), (Node_gab (l2, v2,n2, r2, _) as t2) ->
	let c = compare_elt v1 v2 in
          if c = 0 then
            (n1<=n2) && subset_g l1 l2 && subset_g r1 r2
          else if c < 0 then
            subset_g (Node_gab (l1, v1,n1, Empty_gab, 0)) l2 && subset_g r1 t2
          else
            subset_g (Node_gab (Empty_gab, v1,n1, r1, 0)) r2 && subset_g l1 t2
;;*)


let rec iter_g f = function
    Empty_gab               -> ()
  | Node_gab(l, v, n, r, _) -> iter_g f l; f v; iter_g f r
;;

let rec iter_occurences_g f = function
    Empty_gab               -> ()
  | Node_gab(l, v, n, r, _) -> iter_occurences_g f l; f n v ; iter_occurences_g f r
;;

let rec fold_g f s accu =
  match s with
      Empty_gab -> accu
    | Node_gab(l, v, n, r, _) -> fold_g f l (f v (fold_g f r accu))
;;

let rec fold_occurences_g f s accu =
  match s with
      Empty_gab -> accu
    | Node_gab(l, v,n, r, _) -> fold_occurences_g f l (f n v (fold_occurences_g f r accu))
;;
 
let rec for_all_g p = function
    Empty_gab -> true
  | Node_gab(l, v, n, r, _) -> p v && for_all_g p l && for_all_g p r
;;

let rec exists_g p = function
    Empty_gab -> false
  | Node_gab(l, v, n, r, _) -> p v || exists_g p l || exists_g p r
;;

let filter_g p s =
  let rec filt accu = function
    | Empty_gab -> accu
    | Node_gab(l, v, n, r, _) when (n > 0) ->
	filt (filt (if p v then addn_g v n  accu else accu) l) r
  in
    filt Empty_gab s
;;

let partition_g p s =
  let rec part (t, f as accu) = function
    | Empty_gab -> accu
    | Node_gab(l, v, n, r, _) ->
	part (part (if p v then (addn_g v n t, f) else (t, addn_g v n f)) l) r
  in
    part (Empty_gab, Empty_gab) s
;;

let rec cardinal_g = function
    Empty_gab -> 0
  | Node_gab(l, v, n, r, _) -> cardinal_g l + n + cardinal_g r
;;

let rec set_cardinal_g = function
    Empty_gab -> 0
  | Node_gab(l, v,_, r, _) -> set_cardinal_g l + 1 + set_cardinal_g r
;;

let rec elements_aux_g accu = function
    Empty_gab -> accu
  | Node_gab(l, v, n, r, _) -> elements_aux_g (v :: (elements_aux_g accu r)) l
;;

let elements_g s = elements_aux_g [] s
;;

let rec elements_occ_aux_g accu = function
    Empty_gab -> accu
  | Node_gab(l, v, n, r, _) -> elements_occ_aux_g ((n,v) :: (elements_occ_aux_g accu r)) l
;;

let elements_occ_g s = elements_occ_aux_g [] s
;;

let rec min_elt_g = function
    Empty_gab -> raise Not_found
  | Node_gab(Empty_gab, v, n, r, _) -> v
  | Node_gab(l, v,n, r, _) -> min_elt_g l
;;

let rec max_elt_g = function
    Empty_gab -> raise Not_found
  | Node_gab(l, v, n, Empty_gab, _) -> v
  | Node_gab(l, v, n, r, _) -> max_elt_g r
;;

let choose_g = min_elt_g
;;



let _compareMgsVal_gab_gab ((s1, t1), (s2, t2)) = 
  let tdiff = compare_type t1 t2 
  in
    if tdiff <> 0 
    then VAL_int(tdiff) 
    else  VAL_int(compare_g s1 s2)
;;


