/*=======================================================-*-c++-*-======
 *
 * MGS : un modele general de systeme
 *
 * (U)TopoIa  : Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *            : Olivier Michel      : michel@lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: smith.C,v $ 
 * Path: %P%
 * Description:
 * Created:
 * Author: OM & JLG
 * Modified: $Date: 2004-12-14 14:58:24 $ at 
 * Version: $Revision: 1.1 $
 * Status: 
 * -------------------------------------------------------------------
 * Comments:  
 *
 *  Calcul de la forme normale de Smith et
 *  equivalence de mot dans un groupe abelien.
 *
 * Le calcul de la forme normale de Smith est la traduction en C++ de
 * l'algo utilise par le package smithform.m en Mathematica. Dans
 * UTopIa/MGS/Stuff-SmithNormalForm/Recode/, il y a la version C++ de
 * l'algo utilise dans Maple, et qui utilise les entiers en precisions
 * arbitraires (packetage gnu: gmp). Mais le probleme est que la
 * version Maple calcule les inverses des matrices dont j'ai besoin.
 *
 * ===================================================================
 */



#ifdef __GNUG__
#pragma implementation
#endif


# define  __USE_ISOC99  1 // Pour avoir les constantes LLONG_xxx
#include <limits.h>       // et pour recuperer LONG_MAX et LONG_MIN

#include <vector>

#include "smith.h"



// Si debug, recuperer la valeur dans limits.h, car ca veut pas definir
#ifdef DEBUG_SNF

# define UPPER   LLONG_MAX /2
# define LOWER   LLONG_MIN /2
# define check_sz(nat)  assert((LOWER < nat) && (nat < UPPER))

# else
# define check_sz(nat) 
# endif



namespace MGS {

# ifdef EXPORT_MGS_C_DEVELOPMENT

ostream& operator<<(ostream & o, const CoefVect& v) 
{ 
  o << "<" ;
  bool separe = false; 
  for (unsigned int i = 0; i < v.size(); i++) 
    { 
      if (separe) o << ", "; 
      o << v[i]; 
      separe = true; 
    } 
  return o << ">"; 
} 

#endif



template <class T>
class Matrix
{
   public:
      Matrix(unsigned int r, unsigned int c) 
	 : _row(r),  _col(c),  _data(new T[r*c])
      { 
	 for (unsigned int i = 0; i < r; i++)
	    for (unsigned int j = 0; j < c; j++)
	       elt(i, j) = 0;
      }


      Matrix(const Matrix<T>& m) 
	 : _row(m.row()), _col(m.col()), _data(new T[m.row() * m.col()])
      {
	 assert(data() != NULL);	 
	 (*this) = m;
      }

      Matrix(unsigned int n)
	 : _row(n),  _col(n),  _data(new T[n*n])
      {
	 for (unsigned int i = 0; i < n; i++)
	    for (unsigned int j = 0; j < n; j++)
	       if (i == j) 
		  elt(i, j) = 1;
	       else
		  elt(i, j) = 0;
      }


      ~Matrix() { delete [] _data; }


      unsigned int row() const { return _row; }
      unsigned int col() const { return _col; }  


      T& elt(unsigned int i, unsigned int j) const
      { 
	 assert((i < row()) && (j < col()));
	 check_sz(_data[i * col() + j]);
	 return _data[i * col() + j]; 
      }

      T& operator()(unsigned int i, unsigned int j) const
      { return elt(i, j); }

      void alpha_ligne(unsigned int i, T alpha)
      {  
	 assert(i < row());
	 for(unsigned int k=0; k<col(); k++) 
	    elt(i, k) = alpha * elt(i, k);
      }

      void alpha_col(unsigned int i, T alpha)
      {     
	 assert(i < col());
	 for(unsigned int k=0; k<row(); k++)
	    elt(k, i) = alpha * elt(k, i);
      }

      void add_alpha_ligne(T alpha, unsigned int i, unsigned int j)
      {     
	 assert((j < row()) && (i < col()));
	 for(unsigned int k=0; k<col(); k++)
	    elt(j, k) = elt(j, k) + alpha * elt(i, k);
      }

      void add_alpha_col(T alpha, unsigned int i, unsigned int j)
      {     
	 assert((j < col()) && (i < col()));
	 for(unsigned int k=0; k<row(); k++)
	    elt(k, j) = elt(k, j) + alpha * elt(k, i);

      }

      void swap_ligne(unsigned int i, unsigned int j)
      {  
	 assert((j < row()) && (i < row()));
	 for(unsigned int k=0; k<col(); k++)
	 {
	    T tmp = elt(i, k);
	    elt(i, k) = elt(j, k);
	    elt(j, k) = tmp;
	 }
      }


      void swap_col(unsigned int i, unsigned int j)
      {  
	 assert((j < col()) && (i < col()));
	 for(unsigned int k=0; k<row(); k++)
	 {
	    T tmp = elt(k, i);
	    elt(k, i) = elt(k, j);
	    elt(k, j) = tmp;
	 }
      }

      void operator=(const Matrix<T>& m)
      {
	 assert((row() == m.row()) && (col() == m.col()));
    
	 for(unsigned int i = 0; i < row(); i++)
	    for(unsigned int j = 0; j < col(); j++)
	       elt(i, j) = m(i, j);
      }

    
      void print(ostream& os) const
      {
	 assert(col() > 0);
	 unsigned int tmp_col = col() - 1;
	 
	 os << "Matrix<\n";	 
	 for(unsigned int i = 0; i < row(); i++)
	 {
	    os << "\t";

	    for(unsigned int j = 0; j < tmp_col; j++)
	       os << elt(i, j) << '\t';
	    
	    os << "\t" << elt(i, tmp_col) << endl;	
	 }
	 os << ">";
      }


   private:
      unsigned int  _row;
      unsigned int  _col;
      T* _data;
    
      const T* data() const { return _data; }
};


template <class T>
ostream& operator<< (ostream& os, const Matrix<T>& m)
{
   m.print(os);
   return os;
}


// 
// x & y    = arguments
// u & v    = couple de Bezout
// resultat = pgcd
//

template <class T>
T pgcd(T x, T y, T& u, T& v)
{
   if (!x)
   {
      // A VERIFIER: je veux rendre g, u et v tel que:
      //    g = u.x + v.y
      // Comme x == 0, je prend u = 0, g = abs(y) et v = sgn(y)
      // Ca verifie la propriete et ca a l'air en accord avec le
      // traitement de Mathematica
      assert(y != 0);

      v = (y < 0 ? -1 : 1);
      u = 0;
      return abs(y);
   }

   if (!y)
   {
      u = (x < 0 ? -1 : 1);
      v = 0;
     return abs(x);
   }



   T q;

   u  = 1;
   T u1 = 0;
    
   v  = 0;
   T v1 = 1;
    
   T r  = abs(x);
   T r1 = abs(y);
   T sgn_x = (x < 0 ? -1 : 1);
   T sgn_y = (y < 0 ? -1 : 1);
    
   if (r == r1)
   {
      u = sgn_x;
      v = 0;
      return r;
   }


   T tmp;
    
   while(r1 != 0)
   {
      q = r/r1;

      tmp = u;
      u = u1;          
      u1 = tmp - q*u1; 

      tmp = v;
      v = v1;          
      v1 = tmp - q*v1; 

      tmp = r;
      r = r1;          
      r1 = tmp - q*r1; 
   }
    
   u = u*sgn_x;
   v = v*sgn_y;

   q = x*u + y*v;  

   // Introduit pour avoir le meme couple de Bezout que Mathematica.
   assert(q >= 0);
   if (q < 0)
   {
      cerr << "q negatif" << endl;
      q = - q;
      u = - u - y/q;
      v = - v + x/q;
   }

   return q;
}

namespace{ // --- test de la version etendue du pgcd ------------------
struct TestPgcd
{
      int test(int x, int y, int& u, int& v)
      {
	 int q = pgcd(x, y, u, v);
	 assert(q == x*u + y*v);
	 return q;
      }

      TestPgcd() {
	 int u, v;
	 int q;

	 q = test(1, 1, u, v); assert(q == 1 && u == 1 && v == 0);
	 q = test(13, 13, u, v); assert(q == 13 && u == 1 && v == 0);
	 q = test(-12, 12, u, v); assert(q == 12 && u == -1 && v == 0);
	 q = test(11, -11, u, v); assert(q == 11 && u == 1 && v == 0);
	 q = test(-7, -7, u, v); assert(q == 7 && u == -1 && v == 0);

	 q = test(5, 3, u, v); assert(q == 1 && u == -1 && v == 2);
	 q = test(3, 5, u, v); assert(q == 1 && u == 2  && v == -1);
	 q = test(2, 4, u, v); assert(q == 2 && u == 1  && v == 0);
	 q = test(4, 2, u, v); assert(q == 2 && u == 0  && v == 1);
	 q = test(4, 0, u, v); assert(q == 4 && u == 1  && v == 0);

	 q = test(-5, 3, u, v); assert(q == 1 && u == 1  && v == 2);
	 q = test(-3, 5, u, v); assert(q == 1 && u == -2 && v == -1);
	 q = test(-2, 4, u, v); assert(q == 2 && u == -1 && v == 0);
	 q = test(-4, 2, u, v); assert(q == 2 && u == 0  && v == 1);
	 q = test(-4, 0, u, v); assert(q == 4 && u == -1 && v == 0);

	 q = test(5, -3, u, v); assert(q == 1 && u == -1 && v == -2);
	 q = test(3, -5, u, v); assert(q == 1 && u == 2  && v == 1);
	 q = test(2, -4, u, v); assert(q == 2 && u == 1  && v == 0);
	 q = test(4, -2, u, v); assert(q == 2 && u == 0  && v == -1);

	 q = test(-5, -3, u, v); assert(q == 1 && u == 1  && v == -2);
	 q = test(-3, -5, u, v); assert(q == 1 && u == -2 && v == 1);
	 q = test(-2, -4, u, v); assert(q == 2 && u == -1 && v == 0);
	 q = test(-4, -2, u, v); assert(q == 2 && u == 0  && v == -1);

	 q = test(5,  3*5,  u, v); assert(q == 5 && u == 1  && v == 0);
	 q = test(-5, 3*5,  u, v); assert(q == 5 && u == -1 && v == 0);
	 q = test(5,  3*5,  u, v); assert(q == 5 && u == 1  && v == 0);
	 q = test(-5, -3*5, u, v); assert(q == 5 && u == -1 && v == 0);
      }
} one;

}



template <class T>
void lig4(unsigned int i, unsigned int j, T x, T y, Matrix<T>& a)
{
   T  u, v;
  
   T d = pgcd(x, y, u, v);
  
   for(unsigned int k=0; k< a.col(); k++)
   {
      T tmp = a(i, k);
      a(i, k) = u*a(i, k) + v*a(j, k);
      a(j, k) = -(y/d)*tmp + (x/d)*a(j, k);
   }
}



template <class T>
void col4(unsigned int i, unsigned int j, T x, T y, Matrix<T>& a)
{
   T u, v;
  
   T d = pgcd(x, y, u, v);
  
   for(unsigned int k=0; k < a.row(); k++)
   {
      T tmp = a(k, i);
      a(k, i) = u*a(k, i) + v*a(k, j);
      a(k, j) = -(y/d)*tmp + (x/d)*a(k, j);
   }
}


template <class T>
void elimine_col(unsigned int col, Matrix<T>& a, Matrix<T>& q)
{
   for(unsigned int k=col+1; k<a.col(); k++)
   {
      if (a(col, k) != 0)
      {
	 T x = a(col, col); 
	 T y = a(col, k);
	 col4(col, k, x, y, a);
	 col4(col, k, x, y, q);
      }
   }
}


template <class T>
void elimine_lig(unsigned int lig, Matrix<T>& a, Matrix<T>& p)
{
   for(unsigned int k=lig+1; k<a.row(); k++)
   {
      if (a(k, lig) != 0)
      {
	 T x = a(lig, lig); 
	 T y = a(k, lig);
	 lig4(lig, k, x, y, a);
	 lig4(lig, k, x, y, p);
      }
   }
}


template <class T>
T modulo(T x, T y) { return (y == 0 ? x : x % y); }




template <class T>
void etape3(unsigned int m, Matrix<T>& a, Matrix<T>& p, Matrix<T>& q)
{
   bool chgQ = false;
    
   for(unsigned int i=m+1; i<a.row(); i++)
   {
      for(unsigned int j=m+1; j<a.col(); j++)
      {
	 if (modulo(a(i, j), a(m, m)) != 0)
	 {
	    a.add_alpha_ligne(T(1), i, m); 
	    p.add_alpha_ligne(T(1), i, m);
	    elimine_col(m, a, q);
	    chgQ = true;
	    break;
	 }
      }
   }
    
   if (chgQ) 
      elimine_lig(m, a, p);
}




template <class T>
bool check_smith(Matrix<T>& S,  Matrix<T>& L,  Matrix<T>& B,  Matrix<T>& R)
{
   unsigned int n = min(S.row(), S.col());
   for (unsigned int i = 0; i < n; i++)
   {
      if (S(i,i) < 0) 
      {
	 cerr << "Error: internal: forme de Smith pas positive" << endl;
	 return false;
      }
   }

   for (unsigned int i = 0; i < S.row(); i++)
      for (unsigned int j = 0; j < S.col(); j++)
      {
	 if ((i != j) && (S(i, j) != 0))
	 {
	    cerr << "Error: internal: forme de Smith"
		 << "non-diagonale pas nulle " << S(i, j) << endl;
	    return false;
	 }
      }


   // On verifie la multiplication:

   Matrix<T> r(L.row(), B.col());

   for (unsigned int i = 0; i < L.row(); i++)
      for (unsigned int j = 0; j < B.col(); j++)
      {
	 r(i, j) = 0;
	 for (unsigned int k = 0; k < B.row(); k++)
	 {
	    r(i, j) += L(i, k) * B(k, j);
	 }
      }



   Matrix<T> t(r.row(), R.col());

   for (unsigned int i = 0; i < r.row(); i++)
      for (unsigned int j = 0; j < R.col(); j++)
      {
	 t(i, j) = 0;
	 for (unsigned int k = 0; k < R.row(); k++)
	 {
	    t(i, j) += r(i, k) * R(k, j);
	 }
      }

   assert(t.row() == S.row());
   assert(t.col() == S.col());

   for (unsigned int i = 0; i < S.row(); i++)
      for (unsigned int j = 0; j < S.col(); j++)
      {
	 if (S(i,j) != t(i, j))
	 {
	    cerr << "Error: internal: forme de Smith: "
		 << "multiplication ne redonne pas S" << endl
		 << t << endl;
	    return false;
	 }
      }

   return true;
}



template <class T>
Matrix<T>* fns(Matrix<T>*& L, Matrix<T>& B, Matrix<T>*& R)
{
   // On cree les matrices qui contiendront le resultat.
   L = new Matrix<T>(B.row());
   Matrix<T>& p = *L;

   R = new Matrix<T>(B.col());
   Matrix<T>& q = *R;

   Matrix<T>* AA = new Matrix<T>(B);
   Matrix<T>& a = *AA;

   // La grande boucle

   for(unsigned int m = 0; (m<a.row()) && (m<a.col()); m++)
   {
      T pivot = abs(a(m, m));

      unsigned int x = m; 
      unsigned int y = m;
      bool zero = true;
   
      for(unsigned int i = m; i < a.row(); i++)
      {
	 for(unsigned int j = m; j < a.col(); j++)
	 {
	    if (a(i, j)) 
	    {
	       zero = false;
		    
	       if (abs(a(i, j)) < pivot) { x=i;y=j; pivot = abs(a(i, j)); }
	    }
	 }
      }
	

      if (zero) break;
   
      if (x!=m) { a.swap_ligne(m, x); p.swap_ligne(m, x); }
      if (y!=m) { a.swap_col(m, y);   q.swap_col(m, y); }
   
      if (a(m, m) < 0) 
      {
	 a.alpha_ligne(m, (T)(-1)); 
	 p.alpha_ligne(m, (T)(-1)); 
      }


      if ((m < (a.row() - 1)) && (m < (a.col() - 1)))
      {
	 bool nzerol = true;
	 bool nzeroc = true;

	 for (;;)
	 {
	    // calcul de la condition du while mma
	    {
	       for(unsigned int i=m+1; i<a.row(); i++)
		  if (a(i, m)) 
		  {
		     nzerol = true;
		     break;
		  }

	       if (!nzerol)
	       {
		  for(unsigned int i=m+1; i<a.col(); i++)
		     if (a(m, i))
		     {
			nzeroc = true;
			break;
		     }
	       }	       
	    }

	    // corps du while en mma
	    if (nzerol || nzeroc)
	    {
	       if (nzerol) elimine_lig(m, a, p);
	       if (nzeroc) elimine_col(m, a, q);
	       nzerol = false;
	       nzeroc = false;
	    }
	    else break;
	 }


	 etape3(m, a, p, q);
      }
      else 
      {
	 if (a.row() < a.col()) 
	    elimine_col(m, a, q); 
	 else 
	    elimine_lig(m, a, p);
      }
   }

   // On verifie que c'est OK:
   assert(check_smith(a, p, B, q));

   return AA;
}


// =======================================================
// Calcul d'un inverse


template <class T>
bool check_inverse(Matrix<T>& inv,  Matrix<T>& m)
{
   Matrix<T> r(inv.row(), m.col());

   for (unsigned int i = 0; i < inv.row(); i++)
      for (unsigned int j = 0; j < m.col(); j++)
      {
	 T r = 0;
	 for (unsigned int k = 0; k < m.row(); k++)
	 {
	    r += inv(i, k) * m(k, j);
	 }

	 if (i == j)
	 {
	    if (r != 1) return false;
	 }
	 else
	 {
	    if (r != 0) return false;
	 }
      }

   return true;
}

// ------------------------------------------------------- 
// Remplissage d'une sous-matrice auquelle on enleve 
// la colone i et la ligne j


template <class T>
void subminor(Matrix<T>& m, 
	   unsigned int xi, unsigned int xj,
	   Matrix<T>& M)
{
   unsigned int ii = 0;
   for (unsigned int i = 0; i < M.row(); i++)
   {
      if (i != xi)
      {
	 unsigned int jj = 0;
	 for (unsigned int j = 0; j < M.col(); j++)
	 {
	    if (j != xj)
	    {
	       m(ii, jj) = M(i, j);
	       jj += 1;
	    }
	 }

	 ii += 1;
      }
   }
}


// -------------------------------------------------------
// Calcul du dterminant par une mthode brutale

template <class T>
T determinant(Matrix<T>& M)
{
   assert(M.row() != 0);

   if (1 == M.row())
      return M(0, 0);

   T det = 0;
   Matrix<T> m(M.row()-1, M.col()-1);

   for (unsigned int i = 0; i < M.row(); i++)
   {
      if (0 == M(i, 0)) continue;

      subminor(m, i, 0, M);
      if (i % 2)
	 det -= M(i, 0)*determinant(m);
      else
	 det += M(i, 0)*determinant(m);
   }

   return det;
}


// -------------------------------------------------------
// Inversion de matrice par une mthode brutale

template <class T>
void inverse(Matrix<T>& res, Matrix<T>& M)
{
   assert(M.row() != 0);
   assert(M.row() == M.col());

   if (M.row() == 1)
   {
      assert(1 == abs(M(0,0)));
      res = M;
      return;
   }

   T det = determinant(M);
   
   assert(det != 0);
   assert(M.row() && M.col());
   
   Matrix<T> m(M.row()-1, M.col()-1);

   for (unsigned int i = 0; i < M.row(); i++)
   {
      for (unsigned int j = 0; j < M.col(); j++)
      {
	 subminor(m, i, j, M);

	 T tmp = determinant(m);

	 // c'est le cas pour les matrices unimodulaires
	 assert(0 == tmp % det); 

	 if ((i+j) % 2)
	    res(j, i) = - tmp/det;
	 else
	    res(j, i) = tmp/det;
      } 
   }

   // On verifie que c'est OK:
   assert(check_inverse(res, M));
}



// =======================================================





inline unsigned int min(unsigned int a, unsigned int b)
{ return (a < b ? a : b); }


SmithData::~SmithData() { delete L; delete L_1; }

SmithData::SmithData(String nm, 
		     const StringVect& gen,
		     const Equations& eqns,
		     bool warnings)
   : L(NULL),
     L_1(NULL),
     image(gen.size()),
     ligne(gen.size()),
     colone(eqns.size()),
     rank0(0),
     rankt(0),
     name(nm),
     warn(warnings)
{
   // S'il y a pas d'equation, y a rien a faire, le groupe est libre
   if (0 == colone)
   {
      rank0 = gen.size();
      rankt = 0;
      // image contient que des 0, ce qui veut dire que tous les coefs
      // sont pertinents. Donc tout va bien.
      return;
   }

   // On commence par construire la matrice des equations
   Matrix<LongInteger> A(ligne, colone);

   for(unsigned int i = 0; i < ligne; i++)
      for(unsigned int j = 0; j < colone; j++)
	 A(i, j) = eqns[j][i];

   Matrix<LongInteger>* left;
   Matrix<LongInteger>* right;
   Matrix<LongInteger>* diag;

   diag = fns(left, A, right);


   // creation du vecteur de torsion.
   make_image(diag);
   
   if (warn)
      cerr << "\n// Normalized form of gbf  " << name 
	   << "  has " << rank0 << " free generator"
	   << (rank0 > 1 ? "s\n" : "\n")
	   << "// and " << rankt << " cyclic generator" 
	   << (rankt > 1 ? "s: " : ": ")
	   << image
	   << endl;

   // creation de la matrice de transformation
   make_left(left);

   // On peut a present liberer les matrices intermediaires
   delete left;
   delete right;
   delete diag;
}



// On remplie le slot image qui contient les coefficients de torsion
// du groupe
void SmithData::make_image(Matrix<LongInteger>* diag)
{
   // On verifie en passant si tous les elements sont a 0 ou a 1
   // et on compte le nombre d'entree differentes de 1 et de 0
   bool all_zero = true;
   bool all_one = true;
   int rk0 = 0;
   int rkt = 0;

   // On parcours la diagonale
   for(unsigned int i = 0; i < min(ligne, colone); i++)
   {
      image[i] = convert((*diag)(i, i));
      if (image[i] == 0) rk0 += 1;
      else
      {
	 all_zero = false;
	 if (image[i] != 1) 
	 {
	    all_one = false;
	    rkt += 1;
	 }
      }
   }

   // La taille de image est le nombre de genrateurs. 
   // Mais la diagonale est peut etre plus courte: le reste c'est des 0
      
   for(unsigned int i = min(ligne, colone); i < ligne; i++)
   {
      image[i] = 0;
      all_one = false;
      rk0 += 1;
   }

   if (all_zero)
   {
      // Le groupe est libre, malgre les equations
      if (warn)
	 cerr << "Warning: GBF " << name 
	      << " is free (equations do not add constraints)" << endl;

      colone = 0;
   }
   else if (all_one)
   {
      // Le groupe est trivial
      if (warn)
	 cerr << "Warning: GBF " << name 
	      << " is trivial" << endl;
   }

   rank0 = rk0;
   rankt = rkt;
}  



void SmithData::make_left(Matrix<LongInteger>* left)
{
   // On stocke simplement la matrice de transformation 
   // sous une forme plus compacte

   L = new Matrix<Integer>(left->row(), left->col());

   for (unsigned int i = 0; i < left->row(); i++)
      for (unsigned int j = 0; j < left->col(); j++)
      {
	 (*L)(i, j) = convert((*left)(i, j));
      }

//   cerr << "## Matrice L == " << *L << endl;

   // On construit l'inverse de L:
   L_1 = new Matrix<Integer>(L->col(), L->row());
   inverse(*L_1, *L);

//   cerr << "## L^(-1) == " << *L_1 << endl;
}



// Normalisation d'une position

void SmithData::normalise(CoefVect& v) const
{
   // groupe libre
   if (0 == colone) return;

   assert(ligne == v.size());

   // groupe pas libre

   CoefVect res(ligne, 0);

   for (unsigned int i = 0; i < ligne; i++)
   {
      if (1 == image[i])
      {
	 res[i] = 0;
      }
      else
      {
	 int e = 0;
	 for (unsigned int j = 0; j < ligne; j++)
	 {
	    e += (*L)(i, j) * v[j];
	 }
	 
	 if (image[i]) 
	 {
	    // C'est une coordonee dans Z/image[i]Z : on la traite
	    // modulo. Faut faire gaffe au signe: on veut normaliser
	    // la coordonnee par un entier positif entre 0 inclus et
	    // image[i] exclus

	    if (e < 0)
	       res[i] = image[i] + e % image[i];
	    else
	       res[i] = e % image[i];
	 }
	 else
	    res[i] = e;
      }
   }

   v = res;
}




// Verifie qu'une position a l'air d'etre normalisee

bool SmithData::check_normal(const CoefVect& v) const
{
   // groupe libre
   if (0 == colone) return true;

   assert(ligne == v.size());

   for (unsigned int i = 0; i < ligne; i++)
   {
      if (1 == image[i])
      {
	 if (v[i] != 0) return false;
      }
      else
      {
	 if (image[i]) 
	 {
	    // C'est une coordonee dans Z/image[i]Z : on la traite
	    // modulo et la normalisation doit rendre un entier
	    // positif entre 0 inclus et image[i] exclus

	    if ((v[i] < 0) || (v[i] >= image[i]))
	       return false;
	 }
      }
   }

   return true;
}



// Inverse de la normalisation: retour d'une position dans le groupe
// de depart, mais faut appliquer ensuite le modulo du groupe de
// depart, si celui-ci n'est pas libre, afin de respecter la
// representation normale de Z/pZ par un element de Z.  
// Comme on ne connait pas le groupe de depart, on ne peut pas faire
// directement ici cette modularisation.

void SmithData::back(CoefVect& v) const
{
   if (colone == 0) 
      return;

   CoefVect res(ligne, 0);

   for (unsigned int i = 0; i < ligne; i++)
   {
      int e = 0;
      for (unsigned int j = 0; j < ligne; j++)
      {
	 e += (*L_1)(i, j) * v[j];
      }
	 
      res[i] = e;
   }

   v = res;
}



// une position est un vecteur de Z^n, alors qu'on a en fait 
// des elements de Z/pZ x ... Z/qZ. Cette fonction normalise 
// la position de Z^n en fonction des coefficients p, ..., q.
void SmithData::modulo(CoefVect& v) const
{
   // groupe libre
   if (0 == colone) return;

   assert(ligne == v.size());

   for (unsigned int i = 0; i < ligne; i++)
   {
      if (1 == image[i])
      {
	 v[i] = 0;
      }
      else
      {
	 if (image[i]) 
	 {
	    // C'est une coordonee dans Z/image[i]Z : on la traite
	    // modulo et la normalisation doit rendre un entier
	    // positif entre 0 inclus et image[i] exclus
	    if (v[i] < 0)
		  v[i] = (image[i] + (v[i] % image[i])) %  image[i];
	    else
		  v[i] = v[i] % image[i];
	 }
      }
   }
}




// calcule pos += tr
// Les arguments sont normalisees et la reponse est normalisee. 

void SmithData::inc(CoefVect& pos, int alpha, const CoefVect& tr) const
{
   if (!alpha) return;

   for (unsigned int i = 0; i < ligne; i++)
   {
      switch(image[i])
      {
	 case 0:
	    pos[i] += alpha * tr[i];
	    break;

	 case 1:
	    break;

	 default:
	    pos[i] += alpha * tr[i];
	    // Si on passe les bornes, on recadre
	    if (pos[i] >= image[i]) 
	       pos[i] = pos[i] % image[i];
	    else if (pos[i] < 0)
	    {
	       pos[i] = image[i] + (pos[i] % image[i]);
	    }
      }
   }
}



// Le nieme coef non nul

int SmithData::nth_coeff(int n) const
{
   assert(n>0);

   int k = 0;
   for (unsigned int i = 0; i < ligne; i++)
   {
      if (image[i] != 1) k += 1;
      if (k == n) return i;
   }

   return -1;
}



namespace { // -- local ---

struct ComputeHash2Short { short s1; short s2; };
struct ComputeHash4Char  { char c1; char c2; char c3; char c4; };

static union
{
      size_t retour;
      int h1;
      ComputeHash2Short h2;
      ComputeHash4Char h4;
} hfield;

struct CheckSize {
      CheckSize() { assert(sizeof(size_t) == sizeof(hfield)); }
} one_check_size;

} // -- fin du namespace local ---




// On supose dans la fonction de hachage que l'argument 
// est deja normalise.

size_t SmithData::hash(const CoefVect& p) const
{
   assert(p.size() == ligne);

   switch(ligne)
   {
      case 0:
	 cerr << "Error: ? internal: gbf with 0 generators" << endl;
	 assert(0);

      case 1:
	 hfield.h1 = p[0];
	 return hfield.retour;

      case 2:
	 hfield.h2.s1 = (short)p[0];
	 hfield.h2.s2 = (short)p[1];
	 return hfield.retour;

      case 3:
	 return (size_t)(p[0] + 73*p[1] + 337*p[2]);

      case 4:
	 hfield.h4.c1 = (char)(p[0] % 256);
	 hfield.h4.c2 = (char)(p[1] % 256);
	 hfield.h4.c3 = (char)(p[2] % 256);
	 hfield.h4.c4 = (char)(p[3] % 256);
	 return hfield.retour;

      default:
      {
	 size_t ret = 0;
	 for (unsigned int i = 0; i < ligne; i++)
	    ret = 7*ret + (size_t)(p[i]);
	 return ret;
      }
   }
   assert(0);
   return 0;
}

// =======================================================

namespace { // --- test fns ------------------------------------------

void test_smith_1()
{
   Matrix<long> A(3, 3);
   Matrix<long>* L;
   Matrix<long>* R;
   Matrix<long>* B;

   A(0,0) = 9;    A(1,0) = -36;   A(2,0) = 30; 
   A(0,1) = -36;  A(1,1) = 192;   A(2,1) = -180; 
   A(0,2) = 30;   A(1,2) = -180;  A(2,2) = 180; 

   cerr << "Matrice A " << A << endl;

   B=fns(L,A,R);

   cerr << "Matrice L = " << *L << endl
	<< "Matrice R = " << *R << endl
	<< "Matrice B = " << *B << endl;
   assert( 3 == (*B)(0,0) && 12 == (*B)(1,1) && 60 == (*B)(2, 2));
   assert(-3 == (*L)(0,0) &&  0 == (*L)(0,1) &&  1 == (*L)(0,2));
   assert( 6 == (*L)(1,0) && 20 == (*L)(2,0) && 12 == (*L)(2,2));
}


void test_smith_2()
{
   Matrix<long> A(6, 4);
   Matrix<long>* L;
   Matrix<long>* R;
   Matrix<long>* B;

   A(0,0) = -1; A(0,1) = 1; A(0,2) = 0; A(0,3)=0;
   A(1,0) = 0;  A(1,1) = 1; A(1,2) = 0; A(1,3)=0;
   A(2,0) = 1;  A(2,1) = 0; A(2,2) = 1; A(2,3)=0;
   A(3,0) = 0;  A(3,1) = 0; A(3,2) = 1; A(3,3)=0;
   A(4,0) = 1;  A(4,1) = 0; A(4,2) = 0; A(4,3)=1;
   A(5,0) = 0;  A(5,1) = 0; A(5,2) = 0; A(5,3)=1;

   cerr << "Matrice A " << A << endl;

   B=fns(L,A,R);

   cerr << "Matrice L = " << *L << endl
	<< "Matrice R = " << *R << endl
	<< "Matrice B = " << *B << endl;

   assert(1 == (*B)(0,0) && 1 == (*B)(1,1) && 1 == (*B)(2,2) 
	  && 1 == (*B)(3,3));

   Matrix<long> s(6, 6);
   
   s(0,0) = -1; s(0,1) = 0;  s(0,2) = 0;  s(0,3)=0; s(0,4) = 0;  s(0,5) = 0;
   s(1,0) = 0;  s(1,1) = 1;  s(1,2) = 0;  s(1,3)=0; s(1,4) = 0;  s(1,5) = 0;
   s(2,0) = 1;  s(2,1) = -1; s(2,2) = 1;  s(2,3)=0; s(2,4) = 0;  s(2,5) = 0;
   s(3,0) = 1;  s(3,1) = -1; s(3,2) = 0;  s(3,3)=0; s(3,4) = 1;  s(3,5) = 0;
   s(4,0) = -1; s(4,1) = 1;  s(4,2) = -1; s(4,3)=1; s(4,4) = 0;  s(4,5) = 0;
   s(5,0) = -1; s(5,1) = 1;  s(5,2) = 0;  s(5,3)=0; s(5,4) = -1; s(5,5) = 1;

   cerr << "## res = " << s << endl;

   for (unsigned int i = 0; i < 5; i++)
      for (unsigned int j = 0; j < 5; j++)
      {
	 if ((*L)(i,j) != s(i,j)) 
	    cerr << "## i=" << i << ", j=" << j
		 << " L(i,j)=" << (*L)(i,j) << " et s(i,j)=" << s(i,j) 
		 <<  endl;
// la ligne 4 est la meme au signe pres... donc c'est OK
//	 assert((*L)(i,j) == s(i,j));
      }
}


struct Test
{
      Test() {
//	 test_smith_1();
//	 test_smith_2();
      }

} oneTest;

} // -- fin test fns ---



} // Fin de fichier -------------------------------------------------------

