(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: del.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.3 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    sequences.
 *
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> seq (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = gbf
 *
 *=========================================================================
 *)


open Global;;
open Valeur;;
open Type;;
open Auxfct;;
open Dispatch_sig;;
open Type_del ;;
open Core_del ;;
open Def_del ;;
open Printer;;


let seq_desc = Type_seq.seq_desc
and mfst (x, _, _) = x
and msnd (_, x, _) = x
;;


let rec
    _join_del_del (((tab, _), c), ((tab', _), c')) = (let dim, _, vor1  = get_defs_from_delaunay c.name
						      and dim', _, vor2 = get_defs_from_delaunay c'.name  
						      in
							if ((dim = dim') & (vor1 = vor2))
							then (let sommets    = Array.map deco tab
							      and sommets'   = Array.map deco tab'
							      and positions  = Array.map pos tab	
							      and positions' = Array.map pos tab' in
							      let all_sommets   = Array.append sommets sommets'
							      and all_positions = Array.append positions positions' 
							      in
								delaunay_from_all_sommets_positions dim all_sommets all_positions c vor1)

							else raise (Erreur "cannot join two delaunay graphs of different dimensions or without same Voronoi status"))

and _join_any_del (s, ((tab, _), c)) = (let dim, fonc, b         = get_defs_from_delaunay c.name in
					let sommets, positions   = (let tabret = array_of_mgs_value fonc dim s
								    in
								      (Array.map snd tabret), (Array.map fst tabret)) in
					let sommets'      = Array.map deco tab
					and positions'    = Array.map pos  tab in
					let all_sommets   = Array.append sommets sommets'
					and all_positions = Array.append positions positions' 
					in
					  delaunay_from_all_sommets_positions dim all_sommets all_positions c b)

and _delaunayfy_del_any (d, a)  = _join_any_del (a, d)

and _cons_any_del x             = _join_any_del x

and _size_del ((tab, _), _)     = VAL_int(Array.length tab)

and _hd_del ((tab, _), _ )      = Core_graph_generique.hd tab deco

and _tl_del ((tab, _), c)       = (let ntab               = VAL_seq ((Core_graph_generique.remove_one_from_array 0 deco tab), seq_desc) in
				   let dim, fonc, b1      = get_defs_from_delaunay c.name in
				   let sommets, positions = (let tabret = array_of_mgs_value fonc dim ntab
							     in
							       (Array.map snd tabret), (Array.map fst tabret)) in
				     delaunay_from_all_sommets_positions dim sommets positions c b1)

and _take_del_int (((tab, _), c), i) = deco tab.(i)

and _last_del ((tab, _), _)          = Core_graph_generique.last tab deco

and _exists_funct_del (f, ((tab, _), _)) = Core_graph_generique.exists tab deco f
and _forall_funct_del (f, ((tab, _), _)) = Core_graph_generique.forall tab deco f

and _count_any_del (a, ((tab, _), _))    = Core_graph_generique.count tab deco a

and _member_any_del (a, ((tab, _), _))   = Core_graph_generique.member  tab deco a

and _compareMgsVal_del_del (((tab1, _), c1), ((tab2, _), c2)) = VAL_int (Core_graph_generique.compare deco vois (tab1, c1) (tab2, c2))

and _isomorphism_del_del (((tab1, _), c1), ((tab2, _), c2))   = VAL_bool (0 = Core_graph_generique.compare_iso vois (tab1, c1) (tab2, c2))

and _neighbors_del_int (((t, _), _), i)       = VAL_seq ((Core_graph_generique.neighbors deco vois t i), seq_desc)


and _neighborspos_del_int (((t, _), _), i) = (assert ((i >= 0) && (i < (Array.length t)));
					      let _, _, voisins = t.(i)
					      in
						VAL_seq ((Array.of_list (List.map (fun x -> VAL_int x) voisins)), seq_desc))
					       

and _map_funct_del (fonc, ((tab, _), c)) = (let fonction x         = fonc (deco x) in
					    let sequence           = VAL_seq ((Array.map fonction tab), seq_desc) in
					      
					    let dim, fonc, vor     = get_defs_from_delaunay c.name in
					    let sommets, positions = (let tabret = array_of_mgs_value fonc dim sequence
								      in
									(Array.map snd tabret), (Array.map fst tabret)) 
					    in
					      delaunay_from_all_sommets_positions dim sommets positions c vor)

and _mapindexed_funct_del (fonc, ((tab, _), c)) = 
  let sequence           = VAL_seq ((Array.mapi (fun idx v ->  (app_2 fonc (VAL_int idx) (deco v))) tab), seq_desc) in
  let dim, fonction, vor = get_defs_from_delaunay c.name in
  let sommets, positions = (let tabret = array_of_mgs_value fonction dim sequence
			    in
			      (Array.map snd tabret), (Array.map fst tabret)) 
  in
    delaunay_from_all_sommets_positions dim sommets positions c vor

and _iter_funct_del (fonc, ((tab, _), _))               = Core_graph_generique.iter        fonc deco tab
and _iterindexed_funct_del (fonc, ((tab, _), _))        = Core_graph_generique.iterindexed fonc deco tab
and _fold_funct_any_del (fonc, zero, ((tab, _), _))     = Core_graph_generique.fold        fonc deco tab zero
and _foldindexed_funct_any_del (f, zero, ((tab, _), c)) = (let idx           = ref ((Array.length tab) - 1)
							   and dim, fonction, vor = get_defs_from_delaunay c.name
							   in
							     Array.fold_right (fun v s -> (let val_at_cur = deco v
											   in
											     decr idx;
											     (app_3 f (fonction val_at_cur) val_at_cur s))) tab zero)

and _setify_del ((tab, _), _)        = VAL_set(Core_graph_generique.setify  tab deco, Type_set.set_desc)
and _bagify_del ((tab, _), _)        = VAL_bag(Core_graph_generique.bagify  tab deco, Type_bag.bag_desc)
and _sequify_del ((tab, _), _)       = VAL_seq(Core_graph_generique.sequify tab deco, Type_seq.seq_desc)
and _graphify_del ((tab, _), _)      = (let tbl = Hash.create 1
					and decos_vois = Array.map (function (deco, pos, vois) -> (deco, vois)) tab
					in
					  VAL_grf ((tbl, decos_vois), Type_graph.graph_desc))


and _empty_del ((tab, _), _)         = (VAL_bool false) (* Impossible d'avoir un Graphe de Delaunay vide... Minimum, 3 sommets *)


and _voronoiarea_del ((tab, (voronoi, open_d)), c) = 
  (let _, _, vor = get_defs_from_delaunay c.name
   in
     if vor
     then 
       begin
	 let infos    = Array.init (Array.length tab) (fun i -> (let _, xy, _ = tab.(i)
								 in
								   assert(2 = Array.length(xy));
								   { c_hv = {h = xy.(0); v = xy.(1) }; adj = []; area = 0.0 })) in
	   
	 let voronois = Array.mapi  (fun i (h, v, l) -> {id = i; hv = { h = h; v = v}}) voronoi 
	 in
	   Array.iteri (fun idx (_, _, l) -> List.iter (fun i -> (let old = (infos.(i)).adj 
								  in
								    (infos.(i)).adj <- (voronois.(idx)) :: old)) l) voronoi;
	   Array.iteri (fun idx ({c_hv = origine; adj = l} as f) ->
			  if open_d.(idx)
			  then f.area <- max_float                                       (* Une cellule ouverte a comme surface \infty *)
			  else (if (0 = List.length l)
				then 
				  begin
				    f.area <- max_float;
				    Printf.fprintf stderr "***Warning : delaunay vertex (%f, %f) with no Voronoi vertices... ***\n" origine.h origine.v
				  end
				else
				  begin
				    let d    = (List.hd l).hv in
				    let trie = Sort.list (inferieur_voronoi origine d) l
				    in
				      f.adj  <- trie;
				      f.area <- surface {id=0; hv = origine} trie
				  end)) infos; (* Tri selon le sens horaire *)
	   
(*	       Array.iteri (fun idx {c_hv = { h = x; v = y }; adj = l; area = a} ->
			      Printf.printf "Sommet %d en (%f, %f) S(%f)=> %s\n"
			      idx x y a (clean_comma_string_of_general_list (fun {id = i; hv = {h = x; v = y}} ->
									       Printf.sprintf "(%d - (%f, %f))" i x y) l)) infos;
*)
	   
	   VAL_seq ((Array.map (fun {area = a} -> VAL_float a) infos), Type_seq.seq_desc)
       end
     else raise (Erreur ("cannot get Voronoi informations from a Delaunay that doesn't carry any : "^ c.name)))

and _voronoi_del_bool(((t, (voronoi, open_d)), c), b)  =
  (let _, _, vor = get_defs_from_delaunay c.name 
   in
     if vor
     then
       (let make_seq l             = VAL_seq ((Array.of_list (List.map (function x -> VAL_int x) l)), Type_seq.seq_desc) in
	let make_record3 (x, y, l) = VAL_rec ((Core_rec.add (VAL_string "x") (VAL_float x)
						 (Core_rec.add (VAL_string "y") (VAL_float y)
						    (Core_rec.add (VAL_string "l") (make_seq l)
						       Core_rec.empty))), Type_rec.record_desc)
				       
	and make_record2 (x, y)    = VAL_rec ((Core_rec.add (VAL_string "x") (VAL_float x)
						 (Core_rec.add (VAL_string "y") (VAL_float y)
						    Core_rec.empty)), Type_rec.record_desc)
	in
	  if b
	  then VAL_seq((Array.map make_record3 voronoi), Type_seq.seq_desc)
	  else 
	    let records    = Array.make (Array.length voronoi) (VAL_undef "_") in
	    let voisins    = Array.make (Array.length t) [] in
	    let make_seq v = VAL_seq (v, Type_seq.seq_desc)
	    in
	      Array.iteri (fun idx (x, y, l) -> (List.iter (fun i -> voisins.(i) <- {id = idx; hv = {h = mfst(voronoi.(idx)); 
												     v = msnd(voronoi.(idx))} } :: voisins.(i)) l;
						 records.(idx) <- make_record2(x, y))) voronoi;
	      VAL_rec ((Core_rec.add (VAL_string "coords") (make_seq records)
			  (Core_rec.add (VAL_string "facets") (make_seq (Array.mapi (fun idx l -> (let origine  = (let _, xy, _ = t.(idx)   (* On les trie dans l'ordre horaire *)
												      in
													{h = xy.(0); v = xy.(1)})
										      in
											if (0 = List.length l)
											then 
											  begin
	  Printf.fprintf stderr "***Warning : delaunay vertex (%f, %f) with no Voronoi vertices... ***\n" origine.h origine.v;
											    make_seq [||]
											  end
											else 
											  begin
											    let d        = (assert( 0 < List.length(l) ); (* Sinon, on est mal ! *)
													    (List.hd(l)).hv) in
											    let trie     = Sort.list (inferieur_voronoi origine d) l in
											    let nettoye  = Array.of_list (List.map (fun {id = i} -> VAL_int i) trie)
											    in
											      make_seq nettoye
											  end)) voisins))
			     (Core_rec.add (VAL_string "open") (make_seq (Array.map (fun x -> VAL_bool x) open_d))
				Core_rec.empty))), Type_rec.record_desc))
       
     else raise (Erreur ("cannot get Voronoi informations from a Delaunay that does not have any : "^ c.name)))

;;

