(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: hash.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-29 18:49:03 $
 * Version: $Revision: 1.5 $
 * -------------------------------------------------------------------
 *
 * Status: ajout a partir de la version 1.23 de Doligez de la lib ocaml
 *         pour avoir la taille de la hashtbl + fct supplementaires
 * 
 * 
 *
 *=========================================================================
 *)


Global.__register_version "$RCSfile: hash.ml,v $" "$Revision: 1.5 $" "$Date: 2005-11-29 18:49:03 $";;


(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(***********************************************************************)

(* Hash tables *)

external hash_param : int -> int -> 'a -> int = "caml_hash_univ_param" "noalloc"

let hash x = hash_param 10 100 x

(* We do dynamic hashing, and resize the table and rehash the elements
   when buckets become too long. *)

type ('a, 'b) t =
  { mutable size: int;                        (* number of elements *)
    mutable data: ('a, 'b) bucketlist array } (* the buckets *)

and ('a, 'b) bucketlist =
    Empty
  | Cons of 'a * 'b * ('a, 'b) bucketlist

let create initial_size =
  let s = min (max 1 initial_size) Sys.max_array_length in
  { size = 0; data = Array.make s Empty }

let clear h =
  for i = 0 to Array.length h.data - 1 do
    h.data.(i) <- Empty
  done;
  h.size <- 0

let copy h =
  { size = h.size;
    data = Array.copy h.data }

let resize hashfun tbl =
  let odata = tbl.data in
  let osize = Array.length odata in
  let nsize = min (2 * osize + 1) Sys.max_array_length in
  if nsize <> osize then begin
    let ndata = Array.create nsize Empty in
    let rec insert_bucket = function
        Empty -> ()
      | Cons(key, data, rest) ->
          insert_bucket rest; (* preserve original order of elements *)
          let nidx = (hashfun key) mod nsize in
          ndata.(nidx) <- Cons(key, data, ndata.(nidx)) in
    for i = 0 to osize - 1 do
      insert_bucket odata.(i)
    done;
    tbl.data <- ndata;
  end

let add h key info =
  let i = (hash key) mod (Array.length h.data) in
  let bucket = Cons(key, info, h.data.(i)) in
  h.data.(i) <- bucket;
  h.size <- succ h.size;
  if h.size > Array.length h.data lsl 1 then resize hash h

let remove h key =
  let rec remove_bucket = function
      Empty ->
        Empty
    | Cons(k, i, next) ->
        if (Pervasives.compare k key) = 0
        then begin h.size <- pred h.size; next end
        else Cons(k, i, remove_bucket next) in
  let i = (hash key) mod (Array.length h.data) in
  h.data.(i) <- remove_bucket h.data.(i)

let rec find_rec key = function
    Empty ->
      raise Not_found
  | Cons(k, d, rest) ->
      if (Pervasives.compare key k) = 0 then d else find_rec key rest

let find h key =
  match h.data.((hash key) mod (Array.length h.data)) with
    Empty -> raise Not_found
  | Cons(k1, d1, rest1) ->
      if (Pervasives.compare key k1) = 0 then d1 else
      match rest1 with
        Empty -> raise Not_found
      | Cons(k2, d2, rest2) ->
          if (Pervasives.compare key k2) = 0 then d2 else
          match rest2 with
            Empty -> raise Not_found
          | Cons(k3, d3, rest3) ->
              if (Pervasives.compare key k3) = 0 then d3 else find_rec key rest3

let find_all h key =
  let rec find_in_bucket = function
    Empty ->
      []
  | Cons(k, d, rest) ->
      if (Pervasives.compare k key) = 0 then d :: find_in_bucket rest else find_in_bucket rest in
  find_in_bucket h.data.((hash key) mod (Array.length h.data))

let replace h key info =
  let rec replace_bucket = function
      Empty ->
        raise Not_found
    | Cons(k, i, next) ->
        if (Pervasives.compare k key) = 0
        then Cons(k, info, next)
        else Cons(k, i, replace_bucket next)
  in
  let i = (hash key) mod (Array.length h.data) in
  let l = h.data.(i) in
  try
    h.data.(i) <- replace_bucket l;
  with Not_found ->
    h.data.(i) <- Cons(key, info, l);
    h.size <- succ h.size;
    if h.size > Array.length h.data lsl 1 then resize hash h

let mem h key =
  let rec mem_in_bucket = function
  | Empty ->
      false
  | Cons(k, d, rest) ->
      (Pervasives.compare k key) = 0 || mem_in_bucket rest in
  mem_in_bucket h.data.((hash key) mod (Array.length h.data))

let iter f h =
  let rec do_bucket = function
      Empty ->
        ()
    | Cons(k, d, rest) ->
        f k d; do_bucket rest in
  let d = h.data in
  for i = 0 to Array.length d - 1 do
    do_bucket d.(i)
  done


(** === Rajout MGS ============================================= **)

(** Rajouter pour des questions d'efficacit : on veut reprendre une 
 ** itration  partir d'une position qui sert "d'optimisation".
 **
 **)



(** La fonction next_non_empty_bucket() permet de chercher un bucket non-vide
 ** a partir d'une position de depart comprise dans les bornes de la table de 
 ** hachage. Si on tire ce nombre au hasard, alors cela permet d'avoir un element 
 ** au hasard. Il faut absolument prendre le bucket non-libre, sinon on a pas un
 ** hasard suffisant. N. mann en a fait la malheureuse experience. si on a une zone 
 ** de bucket libre, on tombe pour n'importe quel element dans cette zone, sur le 
 ** 1er element non libre qui termine la zone. 
 **
 ** La fonction random_key() renvoie une cle au hasard parmi les cls prsentes.
 **)

exception Hfound of int
;;

let next_non_empty_bucket h start =
  let tab = h.data in
    try
      for i = start to Array.length tab - 1 do
        if tab.(i) <> Empty then raise (Hfound i)
      done;
      for i = 0 to start - 1 do
        if tab.(i) <> Empty then raise (Hfound i)
      done;
      start
    with Hfound i -> i
;;

let rec nth_non_empty_bucket h n = 
  let b = ref (next_non_empty_bucket h 0)
  in 
    for i = 1 to n do
      b := next_non_empty_bucket h !b
    done;
    !b
;;




(** La fonction last_iter_index() permet de savoir o on tais 
 ** (en cas d'abandon de l'iteration par un raise) et de se servir 
 ** de cet index comme suggestion de dpart pour le prochain iter.
 **)
let last_index_visited = ref (-1)
let last_iter_index() = !last_index_visited
  
let iter_from start f h =
  let rec do_bucket = function
      Empty ->
        ()
    | Cons(k, d, rest) ->
        f k d; do_bucket rest in
  let d = h.data in
    (** Printf.printf "# iter_from : on part de %d\n" start; **)
    begin
      for i = start to Array.length d - 1 do
        last_index_visited := i;
        do_bucket d.(i)
      done;
      for i = 0 to start - 1 do
        last_index_visited := i;
        do_bucket d.(i)
      done
    end

exception FoundKey
;;
let random_key h =
  let rn = ref (Global.crandom h.size) in
  let rkey = ref None in 
  let rec do_bucket = function
    | Empty -> ()
    | Cons(k, _, rest) ->
	rn := !rn - 1;
	if !rn = 0 then (rkey := k; raise FoundKey)
	else do_bucket rest 
  in 
  let d = h.data in
    try
      for i = 0 to Array.length d - 1 do
	last_index_visited := i;
	do_bucket d.(i)
      done;
      raise Not_found
    with FoundKey -> !rkey
;;

exception Cut;;

let iter_from_random_key f h =
  let rn = ref (Global.crandom h.size) in
  let rkey = ref None in

  let do_bucket = ref (function _ -> ()) in

  let rec do_bucket_before_first = function
      Empty -> ()
    | Cons(k, d, rest) ->
	rn := !rn - 1;
	if !rn <= 0 then (
	  rkey := Some k;
	  do_bucket := do_bucket_after_first;
	  f k d; do_bucket_after_first rest
	) else
	  do_bucket_before_first rest

  and do_bucket_after_first = function
      Empty -> ()
    | Cons(k, d, rest) ->
	if (Pervasives.compare k (match (!rkey) with Some p -> p | None -> failwith "Hash: iter_from_random_key: do_bucket_after_first: internal error: !rkey has to be Some _") = 0) then
	  raise Cut
	else
          f k d; do_bucket_after_first rest
  in

  let d = h.data in

    begin
      do_bucket := do_bucket_before_first ;

      try
	for i = 0 to Array.length d - 1 do
          (!do_bucket) d.(i)
	done;
	for i = 0 to Array.length d - 1 do
          (!do_bucket) d.(i)
	done
      with Cut -> ()
    end





let fold f h init =
  let rec do_bucket b accu =
    match b with
      Empty ->
        accu
    | Cons(k, d, rest) ->
        do_bucket rest (f k d accu) in
  let d = h.data in
  let accu = ref init in
  for i = 0 to Array.length d - 1 do
    accu := do_bucket d.(i) !accu
  done;
  !accu


(* Added MGS : better compile with -noassert ;-) *)
let size h = 
  assert((Pervasives.compare (h.size) (fold (fun _ _ s -> 1+s) h 0)) = 0);
  h.size

(** Compter le nombre max de collisions *)
let max_collision h = 
  let rec do_bucket = function
    | Empty -> 0
    | Cons(_, _, rest) -> 1 + (do_bucket rest)
  in let d = h.data in
    begin
      let acc = ref 0 in 
	for i = 0 to Array.length d - 1 do
	  acc := max !acc (do_bucket d.(i))
	done;
	!acc
    end


let nth n tbl = 
  if (n >= size tbl)
  then raise Not_found
  else (let cur  = ref 0
	and elem = ref None
	in
	  try
	    iter (fun k v -> (if (!cur = n) 
			      then (elem := (Some (k, v)); raise Cut)
			      else incr cur))
	      tbl;
	    raise Cut
	  with
	      Cut -> (match !elem
		      with
			  Some x -> x
			| _      -> failwith "internal error on nth in core_hashtbl"))
	  

let hd h = nth 0 h

let last h = nth ((size h) - 1) h

let tl h = 
  let ret = copy h in
  let k, v = hd ret
  in
    remove ret k;
    ret

let from_list f1 f2 l = 
  let tbl = create (List.length l)
  in
    List.iter (fun v -> add tbl (f1 v) (f2 v)) l;
    tbl

let from_array f1 f2 a =
  let tbl = create (Array.length a)
  in
    Array.iter (fun v -> add tbl (f1 v) (f2 v)) a;
    tbl

let to_list h = fold (fun k v s -> (k, v)::s) h []

let to_array h = fold (fun k v s -> Array.append [|k, v|] s) h [||]

let forall f h =
  try
    iter (fun k v -> if (not (f k v)) then (raise Cut)) h;
    true
  with Cut -> false

let exists f h =
  try
    iter (fun k v -> if (f k v) then (raise Cut)) h;
    false
  with Cut -> true


(** Comparaison de tables de hachage: c'est difficile car les tables
 ** peuvent tre de longueur differentes, et, mme si elles ont la mme
 ** longueur, les listes de collisions peuvent tre ordonnes diffrement.
 ** La solution adopte ici consiste a mettre les couples cle,valeur dans 
 ** un vecteur, a le trier, et a comparer ensuite element par
 ** element. On utilise la fonction systeme de comparaison pour comparer
 ** les cles (c'est des chaines de caracteres ou bien des tableaux
 ** d'entier) et la fct de comparaisons specialement passe en argument
 ** pour les valeurs.  
 **)

exception DiffFound of int
;;

let compare2 keycmp valcmp tbl1 tbl2 = 
  let size_tbl1 = size tbl1 in
  let sz_cmp    = Pervasives.compare size_tbl1 (size tbl2) 
  in
    if (sz_cmp <> 0) 
    then sz_cmp
    else 
      if (size_tbl1 = 0)
      then 0
      else (let (k, v) = hd tbl1 in
	    let a1     = Array.make (size tbl1) (k, v)
	    and a2     = Array.make (size tbl2) (k, v)
	    and cmp    = (fun (k1, v1) (k2, v2) -> let diff = keycmp k1 k2 in if (diff <> 0) then diff else valcmp v1 v2)
	    and rempli = (fun tbl a             -> let i = ref 0 in iter (fun k v -> a.(!i) <- (k, v); incr i) tbl)
	    in
	      rempli tbl1 a1;
	      rempli tbl2 a2;
	      Array.sort cmp a1;
	      Array.sort cmp a2;
	      try
		for i = 0 to (Array.length a1)-1
		do
		  let diff = cmp a1.(i) a2.(i) 
		  in
		    if diff <> 0 
		    then raise (DiffFound diff);
		done;
		(** rien ne differe **)
		0
	      with DiffFound diff -> diff)
;;

let compare valcmp tbl1 tbl2 = compare2 Pervasives.compare valcmp tbl1 tbl2
;; 
