(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: patch.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-08-02 00:11:24 $
 * Version: $Revision: 1.15 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *          
 *  Exemple d'ordonnacement des motifs :
 *
 *     patch P = {
 *     
 *         x0 < y1:[dim=2, y1>x0] > z0 
 *         k2 >> z0
 *         k2 >> x0:[dim = dimension(^k2)-2]
 *         y1 > u0:[dim=0, u0>x0, u0>z0, u0 < y1]
 *         => 
 *         z1, y1, x0
 *     };;
 *     
 *     
 * correspond  la relation face suivante :
 *                 
 *              __k2__
 *             /      \
 *            /        \ 
 *           /  __y1__  \
 *          /  /   |  \  \
 *         /  /    V   \  \
 *         x0 <---u0---> z0
 *     
 *     
 * et le reordonnancement de cela (reordonne) renvoie la description ordonne de motifs :
 *     
 *     y1:
 *             guards=  dim=2 &
 *             faces of= 
 *             transitive faces of= 
 *             cofaces of= 
 *             transitive cofaces of= 
 *     u0:
 *             guards=  ('< u0 y1) & dim=0 &
 *             faces of= y1 
 *             transitive faces of= 
 *             cofaces of= 
 *             transitive cofaces of= 
 *     x0:
 *             guards=  ('> y1 x0) & ('> u0 x0) &
 *             faces of= y1 
 *             transitive faces of= 
 *             cofaces of= 
 *             transitive cofaces of= 
 *     z0:
 *             guards=  ('> u0 z0) &
 *             faces of= y1 
 *             transitive faces of= 
 *             cofaces of= 
 *             transitive cofaces of= 
 *     k2:
 *             guards=  ($test_pred dim ('- (dim k2) 2)) &
 *             faces of= 
 *             transitive faces of= 
 *             cofaces of= 
 *             transitive cofaces of= z0 x0 
 *     
 * chaque item d'un motif elementaire (= une variable qui va filtrer un simplexe)
 * ne contient que des rfrences  des motif elementaire deja filtrs (i.e. dcrit 
 * avant dans cette liste).
 *
 *=========================================================================
 *)

(*
patch P = {
  x0 < y1:[dim=2, y1>x0, in face(k2) ] > z0 
  k2 >> z0
  k2 >> x0:[dim = dim(k2)-2]
  y1 > u0:[dim=0, u0>x0, u0>z0, u0 < y1]
  => 
   z1, y1, x0
};;

patch P = {
  e:[dim=1]
  ~v1 < e
  ~v2 < e
  =>
    e+3
};;
*)

open Printf;;


open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Printer_term;;
open Dispatch_sig;;
open Ptopology;;

open Ptopo_achain;;
open Ptopo_qmf;;

__register_version "$RCSfile: patch.ml,v $" "$Revision: 1.15 $" "$Date: 2006-08-02 00:11:24 $";;
       
(* --- Motif issu du parsing ------------------------------------------------------- 
 *
 * A l'issu du parsing, un motif est reprsent par un couple de 2
 * listes : (mlist, conlist). 
 *
 * mlist est une liste de couples (id, guards) o id est une
 * identificateur (string) et guards est une liste de clauses
 * correspondant au contenu des  brackets ventuelles qui suivent un
 * identificateur x:[cl1, cl2,  ..., cln]. Une clause est de la forme
 * (Test expr1) ou bien (Pred idpred expr2) ce qui correspond 
 * x:[expr1, idpred = expr2, ...]
 *
 * conlist est la liste de connectivit. C'est une liste de triplets: 
 * (connect, id1, id2) o connect est un Terme.connecteur et id1 et
 * id2 les entits impliques. Par exemple(FaceOf, x, y) veut dire 
 * x < y.
 *)

let print_parse_motif ((mlist:patch_pattern), conlist) =
  Printf.printf "regle de patch de motif (a l'issu du parsing) :\n";
  List.iter
    (fun (id, useit, guards) -> 
       Printf.printf "\t%s %s %s\n" id
       (if (useit) then "(use me !!)" else "(don't use me !!)")
       (if guards = []  then "with no guards" 
	else (string_of_int (List.length guards))^" clauses in guards"))
    mlist;
  List.iter
    (fun (connect, id1, id2) ->
       Printf.printf "  %s %s = %s\n" 
       (match connect with
	  | NoConnection -> "no connection" 
	  | FaceOf -> "face of"
	  | CoFaceOf -> "coface of"
	  | TfaceOf -> "transitive face of"
	  | TcoFaceOf -> "transitive coface of")
       id1 id2)
    conlist
;;


(* --- Simplification et normalisation des 2 listes issues du parsing ----------------------- 
 *
 * Le resultats est 2 listes d'associations mlist et conlist. Le nom de la variable de motif
 * sert de cls. Dans la premire liste on place la suite des clauses, dans la seconde on a 
 * les connections representes par un triplet :
 *     (used_by, face_of, transitive_face_of).
 * Par exemple,  partir de (C1, id1, id2), (C2, id1, id3), (C1, id1, id4) est transform en 
 * (id1, ([id2; id4], [id3], [])).
 *)

let fst3 (x,_,_) = x ;;

let toutes_les_variables (mlist:patch_pattern) conlist =
  let l1 = List.map fst3 mlist
  and l2 = List.fold_left (fun acc (_, id1, id2) -> id1::id2::acc) [] conlist
  and addfct set id = Terme.Ss.add id set
  in List.fold_left addfct (List.fold_left addfct Terme.empty_string_set l1) l2

(* predicat qui indique si id apparat dans une liste de gardes *)
and occurs_in_guards id gl =  
  List.exists (function 
		 | Test g 
		 | Pred(_, g) -> Terme.Ss.mem id (Terme.free_var_term g))
    gl

and connected_face id = function
  | (FaceOf, id1, id2) when id = id1 -> [id2]
  | (CoFaceOf, id2, id1) when id = id1 -> [id2]
  | (_, _, _) -> []

and transitive_connected_face id = function
  | (TfaceOf, id1, id2) when id = id1 -> [id2]
  | (TcoFaceOf, id2, id1) when id = id1 -> [id2]
  | (_, _, _) -> []

and connected_coface id = function
  | (FaceOf, id1, id2) when id = id2 -> [id1]
  | (CoFaceOf, id2, id1) when id = id2 -> [id1]
  | (_, _, _) -> []

and transitive_connected_coface id = function
  | (TfaceOf, id1, id2) when id = id2 -> [id1]
  | (TcoFaceOf, id2, id1) when id = id2 -> [id1]
  | (_, _, _) -> []
;;


(* retourne tous les id' tels qu'il existe une garde de id' o id apparait *)
let rec collect_clause (mlist:patch_pattern) id =
  (id,
   List.fold_left (fun acc (id', _, g) -> if id = id' then g@acc else acc) [] mlist)

(* retourne la mme liste mais sans doublon *)
and without_double acc l = match l with
    []   -> acc
  | t::q -> without_double (if (List.mem t acc) then acc else t::acc) q

(* calcul les connections d'une variable de patch id *)
and collect_connection mlist glist conlist id =
  let filtered =
    List.fold_left (fun acc (id',b,_) -> if (id'=id) then (b||acc) else acc) false mlist
  and used_by = 
    List.fold_left
      (fun acc (id', gl) -> if occurs_in_guards id gl then id'::acc else acc) 
      [] glist
  and face_of =
    List.fold_left (fun acc con -> (connected_face id con)@acc) [] conlist
  and transitive_face_of =
    List.fold_left (fun acc con -> (transitive_connected_face id con)@acc) [] conlist
  and coface_of =  
    List.fold_left (fun acc con -> (connected_coface id con)@acc) [] conlist
  and transitive_coface_of =
    List.fold_left (fun acc con -> (transitive_connected_coface id con)@acc) [] conlist
  in (filtered,
      without_double [] used_by, 
      without_double [] face_of, 
      without_double [] transitive_face_of,
      without_double [] coface_of, 
      without_double [] transitive_coface_of)
;;

(*
let rec find_cfaces_in_gl (acc_gl,acc_cons) gl = match gl with
    []                       -> (acc_gl,acc_cons)
  | (Test g)::q              -> ((Test g)::acc_gl , acc_cons)
  | (Pred("faces",   g))::q  ->
  | (Pred("cofaces", g))::q  ->
  | (Pred(s, g))::q          -> ((Pred(s,g))::acc_gl , acc_cons)

and find_cfaces varset (acc_glist,acc_conlist) = function
    [] -> (acc_glist,acc_conlist)
  | (id,gl)::q ->
      let (plus_glist, plus_conslist) = find_cfaces_in_gl 
	(acc_glist,acc_conlist)

;;
*)
(** Retourne un couple (glist, clist)
 ** glist est une liste d'associations id <--> liste de Terme.guards 
 ** clist est une liste d'associations id <--> quintuplet 
 **                  (usedby, faceof, transitive_faceof, cofaceof, transitive_cofaceof)
 **)
let normalise_patch ((mlist:patch_pattern), conlist) = 
  let varset = toutes_les_variables mlist conlist in
  let glist  = Terme.Ss.fold (fun e l -> (collect_clause mlist e)::l) varset [] in
  let clist = Terme.Ss.fold 
		(fun e l -> (e, (collect_connection mlist glist conlist e))::l) varset []
  in (glist, clist)
;;
		  

let print_sorted_motif (glist,clist) =
  Printf.printf "regle de patch de motif apres normalisation:\n";
  List.iter
    (fun (id, guards) -> 
       Printf.printf "\t%s %s\n" id
       (if guards = []  then "with no guards" 
	else (string_of_int (List.length guards))^" clauses in guards")) 
    glist;
  List.iter
    (fun (id, (filtered, usedby, faceof, tfaceof, cofaceof, tcofaceof)) ->
       Printf.printf "  %s :\n" id;
       Printf.printf "   filtered: %s" (if filtered then "true" else "false");
       Printf.printf "\n   usedby: ";
       List.iter (fun x -> Printf.printf "%s " x) usedby;
       Printf.printf "\n   faceof: ";
       List.iter (fun x -> Printf.printf "|%s> " x) faceof;
       Printf.printf "\n   tfaceof: ";
       List.iter (fun x -> Printf.printf "%s " x) tfaceof;
       Printf.printf "\n   cofaceof: ";
       List.iter (fun x -> Printf.printf "%s " x) cofaceof;
       Printf.printf "\n   tcofaceof: ";
       List.iter (fun x -> Printf.printf "%s " x) tcofaceof;
       print_newline())
    clist
;;


(**
 FIXME: generalisation du tri:
  parametrage du tri par trois stratgies orthogonale :
  1) choix de la cellule de depart
  2) a partir de la cell de depart, quelle dependence est prise en compte (face, coface, ou les deux)
  3) itration du tri en profondeur ou en largeur

  1) choix de la cell de depart :
        par la composition lexicographique de 4 criteres :
           A : nb faces
           B : nb cofaces
           C : nb faces + cofaces
           D : nb de cell dans la mme dimension (niveau horizontale du graphe)
           E : nb d'lement incident ((co)faces) dj filtr
     une fonction de choix par exemple : "+D-A-B" qui veut dire qu on trie d abord par le nombre
     croissant de cells de la meme dim, puis par le nombre decroissant de face, et par le nombre
     decroisant de cofaces. On finit par choisir l ordre alphabetique du nom.

  2) On a une cell de depart. On veut choisir ses successeurs. On peut les prendre parmi 5 ensembles:
     les faces, les cofaces, l'ensemble des faces+cofaces, ttes les cells, ou aucune.
     Chacun des 5 ensembles, est considr suivant l ordre 1).

  3) pour chercher le succ suivant :
       largeur : on epuise tous les elts de l ensemble choisi en 2) puis on recommence en 1)
       profondeur : on prend le premier de l ensemble du 2), et on ritre 2) en le considrant
                    comme cell de depart. Jusqu' epuisement de l ensemble du 2).

Stratgie par dfaut:
  "+B-A" / ensemble des faces / profondeur

 **)

let patch_criterion  = "-E+B-A" ;;
let patch_successeur = "faces" ;; (*dans "faces", "cofaces", "(co)faces", "all", "nothing" *)
let patch_parcours   = "profondeur" ;;

let make_list_criterion () =
  let lgth = String.length patch_criterion in
    if (lgth mod 2 = 0) then
      let ans = ref [] in
	for i = (lgth / 2 - 1) downto 0 do
	  Printf.printf "%s\n" (String.sub patch_criterion (2*i) 2) ;
	  ans := (String.sub patch_criterion (2*i) 2)::!ans
	done ;
	!ans
    else
      raise (Erreur "Patch: make_list_criterion (1)")
;;

 (* Pour chaque cellule, on associe dans une table de hashage le quintuplet :
  *   (consomm ou non, nb faces non consommes, nb cofaces non consomme,
  *    nb faces consommes, nb cofaces consommes)
  * 
  * Il fo prendre en considration c cardinaux dans le tri...
  *)
let make_hashtbl clist =
  let tbl = Hashtbl.create 101 in
  let rec make_hashtbl_ = function
      [] -> tbl
    | (id,(_, _, faceof, _, cofaceof, _))::q ->
	Hashtbl.add tbl id (ref false, ref (List.length cofaceof), ref (List.length faceof),
			    ref 0, ref 0
			   );
	make_hashtbl_ q
  in
    make_hashtbl_ clist
;;

let consomme tbl (id,(_, _, faceof, _, cofaceof, _)) =
  let (csm,_,_,_,_) = Hashtbl.find tbl id in
    csm := true ;
    List.iter (fun coface_id ->
		 let (_,nb_faces_restantes,_,nb_faces_filtrees,_) = Hashtbl.find tbl coface_id
		 in
		   nb_faces_restantes := !nb_faces_restantes - 1 ;
		   nb_faces_filtrees := !nb_faces_filtrees + 1		 
	      ) faceof ;
    List.iter (fun face_id ->
		 let (_,_,nb_cofaces_restantes,_,nb_cofaces_filtrees) = Hashtbl.find tbl face_id
		 in
		   nb_cofaces_restantes := !nb_cofaces_restantes - 1;
		   nb_cofaces_filtrees := !nb_cofaces_filtrees + 1
	      ) cofaceof
;;

let rec cmp_A tbl (id1,(_, _, _, _, cofaceof1, _)) (id2,(_, _, _, _, cofaceof2, _)) =
  let (_,nbf1,nbcf1,_,_) = Hashtbl.find tbl id1
  and (_,nbf2,nbcf2,_,_) = Hashtbl.find tbl id2 in
    Pervasives.compare !nbf1 !nbf2

and cmp_B tbl (id1,(_, _, faceof1, _, _, _)) (id2,(_, _, faceof2, _, _, _)) =
  let (_,nbf1,nbcf1,_,_) = Hashtbl.find tbl id1
  and (_,nbf2,nbcf2,_,_) = Hashtbl.find tbl id2 in
    Pervasives.compare !nbcf1 !nbcf2

and cmp_C tbl (id1,(_, _, faceof1, _, cofaceof1, _)) (id2,(_, _, faceof2, _, cofaceof2, _)) =
  let (_,nbf1,nbcf1,_,_) = Hashtbl.find tbl id1
  and (_,nbf2,nbcf2,_,_) = Hashtbl.find tbl id2 in
    Pervasives.compare (!nbf1 + !nbcf1) (!nbf2 + !nbcf2)

and cmp_D _ _ _ =
  raise (Erreur "Patch: cmp_D: not yet implemented")

and cmp_E tbl (id1,(_, _, faceof1, _, cofaceof1, _)) (id2,(_, _, faceof2, _, cofaceof2, _)) =
  let (_,_,_,nbf1,nbcf1) = Hashtbl.find tbl id1
  and (_,_,_,nbf2,nbcf2) = Hashtbl.find tbl id2 in
    Pervasives.compare (!nbf1 + !nbcf1) (!nbf2 + !nbcf2)

;;

let rond f g x = f(g(x)) ;;

let rec cree tbl = function
  | [] -> Printf.printf "coucou\n" ; (fun (n,((id1, _) as e1),((id2,_) as e2)) -> (if (n=0) then Pervasives.compare id1 id2 else n),e1,e2 )
  | "+A"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_A tbl c1 c2),c1,c2) else (n,c1,c2) )
  | "-A"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_A tbl c2 c1),c1,c2) else (n,c1,c2) )
  | "+B"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_B tbl c1 c2),c1,c2) else (n,c1,c2) )
  | "-B"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_B tbl c2 c1),c1,c2) else (n,c1,c2) )
  | "+C"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_C tbl c1 c2),c1,c2) else (n,c1,c2) )
  | "-C"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_C tbl c2 c1),c1,c2) else (n,c1,c2) )
  | "+D"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_D tbl c1 c2),c1,c2) else (n,c1,c2) )
  | "-D"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_D tbl c2 c1),c1,c2) else (n,c1,c2) )
  | "+E"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_E tbl c1 c2),c1,c2) else (n,c1,c2) )
  | "-E"::q -> rond (cree tbl q) (fun (n,c1,c2) -> if (n=0) then ((cmp_E tbl c2 c1),c1,c2) else (n,c1,c2) )
  | _       -> raise (Erreur "Patch: cree: (1)")

let second_step_largeur tbl ((id,(_, _, faceof, _, cofaceof, _)) as fst) ordre =
  (* fst::ordre constitue l ensemble des lments qu'il reste  rajouter *)
  let succ_fst,others = match patch_successeur with
    | "faces"     -> List.partition (fun (id',_) -> List.mem id' cofaceof) ordre
    | "cofaces"   -> List.partition (fun (id',_) -> List.mem id' faceof) ordre
    | "(co)faces" -> List.partition (fun (id',_) -> List.mem id' faceof || List.mem id' cofaceof) ordre
    | "all"       -> ordre,[]
    | "nothing"   -> [],ordre
    | _           -> raise (Erreur "Patch: second_step: waiting for \"faces\", \"cofaces\", \"(co)faces\", \"all\", \"nothing\"")
  in
    consomme tbl fst ;
    List.iter (consomme tbl) succ_fst ;
    (fst::succ_fst, others)
;;
  
let rec second_step_profondeur g tbl ((id,(_, _, faceof, _, cofaceof, _)) as fst) ordre =
  (* fst::ordre constitue l ensemble des lments qu'il reste  rajouter *)
  consomme tbl fst ;
  let ret = ref [fst] in
  let succ_fst = match patch_successeur with
    | "faces"     -> List.filter (fun (id',_) -> let (csm,_,_,_,_) = Hashtbl.find tbl id' in (not !csm) && List.mem id' cofaceof) ordre
    | "cofaces"   -> List.filter (fun (id',_) -> let (csm,_,_,_,_) = Hashtbl.find tbl id' in (not !csm) && List.mem id' faceof) ordre
    | "(co)faces" -> List.filter (fun (id',_) -> let (csm,_,_,_,_) = Hashtbl.find tbl id' in (not !csm) && (List.mem id' faceof || List.mem id' cofaceof)) ordre
    | "all"       -> List.filter (fun (id',_) -> let (csm,_,_,_,_) = Hashtbl.find tbl id' in (not !csm)) ordre
    | "nothing"   -> []
    | _           -> raise (Erreur "Patch: second_step: waiting for \"faces\", \"cofaces\", \"(co)faces\", \"all\", \"nothing\"")
  in
    (* Un manque d'optimisation apparat ici  cause de l'itration. En effet,
     * on ne prend pas en compte dans la comparaison du nom d'lment dont on
     * dpend dj filtr... Pour empcher cela, il fo rajout un critre dans
     * la comparaison entre deux lments et une fois toute une passe en
     * profondeur faite, il fo retrier la liste des succ_fst pour retrouver le
     * meilleur candidat suivant sachant les lments filtrs lors de la
     * descente.
     * 
     * Une question reste en suspens : est-ce necessaire que ordre soit tri ?
     * A priori non... On tri en faite toujours succ_fst. Mais en fait si, on
     * a besoin d'avoir le premier !
     *)
    (*
    List.iter (fun ((id',_) as snd) ->
		 let (csm,_,_,_,_) = Hashtbl.find tbl id' in
		   if (not !csm) then
		     let just_csm, _ = second_step_profondeur g tbl snd ordre in
		       ret := !ret @ just_csm
	      ) succ_fst ;
    *)

  let iterlist = ref succ_fst in
    while (!iterlist != []) do
      let ((id',_) as snd) = List.hd (!iterlist) in
      let (csm,_,_,_,_) = Hashtbl.find tbl id' in
	if (not !csm) then (
	  let just_csm, _ = second_step_profondeur g tbl snd ordre in
	    ret := !ret @ just_csm ;
	    iterlist := List.sort ( fun e1 e2 -> let (cmp,_,_) = g (0,e1,e2) in cmp ) (List.tl (!iterlist))
	) else
	  iterlist := List.tl (!iterlist) ;
    done ;

    !ret, List.fold_left (fun acc ((id',_) as o) -> let (csm,_,_,_,_) = Hashtbl.find tbl id' in 
			    if (not !csm) then o::acc else acc) [] ordre
;;



let rec sort_pattern_ tbl clist =
  let g = cree tbl (make_list_criterion ()) in

  let ordre = (List.sort ( fun e1 e2 -> let (cmp,_,_) = g (0,e1,e2) in cmp ) clist) in

  let prems,other = match patch_parcours with
    | "profondeur" -> second_step_profondeur g tbl (List.hd ordre) (List.tl ordre)
    | "largeur"    -> second_step_largeur      tbl (List.hd ordre) (List.tl ordre)
    | _ -> raise (Erreur " dfinir")
  in

    (List.map fst prems)@(if other = [] then [] else (sort_pattern_ tbl other))

;;

let sort_pattern clist =
  let tbl = make_hashtbl clist in
    sort_pattern_ tbl clist
;;

(**

  Vieux code (qui semble marcher...)

 **)
(* --- Aprs normalisation, il faut rordonner les rgles et dplacer
 * --- ventuellement certaine clauses.
 * 
 * On ordonne les rgles suivant la relation face, afin d'avoir en premier les simplexes 
 * qui conditionnent le (font rfrences aux) plus de faces possibles.
 * Ca se fait par un tri topologique (tsort).
 * Dans chaque classe de ce tri, (i.e. l'ensemble des lments non comparables),
 * on affine l'ordonnancement par un critre additionel (cmp_class) en triant les lments 
 * d'une classe (sort_class).
 *
 * On prend l'lment maximal de cette classe, et ses faces transitives, on les enleves des motifs 
 *  trier et on recommence. En effet, l'lment maximal s0 est celui qui conditionne le plus de 
 * faces  trouver ensuite, et donne la liste des faces  examiner. Il est donc normal que ces
 * simplexes apparaissent aprs s0. Il faut juste les trier pour qu'on respecte les critres de tsort.
 * 
 * On itere ce processus jusqu' ce qu'il n'y est plus de simplexes  ordonnancer.
 *)




let nb_occurs ids l = List.fold_left (fun acc (id,_,_) -> if List.mem id l then 1+acc else acc) 0 ids
and nb_occurs2 ids l = List.fold_left (fun acc id -> if List.mem id l then 1+acc else acc) 0 ids
;;

let rec faces_of_s0 ordre clist s0 =
  let l = List.filter (fun (_, (_, _, faceof, _, cofaceof, _)) -> List.mem s0 faceof) clist in
    trie_face l ordre

and trie_face faces ordre = faces (*match ordre with
  | [] -> []
  | id::suite ->
      try
	(id,(List.assoc id faces))::(trie_face faces suite)
      with
	| Not_found -> trie_face faces suite*)

and iter_faces_of_s0 ordre clist l = match l with
  | []        -> []
  | (id,b)::suite ->
      let faces_s0 = faces_of_s0 ordre clist id in
      let ifaces_s0 = iter_faces_of_s0 ordre clist faces_s0 in
	(id,b)::(ifaces_s0)@(iter_faces_of_s0 ordre clist suite)

and keep_first l acc = match l with
  | [] -> acc
  | t::q -> if (List.mem t acc) then keep_first q acc else keep_first q (acc@[t])

and tfaces_of_s0 ordre clist s0 = keep_first (iter_faces_of_s0 ordre clist (faces_of_s0 ordre clist s0)) []
;;


let rec iter_tsort glist clist = 
  let ordre = tsort glist clist in
    (** on prend le premier de la sequence et aussi ses faces. 
     ** On les met en premier et on itere le tsort pour la suite.
     **)
    (match ordre with
       | [] -> []
       | s0::l -> 
(**	   let face_s0 = List.filter (fun (_, (_, _, faceof, _, cofaceof, _)) -> List.mem s0 faceof) clist in  (** Dominante (co)face **) **)
	   let face_s0 =  tfaces_of_s0 ordre clist s0  in
	   let fs0 = List.map (fun (id, _) -> id) face_s0 in
	   let debut = s0::fs0 in(*(trie_face fs0 ordre) in*)
	   let reste = List.filter (fun (id, _) -> not(List.mem id debut)) clist
	   in debut@(iter_tsort glist reste)
    )
    
(* On trie la liste des faces d'un simplexe pour que cela respecte le tri topologique *)
and tsort glist clist =
  (** cl : prise en compte de face et tface  **)
  let cl = List.map 
	     (fun (id, (_, _, fo, tfo, cfo, tcfo)) ->
		let dep = fo@tfo in  (** Dominante (co)face **)
		let nb = nb_occurs2 dep (List.map fst clist)
		in (id, dep, nb))
	     clist
  in let tsorted = tsort_ [] cl in
    (** tri des classes d'quivalence de tsort **)
  let sorted = List.flatten (List.map (sort_class glist) tsorted)
  in List.map (function (id, _, _) -> id) sorted

and tsort_ res = function
  | [] -> res
  | l  ->
      let (zeros, l1) = List.partition (function (_, _, 0) -> true | _ -> false) l in
	if [] = zeros then
	  begin
	    let err = "Error: inconsistent face/coface specification in patch motif" in
	      Printf.printf "%s\n\t" err;
	      List.iter (function (id, _, _) -> Printf.printf "patch variable %s " id) l;
	      print_newline();
	      raise (Erreur err)
	  end
	else
	  let l2 = List.map (function (x, dep, pred) -> (x, dep, pred - (nb_occurs zeros dep))) l1
	  in tsort_ (res@[zeros]) l2

(* sort_class ordonne avec un critre additionnel les motifs qui sont dans la mme classe 
 * d'quivalence par la relation cl1/2 (qui reprsente la relation face-de (et eventuellement
 * coface-de)). Les critres sont les suivants (dans l'ordre du plus ou moins fort) :
 * - le plus de gardes
 * - le plus de faces (ou faces+cofaces)
 * - l'ordre alphabtique des variables
 *)
and sort_class glist l = List.sort (cmp_class glist) l

and cmp_class glist (id1, f1, cf1) (id2, f2, cf2) =
  let nbg1 = List.length (List.assoc id1 glist)
  and nbg2 = List.length (List.assoc id2 glist) in
  let cmp1 = Pervasives.compare nbg2 nbg1
  in if cmp1 <> 0 then cmp1 else
      begin
	let nbf1 = List.length f1
	and nbf2 = List.length f2 in
	let cmp2 = Pervasives.compare nbf2 nbf1
	in if cmp2 <> 0 then cmp2 else Pervasives.compare id1 id2
      end
;;


(* A partir de l'ordre calcul par iter_tsort, on doit rorganiser les gardes car la relation 
 * used_by  (occurence d'un identificateur dans une garde) n'est pas ncessairement
 * anti-symtrique ou compatible avec les faces. On brise les cycles si ncessaire, ce qui 
 * revient  dplacer des gardes d'une variable  une autre.
 *
 * l'ide pour rorganiser les gardes, c'est de parcourir la liste des simplexes dans l'ordre,
 * et de traiter tour a tour le premier simplexe de la liste. On ne garde pour chaque simplex, 
 * que les gardes qui ne contiennent pas une occurence des variables dsignant un simplexe non
 * encore trait. Les gardes qui font rfrence  une variable non encore traite sont mise dans 
 * la liste des gardes de la variable suivante.
 * L'argument gliste reprsente le dico des gardes (nom <-> liste de gardes) qu'on met  jour, 
 * l'argument clist reprsente la liste des simplex.
 *)

let rec reordonne_glist glist clist = reorganise_glist glist [] (sort_pattern clist)(*iter_tsort glist clist*)

and reorganise_glist glist res reste = match reste with
  | [] -> res
  | id::[] -> res@[(id, List.assoc id glist)]
  | id::((id2::_) as suite) ->
      let idrestant = List.fold_left (fun set e -> Terme.Ss.add e set) Terme.Ss.empty suite in
      let (ngl, rgl) = List.partition (occurs_free idrestant) (List.assoc id glist) in
      let moved_guards = move_guards id id2 rgl in
      let nglist = (id2, moved_guards@(List.assoc id2 glist))::glist
      in reorganise_glist nglist (res@[(id, ngl)]) suite
  
and occurs_free set = function 
  | Test e -> Terme.Ss.is_empty (Terme.Ss.inter set (Terme.free_var_term e))
  | Pred(p, e) -> Terme.Ss.is_empty (Terme.Ss.inter set (Terme.free_var_term e))

and move_guards idsource iddest guards = List.map (mvguard idsource iddest) guards

and mvguard idsource iddest = function
  | Pred(p, e) -> Test (Terme.App (Terme.Id "$test_pred", [Cte(VAL_string p); Cte(VAL_string(idsource)); e]))
  | g -> g
;;


(* --- Forme finale:
 * Finalement on produit une liste (id, (guards, faceof, tfaceof, cofaceof, tcofaceof))
 * avec dans les ?(co)?faceof que les id qui sont dj ordonnancs.
 *)

let rec reordonne (glist, clist) = 
  let (glist2, clist2) = normalise_patch (glist, clist) in
  let nglist = reordonne_glist glist2 clist2 in
  let completeglist = List.map (find_id clist2) nglist in
  let purgeglist = purge [] completeglist
  in purgeglist

and find_id clist = function (id, gl) -> 
  let (filtered, usedby, face, tface, coface, tcoface) = List.assoc id clist
  in (id, (filtered, gl, face, tface, coface, tcoface))

and purge debut = function 
  | [] -> []
  | (id, (f, gl, face, tface, coface, tcoface))::suite ->
      let ndebut = id::debut
      in (id, (f, gl, 
	       purgel ndebut face, 
	       purgel ndebut tface, 
	       purgel ndebut coface, 
	       purgel ndebut tcoface))::(purge ndebut suite)

and purgel deja l = List.filter (fun x -> List.mem x deja) l
;;



let print_motif plist =
  List.iter 
    (fun (id, (filtered, gl, face, tface, coface, tcoface)) ->
       Printf.printf "%s:\n\tfiltered= " id;
       Printf.printf "%s" (if (filtered) then "true" else "false") ;
       Printf.printf "\n\tfaces of= ";
       Printf.printf "\n\tguards= ";
       List.iter (function 
		    | Test e -> Printf.printf " %a &" Printer_term.prfterm2 e
		    | Pred(s, e)-> Printf.printf " %s=%a &" s Printer_term.prfterm2 e)
	 gl;
       Printf.printf "\n\tfaces of= ";
       List.iter (function x -> Printf.printf "%s " x) face;
       Printf.printf "\n\ttransitive faces of= ";
       List.iter (function x -> Printf.printf "%s " x) tface;
       Printf.printf "\n\tcofaces of= ";
       List.iter (function x -> Printf.printf "%s " x) coface;
       Printf.printf "\n\ttransitive cofaces of= ";
       List.iter (function x -> Printf.printf "%s " x) tcoface;
      Printf.printf "\n")
    plist
;;


(******************************************************************************)
(******************************************************************************)
(******************************************************************************)
(*                                                                            *)
(*  cration des patchs                                                       *)
(*                                                                            *)
(******************************************************************************)
(******************************************************************************)
(******************************************************************************)

let predefined_rule_vars = [
  ("'self", 0);
  ("self", 0)
];;

let indexation pvar =
  let index = ref (0)
  in List.fold_left (fun varindex var -> incr index; (var, !index)::varindex) predefined_rule_vars  pvar
;;

let rec pattern_var = function
  | []                                                    -> []
  | (id, (filtered, gl, face, tface, coface, tcoface))::q -> id::(pattern_var q)
;;

let env_argname = "$rule_env" ;;

let rec prepare_rhs varindex exp = Abs(env_argname, bind_pvar varindex exp)

and bind_pvar varindex e = 
  let cond s = List.mem_assoc s varindex 
  in
    match e
    with

      (** variable du pattern **)
      | (Id s) when cond s ->
	  op_apply1 env_argname (Cte (VAL_int (List.assoc s varindex)))
	    
      (** fonctions speciales: pos, left, neighbor, etc. qui n'apparaissent que dans une transformation **)
	    
      | App(Id ("pos"|"'pos"), [Id s])  when cond s -> 
	  op_apply1 env_argname (Cte(VAL_int(-(List.assoc s varindex))))
	    
      (** Propagation du traitement **)
	       
      | App(f, a)   -> App(bind_pvar varindex f, List.map (bind_pvar varindex) a) 
      | Abs (s, t)  -> Abs(s, bind_pvar (List.remove_assoc s varindex) t) 
      | If(a, b, c) -> If(bind_pvar varindex a, bind_pvar varindex b, bind_pvar varindex c) 
      | Fby(a, b)   -> Fby(bind_pvar varindex a, bind_pvar varindex b) 
      | Set(s, t)   -> Set(s, bind_pvar varindex t) 
      | x -> x 
;;

let compile_rhs_rule x = (if !Global.trace_input then Compile.traduit_evalue else Compile.eval)(x) 
;;


let rec instancie_ppattern rank varindex =
  let indexvar_one x  = List.assoc x varindex
  and indexvar_list l = List.map (fun x -> List.assoc x varindex) l
  and patvar = fst (List.split varindex)
  in function cont -> function
	[] -> cont
    | (id, (filtering, gl, face, tface, coface, tcoface))::q ->
	let cont' = instancie_ppattern (rank+1) varindex cont q in
	let funct = function
	  | Test (exp)    -> CTest(compile_rhs_rule(prepare_rhs varindex exp))
	  | Pred (id,exp) -> CPred(id,compile_rhs_rule(prepare_rhs varindex exp))
	in
	  (* (id, (filtered, List.map funct gl, face, tface, coface, tcoface))::cont' *)
	let cgl = List.map funct gl in
	let (cgl1,cgl2) = List.partition (function (CTest _) -> true | _ -> false) cgl in
	let cgl = cgl2@cgl1 in
(*	  if ( (face = []) && (tface = []) && (coface = []) && (tcoface = []) ) then
	    ((new indpt_ppattern
	       filtering
	       (indexvar_one  id)
	       id
	       gl
	       cgl
	       cont'
	    ) :> 'position ppattern)
	  else
	    ((new dpt_ppattern
	       filtering
	       (indexvar_one  id)
	       id
	       (indexvar_list face)
	       face
	       (indexvar_list tface)
	       tface
	       (indexvar_list coface)
	       coface
	       (indexvar_list tcoface)
	       tcoface
	       gl
	       cgl
	       cont'
	    ) :> 'position ppattern)
*)
	  ((new select_ppattern
	      rank
	      filtering
	      (indexvar_one  id)
	      id
	      (indexvar_list face)
	      face
	      (indexvar_list coface)
	      coface
	      gl
	      cgl
	      cont'
	   ) :> 'position ppattern)
;;

let rec instancie_rules rules = List.map instancie_one_rule rules

and instancie_one_rule (name, motif, rtype, exp) =
  let pvar = pattern_var motif in 
  let varindex = indexation pvar in
    begin
      assert begin if !trace_opt then 
	(Printf.printf "____ indexations des variables:\n";
	 List.iter (fun (id, idx) -> Printf.printf "\t%s -> %d\n" id idx) varindex);
	true
      end;
      { rname = name;
	motif = instancie_ppattern 0 varindex (new end_ppattern) motif;
	rtype = rtype;
	rhs = compile_rhs_rule(prepare_rhs varindex exp);
	nbvar = List.length pvar
      }
    end
;;

let patch rules =
  let rr_acell_gen       = instancie_rules rules
  and rr_ncell_gen       = instancie_rules rules
  in
    assert begin if !trace_opt then 
      List.iter (fun rr -> Printf.printf "\n" ; rr.motif#print ) rr_acell_gen;
      true
    end;
    function coll -> 
      match coll with
	| VAL_achain c -> (new ptopo_achain c coll)#patch rr_acell_gen
	| VAL_qmf c    -> (new ptopo_qmf c coll)#patch rr_ncell_gen
	| v           -> (let mesg = "Error: patch: bad argument type "^(constructor_string v)
			  in prerr_endline mesg; print_string "\t"; prvalue v; 
			    prerr_endline "\nis not yet handled in a patch (only aggregate types)"; 
			    raise (Erreur mesg))
;;

