(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: rec.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-02-02 23:23:52 $
 * Version: $Revision: 1.7 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    rec. Voir commentaire dans Seq/seq.ml
 *
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> rec (sauf extension)
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = seq
 *
 *=========================================================================
 *)


open Global;;
open Def_rec;;
open Valeur;;
open Core_rec;;
open Type;;
open Type_rec;;
open Dispatch_sig;;



__register_version "$RCSfile: rec.ml,v $" "$Revision: 1.7 $" "$Date: 2005-02-02 23:23:52 $";;



(** === Constructeurs ======================================================- **)

(* Etant donne la taille d'un record, la liste est une bonne structure de donnees
 * pour stocker les cls.
 *)
let _shape_rec (r, deco) = 
  let taille_r, pos = Core_rec.fold (fun k _ (nb, suite) -> (nb+1), k::suite) r (0, []) in
    VAL_funct (function 
		 | VAL_seq (s', _) -> 
		     if (taille_r <> Array.length s')
		     then raise_error "shape (record) : argument is not of the same size of the original"
		     else (let l = List.combine pos (Array.to_list s')
			   in
			     VAL_rec ((List.fold_right (fun (n, v) s -> Core_rec.add n v s) l Core_rec.empty), deco))
		 | _ -> raise_error "shape (record) : a sequence as argument is required")
;;

let _sequify_rec r =
  VAL_seq(Array.of_list (Core_rec.fold (fun _ v s -> v::s) (fst r) []), Type_seq.seq_desc) ;;




(** Extension --------------------------------- **)



(** --- record generic ----------------------------------------------------------- **)

let rec extension_record_generic desc a =
  check_vplist desc.name (to_funct desc.pospredicate) a;
  VAL_rec(Array.fold_left add_elem Core_rec.empty a, desc)

and add_elem map =  function
  | VAL_seq(aa, _) -> Core_rec.add aa.(0) aa.(1) map
  | _ -> failwith "Internal Error: extension: add_elem in extension_rec_generic"
;;


(** --- record avec spec. de champs ------------------------------------------------- **)

(** On ne rajoute que les champs qui sont prsent dans le type de
 ** record, en partant de la valeur par dfaut du type. Mais on ne 
 ** vrifie pas que les valeurs donnes sont correctement types.
 **)

let rec extension_record desc defaultval a = 
  check_vplist desc.name (to_funct desc.pospredicate) a;
  VAL_rec(Array.fold_left add_elem defaultval a, desc)

and add_elem map =  function
  | VAL_seq(aa, _) -> if Core_rec.mem aa.(0) map then Core_rec.add aa.(0) aa.(1) map else map
  | _ -> failwith "Internal Error: extension: add_elem in extension_record"
;;



(** === Prdicats, accs et observateurs =========================================== **)

(** _compareMgsVal_bag_bag est dfini dans core_bag *)

let _val2bool_rec (rc, _) =  VAL_bool(rc <> Core_rec.empty)
;;


let _size_rec x = VAL_int(Core_rec.size (fst x))
and  _empty_rec x = VAL_bool(Core_rec.empty = fst x)
;;




let _hd_rec s =
  let ss = fst s in if (0 = Core_rec.size ss) then VAL_undef ("hd on empty rec") else snd (Core_rec.hd ss)

and  _tl_rec s = 
  let ss = fst s
  in
    if (0 = Core_rec.size ss)
    then VAL_undef("tl on empty bag") 
    else VAL_rec((Core_rec.tl ss), snd s)
;;



(* === Fold et autre itrations ================================ *)

let _fold_funct_any_rec (fonc, zero, (rc, _)) =  Core_rec.fold (fun _ v suite -> (app_2 fonc v suite)) rc zero

and _map_funct_rec (fonc, (rc, c)) = VAL_rec ((Core_rec.map fonc rc), c)

and _iter_funct_rec (fonc, (rc, _)) = (Core_rec.iter   (fun _ x -> fonc x; ()) rc); VAL_undef "iter rec"

and _mapindexed_funct_rec (fct, (rc, t)) = VAL_rec((Core_rec.mapi (fun k v -> app_2 fct k v) rc), t)

and _iterindexed_funct_rec (fonc, (rc, _)) =
  (Core_rec.iter (fun k x -> app_2 fonc k x) rc); VAL_undef "iterindexed rc"

and _foldindexed_funct_any_rec (fct, zero, (tmap, _)) = 
  Core_rec.fold (fun k v acc -> app_3 fct k v acc) tmap zero
  
and _forall_funct_rec (p, (rc, _)) = 
  VAL_bool(Core_rec.fold (fun _ v acc -> acc && Auxfct.bool_from_val(p v)) rc true)

and _exists_funct_rec (p, (rc, _)) = 
  VAL_bool(Core_rec.fold (fun _ v acc -> acc or Auxfct.bool_from_val(p v)) rc false)

and _filter_funct_rec (p, (rc, t)) =
  VAL_rec((Core_rec.fold (fun k v acc -> 
			    if Auxfct.bool_from_val (p v) 
			    then  Core_rec.add k v acc else acc) 
	     rc Core_rec.empty), 
	  t)
;;





(* === Neighbors, neighbor ================================ *)

(** Topologie d'un record: rien n'est connect *)
let _neighbors_rec_any ((_, _), _) =  Seq.empty_seq
;;



(* === Fonctions spcifiques ======================================================- *)

 (** Attention: l'orde des arguments ets important car l'addition des records est asymtrique 
 ** c'est pourquoi il faut le faire squentiellement.
 **)

let _add_rec_rec ((x, tx), (y, ty)) = 
  VAL_rec(Core_rec.fold (fun k v m -> Core_rec.add k v m) y x, 
	  commontype tx ty)
;;

let _equal_rec_rec ((x,_), (y,_)) = VAL_bool(Core_rec.equal (fun a b -> to_bool(equal a b)) x y)
;;

let _compareMgsVal_rec_rec ((x, tx), (y, ty)) =
  let tdiff = compare_type tx ty in if tdiff <> 0 then VAL_int(tdiff) else VAL_int(Core_rec.compare x y)
;;

let _setfield_rec_any_any ((r, color), k, v) =  VAL_rec(Core_rec.add k v r, color)
;;

let _removeField_any_rec (k, (r, color)) = 
   VAL_rec((Core_rec.fold (fun k' v m -> if 0 = compare_mgs_value  k k' then m else Core_rec.add k' v m)
	      r Core_rec.empty),
	   color)
;;


let _take_rec_any ((r, t), s) = 
  try Core_rec.find s r 
  with Not_found -> VAL_undef("take: unknown field '"^(to_string (Printer.uprsval s))^"' in record")
;;







(* ================================================================ *)

(*** !!! Ne pas oublier de mettre  jour les champs
 ***     _ make 
 ***     _ xxxify
 ***     _ empty
 *** du descripteur de type 
 ***)

record_desc.predicate <- wrap_pred(coll_pred record_desc);;
record_desc.empty <- VAL_rec(Core_rec.empty, record_desc);;
record_desc.make <- extension_record_generic record_desc;;
