(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: ring.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-10-17 16:25:14 $
 * Version: $Revision: 1.6 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    rings.
 *
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml (ce fichier) suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> xxx (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = ring
 *
 *=========================================================================
 *)


open Global;;
open Def_ring;;
open Valeur;;
open Type;;
open Dispatch_sig;;



__register_version "$RCSfile: ring.ml,v $" "$Revision: 1.6 $" "$Date: 2007-10-17 16:25:14 $";;

(** alias **)
let ring_desc = Type_ring.ring_desc
;;
let empty = [||]
;;
let empty_ring = VAL_ring(empty, ring_desc)
;;


(** === Constructeurs ======================================================- **)

let _cons_any_ring (e, s) = VAL_ring(Array.append [|e|] (fst s), snd s)
and _cons_undef_ring (_, s) = VAL_ring s
;;


let _append_ring_ring (s1, s2) = VAL_ring(Array.append (fst s1) (fst s2), commontype (snd s1) (snd s2))
;;



let _join_ring_ring (s1, s2) = VAL_ring(Array.append (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _join_ring_undef (s, _) = VAL_ring s

and _join_undef_ring (_, s) = VAL_ring s

and _join_any_ring (e,s) = _cons_any_ring (e, s)
;;



(* pour tout xxx, shape_xxx (c, sequify (c)) = c *)

let _shape_ring (s, deco) = 
  let taille_s = Array.length s
  in
    VAL_funct (function 
		 | VAL_seq (s', _) -> 
		     (if (taille_s <> Array.length s')
		      then raise (Erreur "shape (seq) : argument is not of the same size of the original")
		      else VAL_seq (s', deco))
		 | _  -> raise (Erreur "shape (seq) : a sequence as argument is required"))
;;



let _sequify_ring s = VAL_seq(fst s, Type_seq.seq_desc)
;;


(* s : ('a array, desc), ie contenu d'une squence * couleur *)
let _ringify_seq s = VAL_ring (fst s,ring_desc) ;;

let _iota_int_ring (n, (v, col)) = 
  if (n <= 0)
  then raise (Erreur "iota_ring with a negative argument")
  else (let res = Array.make n (VAL_int 0)
	in
	  for i = 1 to (n-1)
	  do
	    res.(i) <- VAL_int i
	  done;
	  VAL_ring ((Array.append res v), col))
;;



(** Extension --------------------------------- **)

(* definition d'une collection par extension *)

let rec extension_ring desc a = 
  check_vplist desc.name (to_funct desc.pospredicate) a;
  (*verifie que les positions dans la definition par extension ont le bon type *)
  
  let sup = Array.fold_left maxsize 0 a in
  let tab = Array.make (1+sup) (VAL_undef "default value in ring extension")
  in Array.iter (set_tab tab) a;
    VAL_ring(tab, desc)
      
and maxsize m = function
| VAL_seq(aa, _) -> max m (to_int aa.(0))
| _ -> raise (Erreur "Internal Error: extension: maxsize in extension_ring")

and set_tab tab = function
  | VAL_seq(aa, _) -> tab.(to_int aa.(0)) <- aa.(1)
  | _ -> raise (Erreur "Internal Error: extension: set_tab in extension_ring")
;;



(** === Prdicats, accs et observateurs =========================================== **)

let rec _compareMgsVal_ring_ring ((x,tx), (y,ty)) = 
  let tdiff = compare_type tx ty
  in if tdiff <> 0 then VAL_int(tdiff) else
      begin
	let lx = Array.length x
	and ly = Array.length y in
	let diff = lx - ly 
	in (if diff < 0 then VAL_int(-1)
	    else if diff > 0 then VAL_int(1)
	    else compare_ring_aux lx x y)
      end

and compare_ring_aux lg x y =
  let res = ref 0 in
  try
    for i = 0 to lg - 1 do
      res := compare_mgs_value x.(i) y.(i);
      if 0 <> !res then raise BreakLoop
    done;
    mgszero
  with BreakLoop -> VAL_int(!res)


and _val2bool_ring x = VAL_bool(0 <> Array.length (fst x))
;;


let _size_ring x = VAL_int(Array.length (fst x))
and _empty_ring x = VAL_bool(0 = Array.length (fst x))
;;


exception Found of mgs_value;;
exception BreakLoop;;

(* compte le nombre d'occurences d'un el dans un ring *)
let _count_any_ring (a, (s, _)) = 
  let nb = ref 0
  in
    Array.iter (function e -> if (0 = compare_mgs_value e a) then (incr nb)) s;
    VAL_int (!nb)

let _member_any_ring (a, (s, _)) = 
  try
    Array.iter (function e -> if (0 = compare_mgs_value e  a) then raise BreakLoop) s;
    VAL_bool false
  with BreakLoop -> VAL_bool true
;;



let _random_ring (s, _) = 
  let n = Array.length(s)
  in if n == 0 then VAL_undef "random: bad argument: empty ring"  else s.(Random.int(n))
;;



let _hd_ring s = 
  let ss = fst s in if (0 = Array.length ss) then VAL_undef("hd on empty ring") else ss.(0)

and _tl_ring s = 
  let ss  = fst s in
  let len = Array.length ss
  in
    if (0 = len)
    then VAL_undef("tl on empty ring") 
    else VAL_ring(Array.sub ss 1 (len - 1), snd s)
;;

let _take_ring_int ((s, col), n) = let i = Array.length s in s.( (n+i) mod i)
(*  begin
    if n >= 0 then 
      (if n < Array.length s then s.(n) else VAL_undef "take out of range")
    else
      let idx = (Array.length s) + n in
	(if idx <  Array.length s then s.(idx) else VAL_undef "take out of range")
  end*)
;;



(* === Fold et autre itrations ================================ *)

let _fold_funct_any_ring (fonc, zero, (vect, _)) = Array.fold_right (fun v suite -> (app_2 fonc v suite)) vect zero
and _map_funct_ring (fonc, (vect, c)) = VAL_ring ((Array.map fonc vect), c)
and _iter_funct_ring (fonc, (vect, _)) = (Array.iter  (fun x   -> fonc x; ()) vect); VAL_undef "iter ring"
and _mapindexed_funct_ring (fct, (vect, t)) =  VAL_ring ((Array.mapi (fun i x -> (app_2 fct (VAL_int i) x)) vect), t)
and _iterindexed_funct_ring (fonc, (vect, _)) = (Array.iteri (fun idx x -> ignore(app_2 fonc (VAL_int idx) x)) vect); VAL_undef "iter"
and _foldindexed_funct_any_ring (fct, zero, (vect, t)) = 
  let acc = ref zero 
  and sz = Array.length vect in
    begin
      for i = sz-1 downto 0 do
	acc := app_3 fct (VAL_int i) vect.(i) !acc
      done;
      !acc
    end  

and _forall_funct_ring (p, s) = VAL_bool(Global.array_forall (fun x -> Auxfct.bool_from_val (p x)) (fst s))
and _exists_funct_ring (p, s) = VAL_bool(Global.array_exists (fun x -> Auxfct.bool_from_val (p x)) (fst s))
and _filter_funct_ring (p, s) = 
  VAL_ring(Array.of_list (List.filter (fun x -> Auxfct.bool_from_val(p x)) (Array.to_list (fst s))),
	  snd s)
;;







(* === Neighbors, neighbor ================================ *)

(* i : position considre *)

let _neighbors_ring_int ((t, _), i) = 
  let sz         = Array.length t 
  and retourne v = VAL_seq (v, Type_seq.seq_desc)
  in
    assert((i >= 0) && (i < sz));             (* Acces hors des bornes    *)
    if (sz = 1)                               (* Taille de 1 => 0 voisins *)
    then (retourne [||])
    else 
      if (i = 0)                              (* Index 0     => 1 voisins *)
      then (retourne [| t.(1) |])
      else 
	if (i = (sz-1))                       (* Index max   => 1 voisins *)
	then (retourne [| t.(0) |])
	else (retourne [| t.(i+1) |])


and _left_ring_int ((t, _), i) = 
  let sz = Array.length t
  in
    assert((i >= 0) && (i < sz));
    if (i > 0)
    then t.(i-1)
    else if i = 0 then t.(sz-1) else VAL_undef "negative position"

and _right_ring_int ((t, _), i) = 
  let sz = Array.length t
  in
    assert((i >= 0) && (i < sz));
    if (i < sz-1)
    then t.(i+1)
    else if i = sz-1 then t.(0) else VAL_undef "right out of bound"




(* === Fonctions spcifiques ======================================================- *)

(* tri : pas de sens donc pas implment *)




let _last_ring (v, col) =
  let len = Array.length v in 
    if (len > 0) 
    then v.(len-1)
    else VAL_undef "last ring"
;;


(* Flatten ----------------------------------------------------------- *)


let rec _flatten_ring (s, t) =  VAL_ring((flatten_ring_aux s), t)

and flatten_ring_aux s =
   let l = Array.map (function VAL_ring (a,_) -> Array.to_list a | y -> [y]) s in
   let ll = Array.fold_right (fun x acc -> x@acc) l []
   in Array.of_list ll
;;



(** Une fonction partition sur les monoides ----------------------------------- *)


let _partition_int_ring (n, (s, t)) =
  if n <= 0 then VAL_undef "partition: bad negative argument" else
    begin
      let lg = Array.length s in
      let sz = 1+(lg-1)/n in
      let a = Array.init sz 
		(fun i -> 
		   let debut = i*n in let len = (Pervasives.min lg (debut+n)) - debut 
		   in VAL_ring(Array.sub s debut len, t))
      in VAL_seq(a, Type_seq.seq_desc)
    end
;;



(** Partition en 2 classes suivant un predicat ----------------------------------------------- *)


let _partition_funct_ring (f, (s, t)) = 
  let (l1, l2) = Array.fold_left 
		   (fun (ll1, ll2) x -> 
		      if Auxfct.bool_from_val (f x) 
		      then (x::ll1, ll2)
		      else (ll1, x::ll2))
		   ([], [])
		   s
  in let a1 = Array.of_list (List.rev l1)
     and a2 = Array.of_list (List.rev l2)
  in let a = Array.make 2 (VAL_undef "dummy _partition_funct_ring")
  in 
    a.(0) <- VAL_ring(a1, t); 
    a.(1) <- VAL_ring(a2, t); 
    VAL_seq(a, Type_seq.seq_desc)
;;
  




(** Une fonction partition par quivalence sur les monoide ----------------------------------- *)
(* pas fait : qu'est-ce que c'est ??? *)


(** Fonctions reverse et rotate sur les sequences ----------------------------------- *)

let _reversering_ring (s, t) =
  let lg = (Array.length s) - 1
  in VAL_ring(Array.init (lg+1) (function i -> s.(lg - i)), t)

and _rotateleft_ring (s, t) =
  let lg = (Array.length s)
  in VAL_ring(Array.init lg (function i -> if i+1 < lg then s.(i+1) else s.(0)), t)

and _rotateright_ring (s, t) =
  let lg = (Array.length s)
  in VAL_ring(Array.init lg (function i -> if i >= 1 then s.(i-1) else s.(lg-1)), t)
;;


(** Fonctions pad sur les sequences ----------------------------------- *)
(* pour ajouter n occurences d'une valeur en tte *)

let _padLeft_ring_int_any ((s, t), n, v0) =
  let lg = Array.length s in
    if lg == n then VAL_ring(s, t)
    else if lg > n then let deb = lg - n in VAL_ring(Array.init n (function i -> s.(i+deb)), t)
    else (let deb = n - lg 
	  in VAL_ring(Array.init n (function i -> if i < deb then v0 else s.(i - deb)), t))

let _padRight_ring_int_any ((s, t), n, v0) =
  let lg = Array.length s in
    if lg == n then VAL_ring(s, t)
    else if lg > n then VAL_ring(Array.init n (function i -> s.(i)), t)
    else VAL_ring(Array.init n (function i -> if i >= lg then v0 else s.(i)), t)
;;




(* ================================================================ *)

(*** !!! Ne pas oublier de mettre  jour les champs
 ***     _ make 
 ***     _ xxxify
 ***     _ empty
 *** du descripteur de type ici
 ***)
ring_desc.make <- extension_ring ring_desc;;  
ring_desc.xxxify <- VAL_funct(ringify);;
ring_desc.empty <- empty_ring;;

























