(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: type_scalar.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Voir le fichier Common/type.ml
 * 
 *=========================================================================
 *)



open Global;;
open Valeur;;
open Type;;

__register_version "$RCSfile: type_scalar.ml,v $" "$Revision: 1.2 $" "$Date: 2005-11-23 11:10:57 $";;


(*!*)

(** 
       @module scalar
       @name 'scalar
       @type any_bool
       Scalar type and predicate.
*)

let scalar_desc = create_type
		 { name = "'scalar";
		   concrete = false;
		   leibniz = false;
		   predicate = wrap_pred(fun x -> not(is_collection x));
		   empty = VAL_undef "Scalar empty value = scalar default value";
		   pospredicate = VAL_undef "Scalar pospredicate not available: what does this mean?";
		   make = (fun _ -> VAL_undef "Scalar make not available: what does this mean?");
		   xxxify = VAL_undef "Scalar xxxify not available: what does this mean?";
		   derive_from = [any_desc];
		   father_of = [];
		   uid = get_new_uid();
		 }
;;



(**
  @module scalar
  @name 'undef
  @type any_bool 
  Returns [true] if the argument is an undefined value 
  and [false] for any other value. 

  [undef] is also the type of the undefined values. The undefined
  values are denoted by {[ <undef:xxx> ]} where [xxx] is any sequence
  of characters not containing a carriage return [\n]. This string is
  supposed to be an informative message. This message can be omitted,
  simply write: [<undef>]. However, this message does not differentiate
  undefined values: there is essentially only one undefined value:
  {[ <undef: bad argument> == <undef: divided by 0> ]}
  returns [true].
*)
let undef_desc = create_type
	        { name = "'undef";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_undef);
		  empty = VAL_undef("undef default value");
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on undef";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;

(** 
  @module scalar
  @name 'bool
  @type any_bool 
  Returns [true] if the argument is a boolean and [false] for any
  other value. 

  [bool] is also the type of booleans (truth values). There is only
  two booleans denoted by [true] and [false].

  Any MGS value can be converted into a boolean. This
  conversion occurs {i implicitly} for instance in the condition of
  a [if ... then ... else ... fi] expression. This conversion can be
  done {i explicitly} using the function [val2bool]. 
  See the the function [val2bool] for the conversion
  rules.

  @voir scalar.'val2bool the[val2bool] predicate.

*)
let bool_desc = create_type
	        { name = "'bool";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_bool);
		  empty = VAL_bool(false);
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on bool";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;


(**
  @module scalar  
  @name 'int
  @type any_bool 
  Returns [true] if the argument is an integer and [false] for any other value. 

  [int] is also the type of the integer numbers. Integers are 31 bits wide.
   All arithmetic operations are taken modulo 2{^31}. They do not fail
  on overflow.
*)
let int_desc = create_type
	        { name = "'int";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_int);
		  empty = VAL_int(0);
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on int";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;



(**
  @module scalar  
  @name 'bint
  @type any_bool 
  Returns [true] if the argument is a big integer and [false] for any other value. 

  [bint] is also the type of the big integer numbers, that are arbitrary-precision
  intergers.
*)
let bint_desc = create_type
	        { name = "'bint";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_bint);
		  empty = VAL_bint(Big_int.zero_big_int);
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on bint";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;



(**
  @module scalar
  @name 'symbol
  @type any_bool 
  Returns [true] if the argument is a symbol 
  and [false] for any other value. 

  [symbol] is the type of symbols. An MGS's symbol begins
  with a backquote [`]. Symbols can only be compared with other
  symbols.
*)

let symbol_desc = create_type
	        { name = "'symbol";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_symbol);
		  empty = VAL_symbol "`Default";
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on symbol";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;


(**
  @module scalar
  @name 'string
  @type any_bool 
  Returns [true] if the argument is a string 
  and [false] for any other value. 

  [string] is the type of character strings. An MGS's string begins
  with a quote ["] and ends with another ["]. They can include
  carriage-returns. Some characters are represented by a [ \ ] followed
  by a letter: [ \t ] represents a tabulation and [ \n ] represents a
  carriage-returns. Any ascci character can be denoted with a  [ \ ]
  followed by its ascci code. For instance [ "\65" ] corresponds to
  [ "A" ].

  @voir scalar string_length 
  @voir math int_of 
  @voir math float_of 
  @voir scalar string_get 
  @voir scalar string_make
  @voir scalar string_index
  @voir scalar string_rindex
  @voir scalar string_index_from
  @voir scalar string_rindex_from
  @voir scalar string_rindex_from
  @voir scalar string_contains
  @voir scalar string_contains_from
  @voir scalar string_rcontains_from
  @voir scalar string_to_seq
*)
let string_desc = create_type
	        { name = "'string";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_string);
		  empty = VAL_string "";
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on string";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;


(**
  @module functional
  @name 'funct
  @type any_bool 
  Returns [true] if the argument is a functional value 
  and [false] for any other value. 

  [funct] is the type of functionnal values. The denotation of
  functional values relies on a syntax similar to the lambda-calulus
  syntax:
  {[
  \x.x+1
  ]}
  denotes the function that add one to its argument.

  <b>Note:</b> a <i>transformation</i> is also a function (defined by
  cases on the paths in a collection and specified through a special
  syntax). So:
  {[
  funct(trans { x => x+1 });
  ]}
  returns true because [trans { x => x+1 }] is an anonymous transformation
  and hence, a functional value.


  @voir Functional functional

*)
let funct_desc = create_type
	        { name = "'funct";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_funct);
		  empty = VAL_funct(fun x -> x);
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on funct";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;





(**
  @module scalar
  @name float
  @type any_bool 
  Returns [true] if the argument is a floating point value 
  and [false] for any other value.

  [float] is the type of floating-point numbers. MGS's floating-point
  numbers follow the Caml's floating-point numbers which, in turn,
  follow the IEEE 754 standard, using double precision (64 bits)
  numbers.  

  Floating-point operations never raise an exception on overflow,
  underflow, division by zero, etc. Instead, special IEEE numbers are
  returned as appropriate, such as [infinity] for [1.0/0.0],
  [neg_infinity] for [-1.0/0.0], and [nan] (``not a number'') for
  [0.0/0.0].  These special numbers then propagate through
  floating-point computations as expected: for instance, [1.0/infinity] 
  is [0.0], and any operation with [nan] as argument
  returns [nan] as result.
  
  @voir math math
*)
let float_desc = create_type
	        { name = "'float";
		  concrete = true;
		  leibniz = false;
		  predicate = wrap_pred(is_float);
		  empty = VAL_float 0.0;
		  pospredicate = scalar_desc.pospredicate;
		  make = scalar_desc.make;
		  xxxify = VAL_undef "Type_scalar: undef_desc: not yet implemented on float";
		  derive_from = [scalar_desc];
		  father_of = [];
		  uid = get_new_uid();
		}
;;


