(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_seq.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.3 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *          
 * Ici on developpe les instance de generic_topo qui correspondent 
 * aux collections monoidales.
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Topology;;

__register_version "$RCSfile: topo_seq.ml,v $" "$Revision: 1.3 $" "$Date: 2005-11-23 11:10:57 $";;



(* =================================================================
 * === Sequence ====================================================
 * =================================================================
 *
 * POSITION :
 * Les positions dans une sequence sont representes par des entiers,
 * representant l'index dans le tableau (array.mli) qui implemente la
 * sequence. Ces index debutent en 0.
 *
 * MARQUAGE :
 * Le marquage se gere avec un tableau _used de
 * "usage_position". L'usage de l'element de position p est donne par 
 * la valeur _used.(p).
 *
 * POSITION INITIALE :
 * La position initiale utilisee pour commencer a filtrer est une des
 * positions tire au sort (entre 0 et la taille du tableau) avec la 
 * fonction Global.crandom. Cette fonction retourne un vrai nombre 
 * aleatoire ou bien 0 suivant le parametre boolean match_random. Cela
 * afin de "melanger un peu" l'application des rgles, sauf si
 * match_random est a false (pour le debug). Essayer avec le programme 
 * MGS suivant :
 *      !set match_random := true;;
 *      a:= 3;;
 *      trans Q = { x => (a := a + 1; a) };;
 *      Q((1, 2, 3));;
 * Le resultat doit etre 4, 5, 6 mais dans des ordres differents si on
 * value plusieurs fois la derniere expression.
 *
 * BATCH :
 * Un batch de sequence est une liste de triplet (pt, p, v) ou
 *   - pt est le type de path: flat ou pas flat
 *   - p est un path, i.e. une liste de position
 *
 * Pour reconstruire une sequence  partir d'un array c et d'un batch
 * b, on utilise la method build_result qui appelle la fonction
 * merge_seq avec un batch trie, ce qui consiste a resequencer les
 * triplets de la liste. Un batch se trie suivant l'indice du
 * premier element dans le path.
 *
 * La fonction  auxilliaire mergeseq fait le travail de fusion des
 * donnees dans le batch et dans le tableau c
 *
 * ================================================================-
 *)





exception FoundIntPos of int;;
exception FoundMgsValue of mgs_value;;
exception FoundMgsValue2 of mgs_value;;

(** --- Print d'un batch de Sequence ----------------------- **)

let rec print_seq_batch chan valpos batch = 
  Printf.fprintf chan "-- batchSEQ : begin\n";
  List.iter (print_one_occurence chan valpos) batch;
  Printf.fprintf chan "-- batchSEQ : end\n";
  flush chan

and print_one_occurence chan valpos  (pt, pl, v) =
  Printf.fprintf chan "\t%s %a --> %a\n"
    (if pt.a_flat = FlatPath then "flat" else "noFlat")
    (fun o l -> List.iter (function (p) -> Printf.fprintf chan " (pos=%d, val=%a)" p prfvalue2 (valpos p)) l)
    pl
    prfvalue2 v;
  flush chan
;;



(** --- Tri d'un batch de Sequence par position croissante ----------------------- **)

let rec compare_seq_occurence posval q1 q2 = cso posval (q1, q2)

and cso posval = 
  let err = "Error: internal error: compare_seq_occurence: bad batch\n"
  in (function (_, p1::_, _),  (_, p2::_, _) -> p1 - p2 
	| (b1, b2) -> 
	    Printf.fprintf stderr "%s" err;
	    print_one_occurence stderr posval b1; print_one_occurence stderr posval b2; 
	    raise (Erreur err))
;;




(** --- Construction d'une sequence a partir d'un batch et d'une sequence ----------------------- **)

let rec merge_seq couleur c = function
  | [] -> c   (* batch vide: rien a faire *)
  | b -> Array.concat (mergeseq couleur c 0 b)

and mergeseq couleur c i = function 
  (** batch vide **)
  | [] -> let last_index = Array.length c 
          in if i > last_index then [] else [Array.sub c i (last_index - i)]

  (** traitement d'un chemin vide et cas non-flat: ca arrive avec x* **)
  | (rp, [], v)::b
      when (rp.a_flat = NoFlatPath) 
	or not(is_collection v)
	or not(Type.is_subtype (collection_type v) couleur) ->
      (Array.make 1 v)::(mergeseq couleur c i b)

  (** traitement d'un chemin vide et cas flat: ca arrive avec x* **)
  | (_, [], v)::b ->
      assert(is_collection v);
      assert(Type.is_subtype (collection_type v) couleur);
      let v = Array.of_list
	(List.filter (function (VAL_undef _) -> false | _ -> true) (Array.to_list (to_seq v)))
      in
	v::(mergeseq couleur c i b)
      
  (** traitement de la valeur undef : a s'limine (seq est un leibnizien) **)
  | (_, p, VAL_undef _)::b  ->
      let i0 = List.hd p and i1 = last p in
	if i < i0 then (Array.sub c i (i0 - i))::(mergeseq couleur c (i1+1) b)
	else if i0 <= i then (mergeseq couleur c (i1+1) b) 
	else raise (Erreur "Internal error: batch file for seq: impossible case (c)")

  (** traitement du cas non-flat: on insere v comme 1 seul element **)
  | (rp, p, v)::b 
      when (rp.a_flat = NoFlatPath) 
	or not(is_collection v)
	or not(Type.is_subtype (collection_type v) couleur) ->
      let i0 = List.hd p and i1 = last p in
	if i < i0 then (Array.sub c i (i0 - i))::(Array.make 1 v)::(mergeseq couleur c (i1+1) b)
	else if i0 <= i then (Array.make 1 v)::(mergeseq couleur c (i1+1) b) 
	else raise (Erreur "Internal error: batch file for seq: impossible case (a)")

  (** traitement du cas flat: v est une collection compatible et il faut l'inserer **)
  |  (_, p, v)::b -> 
       assert(is_collection v);
       assert(Type.is_subtype (collection_type v) couleur);
       let i0 = List.hd p and i1 = last p in
       let v = Array.of_list
	 (List.filter (function (VAL_undef _) -> false | _ -> true) (Array.to_list (to_seq v)))
       in
	 if i < i0 then (Array.sub c i (i0 - i))::(v)::(mergeseq couleur c (i1+1) b)
	 else if i0 <= i then (v)::(mergeseq couleur c (i1+1) b)
	 else raise (Erreur "Internal error: batch file for seq: impossible case (b)")

and last = function
  | a::[] -> a
  | _::l -> last l
  | _ -> raise (Erreur "Internal error: last (file topology.ml)\n")

;;





(** --- La topologie d'une sequence ------------------------------ **)

class topo_sequence (c, couleur) mgsval =
object(self)
  inherit [int] generic_topo 
    string_of_int                     (** afficheur de position    **)
    (c, couleur)                      (** collection a transformer **)
    mgsval                            (** la valeur mgs complete sur laquelle on travaille **)
    (Global.crandom (Array.length c)) (** position de depart       **)

  val _lg = Array.length c
  val _used = Array.make (Array.length c) PosFree
  val mutable _batch = ([] : (rule_property * 'position path * mgs_value) list)

  method state p          = _used.(p)
  method update_state f p = _used.(p) <- f(_used.(p))

  method unvisit_all = for i = 0 to _lg-1 do (if self#visited(i) then self#unvisit(i)) done

  method next_usable p =
    let delta = if !Global.match_random >= 2 then (Global.crandom _lg) else 0 in
      try 
	for i = 0 to _lg - 1 do
	  assert begin if !trace_opt then 
	    (Printf.printf "---- topoSEQ: next_usable: try %d\n" ((p+i) mod _lg); flush stdout);
	    true
	  end;
	  let pp = (p+i+delta) mod _lg 
	  in if not(self#used pp) & not(self#visited pp) then raise (FoundIntPos pp)
	done;
	raise NoMoreUsable
      with FoundIntPos p -> 
	assert begin if !trace_opt then 
	  (Printf.printf "---- topoSEQ: next_usable: found %d\n" p); 
	  true 
	end;
	p

  method add_match pt p v = _batch <- (pt, p, v)::_batch

  method get_value_from_pos p = c.(p)
  method mgs_value_from_pos p = VAL_int(p)

  method some_neighbors p = function
    | AllDirections -> (if (p >= (_lg-1)) or self#used(p+1) 
			then (0, [||])
			else (assert(p+1 < _lg); (1, [| p+1 |])))

    | SomeDirections(d) -> (let err = "Error: cannot use directions in seq\n"
				      ^(Array.fold_left (fun s x -> s^"\t"^(to_string (prsvalue x))^"\n") "" d)
			    in 
			      prerr_endline err; 
			      raise (Erreur err))

  method direction_from_to p1 p2 = VAL_int(p2 - p1) (** Ou VAL_undef, je ne sais pas ce qui est le mieux **)

  method build_result =
    let sorted_batch = List.sort (compare_seq_occurence (self#get_value_from_pos)) _batch
    in 
      VAL_seq(merge_seq couleur c sorted_batch, couleur)

end;;

