(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_set.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-10-08 11:57:37 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *     La representation d'un ensemble (ou set)
 *     reprsentation adapte  partir des arbres binaires balances (cf. set).
 *
 *=========================================================================
 *)


open Def_set;;
open Valeur;;
open Dispatch_sig;;


Global.__register_version "$RCSfile: core_set.ml,v $" "$Revision: 1.2 $" "$Date: 2004-10-08 11:57:37 $";;



let compare_elt = compare_mgs_value
;;

let empty = Def_set.Empty_set
;;


(* Sets are represented by balanced binary trees (the heights of the
   children differ by at most 2 *)

let set_height = function
    Empty_set -> 0
  | Node_set(_, _, _, h) -> h
;;

(* Creates a new node with left son l, value x and right son r.
   l and r must be balanced and | height l - height r | <= 2.
   Inline expansion of height for better speed. *)

let set_create l x r =
  let hl = match l with Empty_set -> 0 | Node_set(_,_,_,h) -> h in
  let hr = match r with Empty_set -> 0 | Node_set(_,_,_,h) -> h in
    Node_set(l, x, r, (if hl >= hr then hl + 1 else hr + 1))
;;

(* Same as create, but performs one step of rebalancing if necessary.
   Assumes l and r balanced.
   Inline expansion of create for better speed in the most frequent case
   where no rebalancing is required. *)

let set_bal l x r =
  let hl = match l with Empty_set -> 0 | Node_set(_,_,_,h) -> h in
  let hr = match r with Empty_set -> 0 | Node_set(_,_,_,h) -> h in
    if hl > hr + 2 then begin
      match l with
          Empty_set -> invalid_arg "Set.set_bal"
        | Node_set(ll, lv, lr, _) ->
            if set_height ll >= set_height lr then
              set_create ll lv (set_create lr x r)
            else begin
              match lr with
                  Empty_set -> invalid_arg "Set.set_bal"
		| Node_set(lrl, lrv, lrr, _)->
                    set_create (set_create ll lv lrl) lrv (set_create lrr x r)
            end
    end else if hr > hl + 2 then begin
      match r with
          Empty_set -> invalid_arg "Set.set_bal"
        | Node_set(rl, rv, rr, _) ->
            if set_height rr >= set_height rl then
              set_create (set_create l x rl) rv rr
            else begin
              match rl with
                  Empty_set -> invalid_arg "Set.set_bal"
		| Node_set(rll, rlv, rlr, _) ->
                    set_create (set_create l x rll) rlv (set_create rlr rv rr)
            end
    end else
      Node_set(l, x, r, (if hl >= hr then hl + 1 else hr + 1))
;;


(* Same as bal, but repeat rebalancing until the final result
   is balanced. *)

let rec set_join l x r =
  match set_bal l x r with
      Empty_set -> invalid_arg "Set.set_join"
    | Node_set(l', x', r', _) as t' ->
        let d = set_height l' - set_height r' in
          if d < -2 or d > 2 then set_join l' x' r' else t'
;;


(* Merge two trees l and r into one.
   All elements of l must precede the elements of r.
   Assumes | height l - height r | <= 2. *)

let rec set_merge t1 t2 =
  match (t1, t2) with
      (Empty_set, t) -> t
    | (t, Empty_set) -> t
    | (Node_set(l1, v1, r1, h1), Node_set(l2, v2, r2, h2)) ->
        set_bal l1 v1 (set_bal (set_merge r1 l2) v2 r2)
;;

(* Same as merge, but does not assume anything about l and r. *)

let rec set_concat t1 t2 =
  match (t1, t2) with
      (Empty_set, t) -> t
    | (t, Empty_set) -> t
    | (Node_set(l1, v1, r1, h1), Node_set(l2, v2, r2, h2)) ->
        set_join l1 v1 (set_join (set_concat r1 l2) v2 r2)
;;

(* Splitting *)

let rec set_split x = function
    Empty_set ->
      (Empty_set, None, Empty_set)
  | Node_set(l, v, r, _) ->
      let c = compare_mgs_value x v in
        if c = 0 then (l, Some v, r)
        else if c < 0 then
          let (ll, vl, rl) = set_split x l in (ll, vl, set_join rl v r)
        else
          let (lr, vr, rr) = set_split x r in (set_join l v lr, vr, rr)
;;

(* Implementation of the set operations *)

let set_empty = Empty_set
;;

let set_is_empty = function Empty_set -> true | _ -> false
;;

let rec set_mem x = function
    Empty_set -> false
  | Node_set(l, v, r, _) ->
      let c = compare_mgs_value x v in
        c = 0 || set_mem x (if c < 0 then l else r)
;;

let rec set_add x = function
    Empty_set -> Node_set(Empty_set, x, Empty_set, 1)
  | Node_set(l, v, r, _) as t ->
      let c = compare_mgs_value x v in
        if c = 0 then t else
          if c < 0 then set_bal (set_add x l) v r else set_bal l v (set_add x r)
;;

let set_singleton x = Node_set(Empty_set, x, Empty_set, 1)
;;

let rec set_remove x = function
    Empty_set -> Empty_set
  | Node_set(l, v, r, _) ->
      let c = compare_mgs_value x v in
        if c = 0 then set_merge l r else
          if c < 0 then set_bal (set_remove x l) v r else set_bal l v (set_remove x r)
;;

let rec set_union s1 s2 =
  match (s1, s2) with
      (Empty_set, t2) -> t2
    | (t1, Empty_set) -> t1
    | (Node_set(l1, v1, r1, h1), Node_set(l2, v2, r2, h2)) ->
        if h1 >= h2 then
          if h2 = 1 then set_add v2 s1 else begin
            let (l2, _, r2) = set_split v1 s2 in
              set_join (set_union l1 l2) v1 (set_union r1 r2)
          end
        else
          if h1 = 1 then set_add v1 s2 else begin
            let (l1, _, r1) = set_split v2 s1 in
              set_join (set_union l1 l2) v2 (set_union r1 r2)
          end
;;

let rec set_inter s1 s2 =
  match (s1, s2) with
      (Empty_set, t2) -> Empty_set
    | (t1, Empty_set) -> Empty_set
    | (Node_set(l1, v1, r1, _), t2) ->
        match set_split v1 t2 with
            (l2, None, r2) ->
              set_concat (set_inter l1 l2) (set_inter r1 r2)
          | (l2, Some _, r2) ->
              set_join (set_inter l1 l2) v1 (set_inter r1 r2)
;;

let rec set_diff s1 s2 =
  match (s1, s2) with
      (Empty_set, t2) -> Empty_set
    | (t1, Empty_set) -> t1
    | (Node_set(l1, v1, r1, _), t2) ->
        match set_split v1 t2 with
            (l2, None, r2) ->
              set_join (set_diff l1 l2) v1 (set_diff r1 r2)
          | (l2, Some _, r2) ->
              set_concat (set_diff l1 l2) (set_diff r1 r2)
;;

let rec set_compare_aux l1 l2 =
  match (l1, l2) with
      ([], []) -> 0
    | ([], _)  -> -1
    | (_, []) -> 1
    | (Empty_set :: t1, Empty_set :: t2) ->
        set_compare_aux t1 t2
    | (Node_set(Empty_set, v1, r1, _) :: t1, Node_set(Empty_set, v2, r2, _) :: t2) ->
        let c = compare_mgs_value v1 v2 in
          if c <> 0 then c else set_compare_aux (r1::t1) (r2::t2)
    | (Node_set(l1, v1, r1, _) :: t1, t2) ->
        set_compare_aux (l1 :: Node_set(Empty_set, v1, r1, 0) :: t1) t2
    | (t1, Node_set(l2, v2, r2, _) :: t2) ->
        set_compare_aux t1 (l2 :: Node_set(Empty_set, v2, r2, 0) :: t2)
;;

let set_compare s1 s2 = set_compare_aux [s1] [s2]
;;


let set_equal s1 s2 = set_compare s1 s2 = 0
;;

let rec set_subset s1 s2 =
  match (s1, s2) with
      Empty_set, _ ->
        true
    | _, Empty_set ->
        false
    | Node_set (l1, v1, r1, _), (Node_set (l2, v2, r2, _) as t2) ->
        let c = compare_mgs_value v1 v2 in
          if c = 0 then
            set_subset l1 l2 && set_subset r1 r2
          else if c < 0 then
            set_subset (Node_set (l1, v1, Empty_set, 0)) l2 && set_subset r1 t2
          else
            set_subset (Node_set (Empty_set, v1, r1, 0)) r2 && set_subset l1 t2
;;

let rec set_iter f = function
    Empty_set -> ()
  | Node_set(l, v, r, _) -> set_iter f l; f v; set_iter f r
;;

let rec set_fold f s accu =
  match s with
      Empty_set -> accu
    | Node_set(l, v, r, _) -> set_fold f l (f v (set_fold f r accu))
;;

let rec set_for_all p = function
    Empty_set -> true
  | Node_set(l, v, r, _) -> p v && set_for_all p l && set_for_all p r
;;

let rec set_exists p = function
    Empty_set -> false
  | Node_set(l, v, r, _) -> p v || set_exists p l || set_exists p r
;;

let set_filter p s =
  let rec filt accu = function
    | Empty_set -> accu
    | Node_set(l, v, r, _) ->
        filt (filt (if p v then set_add v accu else accu) l) r in
    filt Empty_set s
;;

let set_partition p s =
  let rec part (t, f as accu) = function
    | Empty_set -> accu
    | Node_set(l, v, r, _) ->
        part (part (if p v then (set_add v t, f) else (t, set_add v f)) l) r in
    part (Empty_set, Empty_set) s
;;

let rec set_cardinal = function
    Empty_set -> 0
  | Node_set(l, v, r, _) -> set_cardinal l + 1 + set_cardinal r
;;

let rec set_elements_aux accu = function
    Empty_set -> accu
  | Node_set(l, v, r, _) -> set_elements_aux (v :: set_elements_aux accu r) l
;;

let set_elements s = set_elements_aux [] s
;;

let rec set_min_elt = function
    Empty_set -> raise Not_found
  | Node_set(Empty_set, v, r, _) -> v
  | Node_set(l, v, r, _) -> set_min_elt l
;;

let rec set_max_elt = function
    Empty_set -> raise Not_found
  | Node_set(l, v, Empty_set, _) -> v
  | Node_set(l, v, r, _) -> set_max_elt r
;;

let set_choose = set_min_elt
;;


(** ======================================================= **)



(** On a ici toute les fonctions du module Set **)

let empty = { cardinal = 0; data = set_empty }
and is_empty ts = (ts.cardinal = 0)
and mem e ts = set_mem e ts.data
and add e ts = if set_mem e ts.data then ts else { cardinal = ts.cardinal + 1; data = set_add e ts.data; }
and singleton e = { cardinal = 1; data = set_singleton e; }
and remove e ts = 
  if set_mem e ts.data then { cardinal = ts.cardinal - 1; data = set_remove e ts.data } else ts
and union ts1 ts2 = let s = set_union ts1.data ts2.data in { cardinal = set_cardinal s; data = s; }
and inter ts1 ts2 = let s = set_inter ts1.data ts2.data in { cardinal = set_cardinal s; data = s; }
and diff ts1 ts2  = let s = set_diff  ts1.data ts2.data in { cardinal = set_cardinal s; data = s; }
and compare ts1 ts2 = 
  if ts1.cardinal = ts2.cardinal 
  then set_compare ts1.data ts2.data
  else Pervasives.compare ts1.cardinal ts2.cardinal
and equal ts1 ts2 = (ts1.cardinal = ts2.cardinal) && (set_equal ts1.data ts2.data)
and subset ts1 ts2 = (ts1.cardinal <= ts2.cardinal) && (set_subset ts1.data ts2.data)
and iter fct ts = set_iter fct ts.data
and fold fct ts a = set_fold fct ts.data a
and for_all pred ts = set_for_all pred ts.data
and exists pred ts = set_exists pred ts.data
and filter pred ts = let d = set_filter pred ts.data in { cardinal = set_cardinal d; data = d }
and partition pred ts = 
  let (t1, t2) = set_partition pred ts.data 
  in  ({cardinal = set_cardinal t1; data = t1;}, {cardinal = set_cardinal t2; data = t2;})
and cardinal ts = (assert(ts.cardinal = set_cardinal ts.data); ts.cardinal)
and elements ts = set_elements ts.data
and min_elt ts = set_min_elt ts.data
and max_elt ts = set_max_elt ts.data
and choose ts = set_choose ts.data
;;

let _compareMgsVal_set_set ((s1, t1), (s2, t2)) = 
  let tdiff = compare_type t1 t2 in if tdiff <> 0 then VAL_int(tdiff) else VAL_int(compare s1 s2)
;;



