(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: input_output.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-09-20 13:56:04 $
 * Version: $Revision: 1.9 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

open Global;;
open Auxfct;;
open Unix;;
open Valeur;;
open Printer;;

__register_version "$RCSfile: input_output.ml,v $" "$Revision: 1.9 $" "$Date: 2007-09-20 13:56:04 $";;

type out_file    = Systeme | File | Process
and  in_out_kind = In | Out 
;;

let string_of_in_out = function
    In -> "input"
  | Out -> "output"
;;


(* Pour faciliter l'ecriture du code (??) out_files_tbl et in_files_tbl comportent
   tous les deux des out_channel *ET* des in_channels. Mais c'est le type
   du 1er argument du triplet qui donne la pertinence : si c'est In, alors
   c'est le in_channel qui est correct, sinon c'est le out_channel qui est
   correct.
*)

let files_tbl   = (Hash.create 7 : (string, (out_file * in_out_kind * Pervasives.out_channel * Pervasives.in_channel)) Hash.t)
;;



let rec close_one k = 
  (try
     (let (typ, ou_tbl, v_out, v_in) = Hash.find files_tbl k
      in
	Hash.remove files_tbl k;
	match typ
	with
	    Systeme -> ()
	  | File    -> (Printf.printf "Closing file descriptor %s opened in %s mode.\n" k (string_of_in_out ou_tbl); flush Pervasives.stdout;
			if (ou_tbl = Out)
			then Pervasives.close_out v_out
			else Pervasives.close_in  v_in)
	  | Process -> (let r = (match ((if (ou_tbl = Out) 
					 then Unix.close_process_out v_out
					 else Unix.close_process_in v_in))
				 with 
				     Unix.WEXITED _   -> "...ok exited"
				   | Unix.WSIGNALED i -> ("...hum, killed by signal "^(string_of_int i))
				   | Unix.WSTOPPED i  -> ("...hum, stopped by signal "^(string_of_int i)))
			in
			  Printf.printf "Closing process descriptor %s with return %s\n" k r))
   with Not_found -> raise (Erreur ("while closing, descriptor for "^k^" not found.")))
  
  
and close_all () = Hash.iter (fun k _ -> close_one k) files_tbl

(* Verifie q'un descripteur de fichier existe pour un fichier en sortie ; si oui, le retourne, si non, l'ouvre et
   le retourne.

   find_or_open_and_add 
     force : booleen. Si vrai, force l'ouverture du fichier
     inout : en entree ou en sortie
     fix   : le nom du fichier.
*)

and find_or_open_and_add = 
  let count = ref 0
  in
    (fun force inout fic -> (if force
			     then (let kind, (fd_out, fd_in) = open_or_pipe fic inout in
				   let r                     = kind, inout, fd_out, fd_in
				   and new_name              = (incr count; Printf.sprintf "%s`%d" fic !count)
				   in
				     Hash.add files_tbl new_name r;
				     new_name, r)
			       else fic, (try
					    let ((_, typ, _, _) as r) = Hash.find files_tbl fic
					    in
					      if (typ <> inout)
					      then raise (Erreur (Printf.sprintf "trying to use %s that was opened in %s mode for a %s action " 
								    fic 
								    (string_of_in_out typ)
								    (string_of_in_out inout)))
					      else r
					  with Not_found -> (let kind, (fd_out, fd_in) = open_or_pipe fic inout in
							     let r                     = kind, inout, fd_out, fd_in
							     in
							       Hash.add files_tbl fic r;
							       r))))


(* Vrai si le bieme bit de v est mis *)
and bitset v b = 
  let k = 1 lsl b 
  in
    (v land k) = k


and user_rx  perm = (bitset perm 8) && (bitset perm 6)
and group_rx perm = (bitset perm 5) && (bitset perm 3)
and other_rx perm = (bitset perm 2) && (bitset perm 0)

(* Fonction retournant un (type_du_fd_en_sortie, fd en sortie) vers : 
   _ un processus si l'argument correponds a un fichier executable,  
   _ un fichier si l'argument ne correspond a rien.  
*)

and open_or_pipe fic inout = 
  let me_gid = Unix.getgid()
  and me_uid = Unix.getuid()
  in
    (try
       let s = Unix.stat fic
       in                                           (* Si on arrive ici c'est que le fichier existe et que l'on a pu faire un stat dessus *)
	 if ((s.st_kind <> S_REG) &&                (* Si le fichier n'est pas un fichier regulier *)
             (s.st_kind <> S_LNK))                  (* ou un lien symbolique *)
	 then raise Not_found
	 else
	   if (((s.st_uid = me_uid) && (user_rx s.st_perm))  or  (* possesseur et rx en position U ou G ou O *)
	       ((s.st_gid = me_gid) && (group_rx s.st_perm)) or  (* meme groupe, et rx en position G ou O    *)
	       (other_rx s.st_perm))                             (* aucune relation et rx en position O      *)
	   then (Process, ((if (inout = Out) 
			    then (Unix.open_process_out fic), Pervasives.stdin
			    else Pervasives.stdout, (Unix.open_process_in  fic))))
	   else raise Not_found

     with 
	 Unix.Unix_error (Unix.ENOENT, _, _) 
       | Not_found                           -> (try
						   File, (if (inout = Out)
							  then (Pervasives.open_out fic), Pervasives.stdin
							  else Pervasives.stdout, (Pervasives.open_in fic))
						 with 
						     Sys_error x -> raise (Erreur (Printf.sprintf "open_or_pipe : %s" x))))


and _out_string_any (s, y) = (let fic, (_, _, fd, _) = find_or_open_and_add false Out s
			      in
				Pervasives.output_string fd (string_of_STRING (to_string(Printer.prsvalue y)));
				Pervasives.flush fd;
				VAL_string fic)

and _openio_string_int (filename, io) = (let fic, _ = find_or_open_and_add true (if (io = 0) then In else Out) filename
					 in
					   VAL_string fic)

and _isopened_string (filename) = (VAL_bool (Hash.mem files_tbl (Printf.sprintf "%s" filename) || Hash.mem files_tbl (Printf.sprintf "%s" filename)))

and _out_any_any (x, y) = 
  let err = "Error: << takes a string at left-hand side"
  in prerr_endline err;
     Printf.fprintf Pervasives.stderr "got:\t%a\n" prfvalue2 x;
     raise (Erreur err)


and _forceout_string_any (s, y) = (let fic, (_, _, fd, _) = find_or_open_and_add true Out s
				   in
				     Pervasives.output_string fd (string_of_STRING (to_string(Printer.prsvalue y)));
				     Pervasives.flush fd;
				     VAL_string fic)

and _forceout_any_any    _      = raise (Erreur "<<< : only possible with a string as left-hand side")


and _readline_string s  = (let fic, (_, _, _, fd) = find_or_open_and_add false In s
			   in
			     try
			       VAL_string (Pervasives.input_line fd)
			     with End_of_file -> raise (Erreur (Printf.sprintf "EOF encountered while reading on %s" fic)))



and _foldio_funct_any_string (fonc, zero, s) = (let fic, (_, _, _, fd) = find_or_open_and_add false In s
						in
						  fold_until_exn End_of_file (app_2 fonc) zero (function () -> VAL_string (Pervasives.input_line fd)))

and _close_string s = (close_one s; 
		       VAL_undef "close")
;;


let _system_string s = (let ss = Global.string_of_STRING s
		      in
			match (Unix.system ss)
			with 
			    Unix.WEXITED _   -> VAL_undef "...ok exited"
			  | Unix.WSIGNALED i -> VAL_undef ("...hum, killed by signal "^(string_of_int i))
			  | Unix.WSTOPPED i  -> VAL_undef ("...hum, stopped by signal "^(string_of_int i)))
						

and _time_any _ = VAL_float (Unix.time()) (* ((Unix.times()).tms_utime) *)
;;

(** _fprintcoll_string_any_funct_any_any_any 
 **        (out, coll,(fonc:Valeur.sommet_decoration ->Valeur.mgs_value), pre, inter, post)
 **
 ** Imprime de faon format [coll] dans [out] avec un prfixe [pre], un suffixe [post], 
 ** un infixe [inter], et en appliquant  chaque lment de la collection la fonction [fonc]
 **)
let rec _printcoll_string_any_funct_any_any_any (out, coll,(fonc:Valeur.mgs_value -> Valeur.mgs_value), pre, inter, post) = 
  if (not (is_collection coll))
  then raise (Erreur ("printcoll : defined on collections only"))
  else
    begin
      let fic, (_, _, fd, _)  = find_or_open_and_add false Out out in
      let bufferf = Format.formatter_of_out_channel fd  
      in
        Printer.pobval2 bufferf pre inter post (VAL_funct fonc) coll;
        Format.pp_print_newline bufferf ();
	VAL_string fic
    end
;;


let rec _uprintcoll_string_any_funct_any_any_any (out, coll,(fonc:Valeur.mgs_value ->Valeur.mgs_value), pre, inter, post) = 
  if (not (is_collection coll))
  then raise (Erreur ("uprintcoll : defined on collections only"))
  else
    begin
      let fic, (_, _, fd, _)  = find_or_open_and_add false Out out in
        Printer.upofvalue2 fd pre inter post (VAL_funct fonc) coll;
        Printf.fprintf fd "\n" ;
	VAL_string fic
    end
;;



(** _fprintcollrecpar_'outBufferF_any_seq ( b , coll , l ) : 
 ** Imprime de faon format [coll] dans le buffer format [b] en utlisant 
 ** un prfixe, un suffixe, un infixe parmi la squence l.
 ** La squence l est une squence de type : 
 **        premier lment   : prdicat
 **        deuxieme lment  : fonction  appliquer  chaque lment
 **        troisieme lment : prfixe
 **        quatrieme lment : infixe
 **        cinquime lment : suffixe
 **
 ** Pour cela, on parcourt la sequence l pour trouver le prfixe, le suffixe, 
 ** l'infixe et la fonction pour la collection b en appliquant  chaque le prdicat 
 ** sur la collection.
 ** Puis, on applique la fonction pzbval2
 **)
let rec _fprintcollrecpar_'outBufferF_any_seq (b , coll , l ) =
  let tab = fst ( l ) in
  let rec parcourt t i m =
    if(i<m) then 
      begin
        match tab.(i) 
	with
          | VAL_seq(toto, _) when bool_from_val ((to_funct toto.(0)) coll) ->
              (* fonc toto.(2)
                 pre toto.(3)
                 inter toto.(4)
                 post toto.(5)
              *)
              pobval2 b (toto.(2)) (toto.(3)) (toto.(4)) toto.(1) (* (VAL_seq l) *) coll;
          | _ -> parcourt t (i+1) m
      end
    else
      (prbvalue2 b coll)
  in
    parcourt tab 0 (Array.length tab);
    coll;;


(** _fprintcollrec_string_any_seq ((out:string), coll,l) : 
 ** mme fonction que _fprintcollrecpar_'outBufferF_any_seq sauf que 
 ** l'on renvoie le rsultat dans out 
 **)

let rec _fprintcollrec_string_any_seq ((out:string), coll,l) = 
  if (not (is_collection coll))
  then raise (Erreur ("printcoll : defined on collections only"))
  else
    begin
      let fic, (_, _, fd, _) = find_or_open_and_add false Out out in
      let bufferf            = Format.formatter_of_out_channel fd  
      in
        _fprintcollrecpar_'outBufferF_any_seq ( bufferf , coll , l );
        Format.pp_print_newline bufferf ();
        VAL_string fic
    end
;;


(*** On initialise les variables representant les i/o standards ***)
begin
  print_string "(* registering file descriptors...";
  flush Pervasives.stdout;
  Environnement.addenv_pure   "'stdin"  (VAL_string "stdin");
  Environnement.addenv_pure   "'stdout" (VAL_string "stdout");
  Environnement.addenv_pure   "'stderr" (VAL_string "stderr");
  List.iter   
    (fun (nom, v) -> Hash.add files_tbl nom v)
    [("stdout", (Systeme, Out, Pervasives.stdout, Pervasives.stdin));
     ("stderr", (Systeme, Out, Pervasives.stderr, Pervasives.stdin));
     ("stdin",  (Systeme, In,  Pervasives.stdout,  Pervasives.stdin)); ];
  print_string " done *)\n";
  flush Pervasives.stdout;
end
;;
