(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: combinatorise.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-09-23 21:24:23 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)


open Global;;
open Valeur;;
open Terme;;
open Printer_code;;
open Byte_code;;
open Environnement;;
open Optimise_combinateur;;

__register_version "$RCSfile: combinatorise.ml,v $" "$Revision: 1.1 $" "$Date: 2004-09-23 21:24:23 $";;




(* ----- Manipulation d'ensembles ------------------------------------ *)

    (** simple = pas de variable globale ni de if **)
let rec simple = function
  | Cupl _ -> false
  | Capp(p, q) -> (simple p) & (simple q)
  | Cabs(s, p) -> simple p
  | _ -> true
;;



let rec foldcode ((cv,cd,cp,cabs,capp,cupl,ci,ck,cb,cc,cs,cs4) as fargs) e = 
  let ff = foldcode fargs in
  match e with
    | Cv v        -> cv v
    | Cd s        -> cd s
    | Cp s        -> cp s
    | Cabs(s,p)   -> cabs s (ff p)
    | Capp(p,q)   -> capp (ff p) (ff q)
    | Cupl(o,n,l) -> cupl o n (List.map ff l)
    | C_I         -> ci()
    | C_K         -> ck()
    | C_B         -> cb()   
    | C_C         -> cc()       
    | C_S         -> cs()   
    | C_S4        -> cs4()  
;;


let free_var_code =
  let cv _       = empty_string_set
  and cd _       = empty_string_set
  and cp s       = string_singleton s
  and cabs s u   = Ss.diff  u (string_singleton s)
  and capp u v   = (Ss.union u v)
  and cupl o n l = List.fold_left Ss.union empty_string_set l
  and ci()       = empty_string_set
  and ck()       = empty_string_set
  and cb()       = empty_string_set
  and cc()       = empty_string_set
  and cs()       = empty_string_set
  and cs4()      = empty_string_set
  in
    foldcode (cv,cd,cp,cabs,capp,cupl,ci,ck,cb,cc,cs,cs4)
;;






(** -------------- Combinatorisation --------------------------- **)

  let rec make_app fct = function
    | [] -> raise (Erreur "internal error: make_abs: 0 args")
    | [a] -> Capp(fct, a)
    | a::l -> make_app (Capp(fct, a)) l
  ;;


(** translate : term -> byte_code
 ** traduit un terme en un byte-code qui contient encore des
 ** abstraction. les fonctions 'propage' et 'fonctorise' ci-dessous
 ** transforment ce terme pour liminer les abstractions au profit de
 ** combinateurs. 
 ** Le vrai point d'entree de la traduction en combinateur est la
 ** fonction 'term2bytecode' ci-dessous.  
 ** 
 ** La traduction es assez classique (Peyton-Johnes) avec neanmoins les points suivants :
 **
 ** - il y a des variables imperatives, et on veut gerer le
 **  sequencement finement. Cela amene a introduire 2 types de constantes
 **  fonctionelles: les "pures" et les "impures". Une fonction "pure" ne
 **  fait pas d'effet de bord. Pour des raisons de simplicites, ne sont
 **  pures que certaines fonctions de bases qui ont ete declarees telles
 **  (par exemple l'addition, mais pas le print). Toutes les autres
 **  fonctions (par exemple celle defini par l'utilisateur) sont
 **  impures. La notion de purete et d'impurete est gere par un attribut
 **  stocke dans l'environnement. 
 **
 ** - il y a des applications n-aires. Une application de fonctions de
 **  la forme e(a1, ..., an) est transforme en un byte_code
 **  Cupl(id, ..., [a1, ..., an]) si e se reduit un nom de fonctions
 **  impures, et sinon l'application se traduit en une serie
 **  d'application unaire imbrique Capp(fct..., arg...).
 **
 **  On garde explicitement a la fois Capp et Cupl car les
 **  imbrications de Capp peuvent etre optimisees et l'evaluation ne
 **  respecte pas toujours le sequencement (par exemple \x.2+3 va se
 **  traduire en \x.5 car + est pure). Ce n'est pas voulu, simplement
 **  on fait pas vraiment attention dans les optimisations.
 **)



let rec translate level params = function
  | Cte v -> Cv v
  | Id s  when List.mem s params -> Cp s
  | Id s  when is_pur(s) -> Cd s
  | Id s -> Cupl(Ofct(index_var s), level, [])
  | App(Id s, l) when not(is_pur(s)) & not(List.mem s params) -> 
      let idvar = find_or_create_var s (* on cree la variable avant la traduction du body *)
      in Cupl(Ofct(idvar), level, List.map (translate level params) l)
  | App(p, l) -> make_app (translate level params p) (List.map (translate level params) l)
  | Abs(s, p) -> Cabs(s, translate (level+1) (s::params) p)
  | Set(s, p) -> 
      (if  List.mem s params 
       then print_endline ("Warning: local variable "^s
			   ^" cannot be set. This affectation refers to the global variable."));
      let idvar = (find_or_create_user_var s) (* on cree la variable avant la traduction du body *)
      in Cupl(Oset(idvar), level,[translate level params p])
  | If(p, q, r) -> Cupl(Oif, level, [translate level params p; 
				     translate level params q;
				     translate level params r])
  | Fby(p, q) ->
      let p1 = translate level params p in
      let q1 =  translate level params q
      in Cupl(Ofby, level, [p1; q1])
  | Cnil -> let msg = "Sorry: translate: Cnil term encountered"
    in print_endline msg; raise (Erreur msg)
      
  | Nyi -> let msg = "Sorry: translate: term not yet implemented"
    in print_endline msg; raise (Erreur msg)
      
;;


(** fonctorise : byte_code -> byte_code sans lambda (sans les Abs)
 ** propage : byte_code -> byte_code sans lambda : lance fonctorise 
 **)

let rec propage = function
  | Cabs(s, p) -> fonctorise s p
  | Capp(p, q) -> Capp(propage p, propage q)
  | Cupl(o, n, l) -> Cupl(o, n, List.map propage l)
  | x -> x


and ttrace t s = 
  if !trace_opt then (print_string(s^": "); prcode t; print_newline())

and fonctorise a = function

    (** parametre de la lambda courante **)
  | Cp s as t when s = a ->   
      ttrace t ("-- rule arg "^s);
      C_I

  (** \x.t = K t si x notin FreeVar(t) **)
  | t when simple(t) & not(Ss.mem a (free_var_code t)) ->
      ttrace t "-- rule \\x.t = K t si x pas libre dans t";
      Capp(C_K, propage t)

  | Cupl(op, level, l) as t ->
      ttrace t "-- rule Cupl generique";
      Cupl(op, level, List.map (fonctorise a) l)

	(** eta-reduction **)
  | Cabs(s1, Capp(p, Cp s2))  as t
    when (s1 = s2) & (simple p) & not(Ss.mem s1 (free_var_code p)) ->
      ttrace t "-- rule eta";
      fonctorise a p

  | Capp(f, Cp s) as t
    when (a = s) & (simple f) & not(Ss.mem a (free_var_code f)) ->
      ttrace t "-- rule eta2";
      propage f

  | Cabs(b, p) as t -> 
      ttrace t "-- rule Abs";
      let tt = (fonctorise b p)
      in ttrace tt "-- rule Abs: retour du body";
	let ttt = fonctorise a tt
	in ttrace ttt ("-- rule Abs: retour du ("^a^")[body]");
	  ttt

  | Capp(f, x) as t -> 
      ttrace t "-- rule apply general";
      Capp(Capp(C_S,  (fonctorise a f)), fonctorise a x) 

  (*** ATTENTION: la regle d'extension  
   ***      Cup(f, n, l)@a =  Cup(f, n, l@[a]) 
   *** est fausse, considerer par exemple: fun fff(x) = 3;; (\x.((\y.fff(0))(2)))(1);;
   ***)

  | t -> 
      ttrace t "-- rule default";
      Capp(C_K, t)
;;


let term2bytecode term = propage (translate 0 [] term);;
