(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: environnement.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-06-07 14:43:12 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * L'environnement = gestion des variables.
 *
 * les variables se divisent en:
 *
 *    - variables systemes, predefinies quand l'utilisateur entre dans
 *      mgs et dont le nom commence par un quote. Exemple: 'beep ou '+
 *
 *      On peut acceder a ces variables par le nom sans quote, tant que
 *      le nom sans quote n'a pas ete affecte comme variable
 *      utilisateur. Exemple:  beep(3) est la meme expression que 'beep(3)
 *      tant qu'on a pas fait un fun beep(x) = 5
 *
 *      La valeur des variables systemes ne peut pas etre redefinie:
 *      i.e., on peut faire fun beep(x) = 3 mais on ne peut pas faire 
 *      fun 'beep(3) = 3; Cette contrainte est syntaxique (erreur au parsing).
 *
 *    - les variables definies par l'utilisateur grace a 
 *       . une affectation, exemple: a := 3
 *       . une definition de fonction, exemple:  fun f(x) = 3
 *      L'idee au depart etait de faire une difference entre les deux formes
 *      de definitions: dans le 1er cas, on a une variable imperative 
 *      qu'on peut reaffecter ; dans le deuxieme cas on a la definition 
 *      d'une constante, i.e. on ne peut pas redefinir la valeur du nom f.
 *      Mais FINALEMENT, toutes les variables utilisateurs sont imperatives.
 *      
 * Les variables sont gerees de la maniere suivante:
 *
 * + une table de hachage (var_tbl) associe un index et un mode a un
 *   identificateur (utilisateur ou systeme).
 *
 * + L'index sert d'entre dans un vecteur (var_env) qui contient la
 *   valeur associee au nom. Ce vecteur grossit au fur et a mesure des
 *   besoins. Ce vecteur est appele aussi 'memoire' et un index de ce
 *   vecteur, une 'adresse memoire'.
 *
 * + Une table de hachage (varname_tbl) qui associe un nom  un index 
 *   (table inverse de var_tbl), utiliser pour retrouver le nom qui
 *   reference une adresse memoire donne.
 *
 * Le mode stocke avec le nom dans la variable dans var_tbl correspond
 * a un booleen (Pur ou Impur) appele le 'mode de la variable'.
 *
 *     pur = pas d'effet de bord. Pertinent si c'est une fonction.
 *          Toutes les fonctions utilisateurs sont
 *          considerees comme impure. Seule certaines
 *          fonctions systemes sont pures: l'arithmetique par
 *          exemple. Ou plus generalement, toutes fonctions systemes
 *          sans effet de bord et dont le temps d'execution
 *          est independant des arguments.
 *
 *         La consequence d'etre pur pour une fonction, c'est qu'on
 *         peut l'appliquer plus tot dans le processus de
 *         traduction en byte-code. Par exemple \x.1+2 sera compile
 *         reellement comme (K 3) parce que + est pur. 
 *
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;



Global.__register_version "$RCSfile: environnement.ml,v $" "$Revision: 1.2 $" "$Date: 2006-06-07 14:43:12 $";;


(** mode de l'affectation **)
type variable_mode = Pur | Impur

(** Description de la variable dans le dico var_tbl:
 ** - un index qui correspond  une entre dans la "mmoire" var_env
 ** - un mode de gestion
 **)
and variable_attributes = int * variable_mode
;;


let mode_is_pur = function
  | Pur -> true
  | _ -> false

and mode_is_impur = function
  | Impur -> true
  | _ -> false
;;

(** Je subodore que c'est inutilis.
 ***
 *** let default_function = 
 ***  let msg = "OUPS: uninitialized default function called (internal error?)"
 ***  in VAL_funct (function x -> prerr_endline msg; raise (Erreur msg))
 *** ;;
 ***)


(** ----------------- Cration de l'environnement -------------------------- **)


(** Creation des dico et dico inverse **)
let var_tbl = ((Hash.create 101) : (string, variable_attributes) Hash.t)
and varname_tbl = ((Hash.create 101) : (int, string) Hash.t)
;;


(** Creation de la mmoire var_env. 
 ** C'est une ref vers un vecteur qui grandit si ncessaire 
 **)


(** index de l'element courant libre **)
let var_env_ptr = ref 0

(** index max dans le tableau **)
and var_env_max = ref 256

(** increment quand le tableau grossi **)
and var_env_increase = 512

(** create ab-nihilo d'un vecteur memoire **)
and create_var_env n = Array.make n (VAL_undef("uninitialized var"))
;;

(** la memoire = le vecteur qui contient les valeurs **)
let var_env = ref (create_var_env !var_env_max)
;;


(** croissance de la memoire **)
let grow_var_env () = 
  var_env_max := !var_env_max + var_env_increase;
  var_env := Array.append !var_env (create_var_env var_env_increase)
;;





(** ----------------- Acces a l'environnement -------------------------- **)

let index_var name = 
  try
    (let (idx, _) = Hash.find var_tbl name in idx)
  with Not_found -> 
    try
      (let (idx, _) = Hash.find var_tbl (String.concat "" ["'"; name]) in idx)
    with Not_found -> 
      begin
	if false == !Global.silent_error then 
	  begin
	    Pervasives.flush_all();
	    Printf.fprintf stderr "Error: unknown global variable: %s\n" name;
	  end
      end;
      let msg = "Unknown global variable: "^name in
	if !Global.abort_on_error then raise (Erreur msg)
 	else begin
	  (* on donne la main au handler, pour lui donner une chance de signaler l'erreur,
	   * mais s'il revient, de toute facon on ne sait pas quoi faire puisqu'il faut renvoyer 
	   * l'index d'une variable non existante.
	   *)
	  ignore((pf1 !current_error_handler) (VAL_string msg));
	  raise (Erreur (msg^" (irrecoverable error)"))
	end


and index_cte name = 
  try
    let (idx, m) = Hash.find var_tbl name 
    in if mode_is_pur(m) then idx
      else raise (Erreur ("Error: identifier "^name^" is a impure, not a pure constant"))
  with Not_found -> 
    try
      let (idx, m) = Hash.find var_tbl (String.concat "" ["'"; name]) 
      in if mode_is_pur(m) then idx
	else raise (Erreur ("Error: identifier "^name^" is an impure variable, not a pure constant"))
    with Not_found -> 
      prerr_string("\nError: unknown constante: "^name);
      prerr_newline();
      raise (Erreur ("Unknown cte: "^name))

and is_defined_var name = 
  (Hash.mem var_tbl name) or 
  (Hash.mem var_tbl (String.concat "" ["'"; name]))

and is_exactly_defined_var name = Hash.mem var_tbl name

and is_pur name = 
  try
    let (_, m) = Hash.find var_tbl name 
    in mode_is_pur(m)
  with Not_found -> 
    (try
       let (_, m) = Hash.find var_tbl (String.concat "" ["'"; name])
       in mode_is_pur(m)
     with not_found -> false)


(** par defaut on cree une variable imperative qui peut contenir 
 ** une fct produisant des effets de bord 
 **)
and find_or_create_var name =
  try  
    let (idx, _) = Hash.find var_tbl name in idx
  with Not_found ->
    try 
      let (idx, _) = Hash.find var_tbl (String.concat "" ["'"; name])
      in idx
    with Not_found ->
      (if is_sys_var name then raise (Erreur ("Error: cannot create system var: "^name)));
      let cur = !var_env_ptr
      in (Hash.add var_tbl name (cur, Impur);
   	  Hash.add varname_tbl cur name;
	  (!var_env).(cur) <- VAL_undef("uninitialized var: "^name^" ");
	  incr var_env_ptr;
	  (if !var_env_ptr >= !var_env_max then grow_var_env());
	  cur)

and find_or_create_user_var name =
  try  
    let (idx, _) = Hash.find var_tbl name
    in idx
  with Not_found ->
    (if is_sys_var name then raise (Erreur ("Error: cannot create system variable: "^name)));
    let cur = !var_env_ptr
    in (Hash.add var_tbl name (cur, Impur);
   	Hash.add varname_tbl cur name;
	(!var_env).(cur) <- VAL_undef("uninitialized var: "^name^" ");
	incr var_env_ptr;
	(if !var_env_ptr >= !var_env_max then grow_var_env());
	cur)

and addenv name m value =
  try  
    let (idx, m') = Hash.find var_tbl name
    in if mode_is_impur(m')
      then ((!var_env).(idx) <- value; idx)
      else raise (Erreur ("Error: cannot add to env existing variable: "^name
			  ^" with pure mode"))
  with Not_found ->
    let cur = !var_env_ptr
    in (Hash.add var_tbl name (cur, m);
	Hash.add varname_tbl cur name;
	(!var_env).(cur) <- value;
      Hash.add varname_tbl cur name;
      incr var_env_ptr;
      (if !var_env_ptr >= !var_env_max then grow_var_env());
      cur)
;;

let addfct name f = addenv name Impur f
and index_pur = index_cte
;; 

(** rajouter une liaison en verifiant qu'elle n'existait pas deja.
 **)
let addenv_unique name m value =
  if Hash.mem var_tbl name then
    raise (Erreur ("Internal error: addenv_unique: "^name^" already exists"))
  else (	
    (* rajout dans la table *)
    Hash.add var_tbl name ((!var_env_ptr), m);	
    Hash.add varname_tbl (!var_env_ptr) name;
    (* update des pointeurs *)
    (!var_env).(!var_env_ptr) <- value;
    incr var_env_ptr;
    if !var_env_ptr >= !var_env_max then grow_var_env())
;;

let addenv_pure   name value = addenv_unique name Pur value
and addenv_impure name value = addenv_unique name Impur value
;;



let name_of_funct f = "[funct]"

and name_of_var idx =
  try
    Hash.find varname_tbl idx
  with Not_found -> "?var?"
;;

let  fct1 name index = 
  function x -> match (!var_env).(index) with
  | VAL_funct f -> f(x)
  | _ -> raise (Erreur ("id "^name^" does not refer a fct1"))

and  acces_var name =
  let index = index_var name in VAL_funct(function _ -> (!var_env).(index))

and get_var name   = (!var_env).(index_var name)
and set_var name v = (!var_env).(index_var name) <- v

and get_varidx index   = (!var_env).(index)
and set_varidx index v = (!var_env).(index) <- v
;;




(** ---------- toujours dans l'environnement -------------------- **)

addenv_pure  "'copyright" (VAL_string
"
MGS   [encore un Modele General de Simulation/Systeme]
and related softwares and libraries.

Copyright (c) 2000--2004, Jean-Louis Giavitto and Olivier Michel.

  LaMI, umr 8042 CNRS - Universite d'Evry Val d'Essone, Genopole-Evry
  Tour Evry 2, 
  523 Place des terrasses de l'Agora, 
  91000 Evry, France. 

  email: giavitto@lami.uni-evry.fr, 
  email: michel@lami.uni-evry.fr, 
  url  : http://mgs.lami.univ-evry.fr 
 

Legal Notice:

No representations are made about the suitability of this 
software for any purpose. It is provided ``as is'' without 
express or implied warranty. 

The authors of this software, the CNRS, the University of Evry,  
Genopole and the `Laboratoire des Methodes Informatiques' (LaMI)  
disclaim all warranties with regard to this software, either  
expressed or implied, including, but not limited to,  
the implied warranties of merchantability and fitness.  
The entire risk as to the quality and performance of the program  
is with you. Should the program prove defective, you assume the  
cost of all necessary servicing, repair or correction. In no  
event shall the authors be liable for any special, indirect or  
consequential damages or any damages whatsoever  
resulting from loss of use, data or profits, whether in an  
action of contract, negligence or other tortious action,  
arising out of or in connection with the use or inability to use  
or performance of this software. 

This software is distributed under a license that conforms to the  
GNU Public License version 2. General permission to copy, distribute,
and modify the software will apply. 

But in addition, if you modify the sources, or use any part of them,
you have to explicitly mention the part of this notice appearing
before the 'Legal Notice' sentence, to the final user of your program
at the start-up time of your program.

No commercial use of elements of this software is allowed.
This software includes various libraries with their own copyrights
and the user of this software must complies with their requirements.
The libraries used include:
- qhull: used for the implementation of Delaunay collections
  Cf. http://www.geom.umn.edu/locate/qhull
      http://www.qhull.org/
- gsl: used for various scientific computations
  Cf. http://www.gnu.org/software/gsl/
- nauty: used for the implementation of the graph isomorphism predicate
  Cf. http://cs.anu.edu.au/people/bdm/
- gmap: used for the implementation of the qmf collections
  Cf. http://www.sic.sp2mi.univ-poitiers.fr/moka

 
For more informations, consult the MGS home page:

             http://mgs.lami.univ-evry.fr
\n")
;;



(** Des fonctions impures: BEEP
 **
 **     beep fait un effet de bord. On peut compter le nombre 
 **     de beep et remettre ce compte  0 grce aux fonctions
 **     'beep_cpt et 'beep_cpt_reset.
 **     Cette fonction est utilise dans les tests, pour vrifiez que 
 **     l'on gre correctement l'ordre des excutions.
 **)
(*!*)(**
       @module system
       @name 'beep
       @type any_any
       Make a sound and print a beep on standard output. 
       The returned value is the argument. The number of calls to this
       function can be monitored. This function is used for debugging
       purposes. 
       @voir system.'beep_cpt       beep_cpt
       @voir system.'beep_cpt_reset beep_cpt_reset
       @voir system.'!!             !!
*)(*!*)
let beep_cpt = ref 0
;;
addenv_impure "'beep" 
  (VAL_funct (fun x ->
		print_string "\n########### BEEP #########\n";
		incr beep_cpt;
                print_newline(); x))
;;
(*!*)(**
       @module system
       @name 'beep_cpt
       @type unit_int
       Beep count.
       Returns the number of calls to the function [beep] since the last
       call to [beep_cpt_reset] (or since the begining of the program
       if no call to [beep_cpt_reset] has occured).
       @voir system.'beep           beep 
       @voir system.'beep_cpt_reset beep_cpt_reset
*)(*!*)
addenv_impure "'beep_cpt" (VAL_funct (fun x -> VAL_int(!beep_cpt)))
;;
(*!*)(**
       @module system
       @name 'beep_cpt_reset
       @type any_int
       Reset the counter of calls to the function [beep] to zero. The
       value [0] is returned.
       @voir system.'beep     beep
       @voir system.'beep_cpt beep_cpt 
*)(*!*)
addenv_impure "'beep_cpt_reset" (VAL_funct (fun x -> beep_cpt := 0; VAL_int(!beep_cpt)))
;;





(** --- Gestion des traces ------------------------------------------------------- **)

(** Pour tracer une fonction, on wrappe la fonction par un stub qui affiche chaque appel.
 ** Pour pouvoir annuler le trac, on stocke la fct originale dans un tableau (on ne peut
 ** pas comparer des fonctions pour les mettre directement dans la table de hachage).
 **)


let tracedfct_tbl = ((Hash.create 101) : (string, int) Hash.t)
and save_tracedfct = ref(Array.make 101 None)
;;




let tabulation = ref 0
and printf = Printf.printf
and sprintf = Printf.sprintf
;;

(* Si une erreur se produit pendant la trace, la tabulation reste decale.
 * Par precaution, on remet la tabulation  0 a chaque tour de top_level.
 *)
let raz_tabulation() = tabulation := 0
;;

let rec print_trace_in id v = 
  let curry = sprintf "%s(%s)" id (!forward_prsvalue v)
  in printf "%s%s ->\n" (stab()) curry; incr tabulation; curry

and  print_trace_out curry v = 
  decr tabulation; 
  printf "%s%s <- %s\n" (stab()) curry (!forward_prsvalue v);
  match v with 
      (** fonction currifye: on trace l'application partielle 
       ** avec comme nom la chaine curry qui est l'application partielle **)
    | VAL_funct fv -> wrap_trace curry fv
    | _ -> v


and stab () = stab2 !tabulation
and stab2 = function 
  | 0 -> "" 
  | 1 -> "   " 
  | n when n mod 2 = 0 -> let s = stab2 (n/2) in s^s
  | n -> (stab2 1)^(stab2 (n-1))

and wrap_trace id f =
  VAL_funct(function x -> 
	      begin
		let curry = print_trace_in id x; 
		in print_trace_out curry (f x)
	      end)
;;

exception Found of int;;

let get_one_slot () =
  let lg = Array.length !save_tracedfct in
    try
      for i = 0 to lg - 1 do
	if None = !save_tracedfct.(i) then raise (Found i)
      done;
      (* pas de slot de libre, faut augmenter la taille de la table *)
      save_tracedfct :=  Array.append !save_tracedfct (Array.make 101 None);
      lg
    with Found i -> i
;;


let get_access id = 
  try
    Hash.find var_tbl id
  with Not_found -> 
    try
      Hash.find var_tbl (String.concat "" ["'"; id])
    with Not_found -> 
      prerr_endline("\nError: cannot trace unknown global variable: "^id);
      raise (Erreur ("Cannot trace unknown variable: "^id))
;;


let rec trace_fct id =
  if Hash.mem tracedfct_tbl id
  then prerr_endline ("Warning: function "^id^" already traced")
  else
    let (idx, mode) = get_access id in
      begin 
	match !var_env.(idx) with
	  | VAL_funct(f) -> save_and_trace id idx mode f
	  | _ -> 
	      let err = "Error: cannot trace the non-fonctional value "^id
              in prerr_endline err; raise (Erreur err)
      end

and untrace_fct id =
  if not(Hash.mem tracedfct_tbl id)
  then prerr_endline ("Warning: function "^id^" is not traced")
  else
    begin
      let idx = index_var id
      and i = Hash.find  tracedfct_tbl id in
	match !save_tracedfct.(i) with
	  | Some(mode, f) -> restore_tracedfct id i idx mode f
	  | None -> raise (Erreur ("Erreur: internal error: "^id^" is both traced and not saved"))
    end
 
and save_and_trace id idx mode f = 
  let i = get_one_slot () in
    begin
      Hash.add tracedfct_tbl id i;             (* marquage de la trace                        *)
      !save_tracedfct.(i) <- Some(mode, f);            (* sauvegarde de la valeur et du mode initial  *)
      Hash.replace var_tbl id (idx, Impur);    (* declaration du mode impur de la fct trace  *)
      !var_env.(idx) <- wrap_trace id f;               (* remplacement de la valeur par la fct trace *)
      prerr_endline ("Function "^id^" is traced")
    end 

and restore_tracedfct id i idx mode f = 
  Hash.replace var_tbl id (idx, mode); (* restauration du mode             *) 
  !var_env.(idx) <- VAL_funct(f);              (* restauration de la valeur        *)
  Hash.remove tracedfct_tbl id;        (* annulation de la trace           *) 
  !save_tracedfct.(i) <- None;                 (* liberation du slot de sauvegarde *)
  prerr_endline ("Function "^id^" is untraced")
;;


(** le wrapping vers une fonction utilisateur **)

let _traceFctCall_string s = (trace_fct s; VAL_undef "'trace")
and _untraceFctCall_string s = (untrace_fct s; VAL_undef "'untrace")
;;



let _setCurrentErrorHandler_funct f = 
  let ret = !Valeur.current_error_handler in
    begin
      Valeur.current_error_handler := (VAL_funct f);
      ret
    end

and default_error_handler = function
  | VAL_string m ->
      Pervasives.flush_all();
      if !Global.exit_on_error
      then begin
	prerr_endline ("\nErrors are fatal ("^m^").\nAbort."); 
	exit(1)
      end
      else begin
	(if not !Global.silent_error 
	 then  prerr_endline "Evaluation continue with an undef value."); 
	VAL_undef m
      end
  | _ -> 
      prerr_endline "Internal Error: bad argument type for the default error handler";
      exit(1)
;;	

Valeur.current_error_handler := VAL_funct default_error_handler
;;

let resetCurrentErrorHandler _ = 
  let ret = !Valeur.current_error_handler in
    begin
      Valeur.current_error_handler := VAL_funct default_error_handler;
      ret
    end
;;

