(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: macro.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:58 $
 * Version: $Revision: 1.7 $
 * -------------------------------------------------------------------
 *
 * Status: Ce fichier contient les fonctions utilisees dans le parseur pour 
 *         convertir la syntaxe externe en un noyau restreint(exemple, le let
 *         est converti en application d'une fonction anonyme).
 *
 * On macro-expanse:
 *  . le let
 *  . les dfinitions de transformations
 *  . la cration d'un record
 *  . le return
 *  . les patchs
 *  . les switchs
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Terme;;
open Environnement;;
open Printer;;
open Options;;
open Dispatch_sig;;
open Dispatch;;
open Transformation;;

__register_version "$RCSfile: macro.ml,v $" "$Revision: 1.7 $" "$Date: 2005-11-23 11:10:58 $";;



(* make_abs body list-arg
 * creation de la fct currifyee a partir de la liste des parametres 
 *)
let rec make_abs body = function
  | [] -> raise (Erreur "internal error: make_abs: 0 args")
  | [a] -> Abs(a, body)
  | a::l -> make_abs (Abs(a, body)) l
;;


(* Vielle approche : creation d'un let en le traduisant par l'application d'une lambda 
 ***
 *** cette approche est trop couteuse car on imbrique des lambda. On s'est tourn vers 
 *** une mthode qui utilise une seule lambda. Cf + bas
 *** 
 * -------------------------------------------------------
 * le let strict (make_let_strict) correspond au let caml, 
 * le let sequentiel (make_let_star) correspond au let* de lisp ou les variables 
 * englobantes peuvent etre utilisees dans les clauses and, dans l'ordre d'imbrication.
 *
 * La fonction  make_let est l'interface utilise: elle dispatche actuellement sur 
 * le let strict  la CAML (plus efficace en cas de and multiple).
 *
 *
let rec make_let_strict body env = make_let_app (make_let_abs body env) (List.rev env)

and make_let_abs body = function
  | [] -> body
  | (a, _)::l -> make_let_abs (Abs(a, body)) l

and make_let_app body l = App(body, snd (List.split l))

and make_let_star body = function
  | [] -> body
  | (a, arg)::l -> make_let_star (App(Abs(a, body), [arg])) l

and make_let x = make_let_strict x
;;
 * -------------------------------------------------------------------------
 *)


(** Nouvelle approche du let:
 **     let x = 1+2 and y = 3 in x+y
 ** devient:
 **     (\env. (env 0) + (env 1)) (f (1+2) 3)
 **
 ** avec f une fonction que l'on gnre au parsing et qui contient un array, qui 
 ** reprsente l'environnement, qui consome ses arguments pour mettre  jour 
 ** l'environnement et qui quand elle a consomm tout ses arguments, retourne une 
 ** fonction qui quand on lui passe un entier idx:
 **     - renvoie la valeur de l'lment d'indice idx du tableau si idx > 0 (lecture)
 **     - renvoie une fct qui attend une valeur pour mettre a jour l'element -idx du tableau
 **       si idx < 0 (criture, mise  jour imprative)
 ** l'element d'indice 0 du tableau n'est pas utilis.
 ** Mais attention, il faut creer l'environnement a l'execution, pas maintenant !
 ** exemple : fun f(x) = let y = x-1 in if x <= 0 then 1 else 1 + f(y) fi;;
 ** Dans cette exemple, l'environnement du let doit etre cr a chaque appel 
 ** recursif de f.
 **)
let rec make_let_strict body env = 
  let envname = new_name "$letenv" in
  let (lvars, largs) = (List.split env) in
  let varindex = indexation lvars in
  let envlength = 1+(List.length varindex) in
  let envaccess = makeletenv envlength varindex in
  let corp = bind_letvar envname varindex body in
  let t = App(Abs(envname, corp), [App(Cte(envaccess), List.rev largs)])
  in t 

and indexation lvar =  
  let index = ref (0) 
  in List.fold_left (fun varindex var -> incr index; (var, !index)::varindex) [] lvar 

and bind_letvar envname varindex t = 
  let bd = bind_letvar envname varindex
  in match t with 
      | Id s when List.mem_assoc s varindex -> 
	  (* le Environnement.wrap_delay est ncessaire pour chercher la valeur de 
	   * la variable du let au bon momment *)
	  wrap_delay(op_apply1 envname (Cte (VAL_int (List.assoc s varindex))))

      | Set(s, t) when List.mem_assoc s varindex -> 
	  wrap_delay(op_apply2 envname 
		       (Cte (VAL_int (0 - (List.assoc s varindex))))
		       (bd t))

      | Set(s, t)   -> Set(s, bd t) 
      | App(f, a)   -> App(bd f, List.map bd a) 
      | Abs (s, t)  -> Abs(s,  bind_letvar envname (List.remove_assoc s varindex) t)
      | If(a, b, c) -> If(bd a, bd b, bd c) 
      | Fby(a, b)   -> Fby(bd a, bd b) 
      | x -> x 

and makeletenv envlength varindex = 
  let acces = mkenv envlength varindex
  and defaultval = VAL_undef "unitialized let var"
  in VAL_funct(function v -> let env = Array.create envlength defaultval in (pf1 (acces env))(v))

and mkenv sz varindex a = match varindex with
  | [] -> VAL_funct(function 
		      | VAL_int(i) ->  
			  assert (Pervasives.abs(i) < sz);
			  if i > 0 then a.(i)
			  else VAL_funct(fun v -> a.(0-i) <- v; v)
		      | _ -> failwith "internal error: mkenv")
  | (_, k)::l ->  
      assert (Pervasives.abs(k) < sz);
      let body = mkenv sz l a in VAL_funct(fun v -> a.(k) <- v; body)
;;

let make_let body env =  make_let_strict body env
;;


(**
 ** le SWITCH
 **)
let make_switch e body =
  let id_name = new_name "id_switch" in
  let ifImbr  = List.fold_left
		  (fun acc (sc, exp) ->
		     match sc with
		       | None -> exp
		       | Some sc -> If((op_apply2 "'==" sc (Id id_name)),exp,acc)
		  )
		  (Cte(VAL_undef "macro: make_switch: switch default case"))
		  body
  in
    make_let ifImbr [(id_name,e)]
;;

(**
 ** le TRY
 **)
let make_try body rules =
  let cons_name = new_name "try_catched_cons" in
  let args_name = new_name "try_catched_args" in
  let lgth_name = new_name "try_catched_lgth" in

  let zero = op_apply2 "'raise" (Id cons_name) (Id args_name) in
  let compiled_rules =
    List.fold_left (fun acc (symb,id_list,rule_body) ->
		      (If(
			 (If((op_apply2 "'==" (Cte(VAL_symbol symb)) (Id cons_name)),
			     (op_apply2 "'==" (Cte(VAL_int (List.length id_list))) (Id lgth_name)),
			     Cte(VAL_bool false)
			    )),
			 List.fold_left (fun acc id -> Abs(id,acc)) rule_body id_list,
			 acc
		       ))
		   ) zero rules
  in
    op_apply2
      "'try"
      (Abs(new_name "dummy_arg_for_try_lambda1",body))
      (Abs(cons_name,
	   (Abs(lgth_name,
		(Abs(args_name,
		     compiled_rules
		    ))
	       ))
	  ))
;;




let eval_option_list l = List.map (function (id, t) -> (id, Compile.eval t)) l
;;

let make_application = Options.make_application
;;

let make_transformation rules = 
  let transf = VAL_funct(Transformation.transform rules) in Cte(transf)
;;

let make_transformation2 (vdim,ndim) rules = 
  let transf = VAL_funct(Transformation.transform2 vdim ndim rules) in Cte(transf)
;;

let make_patch rules =
  let patch = VAL_funct(Patch.patch rules) in Cte(patch)


(** Construire un record  partir de la liste de ses champs c'est long !
 ** la version ci-dessous est squentielle et elle nuit a l'efficacite de l'anabaena par exemple
 **)
let rec make_record_version_sequentielle l couleur =
  mkrecord_version_sequentielle (Cte(VAL_rec((Core_rec.empty : (mgs_value, mgs_value) Def_rec.trec), couleur))) l

and mkrecord_version_sequentielle r = function
  | [] -> r
  | (id, v)::l -> mkrecord_version_sequentielle (op_apply3 "'set_field" r (Cte(VAL_string id)) v) l
;;

(** meilleurs version: on gagne 10% ! :-(
 ** L'ide est de crer une fct nomme qui permettra l'instantiation group des arguments
 ** et on utilise pas setfield, mais on gnre directement...
 **)
let rec make_record l couleur =
  if l = [] then Cte(VAL_rec((Core_rec.empty : (mgs_value, mgs_value) Def_rec.trec), couleur))
  else begin
    let id = new_name "$mkrecord"
    in addenv_impure id (mkrecord couleur l); App(Id id, snd (List.split l))
  end

and mkrecord couleur l = mkrecord2 couleur (ref Core_rec.empty) l

and mkrecord2 couleur r = function
  | [] -> failwith "internal error in mkrecord"
  | [(k, _)] ->  VAL_funct(fun v -> VAL_rec(Core_rec.add k v !r, couleur))
  | (k, _)::l -> let body = mkrecord2 couleur r l in VAL_funct(fun v -> r := Core_rec.add k v !r; body)
;;



(** La fonction wrap_return prend un terme t et en fait un autre 
 ** terme qui leve l'exception BreakTransformation avec pour valeur,
 ** la valeur de t. C'est utilis pour implmenter le return dans une 
 ** regle.
 **)
let raise_BreakTransformation(resultat) = raise (BreakTransformation resultat)
;;
Environnement.addenv_impure "'raise_BreakTransformation" (VAL_funct raise_BreakTransformation)
;;
let wrap_return exp = Terme.op_apply1 "'raise_BreakTransformation" exp
;;



(** La fonction make_valposlist prend une liste CAML de couples de
 ** termes, et la transforme un le terme qui cree une sequence MGS de
 ** couples (un couple est une sequence de 2 elements)
 **)

let rec make_valposlist = function
  | [] -> Cte(Type_seq.seq_desc.empty)
  | (v, p)::l ->  
      op_apply2 "'::" 
      (op_apply2 "'::" v (op_apply2 "'::" p (Cte Type_seq.seq_desc.empty)))
      (make_valposlist l)
;;
      

(** Wrappage des partie droite de patch pour la syntaxe exp:[...]. On utilise ici
 ** un descripteur un peu particulier: pRhsSpec.
 **)

let make_patchRhsSpec exp1 exp2 =
  let dfcf =
    (op_apply2 "'::"
       (try (List.assoc "d" exp2) with Not_found -> Cte(VAL_int 0))
       (op_apply2 "'::"
	  (try (List.assoc "f" exp2) with Not_found -> Cte(VAL_seq([||], Type_seq.seq_desc)))
	  (op_apply2 "'::"
	     (try (List.assoc "cf" exp2) with Not_found -> Cte(VAL_seq([||], Type_seq.seq_desc)))
	     (Cte (VAL_seq([||], Type_seq.seq_desc)))
	  )
       )
    )
  in
    
    op_apply2 "'::"
      exp1
      (op_apply2 "'::"
	 (try (List.assoc "v" exp2) with Not_found -> Cte(VAL_symbol "PatchDefaultVal"))
	 (op_apply2 "'::"
	    dfcf
	    (Cte (VAL_seq([||], Type_seq.pRhsSpec_desc)))
	 )
      )
;;







(**
 ** Macro pour la cration de types rcursifs en MGS
 **)

open Type;;
open Printer;;
open Def_gbf;;

(** Retourne l'identifiant d'un type, ie son nom **)
let get_id_of_recTypeDef = function
  | RecTypeDef_coll  (str, _)       -> str
  | RecTypeDef_gbf   (str, _)       -> str
  | RecTypeDef_qmf   (str, _)       -> str
  | RecTypeDef_rec   (str, _, _)    -> str
  | RecTypeDef_del   (str, _, _, _) -> str
  | RecTypeDef_prox  (str, _)       -> str
  | RecTypeDef_cstrt (str, _)       -> str
;;

(** *******************************************************************
 ** Chaque type peut prsenter des dpendances  d'autres types en
 ** cours de dfinition rcursive. Il faut donc distinguer l'ensemble
 ** des varaibles inconnues (qui doit tre vide pour une dfinition
 ** rcursive correcte) des types rcursifs en crous de dfinition.
 ** *******************************************************************)

let rec check_unknown_names_recTypeDef llhs = function
  | RecTypeDef_coll  (_, nt)        -> (Hash.mem type_tbl nt) or (Hash.mem type_tbl ("'"^nt)) or (List.mem nt llhs)
  | RecTypeDef_gbf   (_, _)         -> true
  | RecTypeDef_qmf   (_, _)         -> true
  | RecTypeDef_rec   (_, rspecl, _) -> List.for_all (Type_rec.check_unknown_names_record_type_specification llhs) rspecl
  | RecTypeDef_del   (_, _, _, _)   -> true
  | RecTypeDef_prox  (_, _)         -> true
  | RecTypeDef_cstrt (_, lcstrt)    -> Type.check_unknown_names_lcstrt lcstrt llhs

let make_rec_type_declarations decList =
  let list_of_lhs = List.map get_id_of_recTypeDef decList in

    (* Vrification des doublons *)
    begin
      if (not(check_no_double list_of_lhs)) then
	raise (Erreur "Macro: rec_type_declarations: some variables are defined several times.")
    end;

    (* Vrification de l'existance des types *)
    begin
      List.iter
	(fun nom ->
	   (try
	      let t = Hash.find type_tbl nom
	      in let err = "Error: type  "^nom^"  already defined."
	      in prerr_endline err; prftype stderr t; raise (Erreur err)
	    with Not_found -> (
	      try
		let t = Hash.find type_tbl ("'"^nom)
		in let err = "Warning: type "^nom^" corresponds to the predifined quoted type '"^nom^"\n         Without quote notation is going to be replaced"
		in prerr_endline err;
	      with Not_found -> ()
	    )
	   );
	   if Hash.mem Environnement.var_tbl nom then (
	     prerr_endline ("Warning: type predicate  "^nom^"  hiddes a function with same name")
	   )
	)
	list_of_lhs
    end;

    (* Vrification des Unknown: ils doivent apparatre dans list_of_lhs *)
    begin
      let chck_unknown = List.for_all (check_unknown_names_recTypeDef list_of_lhs) decList in
	if (not (chck_unknown)) then
	  raise (Erreur "Parser: rec_type_declarations: some variables are not defined in these types declarations.")
    end;
      
    (* Ordonanncement pour la cration des rec_collection *)
    let tdecl = 
      begin
	let rec classe ordre vrac =
	  if (vrac = []) then ordre
	  else (
	    let ordre_id = List.map get_id_of_recTypeDef ordre in
	    let (new_ordre, new_vrac) =
	      List.partition
		(function
		   | RecTypeDef_coll  (_, nt)      -> (Hash.mem type_tbl nt) or (Hash.mem type_tbl ("'"^nt)) or (List.mem nt ordre_id)
		   | RecTypeDef_gbf   (_, _)       -> true
		   | RecTypeDef_qmf   (_, _)       -> true
		   | RecTypeDef_rec   (_, _, dep)  -> List.for_all (fun t -> (Hash.mem type_tbl t) or (Hash.mem type_tbl ("'"^t)) or (List.mem t ordre_id)) dep
		   | RecTypeDef_del   (_, _, _, _) -> true
		   | RecTypeDef_prox  (_, _)       -> true
		   | RecTypeDef_cstrt (_, _)       -> true
		)
		vrac
	    in
	      if (new_ordre = []) then
		raise (Erreur "Parser: rec_type_declarations: circular dependence(s) between type decalarations (inheritance)")
	      else
		classe (ordre @ new_ordre) new_vrac
	  )
	in
	  classe [] decList
      end
    in

    (* Crer des noeuds bidons dans l'arbre en suivant l'ordre !!!
       (c noeuds seront corrects sauf pour rec et cstrt)
       On remarque que les noeuds bidons sont des create_dummy_xxx
       alors que les noeuds directement dfinissables sont des
       create_new_xxx. Ces fonctions doivent se trouver dans
       Type_xxx
    *)
    let lcpl_rtd_n =
      List.map
	(function
	   | RecTypeDef_coll  (id, pere)       as rtd -> (rtd, (Type.create_new_color id pere))
	   | RecTypeDef_gbf   (id, gbfspec)    as rtd -> (rtd, (Type_gbf.create_new_gbf id gbfspec))
	   | RecTypeDef_qmf   (id, qmfspec)    as rtd -> (rtd, (Type_qmf.create_new_qmf id qmfspec))
	   | RecTypeDef_rec   (id, recspec, _) as rtd -> (rtd, (Type_rec.create_dummy_record id))
	   | RecTypeDef_del   (id, i, b, x)    as rtd -> (rtd, (Type_del.create_new_del id i x b))
	   | RecTypeDef_prox  (id, x)          as rtd -> (rtd, (Type_prox.create_new_prox id x))
	   | RecTypeDef_cstrt (id, cstrt)      as rtd -> (rtd, (Type.create_dummy_constraint id))
	)
	tdecl
    in

    let dico =
      List.map
	(function
	   | (RecTypeDef_coll  (id, pere)    as rtd, n)    -> (id, n)
	   | (RecTypeDef_gbf   (id, gbfspec) as rtd, n)    -> (id, n)
	   | (RecTypeDef_qmf   (id, qmfspec) as rtd, n)    -> (id, n)
	   | (RecTypeDef_rec   (id, recspec, _) as rtd, n) -> (id, n)
	   | (RecTypeDef_del   (id, i, b, x) as rtd, n)    -> (id, n)
	   | (RecTypeDef_prox  (id, x)       as rtd, n)    -> (id, n)
	   | (RecTypeDef_cstrt (id, cstrt)   as rtd, n)    -> (id, n)
	)
	lcpl_rtd_n
    in
      (* Mise  jour du prdicat, uniquement pour les create_dummy prcdents.
	 Cela est fait en plusieurs tapes... Tous les identifiants inconnus
	 correspondent maintenant  des noeuds qui ont t crs. L'association
	 entre les noeuds et le nom du noeud se trouve dans dico.

	 On remplace dans un premier temps les inconnus en utilisant le dico et
	 les fonctions replaceDico_unknown_??? dfinies dans les type_xxx.ml.

	 On peut alors construire le prcicat en utilisant build_xxx_pred prsent
	 dans type_xxx.ml.

	 On met  jours les champs des noeuds.

	 Ces la fonction update_new_xxx_node qui fait cela et qui est
	 dfini galement dans type_xxx.ml
      *)
      List.map
	(function
	   | (RecTypeDef_coll  (id, pere)       as rtd, n) -> n
	   | (RecTypeDef_gbf   (id, gbfspec)    as rtd, n) -> n
	   | (RecTypeDef_qmf   (id, qmfspec)    as rtd, n) -> n
	   | (RecTypeDef_rec   (id, recspec, _) as rtd, n) -> (Type_rec.update_new_rec_node dico n id recspec)
	   | (RecTypeDef_del   (id, i, b, x)    as rtd, n) -> n
	   | (RecTypeDef_prox  (id, x)          as rtd, n) -> n
	   | (RecTypeDef_cstrt (id, cstrt)      as rtd, n) -> (Type.update_new_cstrt_node dico n id cstrt)
	)
	lcpl_rtd_n
;;
