/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#include "g-map-generic.hh"
using namespace GMap3d;
//******************************************************************************
int CGMapGeneric::duplicateMarkedDarts(int AMarkNumber, int ADirectInfoIndex,
				       bool ADuplicateEmbeddings,
				       bool A0Sew, bool A1Sew,
				       bool A2Sew, bool A3Sew)
{
  int nbDuplicated = 0;

  int duplicated = getNewMark();

  int directInfoIndex =
    ADirectInfoIndex<0 ? getNewDirectInfo() : ADirectInfoIndex;

  CDynamicCoverageAll it(this);

  // Duplication des brins:
  for (; it.cont(); ++it)
    if (isMarked(*it, AMarkNumber))
      {
	CDart * newDart = addMapDart();

	if (ADuplicateEmbeddings)
	  setMark(newDart, duplicated);

	(*it)->setDirectInfo(directInfoIndex, newDart);

	++nbDuplicated;
      }

  // Coutures:
  sewDuplicatedDarts(&it, AMarkNumber, directInfoIndex,
		     A0Sew, A1Sew, A2Sew, A3Sew);

  // Duplication des plongements:
  if (ADuplicateEmbeddings)
    duplicateEmbeddingsOfDuplicatedDarts(&it,
					 AMarkNumber, duplicated,
					 directInfoIndex);

  // Librations:
  if (ADirectInfoIndex<0)
    freeDirectInfo(directInfoIndex);

  freeMark(duplicated);

  return nbDuplicated;
}
//******************************************************************************
CDart * CGMapGeneric::duplicateDarts(CDart * ADart, TOrbit AOrbit,
				    int ADirectInfoIndex,
				    bool ADuplicateEmbeddings)
{
  assert(ADart!=NULL);

  int selected = getNewMark();
  int duplicated = getNewMark();

  int directInfoIndex =
    ADirectInfoIndex<0 ? getNewDirectInfo() : ADirectInfoIndex;

  CCoverage * cov = getDynamicCoverage(ADart, AOrbit);

  // Duplication des brins:
  for (; cov->cont(); ++(*cov))
    {
      CDart * newDart = addMapDart();
      setMark(newDart, duplicated);
      setMark(**cov, selected);
      (**cov)->setDirectInfo(directInfoIndex, newDart);
    }

  // Coutures:
  sewDuplicatedDarts(cov, selected, directInfoIndex,
		     AND_ORBIT(AOrbit, ORBIT_0) != ORBIT_SELF,
		     AND_ORBIT(AOrbit, ORBIT_1) != ORBIT_SELF,
		     AND_ORBIT(AOrbit, ORBIT_2) != ORBIT_SELF,
		     AND_ORBIT(AOrbit, ORBIT_3) != ORBIT_SELF);

  // Duplication des plongements:
  if (ADuplicateEmbeddings)
    duplicateEmbeddingsOfDuplicatedDarts(cov,
					 selected, duplicated,
					 directInfoIndex);

  // Librations:
  CDart * result = getDirectInfoAsDart(ADart, directInfoIndex);

  for (cov->reinit(); cov->cont(); ++(*cov))
    unsetMark(**cov, selected);

  delete cov;

  if (ADirectInfoIndex<0)
    freeDirectInfo(directInfoIndex);

  freeMark(selected);
  freeMark(duplicated);

  return result;
}
//******************************************************************************
#define DUP(D) (getDirectInfoAsDart(D, ADirectInfoIndex))
//******************************************************************************
void CGMapGeneric::sewDuplicatedDarts(CCoverage * ACoverage,
				      int AMarkNumber, int ADirectInfoIndex,
				      bool A0Sew, bool A1Sew,
				      bool A2Sew, bool A3Sew)
{
  assert(ACoverage!=NULL);
  assert(AMarkNumber>=0);
  assert(ADirectInfoIndex>=0);

  // 0 et 1 coutures:
  for (ACoverage->reinit(); ACoverage->cont(); ++(*ACoverage))
    {
      CDart * d = **ACoverage;

      if (AMarkNumber<0 || isMarked(d, AMarkNumber))
	{
	  if (A0Sew)
	    if (!isFree0(d) && isMarked(alpha0(d), AMarkNumber) &&
		isFree0(DUP(d)))
	      linkAlpha0(DUP(d), DUP(alpha0(d)));

	  if (A1Sew)
	    if (!isFree1(d) && isMarked(alpha1(d), AMarkNumber) &&
		isFree1(DUP(d)))
	      linkAlpha1(DUP(d), DUP(alpha1(d)));
	}
    }

  // 2 and 3 coutures:
  for (ACoverage->reinit(); ACoverage->cont(); ++(*ACoverage))
    {
      CDart * d = **ACoverage;

      if (isMarked(d, AMarkNumber))
	{
	  if (A2Sew)
	    if (!isFree2(d) && isMarked(alpha2(d), AMarkNumber) &&
		canSew2(DUP(d), DUP(alpha2(d))))
	      topoSew2(DUP(d), DUP(alpha2(d)));
	
	  if (A3Sew)
	    if (!isFree3(d) && isMarked(alpha3(d), AMarkNumber) &&
		canSew3(DUP(d), DUP(alpha3(d))))
	      topoSew3(DUP(d), DUP(alpha3(d)));
	}
    }
}
//******************************************************************************
void CGMapGeneric::duplicateEmbeddingsOfDuplicatedDarts(CCoverage * ACoverage,
							int AMarkInitial,
							int AMarkDuplicated,
							int ADirectInfoIndex)
{
  assert(ACoverage!=NULL);
  assert(AMarkInitial>=0);
  assert(AMarkDuplicated>=0);
  assert(ADirectInfoIndex>=0);
  assert(AMarkInitial!=AMarkDuplicated);

  int treated = getNewMark();

  for (int i=ORBIT_0; i<=ORBIT_0123; ++i)
    if (isOrbitUsed(TOrbit(i)))
      {
	// Duplication pour l'orbite i:
	for (ACoverage->reinit(); ACoverage->cont(); ++(*ACoverage))
	  {
	    CDart * d = **ACoverage;

	    if (!isMarked(d, treated))
	      {
		if (isMarked(d, AMarkInitial))
		  {
		    CDart * owner = getEmbeddingOwner(d, TOrbit(i));
		    assert(owner!=NULL);
		
		    CEmbedding * E = owner->getEmbedding(TOrbit(i));
		    assert(E!=NULL);
		
		    CCoverage * cov = getDynamicCoverage(d, TOrbit(i));
		    assert(cov!=NULL);
		
		    for (; cov->cont(); setMark(**cov, treated), ++(*cov))
		      if (isMarked(**cov, AMarkInitial) &&
			  !isMarked(DUP(**cov), treated))
			{
			  CEmbedding * dupE = E->copy();
			  DUP(**cov)->addEmbedding(dupE);
			
			  markOrbit(DUP(**cov), TOrbit(i), treated);
			  unmarkOrbit(DUP(**cov), TOrbit(i), AMarkDuplicated);
			}
		
		    delete cov;
		  }
		else
		  if (!isMarked(d, AMarkDuplicated))
		    setMark(d, treated);
	      }
	  }

	// Dmarquage:
	if (ACoverage->type() == COVERAGE_ALL)
	  negateMaskMark(treated);
	else
	  for (ACoverage->reinit(); ACoverage->cont(); ++(*ACoverage))
	    {
	      if (isMarked(**ACoverage, treated))
		unmarkOrbit(**ACoverage, TOrbit(i), treated);

	      if (isMarked(DUP(**ACoverage), treated))
		unmarkOrbit(DUP(**ACoverage), TOrbit(i), treated);
	    }
      }

  freeMark(treated);
}
//******************************************************************************
#undef DUP
//******************************************************************************
