/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Calcule la plus grande dimension pour laquelle ADart est cousu avec un
 * autre brin. Si le brin n'est cousu ni par alpha0, ni par alpha1, ni par
 * alpha2, ni par alpha3, la valeur retourne est -1.
 *
 * Remarque: Cette mthode n'est plus utilise par les mthodes
 * intuitiveIsolateMarkedCells et intuitiveMergeMarkedCells.
 *
 * @param ADart Un brin de la carte
 * @return La plus grande dimension pour laquelle le brin n'est pas libre
 *
 * @postcondition La valeur retourne vaut 3, 2, 1, 0 ou -1.
 */
int getMergeDimension(CDart* ADart);

/**
 * Teste si les deux cellules de dimension ADim dsignes par les brins
 * ADart1 et ADart2 peuvent tre fusionnes.
 *
 * @param ADart1 Un pointeur sur un premier brin
 * @param ADart2 Un pointeur sur un deuxime brin
 * @param ADim Une dimension
 * @return Un boolen indiquant si les deux cellules peuvent tre fusionnes.
 */
bool canMerge(CDart* ADart1, CDart* ADart2, int ADim);

/**
 * Fusionne les deux cellules de dimension ADim dsignes par les brins
 * ADart1 et ADart2.
 *
 * @param ADart1 Un pointeur sur un premier brin
 * @param ADart2 Un pointeur sur un deuxime brin
 * @param ADim Une dimension
 * @param ADeleteDarts Un boolen indiquant s'il faut supprimer les brins
 *   correspondant  la cellule de dimension ADim-1 situe entre les
 *   cellules de dimension ADim incidentes  ADart1 et ADart2
 *
 * @precondition canMerge(ADart1, ADart2, ADim)
 */
void merge(CDart* ADart1, CDart* ADart2, int ADim, bool ADeleteDarts = false);

/**
 * Isole lorsque cela est possible les cellules de dimension ADim en
 * fusionnant entre elles les cellules de dimension ADim+1 qui leurs sont
 * incidentes.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (0, 1 ou 2)
 * @param ADeleteDarts Un boolen indiquant si les cellules isoles doivent
 *   tre dtruites
 * @return Le nombre de cellules isoles
 */
int isolateMarkedCells(int AMarkNumber, int ADim,
		       bool ADeleteDarts = false);

/**
 * Isole lorsque cela est possible les cellules dont un brin au moins est
 * marqu.
 * La mthode procde en 3 passes, avec priorit aux cellules de dimension
 * leve (faces).
 *
 * @param AMarkNumber Un numro de marque
 * @param ADeleteDarts Un boolen indiquant si les cellules isoles doivent
 *   tre dtruites
 * @return Le nombre de cellules isoles
 */
int intuitiveIsolateMarkedCells(int AMarkNumber, bool ADeleteDarts = false);

/**
 * Fusionne lorsque cela est possible (avec leur cellule incidente) les
 * cellules dont un brin au moins est marqu.
 * Les cellules de dimension ADim-1 qui doivent tre isoles sont
 * dtermines par la mthode protge 'markCellsToBeIsolated'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (1, 2 ou 3)
 * @param ADeleteDarts Un boolen indiquant si la fusion doit se faire en
 *   supprimant des brins
 * @return Le nombre de fusions effectues
 */
int mergeMarkedCells(int AMarkNumber, int ADim,
		     bool ADeleteDarts = false);

/**
 * Fusionne lorsque cela est possible (avec leurs cellule incidente) les
 * cellules dont un brin au moins est marqu.
 * Les fusions sont faites en 3 passes, avec priorit aux fusions de
 * cellules de dimension leve (volumes).
 * Lors de chaque passe, les cellules qui doivent tre isoles sont
 * dtermines par la mthode protge 'markCellsToBeIsolated'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADeleteDarts Un boolen indiquant si la fusion doit se faire en
 *   supprimant des brins
 * @return Le nombre de fusions effectues
 */
int intuitiveMergeMarkedCells(int AMarkNumber, bool ADeleteDarts = false);

protected:

/**
 * Cette mthode auxiliaire isole (uniquement lorsque cela est possible si
 * AVerif=='vrai') les cellules de dimension ADim dont un brin au moins est
 * slectionn, c'est--dire qu'elle fusionne les cellules de dimension
 * ADim+1 incidentes aux cellules de dimension ADim marques.
 * Lorsqu'une cellule est isole, elle est marque comme traite avec la
 * marque AMarkTreated (si AMarkTreated>=0).
 *
 * Si AVerif=='faux', les cellules marques doivent imprativement pouvoir
 * tre isoles.
 *
 * La mthode est utilise par:
 *              isolateMarkedCells(int, int, bool),
 *     intuitiveIsolateMarkedCells(int,      bool),
 *                mergeMarkedCells(int, int, bool) et
 *       intuitiveMergeMarkedCells(int,      bool).
 *
 * @param AMarkNumber Un numro de marque indiquant les brins slectionns
 * @param ADim La dimension des cellules qui doivent tre isoles
 * @param ADeleteDarts Un boolen indiquant si les cellules doivent tre
 *   supprimes une fois isoles
 * @param AVerif Un boolen indiquant si pour chaque cellule marque il
 *   faut vrifier si l'opration d'isolement est possible
 * @param AMarkTreated Un numro de marque indiquant (si positif ou nul)
 *   les brins dj traits et dont il ne faut donc pas se proccuper
 * @return Le nombre de cellules isoles
 *
 * @precondition AMarkTreated!=AMarkNumber
 * @precondition AVerif=='faux' ou
 *   toutes les cellules marques peuvent tre isoles
 */
int isolateMarkedCells(int AMarkNumber, int ADim, bool ADeleteDarts,
		       bool AVerif, int AMarkTreated = -1);

/**
 * Marque avec la marque AMarkDestination tous les brins de chaque cellule
 * de dimension ADim qui peut tre isole (c'est--dire qui est ADim-cousue
 * et de degr infrieur ou gal  2) et qui est incidente  des cellules
 * de dimension ADim+1 dont un brin au moins est marqu avec la marque
 * AMarkSource.
 * Voir le corps de la mthode pour plus de dtail.
 *
 * La mthode est utilise par:
 *                mergeMarkedCells(int, int, bool) et
 *       intuitiveMergeMarkedCells(int,      bool).
 *
 * @param AMarkSource Un numro de marque
 * @param ADim Une dimension (celle des cellules  isoler)
 * @param AMarkDestination Un numro de marque
 * @param AMarkTreated Un numro de marque indiquant (si positif ou nul)
 *   les brins dj traits, et dont il ne faut donc pas se proccuper
 *
 * @precondition Aucun brin n'est marqu avec la marque AMarkDestination
 * @precondition 0 <= ADim <= 2
 */
void markCellsToBeIsolated(int AMarkSource, int ADim, int AMarkDestination,
			   int AMarkTreated = -1);

//******************************************************************************
