/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * @param ADart1 Un brin de la carte
 * @param ADart2 Un brin de la carte
 * @param AOrbit Une orbite quelconque
 *
 * @return true si ADart1 et ADart2 appartiennent  la mme orbite oriente.
 */
bool belongToSameOrientedOrbit(CDart* ADart1, CDart* ADart2, TOrbit AOrbit);

/**
 * Teste si l'objet dtermin par ADart et AOrbit est orientable.
 * Remarque: L'objet peut ne pas tre orientable uniquement pour les
 * orbites ORBIT_VERTEX, ORBIT_VOLUME et ORBIT_CC.
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite (ORBIT_VOLUME ou ORBIT_CC)
 * @return 'vrai' si l'objet est orientable
 *
 * @precondition AOrbit n'est pas une orbite de bord
 */
bool isOrientable(CDart* ADart, TOrbit AOrbit);

/**
 * Compte le nombre de bords incidents aux brins marqus avec la marque
 * AMarkNumber. Pour ne compter que certains bords, mettre  NULL les
 * paramtres ANbi correspondant  une dimension non voulue.
 *
 * @param AMarkNumber Un numro de marque
 * @param ANbi Le nombre de i-bords
 * @param ANb0Coupled Le nombre de faces ayant deux 0-bords
 */
void countBorders(int AMarkNumber,
		  int* ANb0, int* ANb1, int* ANb2, int* ANb3);

/**
 * Compte le nombre de bords incidents  l'orbite (ADart, AOrbit).
 * Pour ne compter que certains bords, mettre  NULL les paramtres ANb-
 * correspondant  une dimension non voulue.
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param ANbi Le nombre de i-bords
 */
void countBorders(CDart* ADart, TOrbit AOrbit,
		  int* ANb0, int* ANb1, int* ANb2, int* ANb3);

/**
 * Compte le nombre de cellules incidentes  un brin marqu avec la marque
 * AMarkNumber. Pour ne compter que certains types de cellules, mettre 
 * NULL les paramtres ANb- correspondant  une dimension non voulue.
 *
 * @param AMarkNumber Un numro de marque
 * @param ANbi Les nombres de i-cellules
 */
void countCells(int AMarkNumber,
		int* ANb0, int* ANb1, int* ANb2, int* ANb3, int* ANb4,
		int* ANbDarts = NULL);

/**
 * Compte le nombre de cellules incidentes ou incluses dans l'orbite
 * (ADart, AOrbit). Pour ne compter que certains types de cellules, mettre
 *  NULL les paramtres ANb- correspondant  une dimension non voulue.
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param ANbi Les nombres de i-cellules
 */
void countCells(CDart* ADart, TOrbit AOrbit,
		int* ANb0, int* ANb1, int* ANb2, int* ANb3, int* ANb4,
		int* ANbDarts = NULL);

/**
 * Calcule les caractristiques globales de la carte (nombre de brins,
 * de cellules, de bords). Pour ne pas calculer certaines caractristiques,
 * mettre  NULL les paramtres correspondants.
 *
 * @param ANbDarts Le nombre de brins de la carte
 * @param ANb{Vertices,...,CC} Le nombre de cellules de la carte
 * @param ANbiBorders Le nombre de i-bords de la carte
 */
void getGlobalCharacteristics(int* ANbDarts,
			      int* ANbVertices, int* ANbEdges,
			      int* ANbFaces, int* ANbVolumes, int* ANbCC,
			      int* ANb0Borders, int* ANb1Borders,
			      int* ANb2Borders, int* ANb3Borders);

/**
 * Calcule les caractristiques d'une surface dont un brin est donn
 * (nombre de brins, de cellules, de bords, caractristique d'Euler,
 * orientabilit, genre). Pour ne pas calculer certaines caractristiques,
 * mettre  NULL les paramtres correspondants.
 *
 * @param ADart Un brin du volume
 * @param ANbDarts Le nombre de brins du volume
 * @param ANb{Vertices,...,CC} Le nombre de cellules du volume
 * @param ANbiBorders Le nombre de i-bords du volume
 * @param ANb2BordersWhenClosed Le nombre total de 2-bords qu'on aurait si
 *   le volume incident  ADart tait 0-ferm puis 1-ferm.
 * @param AEuler La caractristique d'Euler-Poincar du volume
 * @param AOrient Le coefficient d'orientabilit du volume
 * @param AGenus Le genre du volume
 */
void getSurfaceCharacteristics(CDart* ADart,
			       int* ANbDarts,
			       int* ANbVertices, int* ANbEdges,
			       int* ANbFaces,
			       int* ANb0Borders, int* ANb1Borders,
			       int* ANb2Borders,
			       int* ANb2BordersWhenClosed,
			       int* AEuler, int* AOrient, int* AGenus);

/**
 * Dtermine  partir des caractristiques topologiques d'une surface son
 * nom usuel.
 *
 * @param AB2 Le nombre de 2-bords de la surface
 * @param AQ Le coefficient d'orientabilit de la surface
 * @param AG Le genre de la surface
 * @return Le nom de la surface (par exemple, "tore  2 trous", "bouteille
 *   de Klein", etc)
 */
std::string getSurfaceName(int AB2, int AQ, int AG);

/**
 * Teste si la carte ne contient qu'une seule composante connexe.
 * Si c'est le cas, la valeur 'vrai' est retourne.
 *
 * @return 'vrai' si la carte est connexe
 */
bool isConnex();

/**
 * Vrifie l'intgrit de la carte.
 * La carte est cohrente si les orbites 0, 1, 2 et 3 sont des permutations
 * et si les orbites 02, 03 et 13 sont des involutions.
 *
 * @return Un boolen indiquant si la topologie de la carte est cohrente
 */
bool checkTopology();

/**
 * Vrifie si pour chaque cellule d'orbite AOrbit de la carte il existe un
 * seul brin de la cellule qui possde le plongement identifi par
 * AAttributeIdentity.
 * Si AEmbeddingMustExist vaut 'vrai', une cellule non plonge est
 * considre comme un problme d'intgrit de la carte.
 *
 * @param AOrbit l'orbite plonge  tester
 * @param AAttributeIdentity L'identificateur de l'attribut  vrifier
 * @param AEmbeddingMustExist Un boolen indiquant si la dtection d'une
 *   orbite non plonge doit provoquer une erreur
 * @return Un boolen indiquant si les plongements sont cohrents
 *
 * @precondition isOrbitUsed(AOrbit);
 */
bool checkEmbeddings(TOrbit AOrbit, int AAttributeIdentity,
		     bool AEmbeddingMustExist = true);

/**
 * Teste si tous les brins de l'orbite (ADart, AOrbit) sont cousus
 * par alpha ADimension.
 * Si c'est le cas, la mthode retourne vrai.
 *
 * @param ADart Un brin de la carte
 * @param ADimension Une dimension quelconque (0, 1, 2 ou 3)
 * @param AOrbit Une orbite quelconque
 * @return Un boolen
 */
bool isIClosedOrbit(CDart* ADart, int ADimension, TOrbit AOrbit);

/**
 * Teste si tous les brins de l'orbite (ADart, AOrbit) sont non cousus
 * par alpha ADimension.
 * Si c'est le cas, la mthode retourne vrai.
 *
 * @param ADart Un brin de la carte
 * @param ADimension Une dimension quelconque (0, 1, 2 ou 3)
 * @param AOrbit Une orbite quelconque
 * @return Un boolen
 */
bool isIFreeOrbit(CDart* ADart, int ADimension, TOrbit AOrbit);

//******************************************************************************
