/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Extrude la cellule de dimension ADim incidente  ADart.
 * Si AMarkToExtrude est positive ou nulle, cette marque indique les cellules
 * qui doivent tre extrudes. Selon que les cellules voisines de la cellule
 * en cours doivent tre extrudes ou pas et selon la valeur que renvoie la
 * mthode 'CGMapGeneric::getExtrusionWithAdacentSews', les plongements ne
 * sont pas calculs de la mme manire.
 *
 * La mthode appelle 'CGMapGeneric::extrudeByNormal' puis calcule les
 * plongements des sommets topologiques crs.
 *
 * Pour la signification du paramtre AMarkToExtrude, se reporter  la
 * mthode 'CGMapGeneric::extrudeByNormal'.
 * Pour la signification des paramtres AMarkExtruded et
 * AExtrusionWithAdjacentSews, se reporter  la mthode
 * 'CGMapGeneric::topoExtrude'.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param AExtrusionCoef Un TCoordinate fixant le coefficient d'extrusion.
 *   Ce coefficient est utilis pour pondrer le vecteur d'extrusion.
 * @param AMarkToExtrude Une marque indiquant les cellules  extruder
 * @param AMarkExtruded Une marque indiquant les cellules dj extrudes
 * @return Un brin de la cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeCell(ADart, ADim)
 */
CDart* extrudeByNormal(CDart* ADart, int ADim,
		       bool AExtrusionWithAdjacentSews,
		       TCoordinate AExtrusionCoef,
		       int AMarkToExtrude = -1, int AMarkExtruded = -1);

/**
 * Extrude la cellule de dimension ADim incidente  ADart selon le chemin
 * donn par APath (orbite O1).
 *
 * La mthode appelle 'CGMapGeneric::extrudeByPath' puis calcule les
 * plongements des sommets topologiques crs.
 *
 * Pour la signification du paramtre AMarkExtruded, se reporter  la
 * mthode 'CGMapGeneric::topoExtrude'.
 *
 * Pour la signification du paramtre ADirectInfoIndex, se reporter  la
 * mthode 'CGMapGeneric::extrudeByPath'.
 *
 * Si APonderationPath est diffrent de NULL, une homothtie est applique
 * sur chaque section de l'objet obtenu aprs extrusion. Chaque homothtie
 * est centre sur le sommet du chemin d'extrusion lui correspondant.
 * APonderationDart dsigne un chemin dont les sommets (plus prcisment la
 * valeur donne par 'CVertex::getY' applique sur ces sommets) donnent les
 * coefficients d'homothtie.
 *
 * Remarque: Le nombre d'artes du chemin peut tre diffrent du nombre
 * d'artes de la fonction de pondration. Si c'est le cas, le paramtre
 * de pondration est calcul par interpolation linaire.
 *
 * Remarque: Si la fonction de pondration s'annule pour une valeur donne,
 * tous les sommets de la section correspondante sont confondus. Cette
 * section rduite  un sommet se propage alors jusqu' la fin de
 * l'extrusion.
 *
 * @param ADart Un brin de la carte incident  la cellule  extruder
 * @param ADim Une dimension (1 ou 2)
 * @param APath Un brin de la carte incident au chemin d'extrusion
 * @param APonderationPath Un brin de la carte
 * @param AExtrusionWithAdjacentSews Un boolen 
 * @param ARotateExtrudedCells Un boolen indiquant si les objets
 *   extruds doivent tre orients
 * @param AScaleExtrudedCells Un boolen indiquant si les objets
 *   extruds doivent tre redimensionns
 * @param AMarkExtruded Une marque indiquant les cellules dj extrudes
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un brin de la dernire cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeByPath(ADart, ADim, APath)
 * @precondition APonderationPath==NULL || ! isClosedPolyline(APonderationPath)
 */
CDart* extrudeByPath(CDart* ADart, int ADim, CDart* APath,
		     bool AExtrusionWithAdjacentSews,
		     bool ARotateExtrudedCells,
		     bool AScaleExtrudedCells,
		     CDart* APonderationPath = NULL,
		     int AMarkExtruded = -1,
		     int ADirectInfoIndex = -1);

/**
 * Surcharge la mthode 'CGMapGeneric::extrudeByPathMarkedCells'.
 * Les cellules sont ventuellement positionnes perpendiculairement
 * au chemin avant d'effectuer l'extrusion (voir mthodes
 * 'setExtrusionInitialPosition' et 'setExtrusionInitialDirection').
 *
 * Attention: si 'getExtrusionInitialDirection' retourne la valeur 'vrai',
 * le vecteur normal de chaque cellule  extruder est calcul, ce qui
 * implique qu'un brin sur deux au maximum pour chaque cellule doit
 * tre slectionn pour viter toute ambigut.
 *
 * Pour la signification des paramtres APonderationDart,ARotateExtrudedCells
 * et AScaleExtrudedCells, se reporter  la mthode 'extrudeByPath'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (1 ou 2)
 * @param APath Un brin de la carte incident au chemin d'extrusion
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param AInitialTranslateExtrudedCells Un entier indiquant si les
 *   objets extruds doivent tre translats sur le premier
 *   sommet du chemin. 0 : Aucune translation. 1 : Les objets
 *   extruds sont initialement translats sur le chemin.
 *   2 : Le chemin est initialement dplac sur le barycentre
 *   des cellules  extruder. 
 * @param AInitialRotateExtrudedCells Un entier indiquant comment
 *   les objets  extruder et le chemin doivent tre orients
 *   l'un par rapport  l'autre avant l'extrusion. 0 : Aucune
 *   rotation. 1 : Les objets extruds sont initialement orients
 *   perpendiculairement  la premire arte du chemin. 2 : Le
 *   chemin est initialement orient de manire  tre perpendiculaire
 *   aux cellules  extruder.
 * @param ARotateExtrudedCells Un boolen
 * @param AScaleExtrudedCells Un boolen
 * @param APonderationPath Un brin de la carte
 * @return Le nombre de cellules extrudes selon le chemin donn
 */
int extrudeByPathMarkedCells(int AMarkNumber, int ADim, CDart* APath,
			     bool AExtrusionWithAdjacentSews,
			     int AInitialTranslateExtrudedCells,
			     int AInitialRotateExtrudedCells,
			     bool ARotateExtrudedCells,
			     bool AScaleExtrudedCells,
			     CDart* APonderationPath = NULL);

/**
 * Cre un chemin d'extrusion compos de ANbEdges artes, ouvert ou ferm
 * selon la valeur de AAngle. Ce chemin est utilis pour simuler l'extrusion
 * par rvolution autour d'un axe.
 *
 * @param AAxeVertex Un sommet de l'axe de rvolution
 * @param AAxeDirection Un vecteur directeur de l'axe de rvolution
 * @param AFirstVertex La position du 1er sommet du chemin
 * @param AAngle L'angle de rvolution (compris entre -360 et +360)
 * @param ANbEdges Le nombre d'artes du chemin
 * @return Un brin extrmit du chemin
 *
 * @precondition !AAxeDirection.isNull()
 * @precondition ANbEdges >= 3
 */
CDart* createRevolutionPath(const CVertex & AAxeVertex,
			    const CVertex & AAxeDirection,
			    const CVertex & AFirstVertex,
			    TCoordinate AAngle, int ANbEdges);

/**
 * Extrude la cellule de dimension ADim incidente  ADart selon autour d'un
 * axe. La mthode cre un chemin constitu de ANbEdges artes qui est ferm
 * ou ouvert selon la valeur de AAngle puis appelle la mthode
 * 'extrudeByPath'.
 *
 * Pour la signification des paramtres AMarkExtruded, ADirectInfoIndex,
 * APonderationPath, AExtrusionWithAdjacentSews, ARotateExtrudedCells,
 * AScaleExtrudedCells se reporter  la mthode 'extrudeByPath'.
 *
 * @param ADart Un brin de la carte incident  la cellule  extruder
 * @param ADim Une dimension (1 ou 2)
 * @param AAxeVertex Un sommet de l'axe de rvolution
 * @param AAxeDirection Un vecteur directeur de l'axe de rvolution
 * @param AFirstVertex La position du barycentre des cellules  extruder
 * @param AAngle L'angle de rvolution (compris entre -360 et +360)
 * @param ANbEdges Le nombre de sections  crer
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param ARotateExtrudedCells Un boolen
 * @param AScaleExtrudedCells Un boolen
 * @param APonderationPath Un brin de la carte
 * @param AMarkExtruded Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un brin de la dernire cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeByRevolution(ADart, ADim, closed),
 *   avec closed = vrai si AAngle/360 est une valeur entire
 * @precondition !AAxeDirection.isNull()
 * @precondition ANbEdges >= 3
 */
CDart* extrudeByRevolution(CDart* ADart, int ADim,
			   const CVertex & AAxeVertex,
			   const CVertex & AAxeDirection,
			   const CVertex & AFirstVertex,
			   TCoordinate AAngle, int ANbEdges,
			   bool AExtrusionWithAdjacentSews,
			   bool ARotateExtrudedCells,
			   bool AScaleExtrudedCells,
			   CDart* APonderationPath = NULL,
			   int AMarkExtruded = -1,
			   int ADirectInfoIndex = -1);

/**
 * Calcule le chemin correspondant  l'axe donn et appelle la mthode
 * 'extrudeByPathMarkedCells'.
 * Les cellules sont ventuellement positionnes perpendiculairement
 * au chemin avant d'effectuer l'extrusion (voir la mthode
 * 'setExtrusionInitialDirection').
 *
 * Attention: si 'getExtrusionInitialDirection' retourne la valeur 'vrai',
 * le vecteur normal de chaque cellule  extruder est calcul, ce qui
 * implique qu'un brin sur deux au maximum pour chaque cellule doit
 * tre slectionn pour viter toute ambigut.
 *
 * Pour la signification du paramtre APonderationDart,
 * AExtrusionWithAdjacentSews, ARotateExtrudedCells et AScaleExtrudedCells
 * se reporter  la mthode 'extrudeByPath'.
 * Pour la signification des paramtres AInitialTranslateExtrudedCells et
 * AInitialRotateExtrudedCells, voir la mthode 'extrudeByPathMarkedCells'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (1 ou 2)
 * @param AAxeVertex Un sommet de l'axe de rvolution
 * @param AAxeDirection Un vecteur directeur de l'axe de rvolution
 * @param ANbEdges Le nombre de sections  crer
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param AInitialTranslateExtrudedCells Un entier
 * @param AInitialRotateExtrudedCells Un entier   
 * @param APonderationPath Un brin de la carte
 * @return Le nombre de cellules extrudes selon l'axe donn
 *
 * @precondition !AAxeDirection.isNull()
 * @precondition ANbEdges >= 3
 */
int extrudeByRevolutionMarkedCells(int AMarkNumber, int ADim,
				   const CVertex & AAxeVertex,
				   const CVertex & AAxeDirection,
				   TCoordinate AAngle, int ANbEdges,
				   bool AExtrusionWithAdjacentSews,
				   int AInitialTranslateExtrudedCells,
				   int AInitialRotateExtrudedCells,
				   bool ARotateExtrudedCells,
				   bool AScaleExtrudedCells,
				   CDart* APonderationPath = NULL);

protected:

/**
 * Marque un brin sur deux des cellules de dimension ADim qui doivent tre
 * extrudes.
 * Voir CGMapGeneric::markCellsToBeExtruded pour plus d'informations.
 *
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AMarkNumberSrce Un numro de marque source
 * @param AMarkNumberDest Un numro de marque destination
 */
void markCellsToBeExtrudedByNormal(int ADim,
				   int AMarkNumberSrce,
				   int AMarkNumberDest);

//******************************************************************************
