/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef ATTRIBUTE_HH
#define ATTRIBUTE_HH
//******************************************************************************
#include "kernel-types.hh"
#include "inline-macro.hh"

#include <iostream>
//******************************************************************************
/**
 * Dfinition de l'interface attribut.
 * Toute classe concrte implmentant cette classe abstraite peut tre
 * associe  une orbite de la G-Carte.
 *
 * Cinq mthodes virtuelles pures sont  implmenter :
 *  - getType() : l'identificateur de la classe de l'attribut,
 *  - copy()     : le constructeur de recopie pour la propagation de l'attribut,
 *  - destroy()  : le destructeur de l'attribut,
 *  - save()     : sauvegarde de l'attribut dans un flux,
 *  - load()     : lecture de l'attribut depuis un flux.
 *
 * @author MOKA TEAM
 */
class CAttribute
{
public:
  // @name Constructeurs et destructeur.
  // @{

  /// Constructeur par dfaut
  CAttribute();

  /// Constructeur par copie
  CAttribute(const CAttribute& CAttribute);

  /// Destructeur
  virtual ~CAttribute();

  // @}
  // @name Mthodes virtuelles pures  implmenter dans les classes filles.
  // @{

  /**
   * @return L'identificateur de la classe d'attribut.
   */
  virtual TAttributeId getType() const = 0;

  /**
   * Cette mthode permet de spcifier le comportement de l'attribut
   * lorsqu'il est propag.
   *
   * @return Un clone de l'attribut
   */
  virtual CAttribute* copy() const = 0;

  /**
   * Mthode de destruction de l'attribut.
   * Lorsque l'attribut n'est plus prsent dans le modle, cette mthode de
   * destruction est dclenche.
   */
  virtual void destroy() = 0;

  /**
   * Sauvegarde de l'attribut.
   * Les mthodes de sauvegarde des attributs sont appeles lors de la
   * sauvegarde du modle.
   *
   * @param AStream Le flux de sortie.
   */
  virtual void save(std::ostream& AStream) const = 0;

  /**
   * Lecture de l'attribut.
   *
   * Les mthodes de lecture des attributs sont appeles lors du chargement
   * du modle.
   *
   * @param AStream Le flux d'entre.
   */
  virtual void load(std::istream& AStream) = 0;

  // @}
  // @name Oprateurs de flux.
  // @{

  /**
   * @param AOS Le flux de sortie.
   * @param AAttribute L'attribut  envoyer dans le flux.
   */
  friend std::ostream& operator<<(std::ostream& AOS,
				  const CAttribute& AAttribute);

  /**
   * @param AIS Le flux d'entre.
   * @param AAttribute L'attribut  lire depuis le flux.
   */
  friend std::istream& operator>>(std::istream& AIS,
				  CAttribute& AAttribute);

  // @}
  // @name Mthodes de manipulation de la liste des attributs.
  // @{

  /**
   * Rcupration de l'attribut successeur.
   *
   * @return Un pointeur sur l'attribut suivant, ou NULL.
   */
  CAttribute* getNext() const;

  /**
   * Chanage de l'attribut successeur.
   *
   * @param ANext Un pointeur sur l'attribut qui va devenir le suivant.
   */
  void setNext(CAttribute* ANext);

  /**
   * Rcupration de l'attribut prdcesseur.
   *
   * @return Un pointeur sur l'attribut prcdent, ou NULL.
   */
  CAttribute* getPrev() const;

  /**
   * Chanage de l'attribut prdcesseur.
   *
   * @param APrev Un pointeur sur l'attribut qui va devenir le prcdent.
   */
  void setPrev(CAttribute* APrev);

  // @}

private:
  // @name Champs privs (chanage des attributs)
  // @{

  /// Attribut prcdent.
  CAttribute* FPrev;

  /// Attribut suivant.
  CAttribute* FNext;

  // @}
};
//******************************************************************************
#include INCLUDE_INLINE("attribute.icc")
//******************************************************************************
#endif // ATTRIBUTE_HH
//******************************************************************************
