/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef G_MAP_BASIC_HH
#define G_MAP_BASIC_HH
//******************************************************************************
#include "kernel-types.hh"
#include "inline-macro.hh"

#include <bitset>
//******************************************************************************
namespace GMap3d
{
  class CDart;

  /**
   * Cette classe regroupe uniquement des mthodes de base :
   *  - pour la gestion des marques boolennes,
   *  - pour la gestion des champs direct-info,
   *  - pour lier/dlier deux brins entre eux, sans vrification
   *    des contraintes d'intgrit.
   *
   * Les mthodes permettant d'explorer la carte et de grer les plongements
   * sont dfinies dans la classe CGMap.
   *
   * @author MOKA TEAM
   */
  class CGMapBasic
  {
  public:
    // @name Constructeur et destructeur
    // @{

    /// Constructeur
    CGMapBasic();

    /// Destructeur
    virtual ~CGMapBasic();

    // @}
    // @name Rinitialisation de la carte
    // @{

    /**
     * Vide la carte.
     * Tous les brins sont dtruits.
     */
    void removeAllDarts();

    /**
     * Rinitialise la carte.
     * La mthode removeAllDarts est appele, puis toutes les marques, champs
     * direct-info, etc, sont rinitialiss.
     */
    virtual void empty();

    // @}
    // @name Gestion des champs direct-info
    // @{

    /**
     * Rserve un champ direct info libre.
     *
     * @return L'indice du champ direct-info rserv.
     */
    int getNewDirectInfo();

    /**
     * Libre un champ direct-info pralablement rserv avec la
     * mthode getNewDirectInfo.
     *
     * @param ADirectIndex Un indice de champ direct-info.
     */
    void freeDirectInfo(int ADirectIndex);

    /**
     * Retourne la valeur du champ direct-info[ADirectIndex] du brin ADart.
     *
     * @param ADart Un brin de la carte
     * @param ADirectIndex Un indice de champ direct-info
     * @return Un pointeur
     */
    void* getDirectInfo(CDart* ADart, int ADirectIndex) const;

    /**
     * Affecte la valeur APointer au champ direct-info[ADirectIndex] du brin
     * ADart.
     *
     * @param ADart Un brin de la carte
     * @param ADirectIndex Un indice de champ direct-info
     * @param APointer La valeur  affecter
     */
    void setDirectInfo(CDart* ADart, int ADirectIndex, void* APointer);

    // @}
    // @name Accs au premier brin de liste des brins de la carte.
    // @{

    /**
     * Retourne le brin situ en tte de la liste des brins de la carte.
     *
     *  @return Le premier brin de la carte
     */
    CDart* getFirstDart() const;

    /**
     * Dfinit le premier brin de la carte.
     *
     * @param ADart Un brin de la carte.
     */
    void setFirstDart(CDart* ADart);

    // @}
    // @name Gestion des marques
    // @{

    /**
     * Rserve une marque inutilise de la carte.
     *
     * @return Un numro de marque
     */
    int getNewMark();

    /**
     * Retourne l'tat de la marque.
     * Initialement, cette valeur vaut false. Elle est modifie  chaque appel
     *  la mthode negateMaskMark().
     *
     * @param AMarkNumber Un numro de marque
     * @return Un boolen indiquant l'tat de la marque
     */
    bool getMaskMark(int AMarkNumber) const;

    /**
     * Libre la marque AMarkNumber pralablement rserve avec la mthode
     * getNewMark().
     *
     * @param AMarkNumber
     */
    void freeMark(int AMarkNumber);

    /**
     * Inverse l'tat d'une marque. Cette mthode permet de changer l'tat de
     * la marque AMarkNumber de tous les brins de la carte en O(1).
     * C'est utile par exemple pour dmarquer tous les brins  la fin d'un
     * traitement.
     *
     * @param AMarkNumber Un numro de marque
     */
    void negateMaskMark(int AMarkNumber);

    /**
     * Cette mthode est utilise (en assertion) par la mthode freeMark.
     * En effet, lorqu'on libre une marque, il faut ncessairement que tous
     * les brins de la carte soient dmarqus.
     *
     * @param AMarkNumber Un numro de marque
     * @return Un boolen indiquant si tous les brins de la carte sont dmarqus
     */
    bool isWholeMapUnmarked(int AMarkNumber);

    /**
     * Cette mthode permet de savoir  tout moment combien de marques sont
     * utilises simultanment.
     * Sert au debugging.
     *
     * @return Le nombre de marques utilises
     */
    int getNbUsedMarks() const;

    /**
     * Teste si la marque AMarkNumber du brin ADart est active.
     * Cette mthode est redirige sur la mthode quivalente de CDart.
     *
     * @param ADart Un brin de la carte
     * @param AMarkNumber Un numro de marque
     * @return Un boolen indiquant l'tat de la marque AMarkNumber du brin
     *   ADart.
     */
    bool isMarked(const CDart* ADart, int AMarkNumber) const;

    /**
     * Permet de modifier l'tat de la marque AMarkNumber du brin ADart.
     * Cette mthode est redirige sur la mthode quivalente de CDart.
     *
     * @param ADart Un brin de la carte
     * @param AMarkNumber Un numro de marque
     * @param AState Un boolen indiquant s'il faut activer ou dsactiver la
     *   marque
     */
    void setMarkTo(CDart* ADart, int AMarkNumber, bool AState);

    /**
     * Permet d'activer la marque AMarkNumber du brin ADart.
     * Cette mthode quivaut  setMarkTo(ADart, AMarkNumber, true).
     *
     * @param ADart Un brin de la carte
     * @param AMarkNumber Un numro de marque
     */
    void setMark(CDart* ADart, int AMarkNumber);

    /**
     * Permet de dsactiver la marque AMarkNumber du brin ADart.
     * Cette mthode quivaut  setMarkTo(ADart, AMarkNumber, false).
     *
     * @param ADart Un brin de la carte
     * @param AMarkNumber Un numro de marque
     */
    void unsetMark(CDart* ADart, int AMarkNumber);

    /**
     * Cette mthode permet d'initialiser toutes les marques du brin ADart.
     * AMarks est modifi pour prendre en comte le masque des marques.
     *
     * @param ADart Un brin de la carte
     * @param AMarks un tableau indiquant l'tat de toutes les marques
     */
    void setMarks(CDart* ADart, const std::bitset<NB_MARKS> & AMarks) const;

    /**
     * Retourne l'tat des marques du brin, en tenant compte du masque des
     * marques.
     *
     * @param ADart Un brin de la carte
     * @return L'tat des marques du brin
     */
    std::bitset<NB_MARKS> getMarks(CDart* ADart) const;

    // @}
    // @name Gestion des orbites utilises pour les plongements
    // @{

    /**
     * Permet de savoir si une orbite est utilise.
     * Cette mthode est utilise par la classe CGMap pour la gestion des
     * plongements.
     *
     * @param AOrbit Une des 16 orbites classiques (pas les orbites de bord)
     * @return Un boolen indiquant si l'orbite est utilise
     */
    bool isOrbitUsed(TOrbit AOrbit) const;

    /**
     * Permet d'activer l'utilisation d'une orbite.
     * Lorsqu'une orbite est active, les plongements qui lui correspondent
     * sont mis  jour lors des oprations de couture et de dcousure (celles
     * avec gestion des plongements  de la classe CGMap).
     *
     * @param AOrbit Une orbite classique (pas une orbite de bord)
     */
    void setOrbitUsed(TOrbit AOrbit);

    /**
     * Permet de dsactiver l'utilistion d'une orbite.
     * Lorsqu'une orbite est dsactive, auncun plongement qui lui correspond
     * n'est mis  jour par les oprations de couture et de dcousure de la
     * classe CGMap.
     * Initialement, toutes les orbites sont dsactives.
     *
     * @param AOrbit Une orbite classique (pas une orbite de bord)
     */
    void unsetOrbitUsed(TOrbit AOrbit);

    /**
     * Permet de savoir si un brin contient un plongement sans avoir 
     * parcourir la liste de ses attributs.
     *
     * @param ADart Un brin de la carte
     * @param AOrbit Une des 16 orbites classiques
     * @return Un boolen indiquant si l'orbite AOrbit du brin ADart contient
     *   un plongement
     */
    bool isOrbitUsed(CDart* ADart, TOrbit AOrbit) const;

    // @}
    // @name Cration et suppression de brins
    // @{

    /**
     * Permet de crer un nouveau brin.
     * Cette mthode doit tre surcharge dans les classes qui drivent de
     * CGMapBasic (car un constructeur ne peut tre virtuel).
     * N'utilisez pas cette mthode mais la mthode addMapDart() pour insrer
     * un nouveau brin dans la carte.
     *
     * @return Un nouveau brin de la carte
     */
    virtual CDart* newDart();

    /**
     * Permet de dtruire un brin.
     * Cette mthode doit tre surcharge dans les classes qui drivent de
     * CGMapBasic.
     * N'utilisez pas cette mthode mais la mthode delMapDart() pour supprimer
     * un brin de la carte.
     *
     * @param ADart Un brin de la carte
     */
    virtual void delDart(CDart* ADart);

    /**
     * Permet de crer un nouveau brin (avec la mthode addDart()) et de
     * l'insrer dans la carte.
     *
     * @return Le brin insr
     */
    CDart* addMapDart();

    /**
     * Ote un brin de la liste des brins de la carte, puis le supprime avec la
     * mthode delDart().
     *
     * @param ADart Le brin  supprimer
     */
    void delMapDart(CDart* ADart);

    // @}

    /**
     * @name Accs aux involutions
     *
     * Attention, les mthodes d'accs en critures qui suivent ne sont pas
     * sres car elles ne prservent pas forcment les contraintes des
     * cartes. Elles ne mettent pas non plus  jour les ventuels plongements.
     */

    /**
     * Teste si le brin ADart est libre par alpha-i, c'est--dire s'il est sa
     * propre image par alpha-i.
     *
     * @param ADart Un brin de la carte
     */
    // @{
    bool isFree0(const CDart* ADart) const;
    bool isFree1(const CDart* ADart) const;
    bool isFree2(const CDart* ADart) const;
    bool isFree3(const CDart* ADart) const;
    // @}

    /**
     * Teste si le brin ADart est libre en dimension ADimension, c'est--dire
     * s'il est sa propre image par alpha-ADimension.
     *
     * @param ADart Un brin de la carte
     * @param ADimension Une dimension comprise entre 0 et 3 inclus
     */
    bool isFree(const CDart* ADart, int ADimension) const;

    /**
     * @param ADart Un brin de la carte
     * @return L'image par alpha-i de ADart
     */
    // @{
    CDart* alpha0(const CDart* ADart) const;
    CDart* alpha1(const CDart* ADart) const;
    CDart* alpha2(const CDart* ADart) const;
    CDart* alpha3(const CDart* ADart) const;
    // @}

    /**
     * @param ADart Un brin de la carte
     * @param ADimension Une dimension comprise entre 0 et 3 inclus
     * @return l'image par alpha-ADimension de ADart.
     */
    CDart* alpha(const CDart* ADart, int ADimension) const;

    /**
     * Lie par alpha-[i] les brins ADart1 et ADart2.
     *
     * @param ADart1 Un brin de la carte
     * @param ADart2 Un brin de la carte
     */
    // @{
    void linkAlpha0(CDart* ADart1, CDart* ADart2);
    void linkAlpha1(CDart* ADart1, CDart* ADart2);
    void linkAlpha2(CDart* ADart1, CDart* ADart2);
    void linkAlpha3(CDart* ADart1, CDart* ADart2);
    // @}

    /**
     * Lie par alpha-ADimension les brins ADart1 et ADart2.
     *
     * @param ADart1 Un brin de la carte
     * @param ADart2 Un brin de la carte
     * @param ADimension Une dimension entre 0 et 3 inclus.
     */
    void linkAlpha(CDart* ADart1, CDart* ADart2, int ADimension);

    /**
     * Dlie par alpha-[i] les brins ADart et alpha-[i](ADart).
     *
     * @param ADart Un brin de la carte
     */
    // @{
    void unlinkAlpha0(CDart* ADart);
    void unlinkAlpha1(CDart* ADart);
    void unlinkAlpha2(CDart* ADart);
    void unlinkAlpha3(CDart* ADart);
    // @}

    /**
     * Dlie par alpha-ADimension les brins ADart et alpha-ADimension(ADart2).
     *
     * @param ADart1 Un brin de la carte
     * @param ADart2 Un brin de la carte
     * @param ADimension Une dimension entre 0 et 3 inclus.
     */
    void unlinkAlpha(CDart* ADart, int ADimension);

  protected:
    // @name Champs protgs :
    // @{

    /// Premier brin de la liste des brins de la carte
    CDart* FFirstDart;

    /// Tableau indiquant quelles sont les marques prises
    std::bitset<NB_MARKS> FUsedMarks;

    /// Masque des marques (modifi par la mthode negateMaskMark())
    std::bitset<NB_MARKS> FMaskMarks;

    /// Liste des marques libres
    int FMarks[NB_MARKS];

    /// Tableau indiquant quelles sont les orbites utilises
    std::bitset<16> FUsedOrbits;

    /// Tableau indiquant quels sont les champs direct-info utiliss
    std::bitset<NB_DIRECT_INFO> FUsedDirectInfo;

    /// Liste des champs directInfo libres
    int FDirectInfos[NB_DIRECT_INFO];

    // @}

  private:
    // @name Champs privs :
    // @{

    /// Nombre de marques utilises
    int FNbUsedMarks;

    /// Nombre de champs directInfo utiliss
    int FNbUsedDirectInfos;

#ifndef NDEBUG
    /// Nombre maximum de marques utilises simultanment
    int FMaxNbUsedMarks;

    /// Nombre maximum de champs directInfo utiliss simultanment
    int FMaxNbUsedDirectInfos;
#endif // NDEBUG

    // @}
  };

} // namespace GMap3d
//******************************************************************************
#include INCLUDE_INLINE("g-map-basic.icc")
//******************************************************************************
#endif // G_MAP_BASIC_HH
//******************************************************************************
