/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef GEOMETRY_HH
#define GEOMETRY_HH
//******************************************************************************
#include "inline-macro.hh"
#include "math-extension.hh"
#include "vertex.hh"
//******************************************************************************

/**
 * Cette classe fournit des fonctions permettant des calculs
 * gomtriques communment utiliss.
 *
 * Conventions:
 *
 * alpha.b signifie  produit scalaire  entre alpha et b
 *     rsultat: ax*bx + ay*by + az*bz
 *
 * alpha*b signifie  produit vectoriel entre alpha et b
 *     rsultat: { ay*bz-az*by, az*bx-ax*bz, ax*by-ay*bx }
 *
 * @author MOKA TEAM
 */
class CGeometry
{
public:
  // @name Projections
  // @{

  /**
   * Calcule le projet orthogonal de AVertex sur la droite de vecteur
   * directeur ALineDirection passant par ALineVertex.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param ALineVertex Un sommet quelconque de la droite
   * @param ALineDirection Un vecteur directeur de la droite
   * @return Le projet de AVertex sur la droite donne
   *
   * @precondition ! ADirection.isNull()
   */
  static CVertex orthoProjectVertexOnLine(const CVertex& AVertex,
					  const CVertex& ALineVertex,
					  const CVertex& ALineDirection);

  /**
   * Calcule le projet orthogonal de AVertex sur le plan d'quation
   * cartsienne AA*x + AB*y + AC*z + AD = 0.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param A? Les paramtres de l'quation cartsienne du plan
   * @return Le projet de AVertex sur le plan donn
   */
  static CVertex
  orthoProjectVertexOnPlane(const CVertex& AVertex,
			    const TCoordinate& AA, const TCoordinate& AB,
			    const TCoordinate& AC, const TCoordinate& AD);

  /**
   * Calcule le projet orthogonal de AVertex sur le plan de vecteur normal
   * APlaneNormal passant par APlaneVertex.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param APlaneVertex Un sommet quelconque du plan
   * @param APlaneNormal Un vecteur normal au plan
   * @return Le projet de AVertex sur le plan donn
   *
   * @precondition !ANormal.isNull()
   */
  static CVertex orthoProjectVertexOnPlane(const CVertex& AVertex,
					   const CVertex& APlaneVertex,
					   const CVertex& APlaneNormal);

  // @}
  // @name Homothties
  // @{

  /**
   * Calcule l'image de AVertex par l'homothtie de centre ACenter et de
   * coefficient ACoef.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param ACenter Le centre d'homothtie
   * @param ACoef Le coefficient d'homothtie
   * @return L'image de AVertex par l'homothtie donne
   */
  static CVertex centralHomothety(const CVertex& AVertex,
				  const CVertex& ACenter,
				  const TCoordinate& ACoef);

  /**
   * Calcule l'image de AVertex par l'homothtie de centre ACenter et de
   * coefficient ACoef.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param ACenter Le centre d'homothtie
   * @param ACoef Les coefficients d'homothtie, selon les 3 dimensions
   * @return L'image de AVertex par l'homothtie donne
   */
  static CVertex centralHomothety(const CVertex& AVertex,
				  const CVertex& ACenter,
				  const CVertex& ACoef);

  /**
   * Calcule l'image de AVertex par l'homothtie axiale de coefficient ACoef et
   * d'axe la droite passant par ALineVertex et de vecteur directeur
   * ALineDirection.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param ALineVertex Un sommet de l'axe d'homothtie
   * @param ALineDirection Un vecteur directeur de l'axe d'homothtie
   * @param ACoef Le coefficient d'homothtie
   * @return L'image de AVertex par l'homothtie donne
   *
   * @precondition ! ALineDirection.isNull()
   */
  static CVertex axialHomothety(const CVertex& AVertex,
				const CVertex& ALineVertex,
				const CVertex& ALineDirection,
				const TCoordinate& ACoef);

  /**
   * Calcule l'image de AVertex par l'homothtie planaire de coefficient ACoef
   * selon le plan passant par APlaneVertex et ayant APlaneNormal pour vecteur
   * normal.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param APlaneVertex Un sommet du plan d'homothtie
   * @param APlaneNormal Un vecteur normal du plan d'homothtie
   * @param ACoef Le coefficient d'homothtie
   * @return L'image de AVertex par l'homothtie donne
   *
   * @precondition ! APlaneNormal.isNull()
   */
  static CVertex planarHomothety(const CVertex& AVertex,
				 const CVertex& APlaneVertex,
				 const CVertex& APlaneNormal,
				 const TCoordinate& ACoef);

  // @}
  // @name Calcul de la distance d'un sommet  un autre objet
  // @{

  /**
   * Calcule la distance euclidienne du sommet AVertex au sommet AOtherVertex.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param AOtherVertex Un sommet quelconque de l'espace
   * @return La distance entre les deux sommets
   */
  static TCoordinate distanceToVertex(const CVertex& AVertex,
				      const CVertex& AOtherVertex);

  /**
   * Calcule la distance euclidienne du sommet AVertex  la droite de vecteur
   * directeur ALineDirection passant par ALineVertex, c'est--dire la distance
   * de AVertex  son projet orthogonal sur cette droite.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param ALineVertex Un sommet quelconque de la droite
   * @param ALineDirection Un vecteur directeur de la droite
   * @return La distance du sommet  la droite
   *
   * @precondition ! ALineDirection.isNull()
   */
  static TCoordinate distanceToLine(const CVertex& AVertex,
				    const CVertex& ALineVertex,
				    const CVertex& ALineDirection);

  /**
   * Calcule la distance euclidienne du sommet AVertex au plan de vecteur
   * normal APlaneNormal passant par APlaneVertex, c'est--dire la distance de
   * AVertex  son projet orthogonal sur ce plan.
   *
   * @param AVertex Un sommet quelconque de l'espace
   * @param APlaneVertex Un sommet quelconque du plan
   * @param APlaneNormal Un vecteur normal du plan
   * @param Ka distance du sommet au plan
   *
   * @precondition ! APlaneNormal.isNull()
   */
  static TCoordinate distanceToPlane(const CVertex& AVertex,
				     const CVertex& APlaneVertex,
				     const CVertex& APlaneNormal);

  // @}
  // @name Fonctions gomtriques diverses
  // @{

  /**
   * Calcule la bissectrice des vecteurs AVector1 et AVector2.
   * Le sens de la bissectrice est dtermin par le signe de la valeur
   * (AVector1 * AVector2).ANormal.
   *
   * @param AVector1 Un premier vecteur
   * @param AVector2 Un deuxime vecteur
   * @param ANormal  Un vecteur normal aux deux premiers (paramtre facultatif)
   *   permettant de dterminer le signe de l'angle.
   * @return La bissectrice des vecteurs AVector1 et AVector2
   */
  static CVertex getBissectrix(const CVertex& AVector1,
			       const CVertex& AVector2,
			       const CVertex& ANormal = ORIGIN);

  /**
   * Calcule un vecteur normal  AVector, si possible se trouvant dans le plan
   * XY.
   *
   * @param AVector Un vecteur non nul
   * @return Un vecteur normal  AVector et de mme norme que celui-ci
   */
  static CVertex getNormalVector(const CVertex& AVector);

  /**
   *  Calcule un vecteur normal  AVector1 et AVector2 (en effectuant le produit
   *  vectoriel de AVector1 et AVector2).
   *
   *  L'intrt de cette mthode est que si aucun des vecteurs n'est le vecteur
   *  nul et si les vecteurs ne sont pas colinaires, la norme du vecteur
   *  retourn est gale  la moyenne des normes des vecteurs AVector1 et
   *  AVector2.
   *
   *  @param AVector1 Un vecteur quelconque
   *  @param AVector2 Un vecteur quelconque
   *  @return Un vecteur normal  AVector1 et AVector2
   */
  static CVertex getNormalVector(const CVertex& AVector1,
				 const CVertex& AVector2);

  /**
   * Calcule l'intersection entre deux droites.
   * Chaque droite est donne par deux sommets.
   *
   * Le rsultat est imprvisible si les deux droites ne se croisent pas.
   *
   * @param ALine1Vertex1 Un premier sommet de la premire droite
   * @param ALine1Vertex2 Un deuxime sommet de la premire droite
   * @param ALine2Vertex1 Un premier sommet de la deuxime droite
   * @param ALine2Vertex2 Un deuximme sommet de la deuxime droite
   *
   * @precondition ALine1Vertex1 != ALine1Vertex2
   * @precondition ALine2Vertex1 != ALine2Vertex2
   */
  static CVertex getLinesIntersection(const CVertex& ALine1Vertex1,
				      const CVertex& ALine1Vertex2,
				      const CVertex& ALine2Vertex1,
				      const CVertex& ALine2Vertex2);

  /**
   * Teste si deux vecteurs sont colinaires.
   *
   * @param AVector1 Un vecteur quelconque
   * @param AVector2 Un vecteur quelconque
   * @return Un boolen indiquant si AVector1 et AVector2 sont colinaires
   */
  static bool areColinear(const CVertex& AVertex1, const CVertex& AVertex2);

  /**
   * Calcule l'angle entre les vecteurs AVector1 et AVector2.
   * Le signe de l'angle est donn par le signe de la valeur de
   * (AVector1*AVector2).ANormal.
   *
   * @param AVector1 Un premier vecteur
   * @param AVector2 Un deuxime vecteur
   * @param ANormal  Un vecteur normal aux deux premiers, permettant de
   *                 dterminer le signe de l'angle
   * @return Un angle
   */
  static TCoordinate getAngle(const CVertex& AVector1,
			      const CVertex& AVector2,
			      const CVertex& ANormal = ORIGIN);

  /**
   * Calcule les 3 angles de rotation permettant d'obtenir le repre
   * (AVx,AVy,AVz) en trois rotations  partir du repre (OX,OY,OZ). AVy vaut
   * AVz*AVx.  Le rsultat est un triplet (alpha,beta,gamma) tel que si l'on
   * applique la mthode CTransformationMatrix::rotate(AAngles) sur le repre
   * (OX,OY,OZ), on obtient le repre (AVx,AVy,AVz)  la norme des vecteurs
   * prs.
   * Si les vecteurs AVx et AVz ne sont pas orthogonaux, AVx est projet sur
   * le plan orthogonal  AVz et passant par l'origine.
   *
   * @param AVx Un premier vecteur
   * @param AVz Un deuxime vecteur
   * @return Trois angles de rotation
   */
  static CVertex getAngles(const CVertex& AVx, const CVertex& AVz);

  // @}
  // @name Mthodes auxiliaires pour le coraffinement 2D
  // @{

  /**
   * Localise le sommet AVertex par rapport  la droite donne.
   * La valeur retourne vaut:
   *  -1 si le sommet est  gauche de la droite,
   *   0 si le sommet est sur la droite,
   *  +1 si le sommet est  droite de la droite.
   *
   * Remarque: Les calculs tiennent compte des erreurs d'arrondi.
   *
   * @param AVertex Le sommet  localiser
   * @param ALineVertex Un sommet de la droite par rapport  laquelle localiser
   *   le sommet AVertex
   * @param ALineDirection Un vecteur directeur de la droite par rapport 
   *   laquelle localiser le sommet AVertex
   * @param APlaneNormal Un vecteur normal au plan dans lequel on travaille
   * @return Un entier
   *
   * @precondition !ALineDirection.isNull()
   * @precondition !APlaneNormal.isNull()
   */
  static int localiseVertexComparedToLine(const CVertex& AVertex,
					  const CVertex& ALineVertex,
					  const CVertex& ALineDirection,
					  const CVertex& APlaneNormal);

  /**
   * Teste si les segments AB et CD s'intersectent.
   * Les rsultats sont stocks dans les paramtres AIntersection-On.
   * AIntersection-On peut prendre la valeur:
   *   0 : pas d'intersection,
   *   1 : intersection franche sur le segment AB
   *   2 : intersection franche sur le segment CD.
   *
   * Si un paramtre AIntersection-On vaut vrai, alors la mthode affecte les
   * coordonnes du point d'intersection  AIntersection-, sinon
   * AIntersection- n'est pas modifi.
   *
   * Remarque : s'il n'y alpha qu'une seule intersection, celle-ci n'est pas
   * forcment place dans AIntersection1.
   *
   * La mthode fonctionne dans toutes les configurations (artes disjointes,
   * scantes, se chevauchant, confondues, etc...)
   *
   * @param AVertexA Un sommet quelconque
   * @param AVertexB Un sommet quelconque
   * @param AVertexC Un sommet quelconque
   * @param AVertexD Un sommet quelconque
   * @param APlaneNormal Un vecteur normal au plan des segments
   * @param ACutABSegment Un boolen indiquant si l'arte AB est coupe par
   *   l'arte CD de manire franche
   * @param AIntersectionOnAB Une rfrence sur un sommet o sont stockes les
   *   coordonnes du point par lequel CD coupe AB
   * @param ACutCDSegment Un boolen indiquant si l'arte CD est coupe par
   *   l'arte AB de manire franche
   * @param AIntersectionOnCD Une rfrence sur un sommet o sont stockes les
   *   coordonnes du point par lequel AB coupe CD
   *
   * @precondition AVertexA != AVertexB
   * @precondition AVertexC != AVertexD
   * @precondition !APlaneNormal.isNull()
   */
  static void getSegmentsIntersection(const CVertex& AVertexA,
				      const CVertex& AVertexB,
				      const CVertex& AVertexC,
				      const CVertex& AVertexD,
				      const CVertex& APlaneNormal,
				      int& AIntersection1On,
				      CVertex& AIntersection1,
				      int& AIntersection2On,
				      CVertex& AIntersection2);

  /**
   *  Mthode quivalente  getSegmentsIntersection, mais qui travaille dans
   *  le plan.
   */
  static void getSegmentsIntersection2d(const CVertex& AVertexA,
					const CVertex& AVertexB,
					const CVertex& AVertexC,
					const CVertex& AVertexD,
					int& AIntersection1On,
					CVertex& AIntersection1,
					int& AIntersection2On,
					CVertex& AIntersection2);

  // @}
  // @name Mthodes pour le chanfreinage :
  // @{

  /**
   * @param AVertex_ trois sommets A,B,C
   * @return le projet orthogonal du vecteur AC sur le vecteur normal
   *   au vecteur AB dans le plan ABC.
   */
  static CVertex getRoundingNormalVector(const CVertex& AVertex1, 
					 const CVertex& AVertex2,
					 const CVertex& AVertex3);

  // @}
};
//******************************************************************************
#include INCLUDE_INLINE("geometry.icc")
//******************************************************************************
#endif // GEOMETRY_HH
//******************************************************************************
