(*
  Minimum Degree.
  
  Based on the article:
  The Minimum Degree Heuristic and the Minimal
  Triangulation Process
  by A. Berry, Pinar Heggernes & Genevive Simonet.
  
  @author Matthieu Sozeau
  @author Pierre-Loic Garoche
  
  $Id: md.ml,v 1.1 2004/06/28 07:32:55 filliatr Exp $
*)

module MD =
struct
  
  module P(G : Sig.P) =
  struct
    module VertexSet = Set.Make(G.V)
    module CT = Cliquetree.CliqueTree(G)
      
    module Choose = Oper.Choose(G)
    module NG = Neighborhood.Graph(G)

    type edgeset = (G.V.t * G.V.t) list
      
    let md g =
      let gref = ref g in
      let gtri = ref g in
      let n = G.nb_vertex g in
      let tri = ref [] in
      let ord = ref [] in
      let i = ref 0 in
	while not (CT.is_chordal !gtri) && !i < n do
	  let v = 
	    let x = 
	      G.fold_vertex 
		(fun v' x ->
		   let deg' = G.out_degree !gref v' in
		     match x with
			 Some (v,deg) when deg' > deg -> x
		       | _ -> Some (v', deg'))
		!gref None
	    in match x with 
		Some (v,_) -> v
	      | None -> failwith "Expecting some vertex"
	  in
	  let ng = NG.of_vertex !gref v in
	  let g', tri' =
	    VertexSet.fold 
	      (fun v (g, tri) ->
		 let tri' =
		   VertexSet.fold 
		     (fun v' tri ->
			if v <> v' && not (G.mem_edge g v v') then
			  (v, v') :: tri
			else tri)
		     ng tri
		 in
		 let g' = 
		   VertexSet.fold
		     (fun v' g ->
			if v <> v' then
			  G.add_edge g v v'
			else g)
		     ng g
		 in (g', tri'))
	      ng (!gref, [])
	  in
	    ord := v :: !ord;
	    gtri := List.fold_left 
	      (fun g (x,y) -> G.add_edge g x y) 
		      !gtri tri';
	    gref := G.remove_vertex g' v;
	    tri := tri' @ !tri;
	    incr i;
	done;
	(!gtri, !tri, !ord)
	  
    let triangulate g = 
      let gtri, _, _ = md g in 
	gtri
  end

  module I(G : Sig.I) =
  struct
    module VertexSet = Set.Make(G.V)   
    module CT = Cliquetree.CliqueTree(G)
      
    module Choose = Oper.Choose(G)
    module NG = Neighborhood.Graph(G)
            
    type edgeset = (G.V.t * G.V.t) list
      
    module Copy = Copy.Copy.I(G)(G)
      (struct 
	 let vertex g v = 
	   G.add_vertex g v;
	   v
	       
	 let edge g x y =
	   G.add_edge g x y	     
       end)


    let md g =
      let gtri = Copy.copy g in
      let gcur = Copy.copy g in
      let n = G.nb_vertex g in
      let tri = ref [] in
      let ord = ref [] in
      let i = ref 0 in
	while not (CT.is_chordal gtri) && !i < n do
	  let v = 
	    let x = 
	      G.fold_vertex 
		(fun v' x ->
		   let deg' = G.out_degree gcur v' in
		     match x with
			 Some (v,deg) when deg' > deg -> x
		       | _ -> Some (v', deg'))
		gcur None
	    in match x with 
		Some (v,_) -> v
	      | None -> failwith "Expecting some vertex"
	  in
	  let ng = NG.of_vertex gcur v in
	  let tri' =
	    VertexSet.fold 
	      (fun v tri ->
		 VertexSet.fold 
		 (fun v' tri ->
		    let tri' =
		      if v <> v' && not (G.mem_edge g v v') then
			(v, v') :: tri
		      else tri
		    in
		      VertexSet.iter
			(fun v' ->
			   if v <> v' then
			     G.add_edge gcur v v')
			ng;
		      tri')
		 ng tri)
	      ng []
	  in
	    ord := v :: !ord;
	    List.iter
	      (fun (x,y) -> G.add_edge gtri x y) 
	      tri';
	    G.remove_vertex gcur v;
	    tri := tri' @ !tri;
	    incr i;
	done;
	(gtri, !tri, !ord)
	  
    let triangulate g = 
      let gtri, _, _ = md g in 
	gtri
  end
    
    
end
