/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Cre un maillage 1d compos de ASx artes.
 * Si ASx est nul, un seul sommet est cr.
 *
 * AMesh1Corners[0] est positionn sur le brin 1-libre situ  la premire
 * extrmit de la polyligne cre.
 * AMesh1Corners[1] est positionn sur le brin 1-libre situ  la deuxime
 * extrmit de la polyligne cre.
 *
 * Si AModel est NULL, le maillage cr est une polyligne, sinon c'est un
 * faisceau calqu sur l'orbite 23 de AModel.
 *
 * @param ASx Le nombre d'artes de la polyligne
 * @param AMesh1Corners Un tableau de 2 pointeurs sur brin
 * @param AModel Un brin de la carte
 */
void createTopoMesh1(int ASx,
		     CDart* AMesh1Corners[2],
		     CDart* AModel = NULL);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoMesh1' ci-dessus et
 * retourne le brin correspondant  AMesh1Corners[0].
 *
 * @param ASx Le nombre d'artes de la polyligne
 * @param AModel Un brin de la carte
 * @return Un brin 1-libre de la polyligne
 */
CDart* createTopoMesh1(int ASx, CDart* AModel = NULL);

/**
 * Cre une grille de dimension ASx*ASy.
 * Les lments de AFaceCorner sont positionns sur des brins stratgiques
 * de la grille, comme indiqu sur le schma.
 *
 * ASx=3, ASy=2
 *
 * AMesh2Corners[0][1]                 AMesh2Corners[1][1]
 *       	     +== --+-- --+-- ==+
 *        	     |     |     |     |
 *        	     |     |     |     |
 *        	     +-- --+-- --+-- --+
 *        	     |     |     |     |
 *        	     |     |     |     |
 *        	     +== --+-- --+-- ==+
 * AMesh2Corners[0][0]                 AMesh2Corners[1][0]
 *
 * Si A3Sewed est vrai, le maillage est doubl, c'est--dire que 2 grilles
 * identiques sont cres et chaque brin qui constitue la premire est
 * 3-cousu  son homologue sur la deuxime.
 *
 * @param ASx Le nombre (non nul) de subdivisions de la grille selon la
 *   premire dimension
 * @param ASy Le nombre (non nul) de subdivisions de la grille selon la
 *   deuxime dimension
 * @param AMesh2Corners Un tableau de taille 2*2 de pointeurs sur brin
 *   correspondant aux coins du maillage 2d
 * @param A3Sewed Un boolen indiquant si le maillage est doubl
 */
void createTopoMesh2(int ASx, int ASy,
		     CDart* AMesh2Corners[2][2], bool A3Sewed = false);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoMesh2' ci-dessus et
 * retourne le brin correspondant  AFaceCorners[0][0].
 *
 * @param ASx Le nombre (non nul) de subdivisions de la grille selon la
 *   premire dimension
 * @param ASy Le nombre (non nul) de subdivisions de la grille selon la
 *   deuxime dimension
 * @param A3Sewed Un boolen indiquant si le maillage est doubl
 * @return Un brin de la grille
 */
CDart* createTopoMesh2(int ASx, int ASy, bool A3Sewed = false);

/**
 * Cre un maillage 3d de dimension ASx*ASy*ASz.
 * Les lments de AMesh3Corners sont positionns sur des brins
 * stratgiques du maillage, comme indiqu sur le schma.
 *
 * ASx=3, ASy=2, ASz=1
 *
 *            AMesh3Corners[0][1][1]             AMesh3Corners[1][1][1]
 *                               +== --+-- --+-- ==+
 *                              /     /     /     /+
 *                             /     /     /     //|
 *                            +-- --+-- --+-- --+/ |
 *                           /     /     /     /+  +
 *                          /     /     /     //| /+ AMesh3Corners[1][1][0]
 *  AMesh3Corners[0][0][1] +== --+-- --+-- ==+/ AMesh3Corners[1][0][1]
 *                         +-- --+-- --+-- --+  +/
 *                         |     |     |     | /+
 *                         |     |     |     |//
 *                         +-- --+-- --+-- --+/
 *                         +== --+-- --+-- ==+
 *      AMesh3Corners[0][0][0]             AMesh3Corners[1][0][0]
 *
 *
 * Remarque: AMesh3Corners[0][1][0] n'est pas visible.
 *
 * @param ASx Le nombre (non nul) de subdivisions du maillage 3d selon la
 *   premire dimension
 * @param ASy Le nombre (non nul) de subdivisions du maillage 3d selon la
 *   deuxime dimension
 * @param ASz Le nombre (non nul) de subdivisions du maillage 3d selon la
 *   troisime dimension
 * @param AMesh3Corners Un tableau de taille 2*2*2 de pointeurs sur brin
 *   correspondant aux coins du maillage 3d
 */
void createTopoMesh3(int ASx, int ASy, int ASz,
		     CDart* AMesh3Corners[2][2][2]);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoMesh3' ci-dessus et
 * retourne le brin correspondant  AMesh3Corners[0][0][0].
 *
 * @param ASx Le nombre (non nul) de subdivisions du maillage 3d selon la
 *   premire dimension
 * @param ASy Le nombre (non nul) de subdivisions du maillage 3d selon la
 *   deuxime dimension
 * @param ASz Le nombre (non nul) de subdivisions du maillage 3d selon la
 *   troisime dimension
 * @return Un brin du maillage 3d
 */
CDart* createTopoMesh3(int ASx, int ASy, int ASz);

//******************************************************************************
