/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Utilise la mthode createMesh1 avec un CMesh1VectorDiver.
 *
 * @param AVector Un vecteur contenant les sommets de la polyline
 * @return Un brin extrmit de la polyline
 */
CDart* createPolyline(const std::vector<CVertex> & AVector);

/**
 * Appelle la mthode CGMapGeneric::createTopoPole puis plonge le sommet
 * central du ple cr avec l'attribut AVertex.
 *
 * @param An Le nombre de triangles qui composent le ple
 * @param AVertex Le plongement du sommet central du ple
 * @return Un brin du sommet central du ple
 */
CDart* createPole(int An, const CVertex & AVertex);

/**
 * Appelle la mthode CGMapGeneric::createTopoOpenedPole puis plonge le
 * sommet central du ple cr avec l'attribut AVertex.
 *
 * @param An Le nombre de triangles qui composent le ple
 * @param AVertex Le plongement du sommet central du ple
 * @return Un brin 2-libre du sommet central du ple
 */
CDart* createOpenedPole(int An, const CVertex & AVertex);

/**
 * Cre un polygone rgulier centr sur l'origine et de vecteur normal OZ.
 * Le premier sommet de la face cre est aux coordonnes (1/2,0,0),
 * c'est--dire que le polygone est inscrit dans un cercle de diamtre 1.
 *
 * Le brin b retourn est celui du premier sommet de la face tel que le
 * plongement de alpha0(b) ait une ordonne positive.
 *
 * @param An Le nombre d'artes du polygone
 * @return Un brin de la face cre
 *
 * @precondition An>=3
 */
CDart* createRegularPolygon(int An);

/**
 * Appelle la mthode CGMapGeneric::createTopoSphere puis plonge sphre cre.
 * La sphre obtenue est centre sur l'origine, son diamtre vaut 1.
 *
 * Le ple sud  est aux coordonnes (0,0,-1/2).
 * Le ple nord est aux coordonnes (0,0,+1/2).
 *
 * Les brins ASouthPole et ANorthPole en sortie sont situs sur le mridien
 * passant par le sommet (+1/2,0,0).
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la sphre
 * @param AParallels Le nombre (non nul) de parallles de la sphre
 * @param ASouthPole L'adresse d'un pointeur sur brin
 * @param ANorthPole L'adresse d'un pointeur sur brin
 */
void createSphere(int AMeridians, int AParallels,
		  CDart** ASouthPole, CDart** ANorthPole);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createSphere' ci-dessus et
 * retourne le brin correspondant  ASouthPole.
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la sphre
 * @param AParallels Le nombre (non nul) de parallles de la sphre
 * @return Un brin de la sphre
 */
CDart* createSphere(int AMeridians, int AParallels);

/**
 * Appelle la mthode CGMapGeneric::createTopoCylinder puis plonge le
 * cylindre cr.
 * Le cylindre obtenu est ax sur OX, sa hauteur vaut 1, son rayon 1/2.
 *
 * La face incidente  ABorder1 est dans le plan Z=-1/2.
 * La face incidente  ABorder2 est dans le plan Z=+1/2.
 *
 * Les brins AClose1 et AClose2 en sortie sont situs sur le mridien
 * passant par le sommet (+1/2,0,0).
 *
 * @param AMeridians Le nombre (non nul) de mridiens du cylindre
 * @param AParallels Le nombre (non nul) de parallles du cylindre
 * @param ABorder1 L'adresse d'un pointeur sur brin
 * @param ABorder2 L'adresse d'un pointeur sur brin
 * @param AClose1 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder1
 * @param AClose2 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder2
 */
void createCylinder(int AMeridians, int AParallels,
		    CDart** ABorder1, CDart** ABorder2,
		    bool AClose1 = true, bool AClose2 = true);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createCylinder' ci-dessus et
 * retourne le brin correspondant  ABorder1.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du cylindre
 * @param AParallels Le nombre (non nul) de parallles du cylindre
 * @param AClose1 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder1 (voir mthode 'createCylinder' ci-dessus)
 * @param AClose2 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder2 (voir mthode 'createCylinder' ci-dessus)
 * @return Un brin du cylindre
 */
CDart* createCylinder(int AMeridians, int AParallels,
		      bool AClose1 = true, bool AClose2 = true);

/**
 * Appelle la mthode CGMapGeneric::createTopoPyramid puis plonge la
 * pyramide cre.
 * La pyramide obtenue est axe sur OZ, sa hauteur vaut 1, le rayon de sa
 * base 1/2.
 *
 * La base (incidente  ABaseDart) est dans le plan Z=-1/2.
 * La pointe (incidente  APoleDart) est aux coordonnes (0,0,+1/2).
 *
 * Les brins ABaseDart et APoleDart en sortie sont situs sur le mridien
 * passant par le sommet (+1/4,0,0).
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la pyramide
 * @param AParallels Le nombre (non nul) de parallles de la pyramide
 * @param ABaseDart L'adresse d'un pointeur sur brin
 * @param APoleDart L'adresse d'un pointeur sur brin
 * @param ACloseBase Un boolen indiquant s'il faut fermer le bord incident
 *    ABaseDart
 */
void createPyramid(int AMeridians, int AParallels,
		   CDart** ABaseDart, CDart** APoleDart,
		   bool ACloseBase = true);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createPyramid' ci-dessus et
 * retourne le brin correspondant  ABaseDart.
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la pyramide
 * @param AParallels Le nombre (non nul) de parallles de la pyramide
 * @param ACloseBase Un boolen indiquant s'il faut fermer le bord incident
 *    ABaseDart (voir mthode 'createPyramid' ci-dessus)
 * @return Un brin du cylindre
 */

CDart* createPyramid(int AMeridians, int AParallels,
		     bool ACloseBase = true);

/**
 * Appelle la mthode CGMapGeneric::createTopoTorus puis plonge le tore cr.
 * Le tore obtenu est ax sur OZ, c'est--dire que l'axe OZ passe dans le
 * trou du tore sans l'intersecter.
 *
 * Le brin AEquator en sortie est situ sur le sommet de coordonnes (+1/2,0,0).
 *
 * On alpha donc les relations suivantes:
 *
 * k = r/R  ,   r+R = 1/2  ,   r = k/2/(1+k)  ,   R = 1/2/(1+k)
 *
 * @param AMeridians Le nombre (non nul) de mridiens du tore
 * @param AParallels Le nombre (non nul) de parallles du tore
 * @param ARadiusProportion La valeur r/R (petit rayon sur grand grand rayon)
 * @param AEquator L'adresse d'un pointeur sur brin
 */
void createTorus(int AMeridians, int AParallels,
		 TCoordinate ARadiusProportion,
		 CDart** AEquator);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTorus' ci-dessus et
 * retourne le brin correspondant  AEquator.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du tore
 * @param AParallels Le nombre (non nul) de parallles du tore
 * @param ARadiusProportion La valeur r/R (petit rayon sur grand grand rayon)
 * @return Un brin du tore
 */
CDart* createTorus(int AMeridians, int AParallels,
		   TCoordinate ARadiusProportion);

/**
 * Appelle la mthode 'CGMapGeneric::createTopoSquareIMeshed' et plonge le
 * carr maill cr.
 * Le carr est centr sur l'origine et son vecteur normal est OZ.
 * Les coins de la face cre sont aux coordonnes (+-1/2,+-1/2,0).
 *
 * @param ASx Le nombre de subdivisions selon la premire dimension
 * @param ASy Le nombre de subdivisions selon la deuxime dimension
 * @param AMeshDimension La dimension du maillage (0 ou 2)
 * @param ASquareCorners Un tableau de dimension 2*2 de pointeurs sur brin
 */
void createSquareIMeshed(int ASx, int ASy, int AMeshDimension,
			 CDart* ASquareCorners[2][2]);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createSquareIMeshed' ci-dessus
 * et retourne le brin correspondant  ASquareCorners[0][0].
 *
 * @param ASx Le nombre de subdivisions selon la premire dimension
 * @param ASy Le nombre de subdivisions selon la deuxime dimension
 * @param AMeshDimension La dimension du maillage (0 ou 2)
 */
CDart* createSquareIMeshed(int ASx, int ASy, int AMeshDimension);

/**
 * Appelle la mthode 'CGMapGeneric::createCubeIMeshed' puis plonge le cube
 * cr. Le cube obtenu est centr sur l'origine et son ct vaut 1.
 *
 * @param ASx Le nombre de subdivisions selon la premire dimension
 * @param ASy Le nombre de subdivisions selon la deuxime dimension
 * @param ASz Le nombre de subdivisions selon la troisime dimension
 * @param AMeshDimension La dimension du maillage (0, 1 ou 2)
 * @param ACreatedFaces Un tableau de dimension 3*2 de boolens
 * @param AFacesCorners Un tableau de dimension 3*2*2*2 de pointeurs sur brin
 */
void createCubeIMeshed(int ASx, int ASy, int ASz,
		       int AMeshDimension,
		       bool ACreatedFaces[3][2],
		       CDart* AFacesCorners[3][2][2][2]);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createCubeIMeshed'
 * ci-dessus et retourne un brin du cube.
 *
 * ATTENTION: Selon les valeurs de AFace--, l'objet cr peut ne pas tre
 * connexe !
 *
 * Si l'objet cr est connexe, ADart1 est positionn sur un de ses brins et
 * la valeur NULL est affecte  ADart2.
 * Si l'objet cr n'est pas connexe, ADart1 est positionn sur un brin de la
 * premire composante et ADart2 sur un brin de la deuxime composante.
 *
 * @param ASx Le nombre de subdivisions selon la premire dimension
 * @param ASy Le nombre de subdivisions selon la deuxime dimension
 * @param ASz Le nombre de subdivisions selon la troisime dimension
 * @param AMeshDimension La dimension du maillage (0, 1 ou 2)
 * @param ACreatedFaces Un tableau de dimension 3*2 de boolens
 * @param ADart1 Un brin du cube
 * @param ADart2 Un brin du cube
 */
void createCubeIMeshed(int ASx, int ASy, int ASz,
		       int AMeshDimension,
		       bool ACreatedFaces[3][2],
		       CDart** ADart1, CDart** ADart2);

//******************************************************************************
